package ch.thomson.ucs.alarms;


import ch.thomson.ucs.alarms.client.dao.AlarmHistoryDAO;
import ch.thomson.ucs.alarms.client.dao.AlarmPendingDAO;
import ch.thomson.ucs.tine.server.internal.client.PropertyNumericChangeListener;
import ch.thomson.ucs.tine.server.internal.client.PropertyNumericDAO;
import ch.thomson.ucs.tine.server.internal.client.PropertyNumericNotification;

public abstract class AlarmVectorHandler implements PropertyNumericChangeListener {
	public enum Category {
		NONE("None"),
		MESSAGE("Message"),
		WARNING("Warning"),
		TRIP("Trip"),
		FAULT("Fault");
		
		private final String name;
		
		private Category(String name) {
			this.name = name;
		}
		
		public String getName() {return name;}
	}
	
	public enum Descriptor {
		NO("-"),
		COME("Come"),
		GONE("Gone"),
		FAULT("Fault"),
		FIRST_FAULT("First Fault"),
		TRIP("Trip"),
		FIRST_TRIP("First Trip"),
		FATAL_ERROR("Fatal Error");
		private final String name;
		
		private Descriptor(String name) {
			this.name = name;
		}
		
		public String getName() {return name;}
	}
	
	protected PropertyNumericDAO property;
	protected AlarmHistoryDAO alarmHistory;
	protected Category category;
	protected Descriptor descriptor;
	protected int messageListOffset;
	private AlarmVectorHandlerFactory alarmManager;
	protected volatile boolean reset = true;
	
	public AlarmVectorHandler(PropertyNumericDAO property, int messageListOffset, Category category, Descriptor descriptor) throws Exception {
		property.addChangeListener(this);
		this.property = property;
		this.messageListOffset = messageListOffset;
		this.category = category;
		this.descriptor = descriptor;
		alarmManager = AlarmVectorHandlerFactory.getInstance();
	}

	@Override
	public void notifyChanged(PropertyNumericNotification msg) {
		
			int[] oldValues = msg.getOldValuesInt();
			int[] newValues = msg.getNewValuesInt();
			
			for(int i=0; i < oldValues.length && i < newValues.length; ++i) {
				if(oldValues[i] != newValues[i]) {
					
					// Refresh AlarmVector
					if(isAlarmVectorEntry()) {
						try {
							AlarmVectorHandlerFactory.getInstance().setAlarmVectorValue(i + messageListOffset, newValues[i]);
						} catch (Exception e1) {
							e1.printStackTrace();
						}
					}
					
					// Alle Bits abfragen
					for(int j = 0; j < Integer.SIZE; ++j) {
						int oldValue = (oldValues[i] >>> j) & 0x1;
						int newValue = (newValues[i] >>> j) & 0x1;
						String messageId = Alarms.translateToMessageID(i + messageListOffset, j);
						
						// Positive Flanke (gekommen)
						if(oldValue < newValue) {
							try {
								alarmHasCome(messageId);
							} catch (Exception e) {
								e.printStackTrace();
								getAlarmPendingDAO().close();
							}
						}
						// Negative Flanke (gegangen) nach einem Reset nicht anzeigen
						if(newValue < oldValue) {
							try {
								alarmHasGone(messageId);
							} catch (Exception e) {
								e.printStackTrace();
								getAlarmPendingDAO().close();
							}
						}
					}	
				}
			}
			reset = false;
		
	}
	
	public abstract void alarmHasCome(String messageId) throws Exception;
	public abstract void alarmHasGone(String messageId) throws Exception;
	
	public AlarmSumHandler getAlarmSumHandler() {
		return alarmManager;
	}
	
	public AlarmPendingDAO getAlarmPendingDAO() {
		return alarmManager.getAlarmPendingDAO();
	}
	
	public AlarmHistoryDAO getAlarmHistoryDAO() {
		return alarmManager.getAlarmHistoryDAO();
	}
	
	public void reset() {
		this.reset = true;
	}
	
	public boolean hasPendingAlarms() {
		int[] values = property.getValuesAsInt();
		for(int value : values) {
			if(value != 0)
				return true;
		}
		return false;
	}
	
	private boolean isAlarmVectorEntry() {
		// FAULT und TRIP werden zu Gunsten ihrer FIRST_xxx	Deskriptoren unterdrckt
		return descriptor != Descriptor.FAULT && descriptor != Descriptor.TRIP;
	}
}
