package ch.thomson.ucs.alarms;

import java.util.LinkedList;
import java.util.List;

import ch.thomson.ucs.tine.server.internal.client.PropertyNumericChangeListener;
import ch.thomson.ucs.tine.server.internal.client.PropertyNumericDAO;
import ch.thomson.ucs.tine.server.internal.client.PropertyNumericNotification;
import ch.thomson.ucs.tine.server.internal.client.ThreadPool;

public class ResetHandler implements PropertyNumericChangeListener {
	
	protected PropertyNumericDAO property;
	private List<ResetListener> resetListeners = new LinkedList<ResetListener>();
	private static ResetHandler instance;
	
	private ResetHandler(PropertyNumericDAO property) {
		this.property = property;
		property.addChangeListener(this);
	}
	
	@Override
	public void notifyChanged(PropertyNumericNotification msg) {
		if((msg.getNewValuesInt())[0] != 0) {
			ThreadPool.getInstance().execute(new Runnable() {
				@Override
				public void run() {
					// Reset Request rcksetzen.
					property.setInt(0);
					
					// Reset Listeners knstlich toggeln
					for (ResetListener l : resetListeners)
						l.resetAlarms(true);
					
					for (ResetListener l : resetListeners)
						l.resetAlarms(false);
				}
			
			});
		}
	}
	
	public static ResetHandler instance() {
		try {
			return instance(null);
		} catch(Exception e) {
			return null;
		}
	}
	
	public static synchronized ResetHandler instance(String path) throws Exception {
		if(instance == null) {
			PropertyNumericDAO pn = PropertyNumericDAO.createPropertyNumericDAO(path, true);
			instance = new ResetHandler(pn);
		}
		return instance;
	}
	
	public void addResetListener(ResetListener listener) {
		this.resetListeners.add(listener);
	}
	
	public void removeResetListener(ResetListener listener) {
		this.resetListeners.remove(listener);
	}
}
