package ch.thomson.ucs.alarms;


import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.Map;

import ch.thomson.ucs.alarms.Alarms.BitVectorPosition;
import ch.thomson.ucs.alarms.client.dao.entries.AlarmHistoryEntry;
import ch.thomson.ucs.alarms.client.dao.entries.AlarmPendingEntry;
import ch.thomson.ucs.tine.server.internal.client.PropertyNumericDAO;
import ch.thomson.ucs.tine.server.internal.client.PropertyNumericNotification;

public class StatemachineAlarmVectorHandler extends AlarmVectorHandler {
	
	protected Map<String, AlarmPendingEntry> pendingWarnings = Collections.synchronizedMap(new LinkedHashMap<String, AlarmPendingEntry>());
	static private final String eqm = "State Machine";
	private AlarmVectorHandlerFactory factory = AlarmVectorHandlerFactory.getInstance();
	
	@Deprecated
	public StatemachineAlarmVectorHandler(PropertyNumericDAO property, int messageListOffset) throws Exception {
		this();
	}
	
	public StatemachineAlarmVectorHandler() throws Exception {
		super(AlarmVectorHandlerFactory.getInstance().getAlarmVector(), -1, Category.NONE, Descriptor.NO);
		// Melde dich selber bei der Factory an
		factory.setStatemachineVectorHandler(this);
	}
	
	@Override
	public void notifyChanged(PropertyNumericNotification msg) {
		// Mache hier nichts, Notification nicht ntig
	}
	
	public void setWarning(String id, boolean come) throws Exception {
		long timeStamp = System.currentTimeMillis();
		
		BitVectorPosition bvp = Alarms.translateToBitVectorPosition(id);
		
		if(come) {
			if(pendingWarnings.containsKey(id))
				return;
			AlarmPendingEntry ape = new AlarmPendingEntry(timeStamp, id, eqm, Category.WARNING.getName(), Descriptor.COME.getName());
			pendingWarnings.put(id, ape);
			
			factory.setAlarmVectorValueMasked(bvp.getArrayIndex(), bvp.getMask());
			this.getAlarmPendingDAO().setEntry(ape);
			this.getAlarmHistoryDAO().setEntry(new AlarmHistoryEntry(timeStamp, id, eqm, Category.WARNING.getName(), Descriptor.COME.getName()));
			this.getAlarmSumHandler().setSumWarning();
		}
	
		else {
			if(pendingWarnings.remove(id) == null)
				return;
			factory.resetAlarmVectorValueMasked(bvp.getArrayIndex(), bvp.getMask());
			this.getAlarmPendingDAO().setEntry(new AlarmPendingEntry(timeStamp, id, eqm, Category.WARNING.getName(), Descriptor.GONE.getName()));
			this.getAlarmHistoryDAO().setEntry(new AlarmHistoryEntry(timeStamp, id, eqm, Category.WARNING.getName(), Descriptor.GONE.getName()));
			this.getAlarmSumHandler().warningIsGone();
		}
	}
	
	public void setFault(String id) throws Exception {
		long timeStamp = System.currentTimeMillis();
		
		BitVectorPosition bvp = Alarms.translateToBitVectorPosition(id);
		
		factory.setAlarmVectorValueMasked(bvp.getArrayIndex(), bvp.getMask());
		this.getAlarmPendingDAO().setEntry(new AlarmPendingEntry(timeStamp, id, eqm, Category.FAULT.getName(), Descriptor.NO.getName()));
		this.getAlarmHistoryDAO().setEntry(new AlarmHistoryEntry(timeStamp, id, eqm, Category.FAULT.getName(), Descriptor.NO.getName()));
		this.getAlarmSumHandler().setSumFault();
	}

	@Override
	public void reset() {
		factory.resetAlarmVector();
		if(pendingWarnings.size() > 0)
			this.getAlarmSumHandler().setSumWarning();
		
		// Noch anstehende Warnungen wieder setzen
		synchronized(pendingWarnings) {
			for(AlarmPendingEntry ape : pendingWarnings.values()) {
				BitVectorPosition bvp = Alarms.translateToBitVectorPosition(ape.getId());
				factory.setAlarmVectorValueMasked(bvp.getArrayIndex(), bvp.getMask());
				try {
					this.getAlarmPendingDAO().setEntry(ape);
				} catch (Exception e) {
					e.printStackTrace();
				}
			}
		}
		super.reset();
	}

	public boolean isRefreshNeeded() {
		return reset;
	}

	public void setRefreshNeeded(boolean refreshNeeded) {
		this.reset = refreshNeeded;
	}
	
	public boolean hasPendingWarnings() {
		return this.pendingWarnings.size() != 0;
	}

	@Override
	public void alarmHasCome(String messageId) throws Exception {
		// Nichts tun
	}

	@Override
	public void alarmHasGone(String messageId) throws Exception {
		// Nichts tun
	}
}
