package ch.thomson.ucs.alarms.client.dao.impl;

import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.SQLIntegrityConstraintViolationException;
import java.sql.Statement;
import java.sql.Timestamp;
import java.util.LinkedList;
import java.util.List;

import ch.thomson.ucs.alarms.client.dao.AlarmPendingDAO;
import ch.thomson.ucs.alarms.client.dao.entries.AlarmPendingEntry;

public class AlarmPendingDAOImpl extends BasicDAOImpl implements AlarmPendingDAO {
		
	public void setupTable() throws SQLException {
		StringBuilder query = new StringBuilder();
		query.append("CREATE TABLE IF NOT EXISTS alarms_pending (");
		query.append("timestamp TIMESTAMP NOT NULL, ");
		query.append("id VARCHAR(16) NOT NULL, ");
		query.append("eqm VARCHAR(32) NOT NULL, ");
		query.append("category VARCHAR(16) NOT NULL, ");
		query.append("descriptor VARCHAR(16), ");
		// Primary Key: Es drfen nur gleiche Meldungen mit unterschiedlichen Kategorien existieren (Warnung oder Fault)
		//query.append("PRIMARY KEY (id, category),");
		query.append("FOREIGN KEY (id) REFERENCES messages);");
		
		Statement st = null;
		try {
			st = connection.createStatement();
			st.executeUpdate(query.toString());
		} finally {
			if(st != null)
				st.close();
		}
	}

	@Override
	public void setEntry(AlarmPendingEntry entry) throws Exception {
		String insert = "INSERT INTO alarms_pending VALUES (?, ?, ?, ?, ?);";
		PreparedStatement ps = null;
		
		try {
			ps = connection.prepareStatement(insert);
			ps.setTimestamp(1, entry.getTimestamp());
			ps.setString(2, entry.getId());
			ps.setString(3, entry.getEqm());
			ps.setString(4, entry.getCategory());
			ps.setString(5, entry.getDescriptor());
			ps.executeUpdate();
		} catch (SQLIntegrityConstraintViolationException e) { 
			// Nichts tun, Eintrag ist schon vorhanden
		} finally {
			if(ps != null)
				ps.close();
		}
	}

	@Override
	public void deleteEntry(String id, String category) throws Exception {
		String delete = "DELETE FROM alarms_pending WHERE id=? AND category=?;";
		PreparedStatement ps = null;
		
		try {
			ps = connection.prepareStatement(delete);
			ps.setString(1, id);
			ps.setString(2, category);
			ps.executeUpdate();
		}
		finally {
			if(ps != null)
				ps.close();
		}
	}

	@Override
	public void deleteAllEntries() throws Exception {
		Statement st = null;
		try {
			st = connection.createStatement();
			st.execute("DELETE FROM alarms_pending;");
		} finally {
			if(st != null)
				st.close();
		}
	}

	@Override
	public List<AlarmPendingEntry> getAllEntries() throws Exception {
		
		List<AlarmPendingEntry> result = new LinkedList<AlarmPendingEntry>();
		
		Statement st = null;
		
		try {
			st = connection.createStatement();
			ResultSet rs = st.executeQuery("SELECT * FROM alarms_pending AS a, messages AS m WHERE a.id = m.id ORDER BY timestamp DESC;");
		
			while (rs.next()) {
				Timestamp timestamp = rs.getTimestamp("timestamp"); 	
				String id = rs.getString("id");
				String eqm = rs.getString("eqm");
				String category = rs.getString("category");
				String descriptor = rs.getString("descriptor");
				String message = rs.getString("message");
			
				AlarmPendingEntry ape = new AlarmPendingEntry(timestamp, id, eqm, category, descriptor, message);
				result.add(ape); 	
			}
		} finally {
			if(st != null)
				st.close();
		}
		
		return result;
	}
}
