package ch.thomson.ucs.desy.business;

import java.util.LinkedList;
import java.util.List;

import org.osgi.framework.BundleActivator;
import org.osgi.framework.BundleContext;
import org.osgi.framework.BundleEvent;
import org.osgi.framework.BundleListener;

import ch.thomson.ucs.alarms.AlarmVectorHandler.Category;
import ch.thomson.ucs.alarms.AlarmVectorHandler.Descriptor;
import ch.thomson.ucs.alarms.AlarmVectorHandlerFactory;
import ch.thomson.ucs.alarms.ResetHandler;
import ch.thomson.ucs.alarms.ResetListener;
import ch.thomson.ucs.desy.specialfunctions.AnalogArrayConditionsHandler;
import ch.thomson.ucs.desy.specialfunctions.AnalogConditionsHandler;
import ch.thomson.ucs.desy.specialfunctions.CopyPropertyObserved;
import ch.thomson.ucs.desy.specialfunctions.Division;
import ch.thomson.ucs.desy.specialfunctions.DivisionMultiplication;
import ch.thomson.ucs.desy.specialfunctions.FlowDifferenceCalculation;
import ch.thomson.ucs.desy.specialfunctions.IIRFilter;
import ch.thomson.ucs.desy.specialfunctions.MultiplicationDivision;
import ch.thomson.ucs.desy.specialfunctions.PMStatusParser;
import ch.thomson.ucs.desy.specialfunctions.PMVersionTypeParser;
import ch.thomson.ucs.desy.specialfunctions.PSMPulseCounterPersistence;
import ch.thomson.ucs.desy.specialfunctions.PSMSequenceHandler;
import ch.thomson.ucs.desy.specialfunctions.SampleScaleFactorChanger;
import ch.thomson.ucs.desy.specialfunctions.SpecialFunction;
import ch.thomson.ucs.desy.specialfunctions.fastchannel.FastChannelCalculationSet;
import ch.thomson.ucs.desy.specialfunctions.fastchannel.PulseActPowerSelection;
import ch.thomson.ucs.desy.specialfunctions.fastchannel.PulseActSelection;
import ch.thomson.ucs.desy.specialfunctions.fastchannel.PulseAverageCalculation;
import ch.thomson.ucs.desy.specialfunctions.fastchannel.PulseEnergyCalculation;
import ch.thomson.ucs.desy.specialfunctions.fastchannel.PulsePowerCalculations;
import ch.thomson.ucs.tine.client.statemachine.StateMachine;
import ch.thomson.ucs.tine.client.statemachine.util.StateMachineResourceLoader;
import ch.thomson.ucs.tine.server.internal.client.PropertyNumericDAO;

public class Activator implements BundleActivator {

	// States
	public enum States {OFF, AUX, STAND_BY_A, STAND_BY_B, STAND_BY_C, HV_READY};
	
	// Device Paths
	private static final String PATH_PSM = "PSM/Device1/";
	private static final String PATH_SCI = "SCI/Device1/";
	private static final String PATH_SCI_FAST = "FAST_CHANNEL/Device1/";
	private static final String PATH_DUMMIES = "DUMMIES/Device1/";
	private static final String PATH_IOS = "IO_SYSTEM/Device1/";
	private static final String PATH_MCB = "MAIN_CB/Device1/";
	
	// PSM Filter Properties
	private static final String[] IIR_COEFFS_A = new String[] {PATH_PSM + "PSMS_COEF_FILTER_IIR_A0", PATH_PSM + "PSMS_COEF_FILTER_IIR_A1", PATH_PSM + "PSMS_COEF_FILTER_IIR_A2"};
	private static final String[] IIR_COEFFS_B = new String[] {PATH_PSM + "PSMS_COEF_FILTER_IIR_B0", PATH_PSM + "PSMS_COEF_FILTER_IIR_B1", PATH_PSM + "PSMS_COEF_FILTER_IIR_B2"};
	private static final String IIR_D = PATH_DUMMIES + "PSMD_COEF_FILTER_IIR_D";
	private static final String IIR_T0 = PATH_DUMMIES + "PSMD_COEF_FILTER_IIR_T0";
	private static final String IIR_T1 = PATH_DUMMIES + "PSMD_COEF_FILTER_IIR_T1";
	
	// PSM Sequences
	private static final String PSM_SEQUENCE_SELECTION = PATH_DUMMIES + "PSMD_NMBR_SEQUENCE";
	private static final String PSM_SEQUENCE = PATH_PSM + "PSMS_ADDR_PM_SEQUENCE";
	
	// PSM Pulse Counter Persistence
	private static final String PSM_SET_PULSE_COUNTER = PATH_PSM + "PSMS_CNTR_PULSE_INIT";
	private static final String PSM_GET_PULSE_COUNTER = PATH_PSM + "PSMG_CNTR_PULSE";
	private static final String ACTUAL_STATE = PATH_DUMMIES + "ACTUAL_STATE";
	
	// PSM Power Modules
	private static final String PM_VECT_VERS_TYPE = PATH_PSM + "PSMG_VECT_PM_VERS_X_TYPE";
	private static final String PM_VERSION = PATH_DUMMIES + "PSMD_VERS_PM";
	private static final String PM_CONF = PATH_DUMMIES + "PSMD_CONF_PM";
	private static final String PM_VECT_STATUS = PATH_PSM + "PSMG_VECT_PM_STATUS";
	private static final String PM_STATUS_LAST = PATH_DUMMIES + "PSMD_STAT_PM_LAST";
	
	// Alarm Handling
		// PSM
	private static final String RESET_ALARMS = PATH_DUMMIES + "RESET_ALARMS";
	private static final String VECT_PSM_WARNING = PATH_PSM + "PSMG_VECT_WARNING";
	private static final String VECT_PSM_FATAL_ERROR = PATH_PSM + "PSMG_VECT_FATAL_ERROR";
	private static final String VECT_PSM_FAULT = PATH_PSM + "PSMG_VECT_FAULT";
	private static final String VECT_PSM_FAULT_FIRST = PATH_PSM + "PSMG_VECT_FAULT_FIRST";
	private static final String VECT_PSM_TRIP = PATH_PSM + "PSMG_VECT_TRIP";
	private static final String VECT_PSM_TRIP_FIRST = PATH_PSM + "PSMG_VECT_TRIP_FIRST";
	
		// SCI
	private static final String VECT_SCI_WARNING = PATH_SCI + "SCIG_VECT_WARNING";
	private static final String VECT_SCI_FATAL_ERROR = PATH_SCI + "SCIG_VECT_FATAL_ERROR";
	private static final String VECT_SCI_FAULT = PATH_SCI + "SCIG_VECT_FAULT";
	private static final String VECT_SCI_FAULT_FIRST = PATH_SCI + "SCIG_VECT_FAULT_FIRST";
	private static final String VECT_SCI_TRIP = PATH_SCI + "SCIG_VECT_TRIP";
	private static final String VECT_SCI_TRIP_FIRST = PATH_SCI + "SCIG_VECT_TRIP_FIRST";
	
		// Sum
	private static final String SUM_WARNING = PATH_DUMMIES + "SUM_WARNING";
	private static final String SUM_TRIP = PATH_DUMMIES + "SUM_TRIP";
	private static final String SUM_FAULT = PATH_DUMMIES + "SUM_FAULT";
	
		// Alarm Vector
	private static final String ALARM_VECTOR = PATH_DUMMIES + "ALARM_VECTOR";
	
	// Analog Conditions
	private static final String COL_IN_F = PATH_IOS + "COL-IN-F";
	private static final String COL_IN_F_MAX = PATH_DUMMIES + "IOSD_COL-IN-F-MAX";
	private static final String COL_IN_F_MIN = PATH_DUMMIES + "IOSD_COL-IN-F-MIN";
	private static final String COL_IN_F_HIGH = PATH_DUMMIES + "IOSD_COL-IN-F-HIGH";
	private static final String COL_IN_F_LOW = PATH_DUMMIES + "IOSD_COL-IN-F-LOW";
	private static final String COL_IN_F_COND = PATH_DUMMIES + "IOSD_COL-IN-F-COND";
	
	private static final String COL_VENT_F = PATH_IOS + "COL-VENT-F";
	private static final String COL_VENT_F_MAX = PATH_DUMMIES + "IOSD_COL-VENT-F-MAX";
	private static final String COL_VENT_F_MIN = PATH_DUMMIES + "IOSD_COL-VENT-F-MIN";
	private static final String COL_VENT_F_HIGH = PATH_DUMMIES + "IOSD_COL-VENT-F-HIGH";
	private static final String COL_VENT_F_LOW = PATH_DUMMIES + "IOSD_COL-VENT-F-LOW";
	private static final String COL_VENT_F_COND = PATH_DUMMIES + "IOSD_COL-VENT-F-COND";
	
	private static final String COL_MODULE_F = PATH_IOS + "COL-MODULE-F";
	private static final String COL_MODULE_F_MAX = PATH_DUMMIES + "IOSD_COL-MODULE-F-MAX";
	private static final String COL_MODULE_F_MIN = PATH_DUMMIES + "IOSD_COL-MODULE-F-MIN";
	private static final String COL_MODULE_F_HIGH = PATH_DUMMIES + "IOSD_COL-MODULE-F-HIGH";
	private static final String COL_MODULE_F_LOW = PATH_DUMMIES + "IOSD_COL-MODULE-F-LOW";
	private static final String COL_MODULE_F_COND = PATH_DUMMIES + "IOSD_COL-MODULE-F-COND";
	
	private static final String COL_PSMFILTER_F = PATH_IOS + "COL-PSMFILTER-F";
	private static final String COL_PSMFILTER_F_MAX = PATH_DUMMIES + "IOSD_COL-PSMFILTER-F-MAX";
	private static final String COL_PSMFILTER_F_MIN = PATH_DUMMIES + "IOSD_COL-PSMFILTER-F-MIN";
	private static final String COL_PSMFILTER_F_HIGH = PATH_DUMMIES + "IOSD_COL-PSMFILTER-F-HIGH";
	private static final String COL_PSMFILTER_F_LOW = PATH_DUMMIES + "IOSD_COL-PSMFILTER-F-LOW";
	private static final String COL_PSMFILTER_F_COND = PATH_DUMMIES + "IOSD_COL-PSMFILTER-F-COND";
	
	private static final String COL_IN_T = PATH_IOS + "COL-IN-T";
	private static final String COL_IN_T_MAX = PATH_DUMMIES + "IOSD_COL-IN-T-MAX";
	private static final String COL_IN_T_HIGH = PATH_DUMMIES + "IOSD_COL-IN-T-HIGH";
	private static final String COL_IN_T_COND = PATH_DUMMIES + "IOSD_COL-IN-T-COND";
	
	private static final String COL_OUT_T = PATH_IOS + "COL-OUT-T";
	private static final String COL_OUT_T_MAX = PATH_DUMMIES + "IOSD_COL-OUT-T-MAX";
	private static final String COL_OUT_T_HIGH = PATH_DUMMIES + "IOSD_COL-OUT-T-HIGH";
	private static final String COL_OUT_T_COND = PATH_DUMMIES + "IOSD_COL-OUT-T-COND";
	
	private static final String COL_AMB_T = PATH_IOS + "COL-AMB-T";
	private static final String COL_AMB_T_MAX = PATH_DUMMIES + "IOSD_COL-AMB-T-MAX";
	private static final String COL_AMB_T_HIGH = PATH_DUMMIES + "IOSD_COL-AMB-T-HIGH";
	private static final String COL_AMB_T_COND = PATH_DUMMIES + "IOSD_COL-AMB-T-COND";
	
	private static final String HV_TP1_T = PATH_IOS + "HV-TP1-T";
	private static final String HV_TP_T_MAX = PATH_DUMMIES + "IOSD_HV-TP-T-MAX";
	private static final String HV_TP_T_HIGH = PATH_DUMMIES + "IOSD_HV-TP-T-HIGH";
	private static final String HV_TP1_T_COND = PATH_DUMMIES + "IOSD_HV-TP1-T-COND";
	
	private static final String HV_TP2_T = PATH_IOS + "HV-TP2-T";
	private static final String HV_TP2_T_COND = PATH_DUMMIES + "IOSD_HV-TP2-T-COND";
	
	private static final String HV_TP3_T = PATH_IOS + "HV-TP3-T";
	private static final String HV_TP3_T_COND = PATH_DUMMIES + "IOSD_HV-TP3-T-COND";
	
	private static final String COL_DIFF_F = PATH_DUMMIES + "IOSD_COL-DIFF-F";
	private static final String COL_DIFF_F_MAX = PATH_DUMMIES + "IOSD_COL-DIFF-F-MAX";
	private static final String COL_DIFF_F_HIGH = PATH_DUMMIES + "IOSD_COL-DIFF-F-HIGH";
	private static final String COL_DIFF_F_COND = PATH_DUMMIES + "IOSD_COL-DIFF-F-COND";
	
	private static final String POWR_TOTAL_ACTIVE = PATH_MCB + "MCBG_POWR_TOTAL_ACTIVE";
	private static final String POWR_TOTAL_ACTIVE_MAX = PATH_DUMMIES + "MCBD_POWR_TOTAL_ACTIVE_MAX";
	private static final String POWR_TOTAL_ACTIVE_COND = PATH_DUMMIES + "MCBD_POWR_TOTAL_ACTIVE_COND";
	
	private static final String POWR_TOTAL_REACTIVE = PATH_MCB + "MCBG_POWR_TOTAL_REACTIVE";
	private static final String POWR_TOTAL_REACTIVE_MAX = PATH_DUMMIES + "MCBD_POWR_TOTAL_REACTIVE_MAX";
	private static final String POWR_TOTAL_REACTIVE_COND = PATH_DUMMIES + "MCBD_POWR_TOTAL_REACTIVE_COND";
	
	private static final String VOLT_PP = PATH_MCB + "MCBG_VOLT_PP";
	private static final String VOLT_PP_MIN = PATH_DUMMIES + "MCBD_VOLT_PP_MIN";
	private static final String VOLT_PP_COND = PATH_DUMMIES + "MCBD_VOLT_PP_COND";
	
	private static final String FREQ = PATH_MCB + "MCBG_FREQ";
	private static final String FREQ_MIN = PATH_DUMMIES + "MCBD_FREQ_MIN";
	private static final String FREQ_COND = PATH_DUMMIES + "MCBD_FREQ_COND";
	
	private static final String VOLT_RHVPS_MEAS_RANGE = PATH_DUMMIES + "SCID_VOLT_RHVPS_MEAS_RANGE";
	private static final String CURR_RHVPS_MEAS_RANGE = PATH_DUMMIES + "SCID_CURR_RHVPS_MEAS_RANGE";
	private static final String VOLT_KLY_MEAS_RANGE = PATH_DUMMIES + "SCID_VOLT_KLY_MEAS_RANGE";
	private static final String CURR_KLY_MEAS_RANGE = PATH_DUMMIES + "SCID_CURR_KLY_MEAS_RANGE";
	
/*	
	Dies sind die Special Function Definitionen gemss Signalliste V1.9:
	
	SF001	PSMS_TIME_CYCLE	= 1,000,000 / PSMS_FREQ_PULSE_REP_RATE
	SF002	PSMS_TIME_CYCLE_TOLERANCE = PSMS_TIME_CYCLE / 10
	SF003	PSMS_STPS_RISETIME =	1 / ( PSMD_TIME_RISE * PSMG_FREQ_SAMPLE_OF_STEP_SIZE)
	SF004	PSMS_STPS_PWM_FREQUENCY = 1000 * PSMD_FREQ_PWM / (1'000'000 * PSMG_FREQ_FPGA_SYSTEM )
	SF010	SCIS_TIME_PULSE_DELAY	
	SF011	SCIS_VOLT_RHVPS_MAX_SCLD = SCID_VOLT_RHVPS_MAX / SCID_VOLT_RHVPS_MEAS_RANGE
	SF012	SCIS_CURR_RHVPS_MAX_SCLD = SCID_CURR_RHVPS_MAX / SCID_CURR_RHVPS_MEAS_RANGE
	SF013	SCIS_CURR_RHVPS_MAX_SC_SCLD = SCID_CURR_RHVPS_MAX_SC / SCID_CURR_RHVPS_MEAS_RANGE
	SF014	SCIS_VOLT_KLY_MAX_SCLD = SCID_VOLT_KLY_MAX / SCID_VOLT_KLY_MEAS_RANGE
	SF015	SCIS_CURR_KLY_MAX_SCLD = SCID_CURR_KLY_MAX / SCID_CURR_KLY_MEAS_RANGE
	SF016	SCIS_VOLT_DIFF_KLY_MAX_SCLD = SCID_VOLT_DIFF_KLY_MAX / SCID_VOLT_RHVPS_MEAS_RANGE
	SF017	SCIS_POWR_MAX_SCLD = SCID_POWR_MAX * 1'000'000 / ( SCID_VOLT_RHVPS_MEAS_RANGE * 1000 * SCID_CURR_RHVPS_MEAS_RANGE )
	SF018	SCIS_ENGY_PULSE_MAX_SCLD = SCID_ENGY_PULSE_MAX * 10'000'000Hz  /  ( SCID_VOLT_RHVPS_MEAS_RANGE * SCID_CURR_RHVPS_MEAS_RANGE )
	SF019	SCIS_POWR_MEAN_MAX_SCLD = SCID_POWR_MEAN_MAX / ( SCID_VOLT_RHVPS_MEAS_RANGE * SCID_CURR_RHVPS_MEAS_RANGE )
	SF020	SCIS_IMPD_MIN_SCLD = SCID_IMPD_MIN * 64 * SCID_CURR_RHVPS_MEAS_RANGE / (SCID_VOLT_RHVPS_MEAS_RANGE * 1000)
	SF021	SCIS_IMPD_DIFF_MAX_SCLD = SCID_IMPD_DIFF_MAX * 64 * SCID_CURR_RHVPS_MEAS_RANGE / (SCID_VOLT_RHVPS_MEAS_RANGE * 1000)
	SF022	SCIS_CURR_IMPD_MON_ON_SCLD = SCID_CURR_IMPD_MON_ON / SCID_CURR_RHVPS_MEAS_RANGE
	SF023	SCIS_CURR_IMPD_MON_OFF_SCLD = SCID_CURR_IMPD_MON_OFF / SCID_CURR_RHVPS_MEAS_RANGE
	SF024	SCID_POWR_OUT = SCID_ENGY_PULSE * SCID_FREQ_PULSE_REP_RATE (ACHTUNG: Fast Channel Calculation)
	SF025	SCID_FREQ_PULSE_REP_RATE = PSMS_FREQ_PULSE_REP_RATE
*/
	
	// Special Functions PSM
	private static final String[] SF001_OPERATORS = {"1000000", PATH_PSM + "PSMS_FREQ_PULSE_REP_RATE"};
	private static final String SF001_RESULT = PATH_PSM + "PSMS_TIME_CYCLE";
	
	private static final String[] SF002_OPERATORS = {PATH_PSM + "PSMS_TIME_CYCLE", "10"};
	private static final String SF002_RESULT = PATH_PSM + "PSMS_TIME_CYCLE_TOLERANCE";
	
	private static final String[] SF003_OPERATORS = {"1", PATH_DUMMIES + "PSMD_TIME_RISE", PATH_PSM + "PSMG_FREQ_SAMPLE_OF_STEP_SIZE"};
	private static final String SF003_RESULT = PATH_PSM + "PSMS_STPS_RISETIME";
	
	private static final String[] SF004_OPERATORS = {PATH_DUMMIES + "PSMD_FREQ_PWM", PATH_PSM + "PSMG_FREQ_FPGA_SYSTEM"};
	private static final String SF004_RESULT = PATH_PSM + "PSMS_STPS_PWM_FREQUENCY";
	
	// Special Functions SCI
	private static final String[] SF011_OPERATORS = {PATH_DUMMIES + "SCID_VOLT_RHVPS_MAX", VOLT_RHVPS_MEAS_RANGE};
	private static final String SF011_RESULT = PATH_SCI + "SCIS_VOLT_RHVPS_MAX_SCLD";
	
	private static final String[] SF012_OPERATORS = {PATH_DUMMIES + "SCID_CURR_RHVPS_MAX", CURR_RHVPS_MEAS_RANGE};
	private static final String SF012_RESULT = PATH_SCI + "SCIS_CURR_RHVPS_MAX_SCLD";
	
	private static final String[] SF013_OPERATORS = {PATH_DUMMIES + "SCID_CURR_RHVPS_MAX_SC", CURR_RHVPS_MEAS_RANGE};
	private static final String SF013_RESULT = PATH_SCI + "SCIS_CURR_RHVPS_MAX_SC_SCLD";
	
	private static final String[] SF014_OPERATORS = {PATH_DUMMIES + "SCID_VOLT_KLY_MAX", VOLT_KLY_MEAS_RANGE};
	private static final String SF014_RESULT = PATH_SCI + "SCIS_VOLT_KLY_MAX_SCLD";
	
	private static final String[] SF015_OPERATORS = {PATH_DUMMIES + "SCID_CURR_KLY_MAX", CURR_KLY_MEAS_RANGE};
	private static final String SF015_RESULT = PATH_SCI + "SCIS_CURR_KLY_MAX_SCLD";
	
	private static final String[] SF016_OPERATORS = {PATH_DUMMIES + "SCID_VOLT_DIFF_KLY_MAX", VOLT_RHVPS_MEAS_RANGE};
	private static final String SF016_RESULT = PATH_SCI + "SCIS_VOLT_DIFF_KLY_MAX_SCLD";
	
	private static final String[] SF017_OPERATORS = {PATH_DUMMIES + "SCID_POWR_MAX", VOLT_RHVPS_MEAS_RANGE, CURR_RHVPS_MEAS_RANGE};
	private static final String SF017_RESULT = PATH_SCI + "SCIS_POWR_MAX_SCLD";
	
	private static final String[] SF018_OPERATORS = {PATH_DUMMIES + "SCID_ENGY_PULSE_MAX", VOLT_RHVPS_MEAS_RANGE, CURR_RHVPS_MEAS_RANGE};
	private static final String SF018_RESULT = PATH_SCI + "SCIS_ENGY_PULSE_MAX_SCLD";
	
	private static final String[] SF019_OPERATORS = {PATH_DUMMIES + "SCID_POWR_MEAN_MAX", VOLT_RHVPS_MEAS_RANGE, CURR_RHVPS_MEAS_RANGE};
	private static final String SF019_RESULT = PATH_SCI + "SCIS_POWR_MEAN_MAX_SCLD";
	
	private static final String[] SF020_OPERATORS = {PATH_DUMMIES + "SCID_IMPD_MIN", CURR_RHVPS_MEAS_RANGE, VOLT_RHVPS_MEAS_RANGE};
	private static final String SF020_RESULT = PATH_SCI + "SCIS_IMPD_MIN_SCLD";
	
	private static final String[] SF021_OPERATORS = {PATH_DUMMIES + "SCID_IMPD_DIFF_MAX", CURR_RHVPS_MEAS_RANGE, VOLT_RHVPS_MEAS_RANGE};
	private static final String SF021_RESULT = PATH_SCI + "SCIS_IMPD_DIFF_MAX_SCLD";
	
	private static final String[] SF022_OPERATORS = {PATH_DUMMIES + "SCID_CURR_IMPD_MON_ON", CURR_RHVPS_MEAS_RANGE};
	private static final String SF022_RESULT = PATH_SCI + "SCIS_CURR_IMPD_MON_ON_SCLD";
	
	private static final String[] SF023_OPERATORS = {PATH_DUMMIES + "SCID_CURR_IMPD_MON_OFF", CURR_RHVPS_MEAS_RANGE};
	private static final String SF023_RESULT = PATH_SCI + "SCIS_CURR_IMPD_MON_OFF_SCLD";
	
	// SF024
	// Bei dieser Special Function handelt es sich um eine Fast Channel Calculation, die in der entsprechenden Energieberechnung integriert ist (aus Performancegrnden)
	private static final String SF024_OPERATOR = PATH_SCI_FAST + "SCID_FREQ_PULSE_REP_RATE";
	private static final String SF024_RESULT = PATH_SCI_FAST + "SCID_POWR_OUT";
	
	private static final String SF025_OPERATOR = PATH_PSM + "PSMS_FREQ_PULSE_REP_RATE";
	private static final String SF025_RESULT = SF024_OPERATOR;
	
	// Flow Difference
	private static final String[] FLOW_DIFF_OPERATORS = {COL_IN_F, COL_VENT_F, COL_MODULE_F, COL_PSMFILTER_F};
	private static final String FLOW_DIFF_RESULT = COL_DIFF_F;
	
	// Fast Channel Calculations
		// Triggers
	private static final String FAST_CHANNEL_START_TRIG = PATH_SCI_FAST + "SCID_TIME_PULSE_AVG_MEAS_DELAY";
	private static final String FAST_CHANNEL_N_SAMPLES = PATH_SCI_FAST + "SCID_NMBR_AVG_MEAS_SAMPLES";
	private static final String FAST_CHANNEL_ACT_TRIG = PATH_SCI_FAST + "SCID_TIME_PULSE_ACT_MEAS_DELAY";

		// Average Results
	private static final String FAST_CHANNEL_VOLT_RHVPS_AVG = PATH_SCI_FAST + "SCID_VOLT_RHVPS_AVG";
	private static final String FAST_CHANNEL_CURR_RHVPS_AVG = PATH_SCI_FAST + "SCID_CURR_RHVPS_AVG";
	private static final String FAST_CHANNEL_VOLT_KLY_AVG = PATH_SCI_FAST + "SCID_VOLT_KLY_AVG";
	private static final String FAST_CHANNEL_CURR_KLY_AVG = PATH_SCI_FAST + "SCID_CURR_KLY_AVG";
	private static final String FAST_CHANNEL_POWR_RHVPS_AVG = PATH_SCI_FAST + "SCID_POWR_RHVPS_AVG";
	
		// Actual Results
	private static final String FAST_CHANNEL_VOLT_RHVPS_ACT = PATH_SCI_FAST + "SCID_VOLT_RHVPS_ACT";
	private static final String FAST_CHANNEL_CURR_RHVPS_ACT = PATH_SCI_FAST + "SCID_CURR_RHVPS_ACT";
	private static final String FAST_CHANNEL_VOLT_KLY_ACT = PATH_SCI_FAST + "SCID_VOLT_KLY_ACT";
	private static final String FAST_CHANNEL_CURR_KLY_ACT = PATH_SCI_FAST + "SCID_CURR_KLY_ACT";
	private static final String FAST_CHANNEL_POWR_RHVPS_ACT = PATH_SCI_FAST + "SCID_POWR_RHVPS_ACT";
	
		// Pulse Energy
	private static final String FAST_CHANNEL_ENGY_PULSE = PATH_SCI_FAST + "SCID_ENGY_PULSE";

	
		// Samples
	private static final String FAST_CHANNEL_VOLT_RHVPS_SAMPLES = PATH_SCI_FAST + "SCIG_VOLT_RHVPS_FAST";
	private static final String FAST_CHANNEL_CURR_RHVPS_SAMPLES = PATH_SCI_FAST + "SCIG_CURR_RHVPS_FAST";
	private static final String FAST_CHANNEL_VOLT_KLY_SAMPLES = PATH_SCI_FAST + "SCIG_VOLT_KLY_FAST";
	private static final String FAST_CHANNEL_CURR_KLY_SAMPLES = PATH_SCI_FAST + "SCIG_CURR_KLY_FAST";
	
	private static final String FAST_CHANNEL_PWR_RHVPS_SAMPLES = PATH_SCI_FAST + "SCID_POWR_RHVPS_FAST";


	private static BundleContext context;
	private List<SpecialFunction> specialFunctions = new LinkedList<SpecialFunction>();

	static BundleContext getContext() {
		return context;
	}
	
	private void start() throws Exception {
		System.out.println("Hello from business bundle!");
		
		this.startSpecialFunctions();
		this.startAlarmVectorHandlers();
		this.startStateMachine();
	}

	/*
	 * (non-Javadoc)
	 * @see org.osgi.framework.BundleActivator#start(org.osgi.framework.BundleContext)
	 */
	public void start(BundleContext bundleContext) throws Exception {
		Activator.context = bundleContext;
		context.addBundleListener(new BundleListener() {

			@Override
			public void bundleChanged(BundleEvent event) {
				if("ch.thomson.ucs.tine.server.application".equals(event.getBundle().getSymbolicName()) && event.getType() == BundleEvent.STARTED) {
					try {
						start();
					} catch (Exception e) {
						e.printStackTrace();
					}
				}
			}
		});
	}

	/*
	 * (non-Javadoc)
	 * @see org.osgi.framework.BundleActivator#stop(org.osgi.framework.BundleContext)
	 */
	public void stop(BundleContext bundleContext) throws Exception {
		System.out.println("Goodbye from business bundle!");
		Activator.context = null;
		this.stopStateMachine();
		this.stopSpecialFunctions();
	}
	
	
	private void startSpecialFunctions() throws Exception {
		// PSM Special Functions
		this.specialFunctions.add(new Division(SF001_OPERATORS, SF001_RESULT));
		this.specialFunctions.add(new Division(SF002_OPERATORS, SF002_RESULT));
		this.specialFunctions.add(new DivisionMultiplication(SF003_OPERATORS, SF003_RESULT));
		this.specialFunctions.add(new Division((1.0f / 1000.0f), SF004_OPERATORS, SF004_RESULT));
		this.specialFunctions.add(new IIRFilter(IIR_COEFFS_A, IIR_COEFFS_B, IIR_D, IIR_T0, IIR_T1));
		
		// PSM Sequence-Handler
		this.specialFunctions.add(new PSMSequenceHandler(PSM_SEQUENCE_SELECTION, PSM_SEQUENCE));
		
		// PSM Pulse Counter Persistence
		this.specialFunctions.add(new PSMPulseCounterPersistence(PSM_SET_PULSE_COUNTER, PSM_GET_PULSE_COUNTER, ACTUAL_STATE, States.HV_READY));
		
		// Power Modules Special Functions
		this.specialFunctions.add(new PMVersionTypeParser(PM_VECT_VERS_TYPE, PM_VERSION, PM_CONF));
		this.specialFunctions.add(new PMStatusParser(PM_VECT_STATUS, PM_STATUS_LAST));
				
		// SCI Special Functions
		this.specialFunctions.add(new Division(SF011_OPERATORS, SF011_RESULT));
		this.specialFunctions.add(new Division(SF012_OPERATORS, SF012_RESULT));
		this.specialFunctions.add(new Division(SF013_OPERATORS, SF013_RESULT));
		this.specialFunctions.add(new Division(SF014_OPERATORS, SF014_RESULT));
		this.specialFunctions.add(new Division(SF015_OPERATORS, SF015_RESULT));
		this.specialFunctions.add(new Division(SF016_OPERATORS, SF016_RESULT));
		this.specialFunctions.add(new DivisionMultiplication((1000000.0f / 1000.0f), SF017_OPERATORS, SF017_RESULT));
		this.specialFunctions.add(new DivisionMultiplication(10000000.0f, SF018_OPERATORS, SF018_RESULT));
		this.specialFunctions.add(new DivisionMultiplication(SF019_OPERATORS, SF019_RESULT));
		this.specialFunctions.add(new MultiplicationDivision((64.0f / 1000.0f), SF020_OPERATORS, SF020_RESULT));
		this.specialFunctions.add(new MultiplicationDivision((64.0f / 1000.0f), SF021_OPERATORS, SF021_RESULT));
		this.specialFunctions.add(new Division(SF022_OPERATORS, SF022_RESULT));
		this.specialFunctions.add(new Division(SF023_OPERATORS, SF023_RESULT));
		this.specialFunctions.add(new CopyPropertyObserved(SF025_OPERATOR, SF025_RESULT));
		
		// IO System Special Functions
		this.specialFunctions.add(new FlowDifferenceCalculation(FLOW_DIFF_OPERATORS, FLOW_DIFF_RESULT));
		
		// Analog Conditions Special Functions
			// Flows
		this.specialFunctions.add(new AnalogConditionsHandler(COL_IN_F, COL_IN_F_LOW, COL_IN_F_HIGH, COL_IN_F_MIN, COL_IN_F_MAX, COL_IN_F_COND));
		this.specialFunctions.add(new AnalogConditionsHandler(COL_VENT_F, COL_VENT_F_LOW, COL_VENT_F_HIGH, COL_VENT_F_MIN, COL_VENT_F_MAX, COL_VENT_F_COND));
		this.specialFunctions.add(new AnalogConditionsHandler(COL_MODULE_F, COL_MODULE_F_LOW, COL_MODULE_F_HIGH, COL_MODULE_F_MIN, COL_MODULE_F_MAX, COL_MODULE_F_COND));
		this.specialFunctions.add(new AnalogConditionsHandler(COL_PSMFILTER_F, COL_PSMFILTER_F_LOW, COL_PSMFILTER_F_HIGH, COL_PSMFILTER_F_MIN, COL_PSMFILTER_F_MAX, COL_PSMFILTER_F_COND));
		this.specialFunctions.add(new AnalogConditionsHandler(COL_DIFF_F, null, COL_DIFF_F_HIGH, null, COL_DIFF_F_MAX, COL_DIFF_F_COND));
		
			// Temperatures
		this.specialFunctions.add(new AnalogConditionsHandler(COL_IN_T, null, COL_IN_T_HIGH, null, COL_IN_T_MAX, COL_IN_T_COND));
		this.specialFunctions.add(new AnalogConditionsHandler(COL_OUT_T, null, COL_OUT_T_HIGH, null, COL_OUT_T_MAX, COL_OUT_T_COND));
		this.specialFunctions.add(new AnalogConditionsHandler(COL_AMB_T, null, COL_AMB_T_HIGH, null, COL_AMB_T_MAX, COL_AMB_T_COND));
		this.specialFunctions.add(new AnalogConditionsHandler(HV_TP1_T, null, HV_TP_T_HIGH, null, HV_TP_T_MAX, HV_TP1_T_COND));
		this.specialFunctions.add(new AnalogConditionsHandler(HV_TP2_T, null, HV_TP_T_HIGH, null, HV_TP_T_MAX, HV_TP2_T_COND));
		this.specialFunctions.add(new AnalogConditionsHandler(HV_TP3_T, null, HV_TP_T_HIGH, null, HV_TP_T_MAX, HV_TP3_T_COND));
		
			// Input Power
		this.specialFunctions.add(new AnalogConditionsHandler(POWR_TOTAL_ACTIVE, null, null, null, POWR_TOTAL_ACTIVE_MAX, POWR_TOTAL_ACTIVE_COND));
		this.specialFunctions.add(new AnalogConditionsHandler(POWR_TOTAL_REACTIVE, null, null, null, POWR_TOTAL_REACTIVE_MAX, POWR_TOTAL_REACTIVE_COND));
		
			// Input Voltage
		this.specialFunctions.add(new AnalogArrayConditionsHandler(VOLT_PP, null, null, VOLT_PP_MIN, null, VOLT_PP_COND));
		
			// Input Frequency
		this.specialFunctions.add(new AnalogConditionsHandler(FREQ, null, null, FREQ_MIN, null, FREQ_COND));
		
		// Fast Channel Special Functions
		this.specialFunctions.add(new SampleScaleFactorChanger(VOLT_RHVPS_MEAS_RANGE, FAST_CHANNEL_VOLT_RHVPS_SAMPLES));
		this.specialFunctions.add(new SampleScaleFactorChanger(CURR_RHVPS_MEAS_RANGE, FAST_CHANNEL_CURR_RHVPS_SAMPLES));
		this.specialFunctions.add(new SampleScaleFactorChanger(VOLT_KLY_MEAS_RANGE, FAST_CHANNEL_VOLT_KLY_SAMPLES));
		this.specialFunctions.add(new SampleScaleFactorChanger(CURR_KLY_MEAS_RANGE, FAST_CHANNEL_CURR_KLY_SAMPLES));
		
		FastChannelCalculationSet calcs = new FastChannelCalculationSet();

		// Average Calculation
		new PulseAverageCalculation(calcs, FAST_CHANNEL_START_TRIG, FAST_CHANNEL_N_SAMPLES, FAST_CHANNEL_VOLT_RHVPS_SAMPLES, FAST_CHANNEL_VOLT_RHVPS_AVG);
		new PulseAverageCalculation(calcs, FAST_CHANNEL_START_TRIG, FAST_CHANNEL_N_SAMPLES, FAST_CHANNEL_CURR_RHVPS_SAMPLES, FAST_CHANNEL_CURR_RHVPS_AVG);
		new PulseAverageCalculation(calcs, FAST_CHANNEL_START_TRIG, FAST_CHANNEL_N_SAMPLES, FAST_CHANNEL_VOLT_KLY_SAMPLES, FAST_CHANNEL_VOLT_KLY_AVG);
		new PulseAverageCalculation(calcs, FAST_CHANNEL_START_TRIG, FAST_CHANNEL_N_SAMPLES, FAST_CHANNEL_CURR_KLY_SAMPLES, FAST_CHANNEL_CURR_KLY_AVG);
		new PulsePowerCalculations(calcs, FAST_CHANNEL_START_TRIG, FAST_CHANNEL_N_SAMPLES, FAST_CHANNEL_VOLT_RHVPS_SAMPLES, FAST_CHANNEL_CURR_RHVPS_SAMPLES, FAST_CHANNEL_POWR_RHVPS_AVG, FAST_CHANNEL_PWR_RHVPS_SAMPLES);
		
			// Actual Selection
		new PulseActSelection(calcs, FAST_CHANNEL_ACT_TRIG, FAST_CHANNEL_VOLT_RHVPS_SAMPLES, FAST_CHANNEL_VOLT_RHVPS_ACT);
		new PulseActSelection(calcs, FAST_CHANNEL_ACT_TRIG, FAST_CHANNEL_CURR_RHVPS_SAMPLES, FAST_CHANNEL_CURR_RHVPS_ACT);
		new PulseActSelection(calcs, FAST_CHANNEL_ACT_TRIG, FAST_CHANNEL_VOLT_KLY_SAMPLES, FAST_CHANNEL_VOLT_KLY_ACT);
		new PulseActSelection(calcs, FAST_CHANNEL_ACT_TRIG, FAST_CHANNEL_CURR_KLY_SAMPLES, FAST_CHANNEL_CURR_KLY_ACT);
		new PulseActPowerSelection(calcs, FAST_CHANNEL_ACT_TRIG, FAST_CHANNEL_VOLT_RHVPS_SAMPLES, FAST_CHANNEL_CURR_RHVPS_SAMPLES, FAST_CHANNEL_POWR_RHVPS_ACT);
	
			// Energy Calculation
		new PulseEnergyCalculation(calcs, FAST_CHANNEL_VOLT_RHVPS_SAMPLES, FAST_CHANNEL_CURR_RHVPS_SAMPLES, SF024_OPERATOR, FAST_CHANNEL_ENGY_PULSE, SF024_RESULT);
		
		calcs.initialize();
		this.specialFunctions.add(calcs);
	}
	
	private void stopSpecialFunctions() {
		for(SpecialFunction sfb : this.specialFunctions)
			sfb.dispose();
	}
	
	private void startAlarmVectorHandlers() throws Exception {
	
		// Fault Reset Handler erzeugen
		ResetHandler.instance(RESET_ALARMS);
		
		// Reset Listener erzeugen und Fault Reset an die entsprechenden Properties delegieren
		final PropertyNumericDAO pnFaultResetPSM = PropertyNumericDAO.createPropertyNumericDAO("PSM/Device1/PSMS_RST_FAULT", true);
		final PropertyNumericDAO pnFaultResetSCI = PropertyNumericDAO.createPropertyNumericDAO("SCI/Device1/SCIS_RST_FAULT", true);
		ResetHandler.instance().addResetListener(new ResetListener() {

			@Override
			public void resetAlarms(boolean reset) {
				pnFaultResetPSM.setInt(reset ? 1 : 0);
				pnFaultResetSCI.setInt(reset ? 1 : 0);
			}
		});

		
		// Smtliche Alarmvectoren anmelden und somit beobachten
		AlarmVectorHandlerFactory factory = AlarmVectorHandlerFactory.getInstance();
		factory.setSumWarningProperty(SUM_WARNING);
		factory.setSumTripProperty(SUM_TRIP);
		factory.setSumFaultProperty(SUM_FAULT);
		factory.setAlarmVector(ALARM_VECTOR);
		
			// PSM
		factory.createAlarmVectorHandler(VECT_PSM_FAULT, 0, Category.FAULT, Descriptor.FAULT);
		factory.createAlarmVectorHandler(VECT_PSM_FAULT_FIRST, 0, Category.FAULT, Descriptor.FIRST_FAULT);
		factory.createAlarmVectorHandler(VECT_PSM_TRIP, 1, Category.FAULT, Descriptor.TRIP);
		factory.createAlarmVectorHandler(VECT_PSM_TRIP_FIRST, 1, Category.FAULT, Descriptor.FIRST_TRIP);
		factory.createAlarmVectorHandler(VECT_PSM_WARNING, 2, Category.WARNING, Descriptor.NO);
		factory.createAlarmVectorHandler(VECT_PSM_FATAL_ERROR, 3, Category.FAULT, Descriptor.FATAL_ERROR);
		
			// SCI
		factory.createAlarmVectorHandler(VECT_SCI_FAULT, 4, Category.FAULT, Descriptor.FAULT);
		factory.createAlarmVectorHandler(VECT_SCI_FAULT_FIRST, 4, Category.FAULT, Descriptor.FIRST_FAULT);
		factory.createAlarmVectorHandler(VECT_SCI_TRIP, 5, Category.FAULT, Descriptor.TRIP);
		factory.createAlarmVectorHandler(VECT_SCI_TRIP_FIRST, 5, Category.FAULT, Descriptor.FIRST_TRIP);
		factory.createAlarmVectorHandler(VECT_SCI_WARNING, 6, Category.WARNING, Descriptor.NO);
		factory.createAlarmVectorHandler(VECT_SCI_FATAL_ERROR, 7, Category.FAULT, Descriptor.FATAL_ERROR);
	}
	
	private StateMachine stateMachine;
	
	private void startStateMachine() throws Exception {
		stateMachine = StateMachineResourceLoader.createStateMachine();
		stateMachine.run();
	}
	
	private void stopStateMachine() {
		if(stateMachine != null) {
			stateMachine.shutdown();
		}
	}
}
