package ch.thomson.ucs.desy.specialfunctions;

import ch.thomson.ucs.tine.server.internal.client.PropertyNumericDAO;

public class AnalogConditionsHandler extends SpecialFunctionPropertyObserved {
	
	protected static enum AnalogConditions {FAULT_LOW, WARNING_LOW, OK, WARNING_HIGH, FAULT_HIGH};
	protected PropertyNumericDAO pnAnalogValue;
	protected PropertyNumericDAO pnWarningLow;
	protected PropertyNumericDAO pnWarningHigh;
	protected PropertyNumericDAO pnFaultLow;
	protected PropertyNumericDAO pnFaultHigh;
	protected final PropertyNumericDAO pnCondition;
	
	public AnalogConditionsHandler(String analogValue, String warningLow, String warningHigh, String faultLow, String faultHigh, String condition) throws Exception {
		this.pnAnalogValue = this.addObservedProperty(analogValue);
		
		// Schwellen knnen null sein, wenn sie nicht existieren
		if(warningLow != null)
			this.pnWarningLow = this.addObservedProperty(warningLow);
		
		if(warningHigh != null)
			this.pnWarningHigh = this.addObservedProperty(warningHigh);
		
		if(faultLow != null)
			this.pnFaultLow = this.addObservedProperty(faultLow);
		
		if(faultHigh != null)
			this.pnFaultHigh = this.addObservedProperty(faultHigh);
		
		this.pnCondition = PropertyNumericDAO.createPropertyNumericDAO(condition, true);
		doCheck();
	}
	@Override
	protected void refresh(PropertyNumericDAO property, float[] newValues, float[] oldValues) {
		doCheck();
	}
	
	private void doCheck() {
		AnalogConditions result;
		
		// Fehler zuerst checken, hat Vorrang gegenber der Warning
		if(!AnalogConditions.OK.equals(result = checkFault())) {
			pnCondition.setInt(result.ordinal());
			return;
		}
		
		pnCondition.setInt(checkWarning().ordinal());
	}
	
	protected AnalogConditions checkWarning() {
		if(pnWarningLow != null && pnAnalogValue.getFloat() < pnWarningLow.getFloat())
			return AnalogConditions.WARNING_LOW;
		
		if(pnWarningHigh != null && pnAnalogValue.getFloat() > pnWarningHigh.getFloat())
			return AnalogConditions.WARNING_HIGH;
		
		return AnalogConditions.OK;
	}
	
	protected AnalogConditions checkFault() {
		if(pnFaultLow != null && pnAnalogValue.getFloat() < pnFaultLow.getFloat())
			return AnalogConditions.FAULT_LOW;
		
		if(pnFaultHigh != null && pnAnalogValue.getFloat() > pnFaultHigh.getFloat())
			return AnalogConditions.FAULT_HIGH;
		
		return AnalogConditions.OK;
	}
}
