package ch.thomson.ucs.desy.specialfunctions;

import java.util.ArrayList;
import java.util.List;

import ch.thomson.ucs.tine.server.internal.client.PropertyNumericDAO;

public abstract class Calculation extends SpecialFunctionPropertyObserved {
	private List<Float> constOperators = new ArrayList<Float>();
	private List<PropertyNumericDAO> dynOperators = new ArrayList<PropertyNumericDAO>();
	private PropertyNumericDAO pnResult;
	
	public Calculation(String[] operators, String result) throws Exception {
		if(operators.length != this.getNumberOfOperators())
			throw new Exception("Special Function: Illegal number of operators");
		
		// Operatoren ermitteln
				for(int i=0; i<operators.length; ++i) {
					try {
						constOperators.add(Float.parseFloat(operators[i])); // Vielleicht eine Konstante?
						dynOperators.add(null);
					} catch (NumberFormatException nfe) {
						constOperators.add(null); // keine Konstante
						
						PropertyNumericDAO p = PropertyNumericDAO.createPropertyNumericDAO(operators[i], true); // OK, dann muss es eine Property sein...
						dynOperators.add(p);
						this.addObservedProperty(p); // berwachen
					}
				}
				
				pnResult = PropertyNumericDAO.createPropertyNumericDAO(result, true);
	}
	
	@Override
	protected void refresh(PropertyNumericDAO property, float[] newValues, float[] oldValues) {
		float[] operators = new float[getNumberOfOperators()];
		for(int i=0; i<operators.length; ++i) {
			operators[i] = constOperators.get(i) != null ? constOperators.get(i) : (dynOperators.get(i).getFloat());
		}
		pnResult.setFloat(this.calculate(operators));
	}
	
	public abstract float calculate(float[] operators);
	public abstract int getNumberOfOperators();
}
