package ch.thomson.ucs.desy.specialfunctions;

/**
 * Diese Klasse implementiert die Rechenoperation result = factor * operator1 / operator2 als Specialfunction.
 * @author FischerU
 *
 */
public class Division extends Calculation {
	
	private float factor = 1.0f;
	
	/**
	 * Konstruiert ein Division Objekt. Das Objekt fhrt folgende Operation aus,
	 * wenn sich einer der Operatoren ndert: 
	 * <b>operators[0] / operators[1] = resul</b>. 
	 * Das Resultat wird in die entsprechende Property geschrieben.
	 * 
	 * @param operators
	 *            Vollspezifizierter TINE Pfad fr jeden einelnen Operator
	 * @param result
	 *            Vollspezifizierter TINE Pfad
	 * @throws Exception
	 *              Wenn eine Property nicht gefunden werden kann oder die Anzahl der Operatoren nicht stimmt.
	 */
	public Division(String[] operators, String result) throws Exception {
		super(operators, result);
	}
	
	/**
	 * Konstruiert ein Division Objekt. Das Objekt fhrt folgende Operation aus,
	 * wenn sich einer der Operatoren ndert: 
	 * <b>factor * operators[0] / operators[1] = result</b>. 
	 * Das Resultat wird in die entsprechende Property geschrieben.
	 * 
	 * @param factor
	 *            Eine Float-Konstante als Faktor
	 * @param operators
	 *            Vollspezifizierter TINE Pfad fr jeden einzelnen Operator
	 * @param result
	 *            Vollspezifizierter TINE Pfad
	 * @throws Exception
	 *             Wenn eine Property nicht gefunden werden kann oder die Anzahl der Operatoren nicht stimmt.
	 */
	public Division(float factor, String[] operators, String result) throws Exception {
		this(operators, result);
		this.factor = factor;
	}

	@Override
	public float calculate(float[] operators) {
		return factor * operators[0] / operators[1];
	}

	@Override
	public int getNumberOfOperators() {
		return 2;
	}
}
