package ch.thomson.ucs.desy.specialfunctions;

import ch.thomson.ucs.tine.server.internal.client.PropertyNumericDAO;


public class IIRFilter extends SpecialFunctionPropertyObserved {

	private int[] a = new int[3];
	private int[] b = new int[3];
	private PropertyNumericDAO[] coeffPropertiesA = new PropertyNumericDAO[3];
	private PropertyNumericDAO[] coeffPropertiesB = new PropertyNumericDAO[3];
	private PropertyNumericDAO pn_d;
	private PropertyNumericDAO pn_t0;
	private PropertyNumericDAO pn_t1;
	
	public IIRFilter(String[] coeffsA, String[] coeffsB, String d, String t0, String t1) throws Exception {
		if(coeffsA.length != 3 || coeffsB.length != 3) 
			throw new Exception("IIRFilter: Incorrect Coefficient Array Length");
		
		for(int i=0; i<coeffsA.length; ++i)
			this.coeffPropertiesA[i] = PropertyNumericDAO.createPropertyNumericDAO(coeffsA[i], true);
		
		for(int i=0; i<coeffsB.length; ++i)
			this.coeffPropertiesB[i] = PropertyNumericDAO.createPropertyNumericDAO(coeffsB[i], true);
		
		
		this.pn_d = PropertyNumericDAO.createPropertyNumericDAO(d, true);
		this.addObservedProperty(pn_d);
		
		this.pn_t0 = PropertyNumericDAO.createPropertyNumericDAO(t0, true);
		this.addObservedProperty(pn_t0);
		
		this.pn_t1 = PropertyNumericDAO.createPropertyNumericDAO(t1, true);
		this.addObservedProperty(pn_t1);
	}
			
	@Override
	protected void refresh(final PropertyNumericDAO property, float[] newValues, final float[] oldValues) {
		try {
			this.calculate();
			this.updateCoefficients();
		} catch (Exception e) {
			// Das Rcksetzen auf den alten Wert muss in einem eigenen Thread ablaufen.
			new Thread(new Runnable() {
				@Override
				public void run() {
					// Berechnungsregel verletzt, alten Wert wieder setzen
					property.setFloat(oldValues[0]);
				}}).start();
		}
	}
	
	private void updateCoefficients() {
		// Den ersten Wert benutzen um die Transaction zu beginnen.
		coeffPropertiesA[0].startTransaction();
		for(int i=0; i<coeffPropertiesA.length; ++i) {
			coeffPropertiesA[i].setInt(a[i]);
		}
		
		for(int i=0; i<coeffPropertiesB.length; ++i) {
			coeffPropertiesB[i].setInt(b[i]);
		}
		// Den ersten Wert bentzen um die Transaction abzuschliessen.
		try {
			coeffPropertiesA[0].commit();
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}
	
	private void calculate() throws Exception {
		// Aktuelle Werte einlesen
		float d = pn_d.getFloat();
		float t0 = pn_t0.getFloat();
		float t1 = pn_t1.getFloat();
		
		// Skalierung fr Flaoting Point / Skalierung 16 Bit 1.18
		int scale = (int) Math.pow(2, 18);
	 
		// Filter A
		float[] a_f = new float[3];
		a_f[0] = 1 + 4*d*t0 + 4*t0*t0;
		a_f[1] = (2 - 8*t0*t0)/a_f[0];
		a_f[2] = (1 - 4*d*t0 + 4*t0*t0)/a_f[0];

		// Filter B
		float[] b_f = new float[3];
		b_f[0] = (1 + 2 *t1)/a_f[0];
		b_f[1] = 2/a_f[0];
		b_f[2] = (1-2*t1)/a_f[0];
	 
		a_f[0] = 1;

		a[0] = (int) Math.floor(scale+0.5f);
		a[1] = (int) Math.floor(a_f[1]*scale+0.5f);
		a[2] = (int) Math.floor(a_f[2]*scale+0.5f);
	 
		b[0] = (int) Math.floor(b_f[0]*scale+0.5f);
		b[1] = (int) Math.floor(b_f[1]*scale+0.5f);
		b[2] = (int) Math.floor(b_f[2]*scale+0.5f);
	 
	 
		// Gainkompensation
		int suma = a[0] + a[1] + a[2];
		int sumb = b[0] + b[1] + b[2];

		if((sumb-suma) > 3)
			throw new Exception("Calculate IIR Filter failed, delta = " + (sumb-suma));

		// Im Falle von falschem Gain a(1) entsprechend modifizieren
		a[1] = a[1] + (sumb-suma);
	}
}
