package ch.thomson.ucs.desy.specialfunctions;

import java.io.BufferedReader;
import java.io.InputStream;
import java.io.InputStreamReader;

import ch.thomson.ucs.tine.server.internal.client.PropertyNumericDAO;

/**
 * Diese Klasse beobachtet die SelectionProperty und liest bei einer nderung
 * die neue Sequenz mit dem entsprechenden Index aus dem Sequences.csv File aus.
 * 
 * @author FischerU
 * 
 */
public class PSMSequenceHandler extends SpecialFunctionPropertyObserved {
	
	protected PropertyNumericDAO pnSequence;
	
	public PSMSequenceHandler(String selectionPropertyPath, String sequencePropertyPath) throws Exception {
		PropertyNumericDAO pnSelection = PropertyNumericDAO.createPropertyNumericDAO(selectionPropertyPath, true);
		this.addObservedProperty(pnSelection);
		
		pnSequence = PropertyNumericDAO.createPropertyNumericDAO(sequencePropertyPath, true);
		
		// Sequenz muss mit aktueller Selection initialisiert werden 
		pnSequence.setValuesAsFloat(getPSMSequenceValues(pnSelection.getInt()));
	}
	
	@Override
	protected void refresh(PropertyNumericDAO property, float[] newValues, float[] oldValues) {
		try {
			pnSequence.setValuesAsFloat(this.getPSMSequenceValues((int) newValues[0]));
		} catch (Exception e) {
			e.printStackTrace();
			property.setInt((int) oldValues[0]); // Alten Wert wieder setzen
		}
	}
	
	private float[] getPSMSequenceValues(int index) throws Exception {
		InputStream is = this.getClass().getResourceAsStream("/Sequences.csv");
		BufferedReader br = new BufferedReader(new InputStreamReader(is));
		String s;
		int i = 0;
		while((s = br.readLine()) != null) {
			// Kommentarzeilen beginnen mit # und werden ignoriert
			if(s.charAt(0) != '#') {
				if(i == index) {
					break; // Index gefunden, Schlaufe abbrechen
				}
				i++;
			}
		}
		
		br.close();
		
		if(s == null)
			throw new Exception("PSM Sequence #" + index + " not found");
		
		String[] sValues =  s.split("[,;]");
		
		float[] values = new float[sValues.length];
		for(int j=0; j<sValues.length; ++j)
			values[j] = Float.parseFloat(sValues[j]);
			
		return values;
	}
}
