package ch.thomson.ucs.desy.specialfunctions;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import ch.thomson.ucs.tine.server.internal.client.PropertyNumericChangeListener;
import ch.thomson.ucs.tine.server.internal.client.PropertyNumericDAO;
import ch.thomson.ucs.tine.server.internal.client.PropertyNumericNotification;

public abstract class SpecialFunctionPropertyObserved implements SpecialFunction, PropertyNumericChangeListener {
	
	private List<PropertyNumericDAO> observedProperties = new LinkedList<PropertyNumericDAO>();
	
	@Override
	public void notifyChanged(PropertyNumericNotification msg) {
		PropertyNumericDAO property = msg.getNotifier();
		float[] oldValues = msg.getOldValuesFloat();
		float[] newValues = msg.getNewValuesFloat();
		this.refresh(property, newValues, oldValues);
	}
	
	public void addObservedProperty(PropertyNumericDAO property) {
		property.addChangeListener(this);
		this.observedProperties.add(property);
	}
	
	public PropertyNumericDAO addObservedProperty(String path) throws Exception {
		PropertyNumericDAO property = PropertyNumericDAO.createPropertyNumericDAO(path, true);
		this.addObservedProperty(property);
		return property;
	}
	
	public void removeObservedProperty(PropertyNumericDAO property) {
		property.removeChangeListener(this);
		this.observedProperties.remove(property);
	}
	
	public void removeObservedProperty(String path) throws Exception {
		PropertyNumericDAO property = PropertyNumericDAO.createPropertyNumericDAO(path, true);
		this.removeObservedProperty(property);
	}
	
	@Override
	public void dispose() {
		// Alle Adapter vom EMF-Model abmelden
		for (Iterator<PropertyNumericDAO> it = observedProperties.iterator(); it.hasNext(); ) { 	
			PropertyNumericDAO pn = it.next();
			pn.removeChangeListener(this);
			it.remove();
		}
	}
	
	protected abstract void refresh(PropertyNumericDAO property, float[] newValues, float[] oldValues);
}
