package ch.thomson.ucs.desy.specialfunctions.fastchannel;

import ch.thomson.ucs.desy.specialfunctions.fastchannel.FastChannelCalculationSet.Result;
import ch.thomson.ucs.desy.specialfunctions.fastchannel.FastChannelCalculationSet.Sample;
import ch.thomson.ucs.desy.specialfunctions.fastchannel.FastChannelCalculationSet.Selection;


public class PulseEnergyCalculation extends FastChannelCalculation {

	protected Sample pnSamplesVolt;
	protected Sample pnSamplesCurrent;
	protected Selection pnPulseRepRate;
	protected Result pnResultEnergyPulse;
	protected Result pnResultPowerOut;
	
	public PulseEnergyCalculation(FastChannelCalculationSet parent, String samplesVolt, String samplesCurrent, String pulseRepRate, String resultEnergyPulse, String resultPowerOut) throws Exception {
		super(parent);
		pnSamplesVolt = this.findSample(samplesVolt);
		pnSamplesCurrent = this.findSample(samplesCurrent);
		pnPulseRepRate = this.findSelection(pulseRepRate);
		pnResultEnergyPulse = this.createResult(resultEnergyPulse);
		pnResultPowerOut = this.createResult(resultPowerOut);
	}

	@Override
	public void refresh() {
		float energy = 0.0f;
		float[] samplesVolt = pnSamplesVolt.getCachedValues();
		float[] samplesCurrent = pnSamplesCurrent.getCachedValues();
		
		for(int i=0; i < samplesVolt.length && i < samplesCurrent.length; ++i) 
			energy += samplesVolt[i] /*[kV]*/ * samplesCurrent[i] * 1e-6f;
		
		pnResultEnergyPulse.setResult(energy);
		pnResultPowerOut.setResult(energy * pnPulseRepRate.getCachedValueFloat());
	}
}
