package ch.thomson.ucs.desy.specialfunctions.fastchannel;

import ch.thomson.ucs.desy.specialfunctions.fastchannel.FastChannelCalculationSet.Result;
import ch.thomson.ucs.desy.specialfunctions.fastchannel.FastChannelCalculationSet.Sample;
import ch.thomson.ucs.desy.specialfunctions.fastchannel.FastChannelCalculationSet.Selection;



public class PulsePowerCalculations extends FastChannelCalculation {
	private Selection pnStart;
	private Selection pnLength;
	private Sample pnSamplesVolt;
	private Sample pnSamplesCurrent;
	private Result pnResultAvg;
	private Result pnResultPulse;
	private float[] powerPulse;
	
	public PulsePowerCalculations(FastChannelCalculationSet parent, String start, String length, String samplesVolt, String samplesCurrent, String resultAvg, String resultPulse) throws Exception {
		super(parent);
		pnStart = this.findSelection(start);
		pnLength = this.findSelection(length);
		pnSamplesVolt = this.findSample(samplesVolt);
		pnSamplesCurrent = this.findSample(samplesCurrent);
		pnResultAvg = this.createResult(resultAvg);
		pnResultPulse = this.createResult(resultPulse);
		powerPulse = new float[pnResultPulse.getLength()];
	}
	
	@Override
	public void refresh() {
		
		//long time = System.nanoTime();
		
		float average = 0.0f;
		float[] samplesVolt = pnSamplesVolt.getCachedValues();
		float[] samplesCurrent = pnSamplesCurrent.getCachedValues();
		int start = pnStart.getCachedValueInt();
		int length = pnLength.getCachedValueInt();
		
		
		for(int i=0; i < samplesVolt.length; ++i) {
			powerPulse[i] = samplesVolt[i]/*[kV]*/ * samplesCurrent[i] / 1000.0f; // Resultat in MW
			
			// Aufsummierung fr Average-Berechnung
			if(i >= start && i < (start + length))
				average += powerPulse[i];
		}
		//System.out.println("PowerCalc: " + (((System.nanoTime() - time) / 1000) / 1000.0));
		
		pnResultPulse.setResult(powerPulse);
		//System.out.println("PowerCalc: " + (((System.nanoTime() - time) / 1000) / 1000.0));
		
		
		// Average-Berechnung
		average = average / length;	
		//long time = System.nanoTime();
		pnResultAvg.setResult(average);
		
		
		
		//System.out.println(((System.nanoTime() - time) / 1000) / 1000.0);
		
	}
}
