package ch.thomson.ucs.ftdi;

import java.nio.ByteBuffer;

public class Firmware {

	// 
	private int id = 0;					// adr=0x00 len=4
	private int version = 0;			// adr=0x04 len=4
	private int dateHdl = 0;			// adr=0x08 len=4
	private int dateSw = 0;				// adr=0x0C len=4	
	private int interfaceVer = 0;		// adr=0x10 len=4	
	private int boardNo = 0;			// adr=0x14 len=4	
	private int boardStatus = 0;		// adr=0x18 len=4	
	private int crc = 0;				// adr=0x1C len=4	
	private int progStatus = 0;			// adr=0x20 len=4	

	private byte[] firmwareBuffer = null;	// byte array for Firmware buffer

	
	/**
	 * Creates a new <code>Firmware</code> form FPGA read buffer. 
	 * 
	 */
	public Firmware(byte[] buffer) throws IllegalArgumentException {
		firmwareBuffer = buffer;
		parseBuffer();
	}
	
	/**
	 * parse the eeprom buffer. 
	 * 
	 */
	private void parseBuffer() throws IllegalArgumentException {
		
		if(firmwareBuffer == null)
			throw new IllegalArgumentException("The buffer is empty.");

		if (firmwareBuffer.length != 36)
			throw new IllegalArgumentException("The buffer is not 36 byte long.");

		ByteBuffer bb = ByteBuffer.wrap(firmwareBuffer);
	
		id = bb.getInt();					// adr=0x00 len=4
		version = bb.getInt();				// adr=0x04 len=4
		dateHdl = bb.getInt();				// adr=0x08 len=4
		dateSw = bb.getInt();				// adr=0x0C len=4	
		interfaceVer = bb.getInt();			// adr=0x10 len=4	
		boardNo = bb.getInt();				// adr=0x14 len=4	
		boardStatus = bb.getInt();			// adr=0x18 len=4	
		crc = bb.getInt();					// adr=0x1C len=4	
		progStatus = bb.getInt();			// adr=0x20 len=4	
	}	
	
	/**
	 * @return the id
	 */
	public int getId() {
		return id;
	}

	/**
	 * @return the version
	 */
	public int getVersion() {
		return version;
	}

	/**
	 * @return the dateHdl
	 */
	public int getDateHdl() {
		return dateHdl;
	}

	/**
	 * @return the dateSw
	 */
	public int getDateSw() {
		return dateSw;
	}

	/**
	 * @return the interfaceVer
	 */
	public int getInterfaceVer() {
		return interfaceVer;
	}

	/**
	 * @return the boardNo
	 */
	public int getBoardNo() {
		return boardNo;
	}

	/**
	 * @return the boardStatus
	 */
	public int getBoardStatus() {
		return boardStatus;
	}

	/**
	 * @return the crc
	 */
	public int getCrc() {
		return crc;
	}

	/**
	 * @return the progStatus
	 */
	public int getProgStatus() {
		return progStatus;
	}

	/**
	 * @return the firmwareBuffer
	 */
	public byte[] getFirmwareBuffer() {
		return firmwareBuffer;
	}

	
}
