package ch.thomson.ucs.security.dao.impl;

import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.Collection;

import ch.thomson.ucs.security.dao.RoleDAO;
import ch.thomson.ucs.security.dao.entries.Role;

public class RoleDAOImpl extends BasicDAOImpl implements RoleDAO {
	
	private int update(Role role) throws SQLException {
		String update = "UPDATE roles SET roleName=?, logoffTime=? WHERE logonLevel=?;";
		PreparedStatement ps = connection.prepareStatement(update);
		ps.setString(1, role.getName());
		ps.setInt(2, role.getLogoffTime());
		ps.setInt(3, role.getLogonLevel());
		
		int result = 0;
		
		try {
			result = ps.executeUpdate();
		} finally {
			if(ps != null)
				ps.close();
		}
		
		return result;
	}
	
	private void insert(Role role) throws SQLException {
		String insert = "INSERT INTO roles VALUES (?, ?, ?);";
		PreparedStatement ps = connection.prepareStatement(insert);
		ps.setInt(1, role.getLogonLevel());
		ps.setString(2, role.getName());
		ps.setInt(3, role.getLogoffTime());

		try {
			ps.executeUpdate();
		} finally {
			if(ps != null)
				ps.close();
		}
	}

	@Override
	public void addEntry(Role role) throws SQLException {
		if(update(role) == 0)
			insert(role);
	}

	@Override
	public void deleteEntry(int logonLevel) throws SQLException {
		String delete = "DELETE FROM roles WHERE logonLevel=?;";
		PreparedStatement ps = connection.prepareStatement(delete);
		ps.setInt(1, logonLevel);
		try {
			ps.executeUpdate();
		}
		finally {
			ps.close();
		}
	}

	@Override
	public Collection<Role> getAllEntries() throws SQLException {
		Collection<Role> result = new java.util.LinkedList<Role>();
		
		Statement query = connection.createStatement();
		ResultSet rs = null;
		
		try {
			rs = query.executeQuery("SELECT * FROM roles ORDER BY logonLevel ASC;");
		
			while (rs.next()) {
				Role role = new Role();				
				role.setLogonLevel(rs.getInt("logonLevel"));
				role.setName(rs.getString("roleName"));
				role.setLogoffTime(rs.getInt("logoffTime"));
				result.add(role); 	
			}
		} finally {
			if(rs != null)
				rs.close();
			query.close();
		}
		
		return result;
	}

	@Override
	public void setupTable() throws SQLException {
		StringBuilder query = new StringBuilder();
		query.append("CREATE TABLE IF NOT EXISTS roles (");
		query.append("logonLevel INTEGER NOT NULL, ");
		query.append("roleName VARCHAR(32) NOT NULL, ");
		query.append("logoffTime INTEGER NOT NULL, ");
		query.append("PRIMARY KEY (logonLevel));");
		
		Statement st = connection.createStatement();
		try {
			st.executeUpdate(query.toString());
		} finally {
			st.close();
		}
		
		Collection<Role> roles = this.getAllEntries();
		if(roles.size() < 5) {
			// Standardrollen definieren
			Role role = new Role();
			
			role.setName("Everyone/Remote");
			role.setLogonLevel(0);
			role.setLogoffTime(0);
			this.addEntry(role);
			
			role.setName("Read Only Remote");
			role.setLogonLevel(1);
			role.setLogoffTime(0);
			this.addEntry(role);
			
			role.setName("Operator/Local");
			role.setLogonLevel(2);
			role.setLogoffTime(0);
			this.addEntry(role);
			
			role.setName("Administrator");
			role.setLogonLevel(3);
			role.setLogoffTime(0);
			this.addEntry(role);
			
			role.setName("Factory/IBS");
			role.setLogonLevel(9);
			role.setLogoffTime(0);
			this.addEntry(role);			
		}
	}

	@Override
	public Role getRole(int logonLevel) throws SQLException {
		Role role = null;
		String query = "SELECT * FROM roles WHERE logonLevel = ?;";
		PreparedStatement ps = connection.prepareStatement(query);
		ps.setInt(1, logonLevel);
		
		ResultSet rs = null;
		try {
			rs = ps.executeQuery();
			
			if(rs.next()) {
				role = new Role();
				role.setName(rs.getString("roleName"));
				role.setLogonLevel(rs.getInt("logonLevel"));
				role.setLogoffTime(rs.getInt("logoffTime"));
			}
			
		} finally {
			if(ps != null)
				ps.close();
		}
		
		return role;
	}
}
