package ch.thomson.ucs.security.dao.impl;

import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.Collection;

import ch.thomson.ucs.security.dao.UserDAO;
import ch.thomson.ucs.security.dao.entries.Role;
import ch.thomson.ucs.security.dao.entries.User;

public class UserDAOImpl extends BasicDAOImpl implements UserDAO {

	private int update(User user) throws SQLException {
		String update = "UPDATE users SET fullName=?, password=?, logonLevel=? WHERE userName=?;";
		PreparedStatement ps = connection.prepareStatement(update);
		ps.setString(1, user.getFullName());
		ps.setString(2, user.getPassword());
		ps.setInt(3, user.getRole().getLogonLevel());
		ps.setString(4, user.getName());
		
		int result = 0;		
		try {
			result = ps.executeUpdate();
		} finally {
			if(ps != null)
				ps.close();
		}
		return result;
	}
	
	private void insert(User user) throws SQLException {
		String insert = "INSERT INTO users VALUES (?, ?, ?, ?);";
		PreparedStatement ps = connection.prepareStatement(insert);
		ps.setString(1, user.getName());
		ps.setString(2, user.getFullName());
		ps.setString(3, user.getPassword());
		ps.setInt(4, user.getRole().getLogonLevel());
		
		try {
			ps.executeUpdate();
		} finally {
			if(ps != null)
				ps.close();
		}
	}
	
	@Override
	public void addUser(User user) throws SQLException {
		if(update(user) == 0)
			insert(user);
	}

	@Override
	public void deleteUser(String name) throws SQLException {
		String delete = "DELETE FROM users WHERE userName=?;";
		PreparedStatement ps = connection.prepareStatement(delete);
		ps.setString(1, name);
		try {
			ps.executeUpdate();
		}
		finally {
			if(ps != null)
				ps.close();
		}
	}

	@Override
	public Collection<User> getAllEntries() throws SQLException {
		
		Collection<User> result = new java.util.LinkedList<User>();
		
		Statement query = connection.createStatement();
		ResultSet rs = null;
		
		try {
			rs = query.executeQuery("SELECT * FROM users AS u, roles AS r WHERE u.logonLevel = r.logonLevel ORDER BY userName ASC;");
		
			while (rs.next()) {
				User user = new User();				
				user.setName(rs.getString("userName"));
				user.setFullName(rs.getString("fullName"));
				user.setPassword(rs.getString("password"));
				
				Role role = new Role();
				role.setLogonLevel(rs.getInt("logonLevel"));
				role.setName(rs.getString("roleName"));
				role.setLogoffTime(rs.getInt("logoffTime"));
				
				user.setRole(role);
							
				result.add(user); 	
			}
		} finally {
			if(rs != null)
				rs.close();
			query.close();
		}
		
		return result;
	}

	@Override
	public void setupTable() throws SQLException {
		StringBuilder query = new StringBuilder();
		query.append("CREATE TABLE IF NOT EXISTS users (");
		query.append("userName VARCHAR(32) NOT NULL, ");
		query.append("fullName VARCHAR(64) NOT NULL, ");
		query.append("password VARCHAR(64) NOT NULL, ");
		query.append("logonLevel INTEGER NOT NULL, ");
		query.append("PRIMARY KEY (userName), ");
		query.append("FOREIGN KEY (logonLevel) REFERENCES roles);");
		
		Statement st = connection.createStatement();
		try {
			st.executeUpdate(query.toString());
		} finally {
			st.close();
		}
		
		// Standardbenutzer anlegen
		User user = this.getUser("factory");
		if(user == null) {
			user = new User();
			user.setName("factory");
			user.setFullName("Factory");
			user.setPasswordUnencrypted("5300tu14");
			Role role = new Role();
			role.setLogonLevel(9);
			user.setRole(role);
			this.addUser(user);
		}
		
		user = this.getUser("admin");
		if(user == null) {
			user = new User();
			user.setName("admin");
			user.setFullName("Administrator");
			user.setPasswordUnencrypted("secret2");
			Role role = new Role();
			role.setLogonLevel(2);
			user.setRole(role);
			this.addUser(user);
		}
	}

	@Override
	public User getUser(String name) throws SQLException {
		User user = null;
		String query = "SELECT * FROM users AS u, roles AS r WHERE u.logonLevel = r.logonLevel AND u.userName = ?;";
		PreparedStatement ps = connection.prepareStatement(query);
		ps.setString(1, name);
		
		ResultSet rs = null;
		try {
			rs = ps.executeQuery();
			
			if(rs.next()) {
				user = new User();
				user.setName(rs.getString("userName"));
				user.setFullName(rs.getString("fullName"));
				user.setPassword(rs.getString("password"));
				
				Role role = new Role();
				role.setLogonLevel(rs.getInt("logonLevel"));
				role.setName(rs.getString("roleName"));
				role.setLogoffTime(rs.getInt("logoffTime"));
				
				user.setRole(role);
			}
			
		} finally {
			if(ps != null)
				ps.close();
		}
		
		return user;
	}

	@Override
	public void createUser(User user) throws SQLException {
		insert(user);
	}
}
