package ch.thomson.ucs.security.ui;

import org.eclipse.core.databinding.DataBindingContext;
import org.eclipse.core.databinding.beans.BeansObservables;
import org.eclipse.core.databinding.observable.Realm;
import org.eclipse.core.databinding.observable.list.IObservableList;
import org.eclipse.core.databinding.observable.map.IObservableMap;
import org.eclipse.core.databinding.observable.value.IObservableValue;
import org.eclipse.jface.databinding.swt.SWTObservables;
import org.eclipse.jface.databinding.viewers.ObservableListContentProvider;
import org.eclipse.jface.databinding.viewers.ObservableMapLabelProvider;
import org.eclipse.jface.dialogs.Dialog;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.jface.viewers.TableViewer;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.DisposeEvent;
import org.eclipse.swt.events.DisposeListener;
import org.eclipse.swt.events.KeyAdapter;
import org.eclipse.swt.events.KeyEvent;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.layout.FormAttachment;
import org.eclipse.swt.layout.FormData;
import org.eclipse.swt.layout.FormLayout;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Table;
import org.eclipse.swt.widgets.TableColumn;
import org.eclipse.swt.widgets.Text;
import org.eclipse.ui.ISharedImages;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.part.ViewPart;

import ch.thomson.ucs.security.LogonManager;
import ch.thomson.ucs.security.dao.entries.User;
import ch.thomson.ucs.security.ui.dialogs.ChangePassword;
import ch.thomson.ucs.security.ui.dialogs.CreateUser;
import org.eclipse.jface.viewers.ISelectionChangedListener;
import org.eclipse.jface.viewers.SelectionChangedEvent;
import org.eclipse.wb.swt.SWTResourceManager;

public class UserManagement extends ViewPart {
	private DataBindingContext m_bindingContext;

	public static final String ID = "ch.thomson.ucs.security.ui.UserManagement"; //$NON-NLS-1$
	private Text txtUser;
	private Text txtPassword;
	private LogonManager logonManager;
	private Label lblDisplayActualUser;
	private Label lblDisplayActualFullName;
	private Label lblActualRole;
	private Table table;
	private TableViewer tableViewer;
	private Button btnCreateUser;
	private Button btnDeleteUser;
	private Button btnChangePassword;

	public UserManagement() {
		logonManager = LogonManager.getInstance();
		logonManager.setHostDB("localhost");
	}
	
	private void login() {
		try {
			logonManager.logOn(txtUser.getText(), txtPassword.getText());
			txtPassword.setText("");
		} catch (Exception e1) {
			MessageDialog.openError(null, "Error", e1.getMessage());
		}
	}

	/**
	 * Create contents of the view part.
	 * @param parent
	 */
	@Override
	public void createPartControl(Composite parent) {
		// set Folder icon after loading like desy ui
		this.setTitleImage(PlatformUI.getWorkbench().getSharedImages()
				.getImage(ISharedImages.IMG_OBJ_FOLDER));
		
		parent.addDisposeListener(new DisposeListener() {
			public void widgetDisposed(DisposeEvent e) {
				m_bindingContext.dispose();
			}
		});
		
		Composite cmpMain = new Composite(parent, SWT.BORDER);
		cmpMain.setLayout(new GridLayout(5, false));
		
		Composite cmpUserLogon = new Composite(cmpMain, SWT.NONE);
		GridLayout gl_cmpUserLogon = new GridLayout(2, false);
		gl_cmpUserLogon.verticalSpacing = 10;
		gl_cmpUserLogon.horizontalSpacing = 10;
		gl_cmpUserLogon.marginHeight = 10;
		gl_cmpUserLogon.marginWidth = 10;
		gl_cmpUserLogon.marginTop = 10;
		gl_cmpUserLogon.marginLeft = 10;
		cmpUserLogon.setLayout(gl_cmpUserLogon);
		cmpUserLogon.setLayoutData(new GridData(SWT.FILL, SWT.FILL, false, true, 1, 1));
		
		Label lblNewLabel = new Label(cmpMain, SWT.BORDER);
		GridData gd_lblNewLabel = new GridData(SWT.LEFT, SWT.CENTER, false, false, 1, 1);
		gd_lblNewLabel.minimumWidth = 1;
		gd_lblNewLabel.heightHint = 160;
		gd_lblNewLabel.widthHint = 1;
		lblNewLabel.setLayoutData(gd_lblNewLabel);
		
		Composite cmpUser = new Composite(cmpMain, SWT.NONE);
		GridLayout gl_cmpUser = new GridLayout(3, false);
		gl_cmpUser.verticalSpacing = 10;
		gl_cmpUser.horizontalSpacing = 10;
		gl_cmpUser.marginHeight = 10;
		gl_cmpUser.marginWidth = 10;
		gl_cmpUser.marginTop = 10;
		gl_cmpUser.marginLeft = 10;
		cmpUser.setLayout(gl_cmpUser);
		cmpUser.setLayoutData(new GridData(SWT.FILL, SWT.FILL, false, false, 1, 1));
		
		Label lblUser = new Label(cmpUser, SWT.NONE);
		lblUser.setFont(SWTResourceManager.getFont("Tahoma", 10, SWT.NORMAL));
		lblUser.setText("User:");
		new Label(cmpUser, SWT.NONE);
		
		txtUser = new Text(cmpUser, SWT.BORDER);
		txtUser.setFont(SWTResourceManager.getFont("Tahoma", 10, SWT.BOLD));
		GridData gd_txtUser = new GridData(SWT.FILL, SWT.CENTER, true, false, 1, 1);
		gd_txtUser.widthHint = 153;
		txtUser.setLayoutData(gd_txtUser);
		
		Label lblPassword = new Label(cmpUser, SWT.NONE);
		lblPassword.setFont(SWTResourceManager.getFont("Tahoma", 10, SWT.NORMAL));
		lblPassword.setText("Password:");
		new Label(cmpUser, SWT.NONE);
		
		txtPassword = new Text(cmpUser, SWT.BORDER | SWT.PASSWORD);
		txtPassword.setFont(SWTResourceManager.getFont("Tahoma", 10, SWT.BOLD));
		GridData gd_txtPassword = new GridData(SWT.FILL, SWT.CENTER, true, false, 1, 1);
		gd_txtPassword.widthHint = 139;
		txtPassword.setLayoutData(gd_txtPassword);
		txtPassword.addKeyListener(new KeyAdapter() {

			@Override
			public void keyReleased(KeyEvent e) {
				if(e.keyCode == SWT.CR)
					login();
			}	
		});
		new Label(cmpUser, SWT.NONE);
		new Label(cmpUser, SWT.NONE);
		
		Button btnLogIn = new Button(cmpUser, SWT.NONE);
		GridData gd_btnLogIn = new GridData(SWT.LEFT, SWT.CENTER, false, false, 1, 1);
		gd_btnLogIn.widthHint = 111;
		btnLogIn.setLayoutData(gd_btnLogIn);
		btnLogIn.setFont(SWTResourceManager.getFont("Tahoma", 10, SWT.BOLD));
		btnLogIn.addSelectionListener(new SelectionAdapter() {
			@Override
			public void widgetSelected(SelectionEvent e) {
				login();
			}
		});
		btnLogIn.setText("Log In");
		
		Label lblActualUser = new Label(cmpUserLogon, SWT.NONE);
		lblActualUser.setFont(SWTResourceManager.getFont("Tahoma", 10, SWT.NORMAL));
		lblActualUser.setBounds(0, 0, 49, 13);
		lblActualUser.setText("Actual User:");
		
		lblDisplayActualUser = new Label(cmpUserLogon, SWT.NONE);
		lblDisplayActualUser.setFont(SWTResourceManager.getFont("Tahoma", 10, SWT.BOLD));
		GridData gd_lblDisplayActualUser = new GridData(SWT.FILL, SWT.CENTER, false, false, 1, 1);
		gd_lblDisplayActualUser.widthHint = 111;
		lblDisplayActualUser.setLayoutData(gd_lblDisplayActualUser);
		lblDisplayActualUser.setBounds(0, 0, 49, 13);
		lblDisplayActualUser.setText("None");
		
		Label lblFullName = new Label(cmpUserLogon, SWT.NONE);
		lblFullName.setFont(SWTResourceManager.getFont("Tahoma", 10, SWT.NORMAL));
		lblFullName.setText("Full Name:");
		
		lblDisplayActualFullName = new Label(cmpUserLogon, SWT.NONE);
		lblDisplayActualFullName.setFont(SWTResourceManager.getFont("Tahoma", 10, SWT.BOLD));
		lblDisplayActualFullName.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, false, false, 1, 1));
		lblDisplayActualFullName.setText("None");
		
		Label lblRole = new Label(cmpUserLogon, SWT.NONE);
		lblRole.setFont(SWTResourceManager.getFont("Tahoma", 10, SWT.NORMAL));
		lblRole.setText("Role:");
		
		lblActualRole = new Label(cmpUserLogon, SWT.NONE);
		lblActualRole.setFont(SWTResourceManager.getFont("Tahoma", 10, SWT.BOLD));
		lblActualRole.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, false, false, 1, 1));
		lblActualRole.setText("None");
		
		btnChangePassword = new Button(cmpUserLogon, SWT.NONE);
		btnChangePassword.setFont(SWTResourceManager.getFont("Tahoma", 10, SWT.BOLD));
		btnChangePassword.addSelectionListener(new SelectionAdapter() {
			@Override
			public void widgetSelected(SelectionEvent e) {
				ChangePassword dialog = new ChangePassword(null);
				dialog.open();
			}
		});
		btnChangePassword.setBounds(0, 0, 68, 23);
		btnChangePassword.setText("Change Password...");
		
		Button btnLogOff = new Button(cmpUserLogon, SWT.NONE);
		GridData gd_btnLogOff = new GridData(SWT.LEFT, SWT.CENTER, false, false, 1, 1);
		gd_btnLogOff.widthHint = 111;
		btnLogOff.setLayoutData(gd_btnLogOff);
		btnLogOff.setFont(SWTResourceManager.getFont("Tahoma", 10, SWT.BOLD));
		btnLogOff.addSelectionListener(new SelectionAdapter() {
			@Override
			public void widgetSelected(SelectionEvent e) {
				btnDeleteUser.setEnabled(false);
				logonManager.logOff();
			}
		});
		btnLogOff.setBounds(0, 0, 68, 23);
		btnLogOff.setText("Log Off");
		
		Label lblNewLabel_1 = new Label(cmpMain, SWT.BORDER);
		GridData gd_lblNewLabel_1 = new GridData(SWT.LEFT, SWT.CENTER, false, false, 1, 1);
		gd_lblNewLabel_1.heightHint = 160;
		gd_lblNewLabel_1.widthHint = 1;
		lblNewLabel_1.setLayoutData(gd_lblNewLabel_1);
		Composite cmpUserManagement = new Composite(cmpMain, SWT.NONE);
		cmpUserManagement.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, true, 1, 1));
		cmpUserManagement.setLayout(new FormLayout());
		
		tableViewer = new TableViewer(cmpUserManagement, SWT.BORDER | SWT.FULL_SELECTION);
		tableViewer.addSelectionChangedListener(new ISelectionChangedListener() {
			public void selectionChanged(SelectionChangedEvent event) {
				IStructuredSelection selection = (IStructuredSelection) event.getSelection();
				btnDeleteUser.setEnabled(!selection.isEmpty());
			}
		});
		table = tableViewer.getTable();
		table.setFont(SWTResourceManager.getFont("Tahoma", 10, SWT.NORMAL));
		table.setHeaderVisible(true);
		FormData fd_table = new FormData();
		fd_table.top = new FormAttachment(0);
		table.setLayoutData(fd_table);
		
		btnDeleteUser = new Button(cmpUserManagement, SWT.NONE);
		btnDeleteUser.setFont(SWTResourceManager.getFont("Tahoma", 10, SWT.BOLD));
		btnDeleteUser.setEnabled(false);
		btnDeleteUser.addSelectionListener(new SelectionAdapter() {
			@Override
			public void widgetSelected(SelectionEvent e) {
				IStructuredSelection selection = (IStructuredSelection) tableViewer.getSelection();
				if(selection.isEmpty())
					return;
				
				User user = (User) selection.getFirstElement();
				try {
					logonManager.deleteUser(user.getName());
				} catch (Exception e1) {
					MessageDialog.openError(null, "Error", "Delete user failed: " + e1.getMessage());
				}
			}
		});
		fd_table.right = new FormAttachment(btnDeleteUser, 0, SWT.RIGHT);
		fd_table.bottom = new FormAttachment(btnDeleteUser, -6);
		FormData fd_btnDeleteUser = new FormData();
		fd_btnDeleteUser.width = 111;
		fd_btnDeleteUser.top = new FormAttachment(100, -41);
		fd_btnDeleteUser.bottom = new FormAttachment(100, -18);
		fd_btnDeleteUser.right = new FormAttachment(100, -10);
		btnDeleteUser.setLayoutData(fd_btnDeleteUser);
		btnDeleteUser.setText("Delete User");
		
		btnCreateUser = new Button(cmpUserManagement, SWT.NONE);
		btnCreateUser.setFont(SWTResourceManager.getFont("Tahoma", 10, SWT.BOLD));
		fd_table.left = new FormAttachment(btnCreateUser, 0, SWT.LEFT);
		
		TableColumn tblclmnUser = new TableColumn(table, SWT.NONE);
		tblclmnUser.setWidth(100);
		tblclmnUser.setText("User");
		
		TableColumn tblclmnFullName = new TableColumn(table, SWT.NONE);
		tblclmnFullName.setWidth(100);
		tblclmnFullName.setText("Full Name");
		
		TableColumn tblclmnNewColumn = new TableColumn(table, SWT.NONE);
		tblclmnNewColumn.setWidth(100);
		tblclmnNewColumn.setText("Role");
		
		btnCreateUser.addSelectionListener(new SelectionAdapter() {
			@Override
			public void widgetSelected(SelectionEvent e) {
				Dialog dialog = new CreateUser(null);
				dialog.open();
			}
		});
		FormData fd_btnCreateUser = new FormData();
		fd_btnCreateUser.width = 111;
		fd_btnCreateUser.top = new FormAttachment(100, -43);
		fd_btnCreateUser.bottom = new FormAttachment(100, -20);
		fd_btnCreateUser.left = new FormAttachment(0, 10);
		btnCreateUser.setLayoutData(fd_btnCreateUser);
		btnCreateUser.setText("Create User...");
	
		createActions();
		initializeToolBar();
		initializeMenu();
		m_bindingContext = initDataBindings();
	}

	/**
	 * Create the actions.
	 */
	private void createActions() {
		// Create the actions
	}

	/**
	 * Initialize the toolbar.
	 */
	private void initializeToolBar() {
		getViewSite().getActionBars().getToolBarManager();
	}

	/**
	 * Initialize the menu.
	 */
	private void initializeMenu() {
		getViewSite().getActionBars().getMenuManager();
	}

	@Override
	public void setFocus() {
		// Set the focus
	}
	protected DataBindingContext initDataBindings() {
		DataBindingContext bindingContext = new DataBindingContext();
		//
		IObservableValue lblDisplayActualUserObserveTextObserveWidget = SWTObservables.observeText(lblDisplayActualUser);
		IObservableValue logonManagerLoggedInUsernameObserveValue = BeansObservables.observeValue(logonManager, "loggedInUser.name");
		bindingContext.bindValue(lblDisplayActualUserObserveTextObserveWidget, logonManagerLoggedInUsernameObserveValue, null, null);
		//
		IObservableValue lblDisplayActualFullNameObserveTextObserveWidget = SWTObservables.observeText(lblDisplayActualFullName);
		IObservableValue logonManagerLoggedInUserfullNameObserveValue = BeansObservables.observeValue(logonManager, "loggedInUser.fullName");
		bindingContext.bindValue(lblDisplayActualFullNameObserveTextObserveWidget, logonManagerLoggedInUserfullNameObserveValue, null, null);
		//
		IObservableValue lblActualRoleObserveTextObserveWidget = SWTObservables.observeText(lblActualRole);
		IObservableValue logonManagerLoggedInUserrolenameObserveValue = BeansObservables.observeValue(logonManager, "loggedInUser.role.name");
		bindingContext.bindValue(lblActualRoleObserveTextObserveWidget, logonManagerLoggedInUserrolenameObserveValue, null, null);
		//
		IObservableValue btnCreateUserObserveEnabledObserveWidget = SWTObservables.observeEnabled(btnCreateUser);
		IObservableValue logonManagerUserLoggedInObserveValue = BeansObservables.observeValue(logonManager, "userLoggedIn");
		bindingContext.bindValue(btnCreateUserObserveEnabledObserveWidget, logonManagerUserLoggedInObserveValue, null, null);
		//
		IObservableValue btnChangePasswordObserveEnabledObserveWidget = SWTObservables.observeEnabled(btnChangePassword);
		bindingContext.bindValue(btnChangePasswordObserveEnabledObserveWidget, logonManagerUserLoggedInObserveValue, null, null);
		//
		ObservableListContentProvider listContentProvider = new ObservableListContentProvider();
		tableViewer.setContentProvider(listContentProvider);
		//
		IObservableMap[] observeMaps = BeansObservables.observeMaps(listContentProvider.getKnownElements(), User.class, new String[]{"name", "fullName", "role.name"});
		tableViewer.setLabelProvider(new ObservableMapLabelProvider(observeMaps));
		//
		IObservableList logonManagerUsersTestObserveList = BeansObservables.observeList(Realm.getDefault(), logonManager, "users");
		tableViewer.setInput(logonManagerUsersTestObserveList);
		//
		return bindingContext;
	}
}
