package ch.thomson.ucs.security.ui.dialogs;

import org.eclipse.jface.dialogs.Dialog;
import org.eclipse.jface.dialogs.IDialogConstants;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.SWT;
import org.eclipse.swt.widgets.Text;
import org.eclipse.swt.layout.GridData;

import ch.thomson.ucs.security.LogonManager;

public class ChangePassword extends Dialog {
	private Text txtOldPassword;
	private Text txtNewPassword;
	private Text txtNewPasswordConfirm;

	/**
	 * Create the dialog.
	 * @param parentShell
	 */
	public ChangePassword(Shell parentShell) {
		super(parentShell);
	}

	/**
	 * Create contents of the dialog.
	 * @param parent
	 */
	@Override
	protected Control createDialogArea(Composite parent) {
		getShell().setText("Change Password");
		Composite container = (Composite) super.createDialogArea(parent);
		GridLayout gl_container = new GridLayout(2, false);
		gl_container.verticalSpacing = 10;
		gl_container.horizontalSpacing = 10;
		gl_container.marginHeight = 10;
		gl_container.marginWidth = 10;
		gl_container.marginTop = 10;
		gl_container.marginLeft = 10;
		container.setLayout(gl_container);
		
		Label lblOldPassword = new Label(container, SWT.NONE);
		lblOldPassword.setText("Old Password:");
		
		txtOldPassword = new Text(container, SWT.BORDER | SWT.PASSWORD);
		txtOldPassword.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, false, 1, 1));
		
		Label lblNewPassword = new Label(container, SWT.NONE);
		lblNewPassword.setText("New Password:");
		
		txtNewPassword = new Text(container, SWT.BORDER | SWT.PASSWORD);
		txtNewPassword.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, false, 1, 1));
		
		Label lblConfirmNewPassword = new Label(container, SWT.NONE);
		lblConfirmNewPassword.setText("Confirm New Password:");
		
		txtNewPasswordConfirm = new Text(container, SWT.BORDER | SWT.PASSWORD);
		txtNewPasswordConfirm.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, false, 1, 1));

		return container;
	}
	
	@Override
	protected void okPressed() {
		try {
			LogonManager logonManager = LogonManager.getInstance();
			if(!logonManager.isPasswordComplexityOk(this.txtNewPassword.getText()))
				throw new Exception(logonManager.getPasswordComplexityMessage());
			
			if(!this.txtNewPassword.getText().equals(this.txtNewPasswordConfirm.getText()))
				throw new Exception("The passwords do not match.");
			
			logonManager.changePassword(txtOldPassword.getText(), txtNewPassword.getText());
			super.okPressed();
		} 
		catch (Exception e) {
			MessageDialog.openError(null, "Error", e.getMessage());
		}
	}

	/**
	 * Create contents of the button bar.
	 * @param parent
	 */
	@Override
	protected void createButtonsForButtonBar(Composite parent) {
		createButton(parent, IDialogConstants.OK_ID, IDialogConstants.OK_LABEL,
				true);
		createButton(parent, IDialogConstants.CANCEL_ID,
				IDialogConstants.CANCEL_LABEL, false);
	}

	/**
	 * Return the initial size of the dialog.
	 */
	@Override
	protected Point getInitialSize() {
		return new Point(450, 300);
	}
}
