package ch.thomson.ucs.security.ui.dialogs;

import org.eclipse.core.databinding.DataBindingContext;
import org.eclipse.core.databinding.beans.BeansObservables;
import org.eclipse.core.databinding.observable.Realm;
import org.eclipse.core.databinding.observable.list.IObservableList;
import org.eclipse.core.databinding.observable.map.IObservableMap;
import org.eclipse.jface.databinding.viewers.ObservableListContentProvider;
import org.eclipse.jface.databinding.viewers.ObservableMapLabelProvider;
import org.eclipse.jface.dialogs.Dialog;
import org.eclipse.jface.dialogs.IDialogConstants;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.viewers.ComboViewer;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.swt.SWT;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Combo;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.Text;

import ch.thomson.ucs.security.LogonManager;
import ch.thomson.ucs.security.dao.entries.Role;
import ch.thomson.ucs.security.dao.entries.User;
import org.eclipse.swt.events.DisposeListener;
import org.eclipse.swt.events.DisposeEvent;

public class CreateUser extends Dialog {
	private DataBindingContext m_bindingContext;
	private Text txtName;
	private Text txtFullName;
	private Text txtPassword;
	private Text txtPasswordConfirm;
	private LogonManager logonManager;
	private ComboViewer comboViewer;

	/**
	 * Create the dialog.
	 * @param parentShell
	 */
	public CreateUser(Shell parentShell) {
		super(parentShell);
	}

	/**
	 * Create contents of the dialog.
	 * @param parent
	 */
	@Override
	protected Control createDialogArea(Composite parent) {
		parent.addDisposeListener(new DisposeListener() {
			public void widgetDisposed(DisposeEvent e) {
				m_bindingContext.dispose();
			}
		});
		logonManager = LogonManager.getInstance();

		this.getShell().setText("Create User");
		
		Composite container = (Composite) super.createDialogArea(parent);
		GridLayout gl_container = new GridLayout(2, false);
		gl_container.verticalSpacing = 10;
		gl_container.horizontalSpacing = 10;
		gl_container.marginHeight = 10;
		gl_container.marginWidth = 10;
		gl_container.marginTop = 10;
		gl_container.marginLeft = 10;
		container.setLayout(gl_container);
		
		Label lblUser = new Label(container, SWT.NONE);
		lblUser.setText("User:");
		
		txtName = new Text(container, SWT.BORDER);
		txtName.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, false, 1, 1));
		
		Label lblFullName = new Label(container, SWT.NONE);
		lblFullName.setText("Full Name:");
		
		txtFullName = new Text(container, SWT.BORDER);
		txtFullName.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, false, 1, 1));
		
		Label lblPassword = new Label(container, SWT.NONE);
		lblPassword.setText("Password:");
		
		txtPassword = new Text(container, SWT.BORDER | SWT.PASSWORD);
		txtPassword.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, false, 1, 1));
		
		Label lblConfirmPassword = new Label(container, SWT.NONE);
		lblConfirmPassword.setLayoutData(new GridData(SWT.RIGHT, SWT.CENTER, false, false, 1, 1));
		lblConfirmPassword.setText("Confirm Password:");
		
		txtPasswordConfirm = new Text(container, SWT.BORDER | SWT.PASSWORD);
		txtPasswordConfirm.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, false, 1, 1));
		
		Label lblRole = new Label(container, SWT.NONE);
		lblRole.setLayoutData(new GridData(SWT.RIGHT, SWT.CENTER, false, false, 1, 1));
		lblRole.setText("Role:");
		
		comboViewer = new ComboViewer(container, SWT.NONE);
		Combo cmbRole = comboViewer.getCombo();
		cmbRole.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, false, 1, 1));

		return container;
	}

/*	
	private void refreshRoleList() {
		try {
			logonManager.refreshRoles();
		} catch (Exception e1) {
			MessageDialog.openError(null, "Error", "Could not read role list: " + e1.getMessage());
		}
	}
*/
	/**
	 * Create contents of the button bar.
	 * @param parent
	 */
	@Override
	protected void createButtonsForButtonBar(Composite parent) {
		createButton(parent, IDialogConstants.OK_ID, IDialogConstants.OK_LABEL,
				true);
		createButton(parent, IDialogConstants.CANCEL_ID,
				IDialogConstants.CANCEL_LABEL, false);
		m_bindingContext = initDataBindings();
	}

	@Override
	protected void okPressed() {
		try {
			if(!logonManager.isPasswordComplexityOk(this.txtPassword.getText()))
				throw new Exception(logonManager.getPasswordComplexityMessage());
			
			if(!this.txtPassword.getText().equals(this.txtPasswordConfirm.getText()))
				throw new Exception("The passwords do not match.");
			
			User user = new User();
			user.setName(this.txtName.getText());
			user.setFullName(this.txtFullName.getText());
			user.setPasswordUnencrypted(this.txtPassword.getText());
			
			IStructuredSelection selection = (IStructuredSelection) this.comboViewer.getSelection();
			if(selection.isEmpty())
				throw new Exception ("No role is selected");
			user.setRole((Role) selection.getFirstElement());
			
			logonManager.createUser(user);
			super.okPressed();
		}
		
		catch(Exception e) {
			MessageDialog.openError(null, "Error", e.getMessage());
		}
	}

	/**
	 * Return the initial size of the dialog.
	 */
	@Override
	protected Point getInitialSize() {
		return new Point(450, 300);
	}
	protected DataBindingContext initDataBindings() {
		DataBindingContext bindingContext = new DataBindingContext();
		//
		ObservableListContentProvider listContentProvider = new ObservableListContentProvider();
		comboViewer.setContentProvider(listContentProvider);
		//
		IObservableMap observeMap = BeansObservables.observeMap(listContentProvider.getKnownElements(), Role.class, "name");
		comboViewer.setLabelProvider(new ObservableMapLabelProvider(observeMap));
		//
		IObservableList logonManagerRolesObserveList = BeansObservables.observeList(Realm.getDefault(), logonManager, "roles");
		comboViewer.setInput(logonManagerRolesObserveList);
		//
		return bindingContext;
	}
}
