/**
 * 
 */
package ch.thomson.swing.graphics.impl;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Paint;
import java.awt.Polygon;
import java.awt.RenderingHints;
import java.awt.Shape;
import java.awt.geom.Rectangle2D;
import java.util.Arrays;
import javax.swing.SwingConstants;
import javax.swing.SwingUtilities;

import ch.thomson.swing.graphics.IBarGraph;
import ch.thomson.swing.graphics.IScaleLabelProvider;
import ch.thomson.swing.graphics.IValueLabelProvider;


/**
 * <p>
 * Graphische Klasse um eine gerade Skala mit Bargraph zu Zeichnen wie sie auf
 * verschiedenen Zeigerinstrumente vorgefunden wird. Die Skala kann mit
 * verschieden Parameter konfiguriert werden welche die Anzahl Teilstriche,
 * Nebenteilstriche und die Position der Skalenbeschriftung einstellen.
 * </p>
 * 
 * @author Dominic Meier
 * @version $version$ ($date$)
 */
public class BarGraph extends GraphicsScaleRectangular implements IBarGraph {
	
	public enum Marker {WARNING_LOWER, WARNING_UPPER, FAULT_LOWER, FAULT_UPPER};
	
	private int gap = 0;
	private double value = 0;
	private String unit = "";
	private boolean filled = false;
	private boolean filledBackgroundBar = false;
	private boolean visibleMarker = false;
	private Color ColorBackgroundBar = Color.white;
	private boolean showBorder = false;
	private double PositionMarkers[] = new double[0];
	private IScaleLabelProvider[] labelprovider = null;
	private boolean showLabel = false;
	private Font font;
	private String[] MaxString = { "" };
	private boolean showMarkers[]= null;
	private int gapPosition;
	private IValueLabelProvider valueLabelProvider;
	
	private int scaleMarker = 7;

	public BarGraph() {
		super();
		Init();
	}

	public BarGraph(int w, int h) {
		super(w, h);
		Init();
	}

	private void Init() {
		gapPosition = Position;
	}

	@Override
	public void paintGraphics(Graphics g, int x, int y) {

		Graphics2D g2 = (Graphics2D) g;
		g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
				RenderingHints.VALUE_ANTIALIAS_ON);

		int MaxShiftLEFT = this.getHighestLabelStringWidth(g2) + (getWidth() / 10); //GetMaxShiftLEFT();
		int MaxShiftTOP = (g2.getFontMetrics(font).getAscent() * 2); //GetMaxShiftTOP(g2);
		
		int MaxShiftRIGHT = 0;// + MaxShiftLEFT;
		int MaxShiftBOTTOM = g2.getFontMetrics(font).getAscent() + MaxShiftTOP;

		// Das Rechteck mit dem Wert zeichnen
		// MH 05.04.12
//		paintValueRectangle(g2, x + MaxShiftLEFT, y + MaxShiftTOP, getHeight() - MaxShiftBOTTOM, getWidth() - MaxShiftRIGHT);
		paintValueRectangle(g2, x + MaxShiftLEFT, y + MaxShiftTOP - 4, getHeight() - MaxShiftBOTTOM, getWidth() - MaxShiftRIGHT);
		
		// Die Wertanzeige zeichnen
		// MH 05.04.12
//		paintValueString(g2, x + MaxShiftLEFT, y + MaxShiftTOP, getHeight() - MaxShiftBOTTOM, getWidth() - MaxShiftRIGHT);
		paintValueString(g2, x, y + MaxShiftTOP, getHeight() - MaxShiftBOTTOM, getWidth() - MaxShiftRIGHT);
		
		// Das obere Rechteck zeichnen
		if(isFilledBackgroundBar())
			// MH 05.04.12
//			paintBackgroundRectangle(g2, x + MaxShiftLEFT, y + MaxShiftTOP, 
//							getHeight() - MaxShiftBOTTOM, getWidth() - MaxShiftRIGHT);
			paintBackgroundRectangle(g2, x + MaxShiftLEFT, y + MaxShiftTOP - 4, 
							getHeight() - MaxShiftBOTTOM, getWidth() - MaxShiftRIGHT);
		
		// MH 05.04.12
//		paintUnitString(g2, x + MaxShiftLEFT, y + MaxShiftTOP / 2, getHeight()
//				- MaxShiftBOTTOM, getWidth() - MaxShiftRIGHT);
		paintUnitString(g2, x, y + MaxShiftTOP / 2, getHeight()
				- MaxShiftBOTTOM, getWidth() - MaxShiftRIGHT);
		
		if (isMarkerVisible()) {
			// MH 05.04.12
//			paintMarkers(g2, x + MaxShiftLEFT + this.getMaxTickLength()/2, y + MaxShiftTOP, getHeight()
//					- MaxShiftBOTTOM, getWidth() - MaxShiftRIGHT);
			paintMarkers(g2, x + MaxShiftLEFT + this.getMaxTickLength()/2, y + MaxShiftTOP - 4, getHeight()
					- MaxShiftBOTTOM, getWidth() - MaxShiftRIGHT);
		}
		
		if (isScaleLabelVisible()) {
			// MH 05.04.12
//			paintLabels(g2, x, y + MaxShiftTOP, getHeight() - MaxShiftBOTTOM, getWidth() - MaxShiftRIGHT);
			paintLabels(g2, x, y + MaxShiftTOP - 4, getHeight() - MaxShiftBOTTOM, getWidth() - MaxShiftRIGHT);
		}

		// MH 05.04.12
//		DrawScala(g2, x + MaxShiftLEFT, y + MaxShiftTOP, getHeight() - MaxShiftBOTTOM, getWidth() - MaxShiftRIGHT);
		DrawScala(g2, x + MaxShiftLEFT, y + MaxShiftTOP - 4, getHeight() - MaxShiftBOTTOM, getWidth() - MaxShiftRIGHT);

		if (isVisibleAxis()) {
			// Die Hauptlinie zeichnen
			// MH 05.04.12
//			DrawLine(g2, x + MaxShiftLEFT, y + MaxShiftTOP, getHeight()
//					- MaxShiftBOTTOM, getWidth() - MaxShiftRIGHT);
			DrawLine(g2, x + MaxShiftLEFT, y + MaxShiftTOP - 4, getHeight()
					- MaxShiftBOTTOM, getWidth() - MaxShiftRIGHT);
		}

	}
	
	private void paintUnitString(Graphics2D g2, int x, int y, int height, int width) {
		double sizeLine = 0;
		
		if (isShowBorder())
			sizeLine = SizeLine;
		
		Font oldFont = g2.getFont();
		g2.setFont(font);
		g2.drawString(unit, (float) (x + (sizeLine / 2.0)) + gap, (float)(y + (sizeLine / 2.0)));
		g2.setFont(oldFont);
	}
	
	private void paintValueString(Graphics2D g2, int x, int y, int height, int width) {
		if(valueLabelProvider == null)
			return;
		
		double sizeline = 0;
		
		if (isShowBorder()) {
			sizeline = SizeLine;
		}
		
		double x0 = x + (sizeline / 2.0) + gap;
		double y0 = y + (sizeline / 2.0);
		double y1 = y0 + height - (sizeline);
		
		Font oldFont = g2.getFont();
		g2.setFont(font);
		g2.drawString(valueLabelProvider.getText(), (float) x0, (float) y1 + g2.getFontMetrics().getMaxAscent());
		g2.setFont(oldFont);
	}
	
	private double valueRectangleWidth;

	private void paintValueRectangle(Graphics2D g2, int x, int y, int height,
			int width) {
		double x0, y0, x1, y1, sizeline;

		if (isShowBorder()) {
			sizeline = SizeLine;
		} else {
			sizeline = 0;
		}
		
		if (Orientation == SwingConstants.VERTICAL) {
			if (gapPosition == SwingConstants.LEFT) {
				x0 = x + (sizeline / 2.0) + gap;
				y0 = y + (sizeline / 2.0);
				x1 = x0 + width - (sizeline) - 1 - gap;
				y1 = y0 + height - (sizeline);
				y0 = y1 - ((y1 - y0) * (value));
			} else if (gapPosition == SwingConstants.RIGHT) {
				x0 = x + (sizeline / 2.0);
				y0 = y + (sizeline / 2.0);
				x1 = x0 + width - (sizeline) - 1 - gap;
				y1 = y0 + height - (sizeline);
				y0 = y1 - ((y1 - y0) * (value));
			} else {
				x0 = x + (sizeline / 2.0) + gap;
				y0 = y + (sizeline / 2.0);
				x1 = x0 + width - (sizeline) - 1 - gap - gap;
				y1 = y0 + height - (sizeline);
				y0 = y1 - ((y1 - y0) * (value));
			}
		} else {
			if (gapPosition == SwingConstants.TOP) {
				x0 = x + (sizeline / 2.0);
				y0 = y + (sizeline / 2.0) + gap;
				x1 = x0 + width - (sizeline) - 1;
				x1 = x0 + ((x1 - x0) * (value));
				y1 = y0 + height - (sizeline) - 1 - gap;
			} else if (gapPosition == SwingConstants.BOTTOM) {
				x0 = x + (sizeline / 2.0);
				y0 = y + (sizeline / 2.0);
				x1 = x0 + width - (sizeline) - 1;
				x1 = x0 + ((x1 - x0) * (value));
				y1 = y0 + height - (sizeline) - 1 - gap;
			} else {
				x0 = x + (sizeline / 2.0);
				y0 = y + (sizeline / 2.0) + gap;
				x1 = x0 + width - (sizeline) - 1;
				x1 = x0 + ((x1 - x0) * (value));
				y1 = y0 + height - (sizeline) - 1 - gap - gap;
			}
		}
		if (getValue() != 0) {
			Shape shapeValue;
			valueRectangleWidth = x1 - x0;
			shapeValue = new Rectangle2D.Double(x0, y0, valueRectangleWidth, y1 - y0);
			Paint oldpaint = g2.getPaint();
			if (isFilled()) {
				g2.setPaint(getFillColor());
				g2.fill(shapeValue);
			}
			if (isShowBorder()) {
				g2.setStroke(new BasicStroke(SizeLine));
				if (SizeLine != 0) {
					g2.setPaint(getLineColor());
				} else if (isFilled()) {
					g2.setPaint(getFillColor());
				} else {
					return;
				}
			}
			g2.draw(shapeValue);
			g2.setPaint(oldpaint);

		}
	}

	private void paintBackgroundRectangle(Graphics2D g2, int x, int y,
			int height, int width) {
		double x0, y0, x1, y1, sizeline;

		if (isShowBorder()) {
			sizeline = SizeLine;
		} else {
			sizeline = 0;
		}
		
		if(value == 1.00)
			return;

		if (Orientation == SwingConstants.VERTICAL) {
			if (gapPosition == SwingConstants.LEFT) {
				x0 = x + (sizeline / 2.0) + gap;
				y0 = y + (sizeline / 2.0);
				x1 = x0 + width - (sizeline) - 1 - gap;
				y1 = y0 + height - (sizeline);
				y1 = y0 + ((y1 - y0) * (1.00 - value));
			} else if (gapPosition == SwingConstants.RIGHT) {
				x0 = x + (sizeline / 2.0);
				y0 = y + (sizeline / 2.0);
				x1 = x0 + width - (sizeline) - 1 - gap;
				y1 = y0 + height - (sizeline);
				y1 = y0 + ((y1 - y0) * (1.00 - value));
			} else {
				x0 = x + (sizeline / 2.0) + gap;
				y0 = y + (sizeline / 2.0);
				x1 = x0 + width - (sizeline) - 1 - gap - gap;
				y1 = y0 + height - (sizeline);
				y1 = y0 + ((y1 - y0) * (1.00 - value));
			}
		} else {
			if (gapPosition == SwingConstants.TOP) {
				x0 = x + (sizeline / 2.0);
				y0 = y + (sizeline / 2.0) + gap;
				x1 = x0 + width - (sizeline) - 1;
				x0 = x0 + ((x1 - x0) * (value));
				y1 = y0 + height - (sizeline) - 1 - gap;
			} else if (gapPosition == SwingConstants.BOTTOM) {
				x0 = x + (sizeline / 2.0);
				y0 = y + (sizeline / 2.0);
				x1 = x0 + width - (sizeline) - 1;
				x0 = x0 + ((x1 - x0) * (value));
				y1 = y0 + height - (sizeline) - 1 - gap;
			} else {
				x0 = x + (sizeline / 2.0);
				y0 = y + (sizeline / 2.0) + gap;
				x1 = x0 + width - (sizeline) - 1;
				x0 = x0 + ((x1 - x0) * (value));
				y1 = y0 + height - (sizeline) - 1 - gap - gap;
			}
		}

		Shape shapeValue;
		shapeValue = new Rectangle2D.Double(x0, y0, x1 - x0, y1 - y0);
		Paint oldpaint = g2.getPaint();
		if (isFilled()) {
			g2.setPaint(getColorBackgroundBar());
			g2.fill(shapeValue);
		}
		if (isShowBorder()) {
			g2.setStroke(new BasicStroke(SizeLine));
			if (SizeLine != 0) {
				g2.setPaint(getLineColor());
			} else if (isFilled()) {
				g2.setPaint(getColorBackgroundBar());
			} else {
				return;
			}
		}
		g2.draw(shapeValue);
		g2.setPaint(oldpaint);
	}

	private void paintMarker(Graphics2D g2, int x, int y, int height, int width, double position, Color fillColor, boolean upper) {
		
		double y1;
		
		y1 = y + height - 1;
		y1 = y1 - ((y1 - y) * (position));

		int[] xs = { x, (int) (x + (valueRectangleWidth / scaleMarker)), (int) (x + (valueRectangleWidth / scaleMarker)) };
		int[] ys = { (int) y1, (int) y1, (int) (y1 + ((valueRectangleWidth / (2 * scaleMarker)) * (upper ? (-1.0) : (+1.0)))) };
		Polygon triangle = new Polygon(xs, ys, xs.length);

		g2.setColor(fillColor); // Fllfarbe
		g2.fillPolygon(triangle);

		g2.setColor(Color.BLACK); // Umrandungsfarbe
		g2.setStroke(new BasicStroke(1)); // Linienstrke
		g2.draw(triangle);
		
	}

	private void paintMarkers(Graphics2D g2, int x, int y, int height, int width) {
		if(isMarkersVisible(Marker.WARNING_LOWER))
			paintMarker(g2, x, y, height, width, getMarkersPosition(Marker.WARNING_LOWER), Color.ORANGE, false);
		
		if(isMarkersVisible(Marker.WARNING_UPPER))
			paintMarker(g2, x, y, height, width, getMarkersPosition(Marker.WARNING_UPPER), Color.ORANGE, true);
		
		if(isMarkersVisible(Marker.FAULT_LOWER))
			paintMarker(g2, x, y, height, width, getMarkersPosition(Marker.FAULT_LOWER), Color.RED, false);
		
		if(isMarkersVisible(Marker.FAULT_UPPER))
			paintMarker(g2, x, y, height, width, getMarkersPosition(Marker.FAULT_UPPER), Color.RED, true);
	}
	
	private int getHighestLabelStringWidth(Graphics2D g2) {
		int highestStringWidth = 0;
		Font oldFont = g2.getFont();
		g2.setFont(font);
		FontMetrics fm = g2.getFontMetrics();
		
		for(int index = 0; index < getInterval(); index++) {
			IScaleLabelProvider labelProvider = getScaleLabelProvider(index);
			if(labelProvider == null)
				continue;
			
			for(int tick = 0; tick < this.getTicks(index); tick++) {
				
				String s = labelProvider.getText(tick);
				if(s == null)
					continue;
				
				int stringWidth = SwingUtilities.computeStringWidth(fm, s);
				if(highestStringWidth < stringWidth)
					highestStringWidth = stringWidth;
			}
		}
		
		g2.setFont(oldFont);
		
		return highestStringWidth;
	}

	private void paintLabels(Graphics2D g2, int x, int y, int height, int width) {
		double x1, y1, y0 = 0, x0, DeltaX, DeltaY;
		
		// FIXME: Diese Korrektur muss leider sein, damit die Labels auf die
		// Ticks passen. In der momentanen Implementierung geht diese
		// Korrekturvariable davon aus, dass der Tick, welchen sie beschriften
		// soll, eine Linienbreite von 2px hat.
	
		int testCorr = 0;
		
		for (int index = 0; index < getInterval(); index++) {
			IScaleLabelProvider scaleLabelProvider = getScaleLabelProvider(index);

			if (scaleLabelProvider == null)
				continue;

			if ((getOrientation() == SwingConstants.VERTICAL)) {
				y1 = y + height;
				y0 = y;
				if (getPosition() == SwingConstants.LEFT) {
					x0 = x;
				} else if (getPosition() == SwingConstants.CENTER) {
					x0 = x + (width / 2.0D);
				} else {
					x0 = x + this.getHighestLabelStringWidth(g2); //width - margin;
					//x0 = x + width;
				}
				DeltaX = 0;

				if ((getTicks(index) - 1) != 0) {
					DeltaY = (y1 - y0) / (getTicks(index) - 1);
				} else {
					DeltaY = (y1 - y0);
				}
			} else {
				x0 = x;
				x1 = x + width;
				if (getPosition() == SwingConstants.TOP) {
					y1 = y;
				} else if (getPosition() == SwingConstants.CENTER) {
					y1 = y + (height / 2.0D);
				} else {
					y1 = y + height;
				}
				DeltaY = 0;
				if ((getTicks(index) - 1) != 0) {
					DeltaX = (x1 - x0) / (getTicks(index) - 1);
				} else {
					DeltaX = (x1 - x0);
				}
			}
			for (int scale = 0; scale < getTicks(index); scale++) {
				if (scaleLabelProvider.getIcon(scale) != null) {
					scaleLabelProvider.getIcon(scale).paintIcon(getParent(),
							g2, (int) x0,
							(int) y1);
				} else if ((scaleLabelProvider.getText(scale) != null)
						&& (font != null)) {
					Font oldFont = g2.getFont();
					g2.setFont(font);
					FontMetrics met = g2.getFontMetrics();
					String s = scaleLabelProvider.getText(scale);
					
					g2.drawString(s, (float) x0 - SwingUtilities.computeStringWidth(met, s), (float) (y1 + (testCorr++) + met.getAscent()/4.0));
					g2.setFont(oldFont);
				}
				if ((getOrientation() == SwingConstants.VERTICAL)) {
					y1 = y1 - DeltaY;
				} else {
					x0 = x0 + DeltaX;
				}
			}
		}
	}

	@Override
	public int getGap() {
		return gap;
	}

	@Override
	public double getValue() {
		return value;
	}

	@Override
	public void setGap(int value) {
		if (gap != value) {
			if (value < 0) {
				value = 0;
			}
			gap = value;
			repaint();
		}
	}

	@Override
	public void setValue(double newvalue) {
		if (value != newvalue) {
			value = newvalue;
			repaint();
		}
	}

	@Override
	public boolean isFilled() {
		return filled;
	}

	@Override
	public void setFill(boolean fill) {
		if (filled != fill) {
			filled = fill;
			repaint();
		}
	}

	@Override
	public boolean isShowBorder() {
		return showBorder;
	}

	@Override
	public void setShowBorder(boolean value) {
		if (showBorder != value) {
			showBorder = value;
			repaint();
		}
	}

	@Override
	public Color getColorBackgroundBar() {
		return ColorBackgroundBar;
	}

	@Override
	public boolean isFilledBackgroundBar() {
		return filledBackgroundBar;
	}

	@Override
	public void setColorBackgroundBar(Color FillColor) {
		if (!ColorBackgroundBar.equals(FillColor)) {
			ColorBackgroundBar = FillColor;
			repaint();
		}
	}

	@Override
	public void setFillBackgroundBar(boolean fill) {
		if (filledBackgroundBar != fill) {
			filledBackgroundBar = fill;
			repaint();
		}

	}
	
	@Override
	public double getMarkersPosition(Marker marker) {
		int index = marker.ordinal();
		if (PositionMarkers == null)
			return 0;
		if ((index >= 0) && (index < PositionMarkers.length)) {
			return PositionMarkers[index];
		}
		return 0;
	}

	@Override
	public double[] getMarkersPosition() {
		return PositionMarkers;
	}

	@Override
	public void setMarkersPosition(BarGraph.Marker marker, double position) {
		int index = marker.ordinal();
		if (PositionMarkers == null) {
			PositionMarkers = new double[index + 1];
			PositionMarkers[index] = position;
			repaint();
		} else if ((index >= 0) && (index < PositionMarkers.length)) {
			if (PositionMarkers[index] != position) {
				PositionMarkers[index] = position;
				repaint();
			}
		} else if ((index >= 0) && (index >= PositionMarkers.length)) {
			double OldOffset[] = PositionMarkers;
			PositionMarkers = new double[index + 1];
			System
					.arraycopy(OldOffset, 0, PositionMarkers, 0,
							OldOffset.length);
			PositionMarkers[index] = position;
			repaint();
		}
	}

	@Override
	public void setMarkersPosition(double[] position) {
		if (!Arrays.equals(PositionMarkers, position)) {
			PositionMarkers = position;
			repaint();
		}
	}

	@Override
	public boolean isMarkerVisible() {
		return visibleMarker;
	}

	@Override
	public void setMarkerVisible(boolean visible) {
		if (visibleMarker != visible) {
			visibleMarker = visible;
			repaint();
		}
	}

	@Override
	public IScaleLabelProvider getScaleLabelProvider(int index) {
		if(labelprovider == null)
			return null;
		if ((index < labelprovider.length) && (index >= 0)) {
			return labelprovider[index];
		}
		return null;
	}

	@Override
	public IScaleLabelProvider[] getScaleLabelProvider() {
		return labelprovider;
	}

	@Override
	public boolean isScaleLabelVisible() {
		return showLabel;
	}

	@Override
	public void setScaleLabelVisible(boolean visible) {
		if (showLabel != visible) {
			showLabel = visible;
			repaint();
		}
	}

	@Override
	public void setScaleLabelProvider(int index, IScaleLabelProvider provider) {
		if (labelprovider == null) {
			labelprovider = new IScaleLabelProvider[index + 1];
			labelprovider[index] = provider;
			CalculateMaxString();
			repaint();
		} else if ((index >= 0) && (index < labelprovider.length)) {
			if ((labelprovider[index] == null)
					|| (!(labelprovider[index].equals(provider)))) {
				labelprovider[index] = provider;
				CalculateMaxString();
				repaint();
			}
		} else if ((index >= 0) && (index >= labelprovider.length)) {
			IScaleLabelProvider Oldlabelprovider[] = labelprovider;
			labelprovider = new IScaleLabelProvider[index + 1];
			System.arraycopy(Oldlabelprovider, 0, labelprovider, 0,
					Oldlabelprovider.length);
			labelprovider[index] = provider;
			CalculateMaxString();
			repaint();
		}
		if ((index < labelprovider.length) && (index >= 0)) {
			if (labelprovider[index] != provider) {
				labelprovider[index] = provider;
				CalculateMaxString();
				repaint();
			}
		}
	}

	@Override
	public void setScaleLabelProvider(IScaleLabelProvider[] provider) {
		if (!Arrays.equals(labelprovider, provider)) {
			labelprovider = provider;
			CalculateMaxString();
			repaint();
		}
	}

	private void CalculateMaxString() {
		MaxString = new String[getInterval()];
		for (int index = 0; index < getInterval(); index++) {
			MaxString[index] = new String();
			IScaleLabelProvider scaleLabelProvider = getScaleLabelProvider(index);
			if (scaleLabelProvider != null) {
				for (int j = 0; j < getTicks(index); j++) {
					if (scaleLabelProvider.getText(j) == null)
						continue;
					if (MaxString[index].length() < scaleLabelProvider.getText(
							j).length())
						MaxString[index] = scaleLabelProvider.getText(j);
				}
			}
		}

	}

	public void setFont(Font setfont) {
		if ((font == null) || (!font.equals(setfont))) {
			font = setfont;
			repaint();
		}
	}

	@Override
	public void setInterval(int interval) {
		super.setInterval(interval);
		CalculateMaxString();
	}

	@Override
	public void setTicks(int index, int value) {
		super.setTicks(index, value);
		CalculateMaxString();
	}

	@Override
	public void setTicks(int[] ticks) {
		super.setTicks(ticks);
		CalculateMaxString();
	}

	@Override
	public boolean isMarkersVisible(Marker marker) {
		int index = marker.ordinal();
		if(isMarkerVisible()){
			if (showMarkers == null)
				return true;
			if ((index >= 0) && (index < showMarkers.length)) {
				return showMarkers[index];
			}
		}
		return true;
	}

	@Override
	public void setMarkersVisible(Marker marker, boolean visible) {
		int index = marker.ordinal();
		
		if (showMarkers == null) {
			showMarkers = new boolean[index + 1];
			Arrays.fill(showMarkers, true);
			showMarkers[index] = visible;
			repaint();
		} else if ((index >= 0) && (index < showMarkers.length)) {
			if (showMarkers[index] != visible) {
				showMarkers[index] = visible;
				repaint();
			}
		} else if ((index >= 0) && (index >= showMarkers.length)) {
			boolean OldMarkers[] = showMarkers;
			showMarkers = new boolean[index + 1];
			Arrays.fill(showMarkers, true);
			System.arraycopy(OldMarkers, 0, showMarkers, 0, OldMarkers.length);
			showMarkers[index] = visible;
			repaint();
		}
	}

	@Override
	public boolean[] getMarkersVisible() {
		return showMarkers;
	}

	@Override
	public void setMarkersVisible(boolean[] visible) {
		if (visible == null)
			return;
		if (!Arrays.equals(showMarkers, visible)) {
			showMarkers = visible;
			repaint();
		}
	}

	@Override
	public int getGapPosition() {
		return gapPosition;
	}

	@Override
	public void setGapPosition(int value) {
		this.gapPosition = value;
	}

	@Override
	public String getUnit() {
		return this.unit;
	}

	@Override
	public void setUnit(String unit) {
		if((unit != null) && !this.unit.equals(unit)) {
			this.unit = unit;
			repaint();
		}
	}

	@Override
	public void setValueLabelProvider(IValueLabelProvider provider) {
		this.valueLabelProvider = provider;
	}

	@Override
	public IValueLabelProvider getValueLabelProvider() {
		return this.valueLabelProvider;
	}

	@Override
	public int getMarkerScale() {
		return scaleMarker;
	}

	@Override
	public void setMarkerScale(int val) {
		scaleMarker = val;
		
	}
}
