package ch.thomson.swing.graphics.impl;

import java.awt.BasicStroke;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.awt.geom.Line2D;

import javax.swing.SwingConstants;

import ch.thomson.swing.graphics.IGraphicsLine;


/**
 * 
 * @author Dominic Meier
 * @version $version$ ($date$)
 */
public class GraphicsLine extends GraphicsBase implements IGraphicsLine {

	protected int SizeLine = 2;
	protected int Orientation = SwingConstants.HORIZONTAL;
	protected int Position = SwingConstants.CENTER;
	protected int margin = 0;


	public GraphicsLine() {
		super();
	}

	public GraphicsLine(int w, int h) {
		super(w, h);
	}

	/* (non-Javadoc)
	 * @see thgui.thIcons.IGraphicsLine#getSizeLine()
	 */
	public int getSizeLine() {
		return SizeLine;
	}

	/* (non-Javadoc)
	 * @see thgui.thIcons.IGraphicsLine#setSizeLine(int)
	 */
	public void setSizeLine(int sizeLine) {
		if (SizeLine != sizeLine) {
			SizeLine = sizeLine;
			repaint();
		}
	}

	@Override
	public void paintGraphics(Graphics g, int x, int y) {

		Graphics2D g2 = (Graphics2D) g;
		g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
				RenderingHints.VALUE_ANTIALIAS_ON);
				
		DrawLine(g2, x, y, getHeight(), getWidth());

	}
	
	protected void DrawLine(Graphics2D g2, double x, double y, double height, double width){
		double x0 = x, y0 = y, x1 = x, y1 = y;
		double halfline = SizeLine/2.0D;
		
		if ((this.Orientation == SwingConstants.VERTICAL)) {
			y1 = y + height - halfline - 1;
			y0 = y + halfline;
			if (this.Position == SwingConstants.LEFT) {
				x0 = x + halfline + margin;
				x1 = x0;
			} else if (this.Position == SwingConstants.CENTER) {
				x0 = x + ((width-1) / 2.0D) + margin;
				x1 = x0;
			} else if (this.Position == SwingConstants.RIGHT) {
				x0 = x + (width-1) - halfline - margin;
				x1 = x0;
			}
		}
		if (this.Orientation == SwingConstants.HORIZONTAL) {
			x0 = x + halfline;
			x1 = x + width - halfline - 1;
			if (this.Position == SwingConstants.TOP) {
				y0 = y + halfline + margin;
				y1 = y0;
			} else if (this.Position == SwingConstants.CENTER) {
				y0 = y + ((height-1) / 2.0D);
				y1 = y0;
			} else if (this.Position == SwingConstants.BOTTOM) {
				y0 = y + (height-1) - halfline - margin;
				y1 = y0;
			}
		}
		g2.setPaint(getLineColor());
		paintLine(g2, x0, y0, x1, y1);
	}

	/* (non-Javadoc)
	 * @see thgui.thIcons.IGraphicsLine#getOrientation()
	 */
	public int getOrientation() {
		return Orientation;
	}

	/* (non-Javadoc)
	 * @see thgui.thIcons.IGraphicsLine#setOrientation(int)
	 */
	public void setOrientation(int orientation) {
		if ((orientation == SwingConstants.VERTICAL)
				|| (orientation == SwingConstants.HORIZONTAL)) {
			if(Orientation != orientation){
				int OldValue = Orientation;
				Orientation = orientation;
				Position = ChangePosition(Position);
				OrientationChanged(OldValue, orientation);
				repaint();
			}
		}
	}
	
	protected void OrientationChanged(int oldValue, int orientation) {
		
	}

	/**
	 * <p>
	 * ndere die Position anhand der Orientierung im Uhrzeigersinn. Das heisst aus
	 * Links wird Oben und aus Rechts wird Unten und umgekehrt
	 * </p>
	 * 
	 * @param Die alte Position
	 * @return Die neue <b>int</b> Position
	 */
	protected int ChangePosition(int oldposition){
		switch(oldposition){
			case SwingConstants.LEFT:{
				if(Orientation == SwingConstants.HORIZONTAL)
					return SwingConstants.TOP;
				break;
			}
			case SwingConstants.RIGHT:{
				if(Orientation == SwingConstants.HORIZONTAL)
					return SwingConstants.BOTTOM;
				break;
			}
			case SwingConstants.TOP:{
				if(Orientation == SwingConstants.VERTICAL)
					return SwingConstants.LEFT;
				break;
			}
			case SwingConstants.BOTTOM:{
				if(Orientation == SwingConstants.VERTICAL)
					return SwingConstants.RIGHT;
				break;
			}
		}
		return oldposition;
	}

	/* (non-Javadoc)
	 * @see thgui.thIcons.IGraphicsLine#getPosition()
	 */
	public int getPosition() {
		return Position;
	}

	/* (non-Javadoc)
	 * @see thgui.thIcons.IGraphicsLine#setPosition(int)
	 */
	public void setPosition(int position) {
		if ((Orientation == SwingConstants.VERTICAL)) {
			if(position == SwingConstants.TOP){
				position = SwingConstants.LEFT;
			} else if(position == SwingConstants.BOTTOM){
				position = SwingConstants.RIGHT;
			}
			if ((position == SwingConstants.LEFT)
					|| (position == SwingConstants.CENTER)
					|| (position == SwingConstants.RIGHT)) {
				Position = position;
				repaint();
			}
		}
		if (Orientation == SwingConstants.HORIZONTAL) {
			if(position == SwingConstants.LEFT){
				position = SwingConstants.TOP;
			} else if(position == SwingConstants.RIGHT){
				position = SwingConstants.BOTTOM;
			}
			if ((position == SwingConstants.TOP)
					|| (position == SwingConstants.CENTER)
					|| (position == SwingConstants.BOTTOM)) {
				Position = position;
				repaint();
			}
		}
	}

	protected void paintLine(Graphics2D g2, double x0, double y0, double x1,
			double y1) {
		g2.setStroke(new BasicStroke(SizeLine));
		// Linie Rechts Oben
		Line2D lineRT = new Line2D.Double(x0, y0, x1, y1);
		g2.draw(lineRT);
	}
}
