package ch.thomson.swing.graphics.impl;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.awt.Stroke;
import java.awt.geom.Line2D;
import java.awt.geom.Point2D;

import javax.swing.SwingConstants;

import ch.thomson.swing.graphics.IGraphicsScaleRectangular;


/**
 * <p>
 * Graphische Klasse um eine gerade Skala zu Zeichnen wie sie auf verschiedenen
 * Zeigerinstrumente vorgefunden wird. Die Skala kann mit verschieden Parameter
 * konfiguriert werden welche die Anzahl Teilstriche, Nebenteilstriche und die
 * Position der Skalenbeschriftung einstellen.
 * </p>
 * 
 * @author Dominic Meier
 * @version $version$ ($date$)
 */
public class GraphicsScaleRectangular extends GraphicsLine implements IGraphicsScaleRectangular {

	private boolean VisibleAxis = true;
	private boolean CorrectLength = true;
	/* Variablen fr die Skalen */
	private int Interval = 2;
	private int[] Ticks = new int[2];
	private Color[] TicksColor = new Color[2];
	private int[] TicksSize = new int[2];
	private int[] TicksLength = new int[2];
	private int[] TicksPosition = new int[2];
	
	public GraphicsScaleRectangular() {
		super();
		Init();
	}

	public GraphicsScaleRectangular(int w, int h) {
		super(w, h);
		Init();
	}

	private void Init() {
		Ticks[0] = 9;
		Ticks[1] = 5;
		TicksColor[0] = Color.black;
		TicksColor[1] = Color.black;
		TicksSize[0] = 1;
		TicksSize[1] = 2;
		TicksLength[0] = 5;
		TicksLength[1] = 10;
		TicksPosition[0] = SwingConstants.BOTTOM;
		TicksPosition[1] = SwingConstants.CENTER;
	}

	@Override
	public void paintGraphics(Graphics g, int x, int y) {
		if (SizeLine == 0) {
			return;
		}
		Graphics2D g2 = (Graphics2D) g;
		g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
				RenderingHints.VALUE_ANTIALIAS_ON);

		if (isVisibleAxis()) {
			// Die Hauptlinie zeichnen
			super.paintGraphics(g, x, y);
		}
		
		DrawScala(g2, x, y, getHeight(), getWidth());
		
	}
	
	protected double getHighestTickLineSize() {
		double maxline = 0;
		for (int scale = 0; scale < Interval; scale++) {
			if(maxline < getTicksSize(scale)){
				maxline = getTicksSize(scale);
			}
		}
		return maxline;
	}

	protected void DrawScala(Graphics2D g2, double x, double y, double height, double width){
		double maxline = this.getHighestTickLineSize();
		double halfline = maxline / 2;

		if(isCorrectLength()){
			for (int scale = 0; scale < Interval; scale++) {
				DrawDashOnScala(g2, x + halfline, y + halfline, (height-0 - maxline), (width - 1 - maxline), getTicks(scale), getTicksLength(scale),
						getTicksSize(scale), getTicksPosition(scale),
						getTicksColor(scale));
			}
		} else {
			for (int scale = 0; scale < Interval; scale++) {
				DrawDashOnScala(g2, x, y, (height-1), (width - 1), getTicks(scale), getTicksLength(scale),
						getTicksSize(scale), getTicksPosition(scale),
						getTicksColor(scale));
			}
		}
		
	}
	
	protected int getMaxTickLength() {
		int maxTickLength = 0;
		for(int tickLength : this.getTicksLength()) {
			if(maxTickLength < tickLength)
				maxTickLength = tickLength;	
		}
		return maxTickLength;
	}
	
	
	private void DrawDashOnScala(Graphics2D g2, double x, double y, double height, double width, int ticks,
			int ticksLenght, int ticksSize, int ticksposition, Color ticksColor) {
		if ((ticks <= 0) || (ticksColor == null)) {
			return;
		}
		
		int maxTickLength = this.getMaxTickLength();

		Point2D.Double Dashstart = new Point2D.Double();
		Point2D.Double Dashstop = new Point2D.Double();
		Color OldColor = g2.getColor();
		Stroke OldStroke = g2.getStroke();

		g2.setColor(ticksColor);
		g2.setStroke(new BasicStroke((float) ticksSize, BasicStroke.CAP_ROUND,
				BasicStroke.JOIN_MITER));
		/* Abstand und Startposition rechnen */
		double DeltaX = 0, DeltaY = 0;
		double x0 = x, y0 = y, x1 = x, y1 = y;
		double x2 = x, y2 = y;
		double halfline = ticksSize / 2.0D;

		if ((getOrientation() == SwingConstants.VERTICAL)) {
			y1 = y + height;
			y0 = y;
			y2 = y0;
			if (getPosition() == SwingConstants.LEFT) {
				margin = 0;
				x0 = x + margin;
				x1 = x0;
			} else if (getPosition() == SwingConstants.CENTER) {
				margin = (int) ((-1D)*((getWidth() / 2.0D) - maxTickLength / 2));
				x0 = x + (width / 2.0D) + margin;
				x1 = x0;
			} else if (getPosition() == SwingConstants.RIGHT) {
				margin = (int) ((getWidth() - maxTickLength / 2));
				x0 = x + width - margin;
				x1 = x0;
			}
			/* Position des Strichendes Rechnen */
			if (getPosition() == SwingConstants.LEFT) {
				if (ticksposition == SwingConstants.RIGHT) {
					x2 = x0 + ticksLenght;
				} else if (ticksposition == SwingConstants.CENTER) {
					x2 = x0 + (ticksLenght / 2.0D);
				} else {
					return;
				}
			} else if (getPosition() == SwingConstants.CENTER) {
				if (ticksposition == SwingConstants.RIGHT) {
					x2 = x0 + ticksLenght;
				} else if (ticksposition == SwingConstants.CENTER) {
					x2 = x0 + (ticksLenght / 2.0D);
					x0 = x2 - ticksLenght;
				} else if (ticksposition == SwingConstants.LEFT) {
					x2 = x0 - ticksLenght;
				} else {
					return;
				}
			} else if (getPosition() == SwingConstants.RIGHT) {
				if (ticksposition == SwingConstants.CENTER) {
					x2 = x0 - (ticksLenght / 2.0D);
				} else if (ticksposition == SwingConstants.LEFT) {
					x2 = x0 - ticksLenght;
				} else {
					return;
				}
			}
			DeltaX = 0;
			if((ticks-1)!=0){
				DeltaY = (y1 - y0) / (ticks-1);
			} else {
				DeltaY = (y1 - y0);
			}
		} else {
				x0 = x;
				x1 = x + width;
				if (getPosition() == SwingConstants.TOP) {
					y0 = y + halfline;
					y1 = y0;
				} else if (getPosition() == SwingConstants.CENTER) {
					y0 = y + (height / 2.0D);
					y1 = y0;
				} else if (getPosition() == SwingConstants.BOTTOM) {
					y0 = y + height;
					y1 = y0;
				}
//			}
			/* Position des Strichendes Rechnen */
			if (getPosition() == SwingConstants.TOP) {
				if (ticksposition == SwingConstants.BOTTOM) {
					y2 = y0 + ticksLenght;
				} else if (ticksposition == SwingConstants.CENTER) {
					y2 = y0 + (ticksLenght / 2.0D);
				} else {
					return;
				}
			} else if (getPosition() == SwingConstants.CENTER) {
				if (ticksposition == SwingConstants.BOTTOM) {
					y2 = y0 + ticksLenght;
				} else if (ticksposition == SwingConstants.CENTER) {
					y2 = y0 + (ticksLenght / 2.0D);
					y0 = y2 - ticksLenght;
				} else if (ticksposition == SwingConstants.TOP) {
					y2 = y0 - ticksLenght;
				} else {
					return;
				}
			} else if (getPosition() == SwingConstants.BOTTOM) {
				if (ticksposition == SwingConstants.CENTER) {
					y2 = y0 - (ticksLenght / 2.0D);
				} else if (ticksposition == SwingConstants.TOP) {
					y2 = y0 - ticksLenght;
				} else {
					return;
				}
			}
			if((ticks-1)!=0){
				DeltaX = (x1 - x0) / (ticks-1);
			} else {
				DeltaX = (x1 - x0);
			}
			DeltaY = 0;
		}

		// Startpunkt Setzen
		Dashstart.x = x0;
		Dashstart.y = y0;
		// EndPunkt Setzen
		Dashstop.x = x2;
		Dashstop.y = y2;

		for (int scale = 0; scale < ticks; scale++) {
			g2.draw(new Line2D.Double(Dashstart.x, Dashstart.y, Dashstop.x,
					Dashstop.y));
			if ((getOrientation() == SwingConstants.VERTICAL)) {
				Dashstart.y = Dashstart.y + DeltaY;
				Dashstop.y = Dashstop.y + DeltaY;
			} else {
				Dashstart.x = Dashstart.x + DeltaX;
				Dashstop.x = Dashstop.x + DeltaX;
			}
		}

		g2.setStroke(OldStroke);
		g2.setColor(OldColor);
	}

	/* (non-Javadoc)
	 * @see thgui.thIcons.Base.IGraphicsScaleRectangular#isVisibleAxis()
	 */
	public boolean isVisibleAxis() {
		return VisibleAxis;
	}

	/* (non-Javadoc)
	 * @see thgui.thIcons.Base.IGraphicsScaleRectangular#setVisibleAxis(boolean)
	 */
	public void setVisibleAxis(boolean visibleAxis) {
		if (VisibleAxis != visibleAxis) {
			VisibleAxis = visibleAxis;
			repaint();
		}
	}

	/* (non-Javadoc)
	 * @see thgui.thIcons.Base.IGraphicsScaleRectangular#getInterval()
	 */
	public int getInterval() {
		return Interval;
	}

	/* (non-Javadoc)
	 * @see thgui.thIcons.Base.IGraphicsScaleRectangular#setInterval(int)
	 */
	public void setInterval(int interval) {
		if (Interval != interval) {
			Interval = interval;
			repaint();
		}
	}

	/* (non-Javadoc)
	 * @see thgui.thIcons.Base.IGraphicsScaleRectangular#getTicks()
	 */
	public int[] getTicks() {
		return Ticks;
	}

	/* (non-Javadoc)
	 * @see thgui.thIcons.Base.IGraphicsScaleRectangular#getTicks(int)
	 */
	public int getTicks(int index) {
		if ((Ticks.length > index) && (index >= 0)) {
			return Ticks[index];
		}
		return 0;
	}

	/* (non-Javadoc)
	 * @see thgui.thIcons.Base.IGraphicsScaleRectangular#setTicks(int[])
	 */
	public void setTicks(int[] ticks) {
		Ticks = ticks;
		repaint();
	}

	/* (non-Javadoc)
	 * @see thgui.thIcons.Base.IGraphicsScaleRectangular#setTicks(int, int)
	 */
	public void setTicks(int index, int value) {
		if ((Ticks.length > index) && (index >= 0)) {
			if (Ticks[index] != value) {
				Ticks[index] = value;
				repaint();
			}
		}
	}

	/* (non-Javadoc)
	 * @see thgui.thIcons.Base.IGraphicsScaleRectangular#getTicksColor()
	 */
	public Color[] getTicksColor() {
		return TicksColor;
	}

	/* (non-Javadoc)
	 * @see thgui.thIcons.Base.IGraphicsScaleRectangular#getTicksColor(int)
	 */
	public Color getTicksColor(int index) {
		if ((TicksColor.length > index) && (index >= 0)) {
			return TicksColor[index];
		}
		return null;
	}

	/* (non-Javadoc)
	 * @see thgui.thIcons.Base.IGraphicsScaleRectangular#setTicksColor(java.awt.Color[])
	 */
	public void setTicksColor(Color[] ticksColor) {
		TicksColor = ticksColor;
		repaint();
	}

	/* (non-Javadoc)
	 * @see thgui.thIcons.Base.IGraphicsScaleRectangular#setTicksColor(int, java.awt.Color)
	 */
	public void setTicksColor(int index, Color value) {
		if ((TicksColor.length > index) && (index >= 0)) {
			if (TicksColor[index] != value) {
				TicksColor[index] = value;
				repaint();
			}
		}
	}

	/* (non-Javadoc)
	 * @see thgui.thIcons.Base.IGraphicsScaleRectangular#getTicksSize()
	 */
	public int[] getTicksSize() {
		return TicksSize;
	}

	/* (non-Javadoc)
	 * @see thgui.thIcons.Base.IGraphicsScaleRectangular#getTicksSize(int)
	 */
	public int getTicksSize(int index) {
		if ((TicksSize.length > index) && (index >= 0)) {
			return TicksSize[index];
		}
		return 0;
	}

	/* (non-Javadoc)
	 * @see thgui.thIcons.Base.IGraphicsScaleRectangular#setTicksSize(int[])
	 */
	public void setTicksSize(int[] ticksSize) {
		TicksSize = ticksSize;
		repaint();
	}

	/* (non-Javadoc)
	 * @see thgui.thIcons.Base.IGraphicsScaleRectangular#setTicksSize(int, int)
	 */
	public void setTicksSize(int index, int value) {
		if ((TicksSize.length > index) && (index >= 0)) {
			if (TicksSize[index] != value) {
				TicksSize[index] = value;
				repaint();
			}
		}
	}

	/* (non-Javadoc)
	 * @see thgui.thIcons.Base.IGraphicsScaleRectangular#getTicksLength()
	 */
	public int[] getTicksLength() {
		return TicksLength;
	}

	/* (non-Javadoc)
	 * @see thgui.thIcons.Base.IGraphicsScaleRectangular#getTicksLength(int)
	 */
	public int getTicksLength(int index) {
		if ((TicksLength.length > index) && (index >= 0)) {
			return TicksLength[index];
		}
		return 0;
	}

	/* (non-Javadoc)
	 * @see thgui.thIcons.Base.IGraphicsScaleRectangular#setTicksLength(int[])
	 */
	public void setTicksLength(int[] ticksLength) {
		TicksLength = ticksLength;
		repaint();
	}

	/* (non-Javadoc)
	 * @see thgui.thIcons.Base.IGraphicsScaleRectangular#setTicksLength(int, int)
	 */
	public void setTicksLength(int index, int value) {
		if ((TicksLength.length > index) && (index >= 0)) {
			if (TicksLength[index] != value) {
				TicksLength[index] = value;
				repaint();
			}
		}
	}

	/* (non-Javadoc)
	 * @see thgui.thIcons.Base.IGraphicsScaleRectangular#getTicksPosition()
	 */
	public int[] getTicksPosition() {
		return TicksPosition;
	}

	/* (non-Javadoc)
	 * @see thgui.thIcons.Base.IGraphicsScaleRectangular#getTicksPosition(int)
	 */
	public int getTicksPosition(int index) {
		if ((TicksPosition.length > index) && (index >= 0)) {
			return TicksPosition[index];
		}
		return SwingConstants.CENTER;
	}

	/* (non-Javadoc)
	 * @see thgui.thIcons.Base.IGraphicsScaleRectangular#setTicksPosition(int[])
	 */
	public void setTicksPosition(int[] ticksPosition) {
		TicksPosition = ticksPosition;
		repaint();
	}

	/* (non-Javadoc)
	 * @see thgui.thIcons.Base.IGraphicsScaleRectangular#setTicksPosition(int, int)
	 */
	public void setTicksPosition(int index, int value) {
		if ((TicksPosition.length > index) && (index >= 0)) {
			if (TicksPosition[index] != value) {
				TicksPosition[index] = value;
				repaint();
			}
		}
	}

	/* (non-Javadoc)
	 * @see thgui.thIcons.Base.IGraphicsScaleRectangular#isCorrectLength()
	 */
	public boolean isCorrectLength() {
		return CorrectLength;
	}

	/* (non-Javadoc)
	 * @see thgui.thIcons.Base.IGraphicsScaleRectangular#setCorrectLength(boolean)
	 */
	public void setCorrectLength(boolean correctLength) {
		if (CorrectLength != correctLength) {
			CorrectLength = correctLength;
			repaint();
		}
	}

	/* (non-Javadoc)
	 * @see thgui.thIcons.Base.GraphicsLine#setOrientation(int)
	 */
	@Override
	protected void OrientationChanged(int oldValue, int orientation) {
		for(int count = 0; count < TicksPosition.length; count++){
			TicksPosition[count] = ChangePosition(TicksPosition[count]);
		}
		//TextPosition = ChangePosition(TextPosition);
	}
}
