package ch.thomson.ucs.tine.databinding.swt;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Frame;
import java.awt.Panel;

import javax.swing.JRootPane;

import org.eclipse.core.databinding.observable.Realm;
import org.eclipse.emf.common.notify.Notification;
import org.eclipse.emf.common.notify.impl.AdapterImpl;
import org.eclipse.swt.SWT;
import org.eclipse.swt.awt.SWT_AWT;
import org.eclipse.swt.layout.FillLayout;
import org.eclipse.swt.widgets.Composite;

import ch.thomson.ucs.tine.client.Client;
import ch.thomson.ucs.tine.client.ClientPackage.Literals;
import ch.thomson.ucs.tine.client.Property;
import de.desy.acop.chart.Acop;

public class ThAcop extends ThComposite implements IDataBinding{
	private Acop acop;
	private String signalName;
	private String signalCaption;
	private WaveForm waveForm;
	private Realm realm;
	protected ConnectionControlDecoration connectionDecoration;
	
	public class WaveForm {
		private Property property;
		private String caption;
		
		public WaveForm(Property property) {
			this.property = property;		
			caption = property.getName();
			
			property.eAdapters().add(new AdapterImpl() {
				@Override
				public void notifyChanged(final Notification msg) {
					if (Literals.PROPERTY__VALUE == msg.getFeature() && !msg.isTouch()) {
						realm.asyncExec(new Runnable() {

							@Override
							public void run() {
								if(ThAcop.this.isDisposed())
									return;
								
								if(waveForm == WaveForm.this) {
									drawGraph(msg.getNewValue());
								}
							}
						});
					}
				}
			});
		}
		
		private void drawGraph(Object object) {
			acop.ClearScreen();
			acop.Draw(object, null, null, null, 0, 0);
		}
		
		public String getCaption() {
			return caption;
		}

		public void setCaption(String caption) {
			this.caption = caption;
		}

		public void loadAcop() {
			acop.setXMin(0.0);
			acop.setXMax(((float[])property.getValue()).length);
			acop.setYMin(property.getMinimum());
			acop.setYMax(property.getMaximum());
			acop.setYAxisLabel("[" + property.getUnit() + "]");
			acop.setCaption(getCaption());
			drawGraph(property.getValue());
			//acop.AutoScale(false, false, true, true);
		}
	}
	
	public ThAcop(Composite parent, int style) {
		super(parent, SWT.NONE);
		setLayout(new FillLayout(SWT.HORIZONTAL));
		
		Composite composite = new Composite(this, SWT.EMBEDDED);
		
		Frame frame = SWT_AWT.new_Frame(composite);
		
		Panel panel = new Panel();
		frame.add(panel);
		panel.setLayout(new BorderLayout(0, 0));
		
		JRootPane rootPane = new JRootPane();
		panel.add(rootPane);
		
		acop = new Acop();
		acop.setYAxisLabel("[V]");
		acop.setXAxisLabel("[us]");
		acop.setGridColor(new Color(128, 128, 128));
		acop.setYGrid(true);
		acop.setXGrid(true);
		acop.setFrameBottomOffset(10);
		acop.setFrameTopOffset(10);
		acop.setFrameLeftOffset(50);
		acop.setBackground(Color.BLACK);
		acop.setForeground(Color.RED);
		
		realm = Realm.getDefault();
		rootPane.getContentPane().add(acop, BorderLayout.CENTER);
	}	
	
	public String getSignalName() {
		return signalName;
	}

	public void setSignalName(String signalName) {
		this.signalName = signalName;
	}

	public String getSignalCaption() {
		return signalCaption;
	}

	public void setSignalCaption(String signalCaption) {
		this.signalCaption = signalCaption;
	}

	public WaveForm getWaveForm() {
		return waveForm;
	}

	public void setWaveForm(WaveForm waveForm) {
		if(waveForm != null)
			waveForm.loadAcop();
		
		this.waveForm = waveForm;
	}
	
	public WaveForm createWaveForm(Property property) {
		return new WaveForm(property);
	}

	@Override
	public void bindValues(Client client) throws Exception {
		// WaveForm erzeugen, falls Signalname vorhanden
		if(signalName != null) {
			Property property = client.findProperty(signalName);
			waveForm = createWaveForm(property);
			
			// Caption setzen, falls vorhanden
			if(signalCaption != null)
				waveForm.setCaption(signalCaption);
			
			waveForm.loadAcop();
		}		
	}
}
