package ch.thomson.ucs.tine.databinding.swt;

import org.eclipse.jface.viewers.ArrayContentProvider;
import org.eclipse.jface.viewers.ComboViewer;
import org.eclipse.jface.viewers.ISelectionChangedListener;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.jface.viewers.LabelProvider;
import org.eclipse.jface.viewers.SelectionChangedEvent;
import org.eclipse.jface.viewers.StructuredSelection;
import org.eclipse.swt.SWT;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Combo;
import org.eclipse.swt.widgets.Composite;

import ch.thomson.ucs.tine.client.Client;
import ch.thomson.ucs.tine.client.Property;
import ch.thomson.ucs.tine.databinding.swt.ThAcop.WaveForm;
import org.eclipse.wb.swt.SWTResourceManager;

public class ThAcopMultiChannel extends ThComposite implements IDataBinding {
	
	private ThAcop thAcop;
	private ThText text;
	private ComboViewer comboViewer;
	private String[] signalList;
	private String[] signalCaptions;
	private WaveForm[] waveForms;

	public ThAcopMultiChannel(Composite parent, int style) {
		super(parent, style);
		setLayout(new GridLayout(1, false));
		
		Composite cmpTraceSelection = new Composite(this, SWT.NONE);
		cmpTraceSelection.setLayoutData(new GridData(SWT.FILL, SWT.FILL, false, false, 1, 1));
		GridLayout gl_cmpTraceSelection = new GridLayout(2, false);
		gl_cmpTraceSelection.horizontalSpacing = 20;
		cmpTraceSelection.setLayout(gl_cmpTraceSelection);
		
		ThGroup grpTraceSelection = new ThGroup(cmpTraceSelection, SWT.NONE);
		grpTraceSelection.setFont(SWTResourceManager.getFont("Tahoma", 10, SWT.BOLD));
		GridData gd_grpTraceSelection = new GridData(SWT.FILL, SWT.FILL, false, false, 1, 1);
		gd_grpTraceSelection.widthHint = 208;
		grpTraceSelection.setLayoutData(gd_grpTraceSelection);
		grpTraceSelection.setText("Trace Selection");
		GridLayout gl_grpTraceSelection = new GridLayout(1, false);
		grpTraceSelection.setLayout(gl_grpTraceSelection);
		
		comboViewer = new ComboViewer(grpTraceSelection, SWT.READ_ONLY);
		Combo combo = comboViewer.getCombo();
		combo.setFont(SWTResourceManager.getFont("Tahoma", 14, SWT.BOLD));
		GridData gd_combo = new GridData(SWT.FILL, SWT.CENTER, true, false, 1, 1);
		gd_combo.widthHint = 129;
		combo.setLayoutData(gd_combo);
		
		ThGroup grpTraceOfPulse = new ThGroup(cmpTraceSelection, SWT.NONE);
		grpTraceOfPulse.setFont(SWTResourceManager.getFont("Tahoma", 10, SWT.BOLD));
		grpTraceOfPulse.setText("Trace of Event Number");
		GridData gd_grpTraceOfPulse = new GridData(SWT.FILL, SWT.FILL, false, false, 1, 1);
		gd_grpTraceOfPulse.widthHint = 210;
		grpTraceOfPulse.setLayoutData(gd_grpTraceOfPulse);
		grpTraceOfPulse.setLayout(new GridLayout(1, false));
		
		text = new ThText(grpTraceOfPulse, SWT.BORDER | SWT.RIGHT);
		text.setFont(SWTResourceManager.getFont("Tahoma", 14, SWT.BOLD));
		text.setEditable(false);
		text.setText("0000");
		text.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, false, 1, 1));
		
		thAcop = new ThAcop(this, SWT.NONE);
		thAcop.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, true, 1, 1));
		comboViewer.setContentProvider(ArrayContentProvider.getInstance());
		comboViewer.setLabelProvider(new LabelProvider() {
			
			@Override
			public String getText(Object element) {
				return ((WaveForm) element).getCaption();
			}
		});
		comboViewer.addSelectionChangedListener(new ISelectionChangedListener() {

			@Override
			public void selectionChanged(SelectionChangedEvent event) {
				IStructuredSelection selection = (IStructuredSelection) event.getSelection();
				thAcop.setWaveForm((WaveForm) selection.getFirstElement());
			}
		});
	}
	
	public void setSignalList(String[] signalList) {
		this.signalList = signalList;
	}
	
	public String[] getSignalList() {
		return this.signalList;
	}
	
	public void setSignalCaptions(String[] signalCaptions) {
		this.signalCaptions = signalCaptions;
	}
	
	public String[] getSignalCaptions() {
		return this.signalCaptions;
	}
	
	public String getSignalNamePulseCounter() {
		return text.getSignalName();
	}

	public void setSignalNamePulseCounter(String signalNamePulseCounter) {
		text.setSignalName(signalNamePulseCounter);
	}

	@Override
	public void bindValues(Client client) throws Exception {
		// WaveForms erzeugen
		waveForms = new WaveForm[signalList.length];
		for(int i=0; i<signalList.length; ++i) {
			Property property = client.findProperty(signalList[i]);
			waveForms[i] = thAcop.createWaveForm(property);
		}
		
		// Captions setzen, falls vorhanden
		for(int i=0; i < waveForms.length && i < signalCaptions.length; ++i) 
			waveForms[i].setCaption(signalCaptions[i]);
		
		comboViewer.setInput(waveForms);
		comboViewer.setSelection(new StructuredSelection(waveForms[0]));
	}
}
