package ch.thomson.ucs.tine.databinding.swt;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Frame;
import java.awt.Panel;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.text.DecimalFormat;

import javax.swing.Icon;
import javax.swing.JRootPane;
import javax.swing.SwingConstants;

import org.eclipse.core.databinding.DataBindingContext;
import org.eclipse.core.databinding.UpdateValueStrategy;
import org.eclipse.core.databinding.beans.BeansObservables;
import org.eclipse.core.databinding.observable.Realm;
import org.eclipse.core.databinding.observable.value.IObservableValue;
import org.eclipse.emf.databinding.EMFObservables;
import org.eclipse.swt.SWT;
import org.eclipse.swt.awt.SWT_AWT;
import org.eclipse.swt.events.DisposeEvent;
import org.eclipse.swt.events.DisposeListener;
import org.eclipse.swt.widgets.Composite;

import ch.thomson.swing.graphics.IScaleLabelProvider;
import ch.thomson.swing.graphics.IValueLabelProvider;
import ch.thomson.swing.graphics.impl.BarGraph;
import ch.thomson.swing.graphics.impl.BarGraph.Marker;
import ch.thomson.swing.jlabel.JLabelBarGraph;
import ch.thomson.ucs.tine.client.Client;
import ch.thomson.ucs.tine.client.ClientPackage.Literals;
import ch.thomson.ucs.tine.client.Property;
import ch.thomson.ucs.tine.client.PropertyNumeric;
import ch.thomson.ucs.tine.client.databinding.converters.ConverterFactory;
import ch.thomson.ucs.tine.client.databinding.converters.IndexedConverter;
import ch.thomson.ucs.tine.databinding.swt.dialogs.BarGraphMarkerSettings;

public class ThBarGraph extends ThComposite implements IDataBinding , IMarkerBarGraph {
	private JLabelBarGraph labelBarGraph;
	private PropertyChangeSupport pcs = new PropertyChangeSupport(this);
	private String caption;
	private String description;
	private int valueIndex = 0;
	private String pathValue;
	private String pathWarningLower;
	private String pathWarningUpper;
	private String pathFaultLower;
	private String pathFaultUpper;
	private double scaleMin = 0;
	private double scaleMax = 100;
	private boolean scaleAutoEnabled = false;
	private Realm realm;
	private Color fillColor = Color.GREEN;
	private double factor = 1.0;
	@SuppressWarnings("unused")
	private ConnectionControlDecoration connectionDecoration;
	private Property warningLowerProperty;
	private Property warningUpperProperty;
	private Property faultLowerProperty;
	private Property faultUpperProperty;
	private PropertyNumeric valueProperty;
	private IndexedConverter valueConverter;
	private int scaleMarker;
	private DataBindingContext dataBindingContext = new DataBindingContext();
	private BarGraphMarkerSettings bgms = null;
	private boolean markerSingleIndexed = false;
	
	public boolean isMarkerSingleIndexed() {
		return markerSingleIndexed;
	}

	public void setMarkerSingleIndexed(boolean markerSingleIndexed) {
		this.markerSingleIndexed = markerSingleIndexed;
	}

	public ThBarGraph(Composite parent) {
		super(parent, SWT.EMBEDDED);
		
		this.realm = Realm.getDefault();
		
		
		Frame frame = SWT_AWT.new_Frame(this);
		
		Panel panel = new Panel();
		frame.add(panel);
		panel.setLayout(new BorderLayout(0, 0));
		
		JRootPane rootPane = new JRootPane();
		panel.add(rootPane);
		
		labelBarGraph = new JLabelBarGraph();
//		labelBarGraph.setCursor(java.awt.Cursor.getPredefinedCursor(java.awt.Cursor.HAND_CURSOR));
		labelBarGraph.addMouseListener(new java.awt.event.MouseAdapter() {
			@Override
			public void mouseClicked(java.awt.event.MouseEvent e) {
				realm.exec(new Runnable() {
					@Override
					public void run() {
						if((getWarningLowerProperty() != null) || (getWarningUpperProperty() != null) ||
						   (getFaultLowerProperty() != null) || (getFaultUpperProperty() != null)) {
//							BarGraphMarkerSettings bgms = new BarGraphMarkerSettings(null/*getShell()*/, ThBarGraph.this);
							bgms = BarGraphMarkerSettings.instance(null/*ThText.this.getParent().getShell()*/, ThBarGraph.this);
							bgms.open();
						}
					}
				});
			}
		});
		
		this.addDisposeListener(new DisposeListener() {

			@Override
			public void widgetDisposed(DisposeEvent e) {
				// TODO Auto-generated method stub
				if((bgms != null))
					bgms.close();
			}
		});
		
		labelBarGraph.setGap(10);
		labelBarGraph.setGapPosition(SwingConstants.LEFT);
		labelBarGraph.setPosition(SwingConstants.RIGHT);
		labelBarGraph.setFillColor(Color.BLUE);
		labelBarGraph.setFillBackgroundBar(false);
		labelBarGraph.setFill(true);
		labelBarGraph.setValue(0.5);
		labelBarGraph.setOrientation(SwingConstants.VERTICAL);
		labelBarGraph.setTicks(new int[] {21, 5});
		labelBarGraph.setTicksSize(new int[]{1, 2});
		labelBarGraph.setTicksLength(new int[]{10, 20});
		
		IScaleLabelProvider[] labelProviders = new IScaleLabelProvider[1];
		
		labelProviders[0] = new IScaleLabelProvider() {
			private DecimalFormat df = new DecimalFormat("#0.##");
			@Override
			public String getText(int tick) {
				if(tick % 5 == 0)
					return df.format(scaleMin + (scaleMax - scaleMin)/4.0 * (tick / 5));
				
				return null;
			}

			@Override
			public Icon getIcon(int tick) {
				return null;
			}
		};
		
		labelBarGraph.setValueLabelProvider(new IValueLabelProvider() {
			private DecimalFormat df = new DecimalFormat("#0.0#");
			@Override
			public String getText() {
				
				if(valueProperty == null)
					return df.format(0.0);
				float[] fArr  = (float[]) valueProperty.getValue();
				return df.format(fArr[getValueIndex()]);
//				return df.format(valueProperty.getDouble());
			}
		});
		
		labelBarGraph.setScaleLabelProvider(labelProviders);
		labelBarGraph.setScaleLabelVisible(true);
		
		
		double[] markersPosition = {0.2D, 0.4D, 0.6D, 0.8D};
		labelBarGraph.setMarkersPosition(markersPosition);
		labelBarGraph.setMarkerVisible(true);
		
		labelBarGraph.setTicksPosition(new int[]{SwingConstants.CENTER, SwingConstants.CENTER});
		
		rootPane.getContentPane().add(labelBarGraph, BorderLayout.CENTER);
	}
		
	public boolean isScaleAutoEnabled() {
		return scaleAutoEnabled;
	}

	public void setScaleAutoEnabled(boolean scaleAutoEnabled) {
		this.scaleAutoEnabled = scaleAutoEnabled;
	}

	public double getScaleMin() {
		return scaleMin;
	}

	public void setScaleMin(double scaleMin) {
		this.scaleMin = scaleMin;
		// set Fault and warning level for correct fill color
		this.setFaultLower(scaleMin);
		this.setWarningLower(scaleMin);
	}

	public double getScaleMax() {
		return scaleMax;
	}

	public void setScaleMax(double scaleMax) {
		this.scaleMax = scaleMax;
		// set Fault and warning level for correct fill color
		this.setFaultUpper(scaleMax);
		this.setWarningUpper(scaleMax);
	}

	public String getPathValue() {
		return pathValue;
	}

	public void setPathValue(String pathValue) {
		this.pathValue = pathValue;
	}

	public String getPathWarningLower() {
		return pathWarningLower;
	}

	public void setPathWarningLower(String pathWarningLower) {
		labelBarGraph.setCursor(java.awt.Cursor.getPredefinedCursor(java.awt.Cursor.HAND_CURSOR));
		this.pathWarningLower = pathWarningLower;
	}

	public String getPathWarningUpper() {
		return pathWarningUpper;
	}

	public void setPathWarningUpper(String pathWarningUpper) {
		labelBarGraph.setCursor(java.awt.Cursor.getPredefinedCursor(java.awt.Cursor.HAND_CURSOR));
		this.pathWarningUpper = pathWarningUpper;
	}

	public String getPathFaultLower() {
		return pathFaultLower;
	}

	public void setPathFaultLower(String pathFaultLower) {
		labelBarGraph.setCursor(java.awt.Cursor.getPredefinedCursor(java.awt.Cursor.HAND_CURSOR));
		this.pathFaultLower = pathFaultLower;
	}

	public String getPathFaultUpper() {
		return pathFaultUpper;
	}

	public void setPathFaultUpper(String pathFaultUpper) {
		labelBarGraph.setCursor(java.awt.Cursor.getPredefinedCursor(java.awt.Cursor.HAND_CURSOR));
		this.pathFaultUpper = pathFaultUpper;
	}

	public int getValueIndex() {
		if(valueConverter != null)
			valueIndex = valueConverter.getIndex();
		return valueIndex;
	}

	public void setValueIndex(int valueIndex) {
		this.valueIndex = valueIndex;
		if(valueConverter != null)
			valueConverter.setIndex(valueIndex);
		dataBindingContext.updateTargets();
	}

	public double getFactor() {
		return factor;
	}

	public void setFactor(double factor) {
		this.factor = factor;
	}

	public double getValue() {
		return labelBarGraph.getValue();
	}
	
	private boolean isWarningLowerEnabled() {
		return labelBarGraph.isMarkersVisible(Marker.WARNING_LOWER);
	}
	
	private boolean isWarningUpperEnabled() {
		return labelBarGraph.isMarkersVisible(Marker.WARNING_UPPER);
	}
	
	private boolean isFaultLowerEnabled() {
		return labelBarGraph.isMarkersVisible(Marker.FAULT_LOWER);
	}
	
	private boolean isFaultUpperEnabled() {
		return labelBarGraph.isMarkersVisible(Marker.FAULT_UPPER);
	}
	
	private void refreshFillColor() {
		Color color = this.fillColor;
		
		// Balkenfarbe ndern, wenn der Wert ausserhalb des Markerbereichs liegt
		if((isWarningLowerEnabled() && (getValue() < getWarningLower())) || (isWarningUpperEnabled() && (getValue() > getWarningUpper())))
			color = Color.ORANGE;
		
		if((isFaultLowerEnabled() && (getValue() < getFaultLower())) || (isFaultUpperEnabled() && (getValue() > getFaultUpper())))
			color = Color.RED;
		
		labelBarGraph.setFillColor(color);
	}
	
	public void setValue(double value) {
		labelBarGraph.setValue(value);
		refreshFillColor();
	}
	
	public double getWarningLower() {
		return labelBarGraph.getMarkersPosition(BarGraph.Marker.WARNING_LOWER);
	}
	
	public void setWarningLower(double value) {
		labelBarGraph.setMarkersPosition(BarGraph.Marker.WARNING_LOWER, value);
		refreshFillColor();
	}
	
	public double getWarningUpper() {
		return labelBarGraph.getMarkersPosition(BarGraph.Marker.WARNING_UPPER);
	}
	
	public void setWarningUpper(double value) {
		labelBarGraph.setMarkersPosition(BarGraph.Marker.WARNING_UPPER, value);
		refreshFillColor();
	}
	
	public double getFaultLower() {
		return labelBarGraph.getMarkersPosition(BarGraph.Marker.FAULT_LOWER);
	}
	
	public void setFaultLower(double value) {
		labelBarGraph.setMarkersPosition(BarGraph.Marker.FAULT_LOWER, value);
		refreshFillColor();
	}
	
	public double getFaultUpper() {
		return labelBarGraph.getMarkersPosition(BarGraph.Marker.FAULT_UPPER);
	}
	
	public void setFaultUpper(double value) {
		labelBarGraph.setMarkersPosition(BarGraph.Marker.FAULT_UPPER, value);
		refreshFillColor();
	}
	
	public Color getFillColor() {
		return this.fillColor;
	}
	
	public void setFillColor(Color color) {
		this.fillColor = color;
		labelBarGraph.setFillColor(color);
	}
	
	public boolean isFill() {
		return labelBarGraph.isFilled();
	}
	
	public void setFill(boolean filled) {
		labelBarGraph.setFill(filled);
	}
	
	public boolean isFillBackgroundBar() {
		return labelBarGraph.isFilledBackgroundBar();
	}
	
	public void setFillBackgroundBar(boolean filled) {
		labelBarGraph.setFillBackgroundBar(filled);
	}
	
	public void addPropertyChangeListener(PropertyChangeListener l) {
		pcs.addPropertyChangeListener(l);
	}

	public void removePropertyChangeListener(PropertyChangeListener l) {
		pcs.removePropertyChangeListener(l);
	}
	
	public void bindMarker(Client client, BarGraph.Marker marker, Property property, String observedProperty, UpdateValueStrategy modelToTarget) throws Exception{
		
		labelBarGraph.setMarkersVisible(marker, false);
		if(property == null) {
			return; // Es gibt nichts zu binden
		}
		
		labelBarGraph.setMarkersVisible(marker, true);
		IObservableValue widget = BeansObservables.observeValue(this, observedProperty);
		IObservableValue value = EMFObservables.observeValue(property, Literals.PROPERTY__VALUE);
		
		UpdateValueStrategy updateStrategy = null;
		if(isMarkerSingleIndexed()) {
			updateStrategy = new UpdateValueStrategy();
			valueConverter = ConverterFactory.createScalingConverter(property.getValue().getClass(), Float.class, 0, property, scaleMin * factor, scaleMax * factor);
			updateStrategy.setConverter(valueConverter);
		}
		
		dataBindingContext.bindValue(widget, value, null, (updateStrategy == null ? modelToTarget : updateStrategy));
	}

	@Override
	public void bindValues(Client client) throws Exception {
		valueProperty = (PropertyNumeric) client.findProperty(pathValue);
		labelBarGraph.setUnit(valueProperty.getUnit());
		
		if(isScaleAutoEnabled()) {
			setScaleMax(valueProperty.getMaximum());
			setScaleMin(valueProperty.getMinimum());
		}
		
		connectionDecoration = new ConnectionControlDecoration(this, SWT.LEFT | SWT.TOP, valueProperty);
		
		IObservableValue widget = BeansObservables.observeValue(this, "value");
		IObservableValue value = EMFObservables.observeValue(valueProperty, Literals.PROPERTY__VALUE);
		
		UpdateValueStrategy modelToTarget = new UpdateValueStrategy();
		valueConverter = ConverterFactory.createScalingConverter(valueProperty.getValue().getClass(), Float.class, getValueIndex(), valueProperty, scaleMin * factor, scaleMax * factor);
		modelToTarget.setConverter(valueConverter);
		
		dataBindingContext.bindValue(widget, value, null, modelToTarget);
	
		// Binde die Markers
		warningLowerProperty = client.findProperty(getPathWarningLower());
		warningUpperProperty = client.findProperty(getPathWarningUpper());
		faultLowerProperty = client.findProperty(getPathFaultLower());
		faultUpperProperty = client.findProperty(getPathFaultUpper());
		
		bindMarker(client, BarGraph.Marker.WARNING_LOWER, getWarningLowerProperty(), "warningLower", modelToTarget);
		bindMarker(client, BarGraph.Marker.WARNING_UPPER, getWarningUpperProperty(), "warningUpper", modelToTarget);
		bindMarker(client, BarGraph.Marker.FAULT_LOWER, getFaultLowerProperty(), "faultLower", modelToTarget);
		bindMarker(client, BarGraph.Marker.FAULT_UPPER, getFaultUpperProperty(), "faultUpper", modelToTarget);
	}

	@Override
	public Property getWarningLowerProperty() {
		return this.warningLowerProperty;
	}

	@Override
	public Property getWarningUpperProperty() {
		return this.warningUpperProperty;
	}

	@Override
	public Property getFaultLowerProperty() {
		return this.faultLowerProperty;
	}

	@Override
	public Property getFaultUpperProperty() {
		return this.faultUpperProperty;
	}

	public void setCaption(String caption) {
		this.caption = caption;
	}
	
	@Override
	public String getCaption() {
		return this.caption;
	}

	public void setDescription(String description) {
		this.description = description;
	}
	
	@Override
	public String getDescription() {
		return description;
	}

	@Override
	public String getUnit() {
		return labelBarGraph.getUnit();
	}

	public int getScaleMarker() {
		if(labelBarGraph != null)
			scaleMarker = labelBarGraph.getMarkerScale();
		return scaleMarker;
	}

	public void setScaleMarker(int scaleMarker) {
		this.scaleMarker = scaleMarker;
		if(labelBarGraph != null)
			labelBarGraph.setMarkerScale(scaleMarker);
	}
}
