package ch.thomson.ucs.tine.databinding.swt;

import org.eclipse.core.databinding.DataBindingContext;
import org.eclipse.core.databinding.UpdateValueStrategy;
import org.eclipse.core.databinding.observable.Realm;
import org.eclipse.core.databinding.observable.value.IObservableValue;
import org.eclipse.emf.databinding.EMFObservables;
import org.eclipse.jface.databinding.swt.SWTObservables;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.DisposeEvent;
import org.eclipse.swt.events.DisposeListener;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;

import ch.thomson.ucs.security.LogonLevelListener;
import ch.thomson.ucs.security.LogonManager;
import ch.thomson.ucs.tine.client.Client;
import ch.thomson.ucs.tine.client.Property;
import ch.thomson.ucs.tine.client.ClientPackage.Literals;
import ch.thomson.ucs.tine.client.databinding.converters.CheckIndexedBooleanToFloat;
import ch.thomson.ucs.tine.client.databinding.converters.CheckIndexedFloatToBoolean;
import ch.thomson.ucs.tine.client.databinding.converters.ConverterFactory;
import ch.thomson.ucs.tine.client.databinding.converters.IndexedConverter;

public class ThCheckButton extends Button implements IDataBinding, LogonLevelListener {
	private String signalName;
	
	private Property property = null;
	protected DataBindingContext bindingContext = new DataBindingContext();
	protected ConnectionControlDecoration connnectionDecoration;
	protected IndexedConverter modelToTargetConverter;
	protected IndexedConverter targetToModelConverter;
	protected IndexedConverter indexedConverter;	
	protected int signalIndex;
	protected Property internalProperty;
	protected String maskValue;

	public ThCheckButton(Composite parent, int style) {
		super(parent, SWT.CHECK);
		LogonManager.getInstance().addLogonLevelListener(this);
	}

	public String getSignalName() {
		return signalName;
	}

	public void setSignalName(String signalName) {
		this.signalName = signalName;
	}
	
	public int getSignalIndex() {
		return signalIndex;
	}

	public void setSignalIndex(int signalIndex) {
		this.signalIndex = signalIndex;
		if(modelToTargetConverter != null)
			modelToTargetConverter.setIndex(signalIndex);
		
		if(targetToModelConverter != null)
			targetToModelConverter.setIndex(signalIndex);
		
		bindingContext.updateTargets();
	}

	@Override
	protected void checkSubclass() {}

	@Override
	public void bindValues(Client client) throws Exception {
		property = client.findProperty(signalName);
		this.internalProperty = property;
		this.connnectionDecoration = new ConnectionControlDecoration(this, SWT.LEFT | SWT.TOP, property);
		
		this.setEnabled(property.isWritable());

		IObservableValue observeWidget = SWTObservables.observeSelection(this);
		IObservableValue observeValue = EMFObservables.observeValue(property, Literals.PROPERTY__VALUE);

		UpdateValueStrategy targetToModel = new UpdateValueStrategy();
		targetToModelConverter = ConverterFactory.createCheckIndexedConverter(Boolean.class, property.getValue().getClass(), getSignalIndex(), property);
		((CheckIndexedBooleanToFloat)targetToModelConverter).setMaskValue(Integer.decode(maskValue));
		targetToModel.setConverter(targetToModelConverter);
		
		UpdateValueStrategy modelToTarget = new UpdateValueStrategy();
		modelToTargetConverter = (CheckIndexedFloatToBoolean) ConverterFactory.createCheckIndexedConverter(property.getValue().getClass(), Boolean.class, signalIndex, property);
		((CheckIndexedFloatToBoolean)modelToTargetConverter).setMaskValue(Integer.decode(maskValue));
		modelToTarget.setConverter(modelToTargetConverter);

		bindingContext.bindValue(observeWidget, observeValue, targetToModel, modelToTarget);
		
		this.addDisposeListener(new DisposeListener() {
			@Override
			public void widgetDisposed(DisposeEvent e) {
				bindingContext.dispose();
				LogonManager.getInstance().removeLogonLevelListener(ThCheckButton.this);
			}
		});	
		logonLevelChanged(LogonManager.getInstance().getLogonLevel());
	}

	public String getMaskValue() {
		return maskValue;
	}

	public void setMaskValue(String maskValue) {
		this.maskValue = maskValue;
	}
	@Override
	public void logonLevelChanged(int logonLevel) {
		if(property != null) {
			if(property.getLogonLevel() > logonLevel) {
				Realm.getDefault().asyncExec(new Runnable() {
					@Override
					public void run() {
						if(!ThCheckButton.this.isDisposed())
							ThCheckButton.this.setEnabled(false);
					}
				});
			}
			else {
				Realm.getDefault().asyncExec(new Runnable() {
					@Override
					public void run() {
						if(!ThCheckButton.this.isDisposed())
							ThCheckButton.this.setEnabled(true);
					}
				});
			}
		}
	}	
}
