package ch.thomson.ucs.tine.databinding.swt;

import org.eclipse.core.databinding.observable.Realm;
import org.eclipse.emf.common.notify.Notification;
import org.eclipse.emf.common.notify.impl.AdapterImpl;
import org.eclipse.jface.resource.JFaceResources;
import org.eclipse.jface.resource.LocalResourceManager;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Label;
import org.eclipse.wb.swt.ResourceManager;

import ch.thomson.ucs.tine.client.Client;
import ch.thomson.ucs.tine.client.Property;
import ch.thomson.ucs.tine.client.ClientPackage.Literals;

public class ThLedSummary extends Label implements IDataBinding {

	private String[] signalList;
	private String[] maskValues;
	private String[] invertedList;
	private String[] valueCondList;
	
	private ThLedSummary nextLed;
	private ThLedSummary prevLed;

	private Entry[] entries;
	private boolean[] values;
	private Realm realm;

	protected Image greenLed;
	protected Image redLed;
	
	LocalResourceManager resources;	
	
	public class Entry {
		private int id;
		
		public Entry(Property property, int id) {
			this.id = id;
			
			property.eAdapters().add(new AdapterImpl() {
				@Override
				public void notifyChanged(final Notification msg) {
					if (Literals.PROPERTY__VALUE == msg.getFeature() && !msg.isTouch()) {
						realm.asyncExec(new Runnable() {

							@Override
							public void run() {
								if(ThLedSummary.this.isDisposed())
									return;
								
								// set value in array
								float[] fVal = (float[]) msg.getNewValue();
								int val = (int) fVal[0];
								int maskVal = Integer.decode(maskValues[Entry.this.id]);
								boolean inv = Boolean.parseBoolean(invertedList[Entry.this.id]);
								boolean cond = Boolean.parseBoolean(valueCondList[Entry.this.id]);
																
								if(cond) {
									values[Entry.this.id] = ((inv == true) &&  (val != maskVal)) || 
															((inv == false) && (val == maskVal));
								}
								else {
									values[Entry.this.id] = ((inv == true) &&  ((val & maskVal) == 0)) || 
											((inv == false) && ((val & maskVal) == maskVal));
								}
								updateLed();
							}
						});
					}
				}
			});
		}		
	}

	public ThLedSummary(Composite parent, int style) {
		super(parent, style);
		resources = new LocalResourceManager(JFaceResources.getResources(), this);
		greenLed = ResourceManager.getPluginImage("ch.thomson.ucs.tine.client.databinding", "icons/LedGreen32x32.png");
		redLed = ResourceManager.getPluginImage("ch.thomson.ucs.tine.client.databinding", "icons/LedRed32x32.png");
		this.setImage(redLed);
		realm = Realm.getDefault();
	}
	
	public void updateLed() {
		if(this.isSummaryValue() == false)
			this.setImage(redLed);
		else
			this.setImage(greenLed);
		if(nextLed != null)
			nextLed.updateLed();
	}
	
	public boolean isSummaryValue() {
		if((prevLed != null) && (prevLed.isSummaryValue() == false))
			return false;
		if(values != null) {
			for(int i=0; i<values.length; i++) {
				if(values[i] == false)
					return false;
			}
		}
		return true;
	}
	

	public String[] getSignalList() {
		return signalList;
	}

	public void setSignalList(String[] signalList) {
		this.signalList = signalList;
	}

	public ThLedSummary getNextLed() {
		return nextLed;
	}

	public void setNextLed(ThLedSummary nextLed) {
		this.nextLed = nextLed;
	}

	public ThLedSummary getPrevLed() {
		return prevLed;
	}

	public void setPrevLed(ThLedSummary prevLed) {
		this.prevLed = prevLed;
	}

	public String[] getMaskValues() {
		return maskValues;
	}

	public void setMaskValues(String[] maskValues) {
		this.maskValues = maskValues;
	}

	public String[] getInvertedList() {
		return invertedList;
	}

	public void setInvertedList(String[] invertedList) {
		this.invertedList = invertedList;
	}
	
	public String[] getValueCondList() {
		return valueCondList;
	}

	public void setValueCondList(String[] valueCondList) {
		this.valueCondList = valueCondList;
	}	
	
	@Override
	protected void checkSubclass() {}	
	
	@Override
	public void bindValues(Client client) throws Exception {
		if((signalList == null) || (signalList.length == 0)) {
			this.setImage(greenLed);
		}
		else {			
			entries = new Entry[signalList.length];
			values = new boolean[signalList.length];
			
			for(int i=0; i<signalList.length; i++) {
				Property property = client.findProperty(signalList[i]);
				float[] fVal = (float[]) property.getValue();
				int val = (int) fVal[0];
				int maskVal = Integer.decode(maskValues[i]);
				boolean inv = Boolean.parseBoolean(invertedList[i]);
				boolean cond = Boolean.parseBoolean(valueCondList[i]);
				
				if(cond) {
					values[i] = ((inv == true) &&  (val != maskVal)) || 
								((inv == false) && (val == maskVal));
				}
				else {
					values[i] = ((inv == true) &&  ((val & maskVal) == 0)) || 
								((inv == false) && ((val & maskVal) == maskVal));
				}				
				entries[i] = new Entry(property, i);
			}
			updateLed();
		}
	}
	
}
