package ch.thomson.ucs.tine.databinding.swt;


import org.eclipse.core.databinding.DataBindingContext;
import org.eclipse.core.databinding.UpdateValueStrategy;
import org.eclipse.core.databinding.observable.Realm;
import org.eclipse.core.databinding.observable.value.IObservableValue;
import org.eclipse.emf.databinding.EMFObservables;
import org.eclipse.jface.databinding.swt.SWTObservables;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.DisposeEvent;
import org.eclipse.swt.events.DisposeListener;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Spinner;

import ch.thomson.ucs.security.LogonLevelListener;
import ch.thomson.ucs.security.LogonManager;
import ch.thomson.ucs.tine.client.Client;
import ch.thomson.ucs.tine.client.ClientPackage.Literals;
import ch.thomson.ucs.tine.client.Property;
import ch.thomson.ucs.tine.client.databinding.converters.ConverterFactory;
import ch.thomson.ucs.tine.client.databinding.converters.IndexedConverter;


public class ThSpinner extends Spinner implements IDataBinding, LogonLevelListener {
	private String signalName;
	private DataBindingContext bindingContext = new DataBindingContext();
	private Property property = null;
	@SuppressWarnings("unused")
	private ConnectionControlDecoration connnectionDecoration;
	protected IndexedConverter modelToTargetConverter;
	protected IndexedConverter targetToModelConverter;
	protected int signalIndex;

	
	public ThSpinner(Composite parent) {
		this(parent, SWT.BORDER);
		LogonManager.getInstance().addLogonLevelListener(this);
	}
	
	public ThSpinner(Composite parent, int style) {
		super(parent, style);
		LogonManager.getInstance().addLogonLevelListener(this);
	}

	public String getSignalName() {
		return signalName;
	}

	public void setSignalName(String signalName) {
		this.signalName = signalName;
	}

	public int getSignalIndex() {
		return signalIndex;
	}

	public void setSignalIndex(int signalIndex) {
		this.signalIndex = signalIndex;
		if(modelToTargetConverter != null)
			modelToTargetConverter.setIndex(signalIndex);
		
		if(targetToModelConverter != null)
			targetToModelConverter.setIndex(signalIndex);
		
		bindingContext.updateTargets();
	}

	@Override
	protected void checkSubclass() {}

	@Override
	public void bindValues(Client client) throws Exception {
		
		property = client.findProperty(signalName);
		
		int factor = (int) Math.pow(10.0, property.getDecimalPlaces());
		this.setMaximum((int) (property.getMaximum() * factor));
		this.setMinimum((int) (property.getMinimum() * factor));		
		this.setDigits(property.getDecimalPlaces());
		this.setEnabled(property.isWritable());
		this.setIncrement(1); // TODO Increment-Berechnung automatisieren oder in Config?
//		this.pack();
		
		this.connnectionDecoration = new ConnectionControlDecoration(this, SWT.LEFT | SWT.TOP, property);
		
		IObservableValue spinnerObserveSelectionObserveWidget = SWTObservables.observeSelection(this);
		IObservableValue dataTestObjectObserveValue = EMFObservables.observeValue(client.findProperty(signalName), Literals.PROPERTY__VALUE);
		
		UpdateValueStrategy targetToModel = new UpdateValueStrategy();
		targetToModelConverter = ConverterFactory.createSpinnerIndexedConverter(Integer.class, property.getValue().getClass(), getSignalIndex(), property);
		targetToModel.setConverter(targetToModelConverter);
		
		UpdateValueStrategy modelToTarget = new UpdateValueStrategy();
		modelToTargetConverter = ConverterFactory.createSpinnerIndexedConverter(property.getValue().getClass(), Integer.class, getSignalIndex(), property);
		modelToTarget.setConverter(modelToTargetConverter);
		
		bindingContext.bindValue(spinnerObserveSelectionObserveWidget, dataTestObjectObserveValue, targetToModel, modelToTarget);
			
		this.addDisposeListener(new DisposeListener() {
			@Override
			public void widgetDisposed(DisposeEvent e) {
				bindingContext.dispose();
				LogonManager.getInstance().removeLogonLevelListener(ThSpinner.this);				
			}
		});
		logonLevelChanged(LogonManager.getInstance().getLogonLevel());		
	}
	@Override
	public void logonLevelChanged(int logonLevel) {
		if(property != null) {
			if(property.getLogonLevel() > logonLevel) {
				Realm.getDefault().asyncExec(new Runnable() {
					@Override
					public void run() {
						if(!ThSpinner.this.isDisposed())
							ThSpinner.this.setEnabled(false);
					}
				});
			}
			else {
				Realm.getDefault().asyncExec(new Runnable() {
					@Override
					public void run() {
						if(!ThSpinner.this.isDisposed())
							ThSpinner.this.setEnabled(true);
					}
				});
			}
		}
	}	
}
