package ch.thomson.ucs.tine.databinding.swt;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.viewers.ArrayContentProvider;
import org.eclipse.jface.viewers.ColumnLabelProvider;
import org.eclipse.jface.viewers.TableViewer;
import org.eclipse.jface.viewers.TableViewerColumn;
import org.eclipse.jface.viewers.Viewer;
import org.eclipse.swt.SWT;
import org.eclipse.swt.layout.FillLayout;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Table;
import org.eclipse.swt.widgets.TableColumn;
import org.eclipse.ui.progress.UIJob;

import ch.thomson.ucs.alarms.client.ClientUtil;
import ch.thomson.ucs.alarms.client.dao.AlarmHistoryDAO;
import ch.thomson.ucs.alarms.client.dao.entries.AlarmHistoryEntry;
import ch.thomson.ucs.tine.databinding.swt.utils.TableColumnSorter;

public class ThTableAlarmsHistory extends ThComposite {
	private Table table;
	private TableViewer tableViewer;
	private AlarmHistoryDAO dao;
	
	public enum HistoryAlarmColumnTypes {
		TIMESTAMP("Timestamp", 200),
		ID("ID", 100),
		EQUIPMENT("Equipment", 150),
		CATEGORY("Category", 100),
		DESCRIPTOR("Descriptor", 100),
		MESSAGE("Message", 500);
		
		private final String name;
		private final int width;
		
		private HistoryAlarmColumnTypes(String name, int width) {
			this.name = name;
			this.width = width;
		}
		
		public String getName() {return name;}
		public int getWidth() {return width;}
	}

	public ThTableAlarmsHistory(Composite parent, int style) {
		super(parent, SWT.NONE);
		setLayout(new FillLayout(SWT.HORIZONTAL));
		dao = ClientUtil.createAlarmHistoryDAO();
		
		tableViewer = new TableViewer(this, SWT.BORDER | SWT.VIRTUAL);
		table = tableViewer.getTable();
		table.setLinesVisible(true);
		table.setHeaderVisible(true);
		
		for(HistoryAlarmColumnTypes type : HistoryAlarmColumnTypes.values()) {
			TableViewerColumn tvc = new TableViewerColumn(tableViewer, SWT.LEFT);
			TableColumn tc = tvc.getColumn();			
			tc.setText(type.getName());
			tc.setWidth(type.getWidth());
			tc.setResizable(false);
			initTableColumn(type, tvc);
		}

		tableViewer.setContentProvider(ArrayContentProvider.getInstance());
	}
	
	public void refresh() {
		UIJob uiJob = new UIJob("Update History Log") {
			
			@Override
			public IStatus runInUIThread(IProgressMonitor monitor) {
				try {
					dao.open();
					tableViewer.setInput(dao.getAllEntries());
				} catch (Exception e) {
					MessageDialog.openError(null, "Error", "Read History DB failed: " + e.getMessage());
					return Status.CANCEL_STATUS;
				} finally {
					dao.close();
				}
				return Status.OK_STATUS;
			}
		};
		uiJob.schedule();
	}

	private void initTableColumn(HistoryAlarmColumnTypes type, TableViewerColumn tvc) {
		
		switch(type) {
		case TIMESTAMP:
			// Sortierer erzeugen
			new TableColumnSorter(tableViewer, tvc.getColumn()) {

				@Override
				protected int doCompare(Viewer TableViewer, Object e1, Object e2) {
					AlarmHistoryEntry ape1 = (AlarmHistoryEntry) e1;
					AlarmHistoryEntry ape2 = (AlarmHistoryEntry) e2;
					return ape1.getTimestamp().compareTo(ape2.getTimestamp());
				}
			};
			// ColumnLabelProvider erzeugen
			tvc.setLabelProvider(new ColumnLabelProvider() {
				@Override
				public String getText(Object element) {
					return ((AlarmHistoryEntry) element).getTimestamp().toString();
				}
			});
			break;
		case ID:
			// Sortierer erzeugen
			new TableColumnSorter(tableViewer, tvc.getColumn()) {

				@Override
				protected int doCompare(Viewer TableViewer, Object e1, Object e2) {
					AlarmHistoryEntry ape1 = (AlarmHistoryEntry) e1;
					AlarmHistoryEntry ape2 = (AlarmHistoryEntry) e2;
					return ape1.getId().compareTo(ape2.getId());
				}
			};
			// ColumnLabelProvider erzeugen
			tvc.setLabelProvider(new ColumnLabelProvider() {
				@Override
				public String getText(Object element) {
					return ((AlarmHistoryEntry) element).getId();
				}
			});
			break;
		case EQUIPMENT:
			// Sortierer erzeugen
			new TableColumnSorter(tableViewer, tvc.getColumn()) {

				@Override
				protected int doCompare(Viewer TableViewer, Object e1, Object e2) {
					AlarmHistoryEntry ape1 = (AlarmHistoryEntry) e1;
					AlarmHistoryEntry ape2 = (AlarmHistoryEntry) e2;
					return ape1.getEqm().compareTo(ape2.getEqm());
				}
			};
			// ColumnLabelProvider erzeugen
			tvc.setLabelProvider(new ColumnLabelProvider() {
				@Override
				public String getText(Object element) {
					return ((AlarmHistoryEntry) element).getEqm();
				}
			});
			break;
		case CATEGORY:
			// Sortierer erzeugen
			new TableColumnSorter(tableViewer, tvc.getColumn()) {

				@Override
				protected int doCompare(Viewer TableViewer, Object e1, Object e2) {
					AlarmHistoryEntry ape1 = (AlarmHistoryEntry) e1;
					AlarmHistoryEntry ape2 = (AlarmHistoryEntry) e2;
					return ape1.getCategory().compareTo(ape2.getCategory());
				}
			};
			// ColumnLabelProvider erzeugen
			tvc.setLabelProvider(new ColumnLabelProvider() {
				@Override
				public String getText(Object element) {
					return ((AlarmHistoryEntry) element).getCategory();
				}
			});
			break;
		case DESCRIPTOR:
			// Sortierer erzeugen
			new TableColumnSorter(tableViewer, tvc.getColumn()) {

				@Override
				protected int doCompare(Viewer TableViewer, Object e1, Object e2) {
					AlarmHistoryEntry ape1 = (AlarmHistoryEntry) e1;
					AlarmHistoryEntry ape2 = (AlarmHistoryEntry) e2;
					return ape1.getDescriptor().compareTo(ape2.getDescriptor());
				}
			};
			// ColumnLabelProvider erzeugen
			tvc.setLabelProvider(new ColumnLabelProvider() {
				@Override
				public String getText(Object element) {
					return ((AlarmHistoryEntry) element).getDescriptor();
				}
			});
			break;
		case MESSAGE:
			// Sortierer erzeugen
			new TableColumnSorter(tableViewer, tvc.getColumn()) {

				@Override
				protected int doCompare(Viewer TableViewer, Object e1, Object e2) {
					AlarmHistoryEntry ape1 = (AlarmHistoryEntry) e1;
					AlarmHistoryEntry ape2 = (AlarmHistoryEntry) e2;
					return ape1.getMessage().compareTo(ape2.getMessage());
				}
			};
			// ColumnLabelProvider erzeugen
			tvc.setLabelProvider(new ColumnLabelProvider() {
				@Override
				public String getText(Object element) {
					return ((AlarmHistoryEntry) element).getMessage();
				}
			});
			break;
		}
	}
}
