package ch.thomson.ucs.tine.databinding.swt;

import org.eclipse.core.databinding.observable.Realm;
import org.eclipse.jface.viewers.ArrayContentProvider;
import org.eclipse.jface.viewers.ColumnLabelProvider;
import org.eclipse.jface.viewers.TableViewer;
import org.eclipse.jface.viewers.TableViewerColumn;
import org.eclipse.jface.viewers.Viewer;
import org.eclipse.swt.SWT;
import org.eclipse.swt.layout.FillLayout;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Table;
import org.eclipse.swt.widgets.TableColumn;

import ch.thomson.ucs.alarms.client.dao.entries.AlarmPendingEntry;
import ch.thomson.ucs.alarms.models.AlarmPendingModel;
import ch.thomson.ucs.tine.client.Client;
import ch.thomson.ucs.tine.databinding.swt.utils.TableColumnSorter;

public class ThTableAlarmsPending extends ThComposite implements IDataBinding {
	private Table table;
	private TableViewer tableViewer;
	private Realm realm;
	private int pollingTime = 1000;
	
	public enum PendingAlarmColumnTypes {
		TIMESTAMP("Timestamp", 200),
		ID("ID", 100),
		EQUIPMENT("Equipment", 150),
		CATEGORY("Category", 100),
		DESCRIPTOR("Descriptor", 100),
		MESSAGE("Message", 500);
		
		private final String name;
		private final int width;
		
		private PendingAlarmColumnTypes(String name, int width) {
			this.name = name;
			this.width = width;
		}
		
		public String getName() {return name;}
		public int getWidth() {return width;}
	}

	public ThTableAlarmsPending(Composite parent, int style) {
		super(parent, SWT.NONE);
		setLayout(new FillLayout(SWT.HORIZONTAL));
		
		tableViewer = new TableViewer(this, SWT.BORDER | SWT.VIRTUAL);
		table = tableViewer.getTable();
		table.setLinesVisible(true);
		table.setHeaderVisible(true);
		
		for(PendingAlarmColumnTypes type : PendingAlarmColumnTypes.values()) {
			TableViewerColumn tvc = new TableViewerColumn(tableViewer, SWT.LEFT);
			TableColumn tc = tvc.getColumn();			
			tc.setText(type.getName());
			tc.setWidth(type.getWidth());
			tc.setResizable(false);
			initTableColumn(type, tvc);
		}
		
		tableViewer.setContentProvider(ArrayContentProvider.getInstance());
	}
	
	private void refreshInput() {
		try {
			tableViewer.setInput(AlarmPendingModel.getInstance().getEntries());
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	@Override
	public void bindValues(Client client) throws Exception {
		refreshInput();
		
		// Automatisches Refresh anmelden
		realm = Realm.getDefault();
		realm.timerExec(pollingTime, new Runnable() {
			@Override
			public void run() {
				if(table.isDisposed())
					return;
				refreshInput();
				realm.timerExec(pollingTime, this);
			}
		});
	}

	private void initTableColumn(PendingAlarmColumnTypes type, TableViewerColumn tvc) {
		
		switch(type) {
		case TIMESTAMP:
			// Sortierer erzeugen
			new TableColumnSorter(tableViewer, tvc.getColumn()) {

				@Override
				protected int doCompare(Viewer TableViewer, Object e1, Object e2) {
					AlarmPendingEntry ape1 = (AlarmPendingEntry) e1;
					AlarmPendingEntry ape2 = (AlarmPendingEntry) e2;
					return ape1.getTimestamp().compareTo(ape2.getTimestamp());
				}
			};
			// ColumnLabelProvider erzeugen
			tvc.setLabelProvider(new ColumnLabelProvider() {
				@Override
				public String getText(Object element) {
					return ((AlarmPendingEntry) element).getTimestamp().toString();
				}
			});
			break;
		case ID:
			// Sortierer erzeugen
			new TableColumnSorter(tableViewer, tvc.getColumn()) {

				@Override
				protected int doCompare(Viewer TableViewer, Object e1, Object e2) {
					AlarmPendingEntry ape1 = (AlarmPendingEntry) e1;
					AlarmPendingEntry ape2 = (AlarmPendingEntry) e2;
					return ape1.getId().compareTo(ape2.getId());
				}
			};
			// ColumnLabelProvider erzeugen
			tvc.setLabelProvider(new ColumnLabelProvider() {
				@Override
				public String getText(Object element) {
					return ((AlarmPendingEntry) element).getId();
				}
			});
			break;
		case EQUIPMENT:
			// Sortierer erzeugen
			new TableColumnSorter(tableViewer, tvc.getColumn()) {

				@Override
				protected int doCompare(Viewer TableViewer, Object e1, Object e2) {
					AlarmPendingEntry ape1 = (AlarmPendingEntry) e1;
					AlarmPendingEntry ape2 = (AlarmPendingEntry) e2;
					return ape1.getEqm().compareTo(ape2.getEqm());
				}
			};
			// ColumnLabelProvider erzeugen
			tvc.setLabelProvider(new ColumnLabelProvider() {
				@Override
				public String getText(Object element) {
					return ((AlarmPendingEntry) element).getEqm();
				}
			});
			break;
		case CATEGORY:
			// Sortierer erzeugen
			new TableColumnSorter(tableViewer, tvc.getColumn()) {

				@Override
				protected int doCompare(Viewer TableViewer, Object e1, Object e2) {
					AlarmPendingEntry ape1 = (AlarmPendingEntry) e1;
					AlarmPendingEntry ape2 = (AlarmPendingEntry) e2;
					return ape1.getCategory().compareTo(ape2.getCategory());
				}
			};
			// ColumnLabelProvider erzeugen
			tvc.setLabelProvider(new ColumnLabelProvider() {
				@Override
				public String getText(Object element) {
					return ((AlarmPendingEntry) element).getCategory();
				}
			});
			break;
		case DESCRIPTOR:
			// Sortierer erzeugen
			new TableColumnSorter(tableViewer, tvc.getColumn()) {

				@Override
				protected int doCompare(Viewer TableViewer, Object e1, Object e2) {
					AlarmPendingEntry ape1 = (AlarmPendingEntry) e1;
					AlarmPendingEntry ape2 = (AlarmPendingEntry) e2;
					return ape1.getDescriptor().compareTo(ape2.getDescriptor());
				}
			};
			// ColumnLabelProvider erzeugen
			tvc.setLabelProvider(new ColumnLabelProvider() {
				@Override
				public String getText(Object element) {
					return ((AlarmPendingEntry) element).getDescriptor();
				}
			});
			break;
		case MESSAGE:
			// Sortierer erzeugen
			new TableColumnSorter(tableViewer, tvc.getColumn()) {

				@Override
				protected int doCompare(Viewer TableViewer, Object e1, Object e2) {
					AlarmPendingEntry ape1 = (AlarmPendingEntry) e1;
					AlarmPendingEntry ape2 = (AlarmPendingEntry) e2;
					return ape1.getMessage().compareTo(ape2.getMessage());
				}
			};
			// ColumnLabelProvider erzeugen
			tvc.setLabelProvider(new ColumnLabelProvider() {
				@Override
				public String getText(Object element) {
					return ((AlarmPendingEntry) element).getMessage();
				}
			});
			break;
		}
	}
}
