package ch.thomson.ucs.tine.databinding.swt;

import java.util.ArrayList;
import java.util.List;

import org.eclipse.core.databinding.observable.Realm;
import org.eclipse.jface.viewers.ArrayContentProvider;
import org.eclipse.jface.viewers.ColumnLabelProvider;
import org.eclipse.jface.viewers.TableViewer;
import org.eclipse.jface.viewers.TableViewerColumn;
import org.eclipse.swt.SWT;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.layout.FillLayout;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Table;
import org.eclipse.swt.widgets.TableColumn;
import org.eclipse.wb.swt.SWTResourceManager;

import ch.thomson.ucs.tine.client.Client;
import ch.thomson.ucs.tine.client.Property;

public class ThTableViewer extends ThComposite implements IDataBinding {
	private Table table;
	private TableViewer tableViewer;
	private Realm realm;
	
	private String[] signalList;
	private String[] signalCaptions;
	private List<float[]> floatArray;
	private int index;
	private int lineCount = 0;
	
	private Client tineClient = null;
	private int cycle_time = 500;	// in ms
	
	private static final int ID_COLOR_GREY = 	0x000;
	private static final int ID_COLOR_GREEN = 	0x100;
	private static final int ID_COLOR_RED = 	0x200;
	private static final int ID_COLOR_ORANGE = 	0x300;

	public ThTableViewer(Composite parent, int style) {
		super(parent, SWT.NONE);
		setLayout(new FillLayout(SWT.VERTICAL));
		
		floatArray = new ArrayList<float[]>();
		
		tableViewer = new TableViewer(this, SWT.BORDER | SWT.VIRTUAL);
		table = tableViewer.getTable();
		table.setLinesVisible(true);
		table.setHeaderVisible(true);
		table.setFont(SWTResourceManager.getFont("Tahoma", 7, SWT.NORMAL));
		
		realm = Realm.getDefault();
	}
	
	public void setSignalList(String[] signalList) {
		this.signalList = signalList;
	}
	
	public String[] getSignalList() {
		return this.signalList;
	}
	
	public void setSignalCaptions(String[] signalCaptions) {
		this.signalCaptions = signalCaptions;
	}
	
	public String[] getSignalCaptions() {
		return this.signalCaptions;
	}

	@Override
	protected void checkSubclass() {}
	
	@Override
	public void bindValues(Client client) throws Exception {		
		tineClient = client;
		for(index=0; index<signalCaptions.length; index++) {
			TableViewerColumn tvc = new TableViewerColumn(tableViewer, SWT.LEFT);
			TableColumn tc = tvc.getColumn();
			tc.setResizable(true);
			if(index == 0)
			{
				tc.setWidth(30);
				tc.setAlignment(SWT.CENTER);

				String txt = client.findProperty(signalList[index]).getUnit();
				if(("none".contentEquals(txt) == false) && ("1".contentEquals(txt) == false))
					tc.setText(signalCaptions[index] + " [" + txt + "]");
				else
					tc.setText(signalCaptions[index]);


				tvc.setLabelProvider(new ColumnLabelProvider() {
					@Override
					public Color getBackground(Object element) {
						float [] array = (float[]) element;
						int val = ((int)array[0] & 0xF00); // Farb ID ausmaskieren
						switch(val) {
						case ID_COLOR_GREY:
							return SWTResourceManager.getColor(SWT.COLOR_GRAY);
						case ID_COLOR_GREEN:
							return SWTResourceManager.getColor(SWT.COLOR_GREEN);
						case ID_COLOR_ORANGE:
							return SWTResourceManager.getColor(255, 165, 0);
						case ID_COLOR_RED:
							return SWTResourceManager.getColor(SWT.COLOR_RED);
						}
						
						return SWTResourceManager.getColor(SWT.COLOR_GRAY);
					}

					@Override
					public String getText(Object element) {
						float [] array = (float[]) element;
						return Integer.toString((int) array[0] & 0xFF);
					}
				});	
			}
			else
			{
				String txt = client.findProperty(signalList[index]).getUnit();
				if(("none".contentEquals(txt) == false) && ("1".contentEquals(txt) == false))
					tc.setText(signalCaptions[index] + " [" + txt + "]");
				else
					tc.setText(signalCaptions[index]);
//				tc.setWidth(98);
				tc.setWidth(75);
				tc.setAlignment(SWT.RIGHT);
						
				tvc.setLabelProvider(new ColumnLabelProvider() {
					int i = index;
					@Override
					public String getText(Object element) {
						float [] array = (float[]) element;
						return Float.toString(array[i]);
					}
				});			
			}
		}
		
		if(signalList.length > 0)
		{
			float[] fArr  = (float[]) client.findProperty(signalList[0]).getValue();
			lineCount = fArr.length;
		}
		for(int j=0; j<lineCount; j++)
		{
			float[] fLine = new float[signalList.length];
			floatArray.add(fLine);
		}
		tableViewer.setContentProvider(new ArrayContentProvider());
		tableViewer.setInput(floatArray);		
		
		Runnable timer = new Runnable() {

			@Override
			public void run() {
				if((ThTableViewer.this.isDisposed() == false) && ThTableViewer.this.isVisible()) {			
					// read tine variables 
					float[] fArr = null;
					for(int i=0; i<signalList.length; i++) {
						Property property = null;
						try {
							property = tineClient.findProperty(signalList[i]);
						} catch (Exception e) {
							// TODO Auto-generated catch block
							e.printStackTrace();
						}
						float[] floatPropertyArray = (float[]) property.getValue();
						if(i == 0)
						{
							for(int j=0; j<lineCount; j++)
							{
								fArr = floatArray.get(j);
								int val = j + 1;
								
								int propValue = (int) floatPropertyArray[j];
								
								if ((propValue & 0x7) != 0) // Hintere 3 Bits mind. 1 Bit gesetzt?
									val += ID_COLOR_RED;
								else if((propValue & 0x38) == 0) // Vordere 3 Bits keines gesetzt?
									val += ID_COLOR_GREY;
								else if ((propValue & 0x38) == 0x38) // Vordere 3 Bits alle gesetzt?
									val += ID_COLOR_GREEN;
								else
									val += ID_COLOR_ORANGE;
								fArr[i] = val;
							}
						}
						else
						{
							for(int j=0; j<lineCount; j++)
							{
								fArr = floatArray.get(j);
								fArr[i] = floatPropertyArray[j];
							}
						}
					}
					tableViewer.refresh();
				}
				realm.timerExec(cycle_time,this);
			}
		};
		realm.exec(timer);
	}
}
