package ch.thomson.ucs.tine.databinding.swt;


import org.eclipse.core.databinding.observable.Realm;
import org.eclipse.emf.common.notify.Adapter;
import org.eclipse.emf.common.notify.Notification;
import org.eclipse.emf.common.notify.impl.AdapterImpl;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.DisposeEvent;
import org.eclipse.swt.events.DisposeListener;
import org.eclipse.swt.events.FocusAdapter;
import org.eclipse.swt.events.FocusEvent;
import org.eclipse.swt.events.KeyAdapter;
import org.eclipse.swt.events.KeyEvent;
import org.eclipse.swt.events.MouseAdapter;
import org.eclipse.swt.events.MouseEvent;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Text;
import org.eclipse.wb.swt.SWTResourceManager;

import ch.thomson.ucs.security.LogonLevelListener;
import ch.thomson.ucs.security.LogonManager;
import ch.thomson.ucs.tine.client.Client;
import ch.thomson.ucs.tine.client.ClientPackage.Literals;
import ch.thomson.ucs.tine.client.Property;
import ch.thomson.ucs.tine.client.PropertyText;
import ch.thomson.ucs.tine.databinding.swt.dialogs.NumericPad;

public class ThText extends Text implements IDataBinding, INumKey, LogonLevelListener {
	
	private String signalName;
	private int signalIndex;
	private Property property = null;
	@SuppressWarnings("unused")
	private ConnectionControlDecoration connectionDecoration = null;
	private NumericPad numericPad = null; 
	
	private Realm realm;

	
	public ThText(Composite parent, int style) {
		super(parent, style);
		realm = Realm.getDefault();
		
		this.addMouseListener(new MouseAdapter() {
			@Override
			public void mouseUp(MouseEvent e) {
				if(!ThText.this.isDisposed() && ThText.this.getEditable())
				{
					ThText.this.setSelection(ThText.this.getText().length());
					ThText.this.setSelection(3);
					ThText.this.setBackground(SWTResourceManager.getColor(SWT.COLOR_YELLOW));
					numericPad = NumericPad.instance(null/*ThText.this.getParent().getShell()*/, ThText.this);
					numericPad.open();
				}
				super.mouseUp(e);
			}
		});
		this.addKeyListener(new KeyAdapter() {
			public void keyReleased(KeyEvent event) {
				if(!ThText.this.isDisposed() && ThText.this.getEditable())
				{
					if (event.character == SWT.CR) { 
						setValue(ThText.this.getText());
						ThText.this.setSelection(ThText.this.getText().length());
					}
				}
			}
		});
		this.addFocusListener(new FocusAdapter() {

			@Override
			public void focusLost(FocusEvent e) {
				// TODO Auto-generated method stub
				if(!ThText.this.isDisposed() && ThText.this.getEditable())
				{
					if(property.getLinkErrorCode() != 0)
						property.setValueForce(property.getValue());
					ThText.this.setText(property.getValue());
				}
				super.focusLost(e);
			}

			@Override
			public void focusGained(FocusEvent e) {
				// TODO Auto-generated method stub
				if(!ThText.this.isDisposed() && ThText.this.getEditable())
				{
					ThText.this.setSelection(ThText.this.getText().length());
				}
				super.focusGained(e);
			}
			
		});
		LogonManager.getInstance().addLogonLevelListener(this);
	}


	public String getSignalName() {
		return signalName;
	}

	public void setSignalName(String signalName) {
		this.signalName = signalName;
	}

	public int getSignalIndex() {
		return signalIndex;
	}

	public void setSignalIndex(int signalIndex) {
		this.signalIndex = signalIndex;
		if(property != null) {
			setText(property.getValue());
		}
	}
	
	public void setText(Object val) {
		if(property != null) {
			String sVal = "";
			if(property instanceof PropertyText) {
				sVal = (String)val;
			}
			else {
				float[] fVal = (float[]) val;
				// Bei reinen Integer-Typen soll das lstige Komma (X.0) verschwinden
				if(property.getDecimalPlaces() == 0) 
					sVal = Integer.toString((int)fVal[signalIndex]);
				else
					sVal = Float.toString(fVal[signalIndex]);				
			}
			this.setText(sVal);
		}
	}

	public void setValue(String val) {
		if(property != null) {
			if(property instanceof PropertyText) {
				property.setValueForce(val);
			}
			else {
				float[] gVal = (float[]) property.getValue();
				float[] tmpVal = gVal.clone();

				try {
					tmpVal[signalIndex] = Float.valueOf(val);
				}
				catch (NumberFormatException e) {
					this.setText(property.getValue());
				}
				property.setValueForce(tmpVal);
			}
		}
	}

	@Override
	protected void checkSubclass() {}
	
	private Adapter adapter;

	@Override
	public void bindValues(Client client) throws Exception {
				
		property = client.findProperty(signalName);
		if(this.getEditable())
			this.setEditable(property.isWritable());
		
		connectionDecoration = new ConnectionControlDecoration(this, SWT.LEFT | SWT.TOP, property);
		
		if(property != null) {
			setText(property.getValue());
			adapter = new AdapterImpl() {
				@Override
				public void notifyChanged(final Notification msg) {
					if (Literals.PROPERTY__VALUE == msg.getFeature() && !msg.isTouch()) {
						realm.asyncExec(new Runnable() {
	
							@Override
							public void run() {
								if(!ThText.this.isDisposed()) {
									ThText.this.setText(msg.getNewValue());
									ThText.this.setSelection(ThText.this.getText().length());
								}
							}
						});
					}
				}
			};
			
			property.eAdapters().add(adapter);
			
			this.addDisposeListener(new DisposeListener() {
				@Override
				public void widgetDisposed(DisposeEvent e) {
					property.eAdapters().remove(adapter);
					LogonManager.getInstance().removeLogonLevelListener(ThText.this);
					if(ThText.this.numericPad != null)
						ThText.this.numericPad.close();
				}
			});
			
			logonLevelChanged(LogonManager.getInstance().getLogonLevel());
		}		
	}

	@Override
	public void setActive(boolean val) {
		if(this.isDisposed() == false) {
			if(val == true)
				this.setBackground(SWTResourceManager.getColor(SWT.COLOR_YELLOW));
			else
				this.setBackground(SWTResourceManager.getColor(SWT.COLOR_WHITE));
		}
	}

	@Override
	public void setTextValue(String txt) {
		if(!this.isDisposed()) {
			this.setText(txt);
			this.setSelection(txt.length());
		}
	}

	@Override
	public void logonLevelChanged(int logonLevel) {
		if(property != null) {
			if(property.getLogonLevel() > logonLevel) {
				realm.asyncExec(new Runnable() {
					@Override
					public void run() {
						if(!ThText.this.isDisposed())
							ThText.this.setEnabled(false);
					}
				});
			}
			else {
				realm.asyncExec(new Runnable() {
					@Override
					public void run() {
						if(!ThText.this.isDisposed())
							ThText.this.setEnabled(true);
					}
				});
			}
		}
	}
}
