package ch.thomson.ucs.tine.databinding.swt;

import org.eclipse.core.databinding.DataBindingContext;
import org.eclipse.core.databinding.UpdateValueStrategy;
import org.eclipse.core.databinding.observable.Realm;
import org.eclipse.core.databinding.observable.value.IObservableValue;
import org.eclipse.emf.databinding.EMFObservables;
import org.eclipse.jface.databinding.swt.SWTObservables;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.DisposeEvent;
import org.eclipse.swt.events.DisposeListener;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;

import ch.thomson.ucs.security.LogonLevelListener;
import ch.thomson.ucs.security.LogonManager;
import ch.thomson.ucs.tine.client.Client;
import ch.thomson.ucs.tine.client.ClientPackage.Literals;
import ch.thomson.ucs.tine.client.Property;
import ch.thomson.ucs.tine.client.databinding.converters.ConverterFactory;
import ch.thomson.ucs.tine.client.databinding.converters.IndexedConverter;

public class ThToggleButton extends Button implements IDataBinding, LogonLevelListener{
	protected Property property;
	protected String signalName;
	protected DataBindingContext dbc = new DataBindingContext();
	protected IndexedConverter modelToTargetConverter;
	protected IndexedConverter modelToTargetConverter2;
	protected IndexedConverter targetToModelConverter;
	protected String[] textList;
	protected int signalIndex;
	
	public ThToggleButton(Composite parent) {
		super(parent, SWT.TOGGLE);
		LogonManager.getInstance().addLogonLevelListener(this);
		
		this.addDisposeListener(new DisposeListener() {
			@Override
			public void widgetDisposed(DisposeEvent e) {
				dbc.dispose();
				LogonManager.getInstance().removeLogonLevelListener(ThToggleButton.this);
			}
		});
	}
	
	public String[] getTextList() {
		return textList;
	}

	public void setTextList(String[] textList) {
		this.textList = textList;
	}

	public String getSignalName() {
		return signalName;
	}

	public void setSignalName(String signalName) {
		this.signalName = signalName;
	}

	public int getSignalIndex() {
		return signalIndex;
	}

	public void setSignalIndex(int signalIndex) {
		if(modelToTargetConverter != null)
			modelToTargetConverter.setIndex(signalIndex);
		
		if(targetToModelConverter != null)
			targetToModelConverter.setIndex(signalIndex);
		
		dbc.updateTargets();
	}

	@Override
	protected void checkSubclass() {}

	@Override
	public void logonLevelChanged(final int logonLevel) {
		if(property != null) {
			Realm.getDefault().exec(new Runnable() {
				@Override
				public void run() {
					if(!ThToggleButton.this.isDisposed())
						ThToggleButton.this.setEnabled(!(property.getLogonLevel() > logonLevel));
				}
			});
		}
	}

	@Override
	public void bindValues(Client client) throws Exception {
		this.property = client.findProperty(signalName);
		IObservableValue observeWidget = SWTObservables.observeSelection(this);
		IObservableValue observeValue = EMFObservables.observeValue(property, Literals.PROPERTY__VALUE);
		
		UpdateValueStrategy modelToTarget = new UpdateValueStrategy();
		modelToTargetConverter = ConverterFactory.createIndexedConverter(property.getValue().getClass(), Boolean.class, signalIndex, property);
		modelToTarget.setConverter(modelToTargetConverter);
		
		UpdateValueStrategy targetToModel = new UpdateValueStrategy();
		targetToModelConverter = ConverterFactory.createIndexedConverter(Boolean.class, property.getValue().getClass(), signalIndex, property);
		targetToModel.setConverter(targetToModelConverter);
		
		dbc.bindValue(observeWidget, observeValue, targetToModel, modelToTarget);
		
		observeWidget = SWTObservables.observeText(this);
		modelToTarget = new UpdateValueStrategy();
		modelToTargetConverter2 = ConverterFactory.createStringListIndexedConverter(property.getValue().getClass(), String.class, signalIndex, property, getTextList());		
		modelToTarget.setConverter(modelToTargetConverter2);
		
		dbc.bindValue(observeWidget, observeValue, null, modelToTarget);
		
		logonLevelChanged(LogonManager.getInstance().getLogonLevel());
	}
}
