package ch.thomson.ucs.tine.client.util;

import java.net.InetAddress;
import java.net.UnknownHostException;

import org.eclipse.emf.common.util.URI;

import ch.thomson.ucs.tine.client.Client;
import ch.thomson.ucs.tine.client.ClientFactory;
import ch.thomson.ucs.tine.client.GroupedFloats;
import ch.thomson.ucs.tine.client.PropertyFloat;
import ch.thomson.ucs.tine.client.PropertyTine;
import ch.thomson.ucs.tine.server.Device;
import ch.thomson.ucs.tine.server.EquipmentModule;
import ch.thomson.ucs.tine.server.FrontEndController;
import ch.thomson.ucs.tine.server.util.ServerResourceLoader;
import ch.thomson.ucs.tine.server.util.UCSConfig;

public class SignalListParser {
	private static Client client = null;
	private static URI uriServer;
	
	public static synchronized Client getClientInstance() throws Exception {
		if(client == null) {
			// Server Modell laden
			FrontEndController fec = ServerResourceLoader.getFrontEndControllerInstance();
			
			// Client Modell erzeugen
			ClientFactory clientFactory = ClientFactory.eINSTANCE;
			client = clientFactory.createClient();
		
			// Alle Properties erzeugen
			for (EquipmentModule eqm : fec.getEquipmentModules()) {
				for (ch.thomson.ucs.tine.server.Property sprop : eqm.getProperties()) {
					for (Device dev : eqm.getDevices()) {
						createProperty(fec, eqm, dev, sprop);
					}
				}
			}
			
			client.initialize();
		}
		
		return client;
	}
	
	public static URI getURIServer() throws Exception {
		if(uriServer == null) {
			getClientInstance();
			uriServer = ServerResourceLoader.createURI();
		}
		return uriServer;
	}
	
	private static InetAddress inetAddress;
	
	private static InetAddress getInetAddress() throws UnknownHostException {
		if(inetAddress == null) 
			inetAddress = InetAddress.getLocalHost();
		return inetAddress;
	}
	
	private static PropertyTine createProperty(FrontEndController fec, EquipmentModule eqm, Device dev, ch.thomson.ucs.tine.server.Property sProp ) throws Exception {
		UCSConfig config = UCSConfig.getInstance();
		config.usedFecContextName = fec.getContextName();
		
		// Wenn Ini-File vorhanden, den Context Name von da verwenden
		if(config.hasIniFile()) {
			config.usedFecContextName = config.getContextName();
		}
		
		// Sonst den Context Name vom EMF nehmen und wenn da auch keiner definiert wurde, den Hostnamen verwenden 
		else {
			if(config.usedFecContextName == null || "".equals(config.usedFecContextName)) 
				config.usedFecContextName = getInetAddress().getHostName();
		}
		
		// Export- und Devicename wird evtl. durch Ini-File bestimmt.
		String eqmExportName = config.getEQMExportName(eqm.getExportName());
		String deviceName = config.getDeviceName(dev.getName());
		
		String path = "/" + config.usedFecContextName + "/" + eqmExportName + "/" + deviceName;
		ch.thomson.ucs.tine.client.PropertyTine cProp = null;
		
		String propName = config.getPropertyNamePrefixed(sProp);
		
		try {
			// Property bereits vorhanden? Wenn nicht, Exception fangen und Property erzeugen
			cProp = (PropertyTine) client.findProperty(path + "/" + propName);
			
		} catch (Exception e) {

			ClientFactory clientFactory = ClientFactory.eINSTANCE;

			// Alle nummerischen Properties werden als Float behandelt
			if (sProp instanceof ch.thomson.ucs.tine.server.PropertyNumeric) {
				cProp = clientFactory.createPropertyFloat();
				cProp.setDecimalPlaces(((ch.thomson.ucs.tine.server.PropertyNumeric) sProp).getDecimalPlaces());
			}
			
			// Text Properties erzeugen
			else if(sProp instanceof ch.thomson.ucs.tine.server.PropertyText) {
				cProp = clientFactory.createPropertyText();
			}
			
			// Skalierte Properties werden als Float behandelt
			else if (sProp instanceof ch.thomson.ucs.tine.server.ScaledProperty) {
				cProp = clientFactory.createPropertyFloat();
				cProp.setDecimalPlaces(((ch.thomson.ucs.tine.server.ScaledProperty) sProp).getDecimalPlaces());
			}
			
			// Gruppierte Properties und ihre Mitglieder erzeugen
			else if (sProp instanceof ch.thomson.ucs.tine.server.GroupProperty) {
				ch.thomson.ucs.tine.server.GroupProperty gp = (ch.thomson.ucs.tine.server.GroupProperty) sProp;
				GroupedFloats gf = clientFactory.createGroupedFloats();
				// Eingetragene Properties suchen oder erzeugen, falls noch nicht vorhanden
				for (ch.thomson.ucs.tine.server.Property sp : gp.getProperties()) {	
					ch.thomson.ucs.tine.client.PropertyTine cp = createProperty(sp.getEquipmentModule().getFrontEndController(), sp.getEquipmentModule(), dev, sp); // Rekursion
					gf.getProperties().add((PropertyFloat) cp);
					if(cp.getPropertyGroup() == null)
						cp.setPropertyGroup(gf);
					else
						throw new Exception(cp + " is already member of property group " + cp.getPropertyGroup());
				}

				cProp = gf;
			}
			
			if(cProp == null)
				throw new Exception("Can't create Property \"" + path + "/" + sProp + "\"");
			
			cProp.setName(propName);
			// Die Attribute, welche durch den TINE-Server verfgbar sind, werden hier nicht gesetzt (min, max, unit, ...)
			cProp.setPath(path);			
			cProp.setLogonLevel(sProp.getLogonLevel());
			cProp.setScheduled(sProp.isScheduled());
			cProp.setRefreshRate(sProp.getRefreshRate());
			client.getTineProperties().add(cProp);
		} 
		
		return cProp;
	}
}
