package ch.thomson.ucs.tine.server.driver.modbus;

public class AddressParser {

	public final static int MODBUSTCP_TABLE_OUTPUT_COIL = 0;
	public final static int MODBUSTCP_TABLE_INPUT_COIL = 1;
	public final static int MODBUSTCP_TABLE_INPUT_REGISTER = 3;
	public final static int MODBUSTCP_TABLE_OUTPUT_REGISTER = 4;
	
	public final static int MODBUS_TYPE_DEFAULT = 0;
	public final static int MODBUS_TYPE_WAGO = 1;
	

	private int typeCode;
	private int address;
	private int length;
	private int modbusType = MODBUS_TYPE_DEFAULT;

	public AddressParser() {}
	
	public AddressParser(int modbusType) {
		this.modbusType = modbusType;
	}

	public int getModbusType() {
		return modbusType;
	}

	public void setModbusType(int modbusType) {
		this.modbusType = modbusType;
	}

	public void parseAddress(String arg) throws Exception {
	
		String[] parts = arg.split(":");
		if (parts.length != 3) {
			throw new Exception("Address Parser: Wrong length");
		}

		// Parse Type Code
		try {
			int typeCode = Integer.parseInt(parts[0]);
			if (typeCode == MODBUSTCP_TABLE_OUTPUT_COIL
					|| typeCode == MODBUSTCP_TABLE_INPUT_COIL
					|| typeCode == MODBUSTCP_TABLE_INPUT_REGISTER
					|| typeCode == MODBUSTCP_TABLE_OUTPUT_REGISTER) {

				this.typeCode = typeCode;
			}

			else {
				throw new Exception("Address Parser: Illegal type code");
			}

		} catch (NumberFormatException e) {

			if ("DI".equals(parts[0])) {
				typeCode = MODBUSTCP_TABLE_INPUT_COIL;
			}

			else if ("DO".equals(parts[0])) {
				typeCode = MODBUSTCP_TABLE_OUTPUT_COIL;
			}

			else if ("AI".equals(parts[0])) {
				typeCode = MODBUSTCP_TABLE_INPUT_REGISTER;
			}

			else if ("AO".equals(parts[0])) {
				typeCode = MODBUSTCP_TABLE_OUTPUT_REGISTER;
			}

			else {
				throw new Exception("Illegal type code");
			}
		}

		// Parse address
		try {
			int address = Integer.parseInt(parts[1]);
			this.address = address;
		} catch (NumberFormatException e) {
			throw new Exception("Address Parser: Illegal address");
		}

		// Parse length
		try {
			int length = Integer.parseInt(parts[2]);
			this.length = length;
		} catch (NumberFormatException e) {
			throw new Exception("Address Parser: Illegal length");
		}
	}

	public int getTypeCode() {
		return typeCode;
	}

	public int getAddress() {
		return address;
	}
	
	public int getReadAddress() {
		// Bei Wago werden schreibbare Register auf einer anderen Adresse zurckgelesen (Offset: 0x200)
		if(modbusType == MODBUS_TYPE_WAGO && this.isOutput()) {
			return (getAddress() + 0x200);
		}
		
		return address;
	}

	public int getLength() {
		return length;
	}
	
	public boolean isDiscrete() {
		return typeCode == MODBUSTCP_TABLE_OUTPUT_COIL || typeCode == MODBUSTCP_TABLE_INPUT_COIL;
	}
	
	public boolean isOutput() {
		return typeCode == MODBUSTCP_TABLE_OUTPUT_COIL || typeCode == MODBUSTCP_TABLE_OUTPUT_REGISTER;
	}
}
