package ch.thomson.ucs.tine.server.driver.usb;

import org.eclipse.emf.common.util.URI;

import ch.thomson.ucs.ftdi.UsbDevice;
import ch.thomson.ucs.tine.server.DataTable;
import ch.thomson.ucs.tine.server.RealProperty;
import ch.thomson.ucs.tine.server.impl.DriverImpl;

public class USBDriver extends DriverImpl {
	
	private URI uri;
	private UsbDevice device;
	private boolean queued;
	
	@Override
	public void initialize() throws Exception {
		
		// usb://BoardID/Type/BoardNo/InterfaceVersion
		// Type: "FTDI_SPI" oder "FTDI_FIFO"
		
		uri = URI.createURI(getAddress());
		if(uri.segmentCount() < 3)
			throw new Exception("Illegal URI: " + uri);
		
		device = new UsbDevice(typeToID(uri.segment(0)), Integer.parseInt(uri.host()), Integer.parseInt(uri.segment(1)), Integer.parseInt(uri.segment(2)));
		device.init();
	}

	@Override
	public void writeData(DataTable dataTable) throws Exception {
		byte[] buffer = dataTable.getByteBuffer().array();
		device.write(buffer, dataTable.getAbsoluteAddress(), buffer.length);
	}

	@Override
	public void parseAddressString(RealProperty realProperty) throws Exception {
		// Nothing to do
	}
	
	@Override
	public int readData(DataTable dataTable) throws Exception {
		byte[] buffer = dataTable.getByteBuffer().array();
		int result = device.read(buffer, dataTable.getAbsoluteAddress(), buffer.length);
		return queued ? result : 0;
	}
	
	private int typeToID(String type) throws Exception {
		if("FTDI_SPI".equals(type)) {
			queued = false;
			return UsbDevice.FTDI_SPI;
		}
		
		else if("FTDI_FIFO".equals(type)) {
			queued = true;
			return UsbDevice.FTDI_FIFO;
		}
		
		throw new Exception("Unknown Board Type: " + type);
	}
}
