/**
 * <copyright>
 * </copyright>
 *
 * $Id$
 */
package ch.thomson.ucs.tine.server.impl;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.Vector;

import org.eclipse.emf.common.notify.Notification;
import org.eclipse.emf.common.notify.NotificationChain;
import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.InternalEObject;
import org.eclipse.emf.ecore.impl.ENotificationImpl;
import org.eclipse.emf.ecore.impl.EObjectImpl;
import org.eclipse.emf.ecore.util.EObjectContainmentWithInverseEList;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.eclipse.emf.ecore.util.InternalEList;

import ch.thomson.ucs.tine.server.ByteOrder;
import ch.thomson.ucs.tine.server.Device;
import ch.thomson.ucs.tine.server.EquipmentModule;
import ch.thomson.ucs.tine.server.FrontEndController;
import ch.thomson.ucs.tine.server.Persistable;
import ch.thomson.ucs.tine.server.Property;
import ch.thomson.ucs.tine.server.PropertyNumeric;
import ch.thomson.ucs.tine.server.RealProperty;
import ch.thomson.ucs.tine.server.RealPropertyDAO;
import ch.thomson.ucs.tine.server.ServerPackage;
import ch.thomson.ucs.tine.server.TineAccess;
import ch.thomson.ucs.tine.server.util.Activator;
import ch.thomson.ucs.tine.server.util.UCSConfig;
import de.desy.tine.definitions.TErrorList;
import de.desy.tine.server.devices.TDevice;
import de.desy.tine.server.equipment.TEquipmentBackgroundTask;
import de.desy.tine.server.equipment.TEquipmentModule;

/**
 * <!-- begin-user-doc -->
 * An implementation of the model object '<em><b>Equipment Module</b></em>'.
 * <!-- end-user-doc -->
 * <p>
 * The following features are implemented:
 * <ul>
 *   <li>{@link ch.thomson.ucs.tine.server.impl.EquipmentModuleImpl#getFrontEndController <em>Front End Controller</em>}</li>
 *   <li>{@link ch.thomson.ucs.tine.server.impl.EquipmentModuleImpl#getExportName <em>Export Name</em>}</li>
 *   <li>{@link ch.thomson.ucs.tine.server.impl.EquipmentModuleImpl#getLocalName <em>Local Name</em>}</li>
 *   <li>{@link ch.thomson.ucs.tine.server.impl.EquipmentModuleImpl#getTineEquipmentModule <em>Tine Equipment Module</em>}</li>
 *   <li>{@link ch.thomson.ucs.tine.server.impl.EquipmentModuleImpl#getDevices <em>Devices</em>}</li>
 *   <li>{@link ch.thomson.ucs.tine.server.impl.EquipmentModuleImpl#getProperties <em>Properties</em>}</li>
 *   <li>{@link ch.thomson.ucs.tine.server.impl.EquipmentModuleImpl#getTineBackgroundTask <em>Tine Background Task</em>}</li>
 *   <li>{@link ch.thomson.ucs.tine.server.impl.EquipmentModuleImpl#getDevicePollingRate <em>Device Polling Rate</em>}</li>
 *   <li>{@link ch.thomson.ucs.tine.server.impl.EquipmentModuleImpl#getDriverType <em>Driver Type</em>}</li>
 *   <li>{@link ch.thomson.ucs.tine.server.impl.EquipmentModuleImpl#getByteOrder <em>Byte Order</em>}</li>
 *   <li>{@link ch.thomson.ucs.tine.server.impl.EquipmentModuleImpl#getTinePath <em>Tine Path</em>}</li>
 *   <li>{@link ch.thomson.ucs.tine.server.impl.EquipmentModuleImpl#getActualState <em>Actual State</em>}</li>
 *   <li>{@link ch.thomson.ucs.tine.server.impl.EquipmentModuleImpl#getTargetState <em>Target State</em>}</li>
 * </ul>
 * </p>
 *
 * @generated
 */
public class EquipmentModuleImpl extends EObjectImpl implements EquipmentModule {
	// Benutzerdefinierte TINE Error Codes
	public static final int TINE_WRONG_OPERATION_LOCATION = 512;
	public static final int TINE_OPERATING_STATE_TOO_HIGH = 513;
	public static final int TINE_MODULATOR_LOCKED = 514;
	public static final int TINE_REMOTE_WRITE_ACCESS_DENIED = 515;
	
	private static final String TINE_WRONG_OPERATION_LOCATION_STR = "Command not allowed due to wrong control location.";
	private static final String TINE_OPERATING_STATE_TOO_HIGH_STR = "Command not allowed. Current operating state is too high.";
	private static final String TINE_MODULATOR_LOCKED_STR = "Command not allowed. Modulator is locked.";
	private static final String TINE_REMOTE_WRITE_ACCESS_DENIED_STR = "Command from remote control location not allowed.";
	
	/**
	 * Generiert einen benutzerdefinierten TINE-Fehler. Als Error-Code wird eine statische Konstante dieser Klasse mitgeben.
	 * @param eqm Equipment Module Instanz
	 * @param errorCode Error-Konstante aus dieser Klasse
	 * @return TINE Error Konstante
	 **/
	public static int makeUserDefinedTineError(EquipmentModule eqm, int errorCode) {
		switch(errorCode) {
		case TINE_WRONG_OPERATION_LOCATION:
			eqm.getTineEquipmentModule().setCompletionString(TINE_WRONG_OPERATION_LOCATION_STR);
			return errorCode;
		case TINE_OPERATING_STATE_TOO_HIGH:
			eqm.getTineEquipmentModule().setCompletionString(TINE_OPERATING_STATE_TOO_HIGH_STR);
			return errorCode;
		case TINE_MODULATOR_LOCKED:
			eqm.getTineEquipmentModule().setCompletionString(TINE_MODULATOR_LOCKED_STR);
			return errorCode;
		case TINE_REMOTE_WRITE_ACCESS_DENIED:
			eqm.getTineEquipmentModule().setCompletionString(TINE_REMOTE_WRITE_ACCESS_DENIED_STR);
			return errorCode;
		}
		
		return TErrorList.not_defined;
	}
	
	/**
	 * The default value of the '{@link #getExportName() <em>Export Name</em>}' attribute.
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @see #getExportName()
	 * @generated
	 * @ordered
	 */
	protected static final String EXPORT_NAME_EDEFAULT = null;

	/**
	 * The cached value of the '{@link #getExportName() <em>Export Name</em>}' attribute.
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @see #getExportName()
	 * @generated
	 * @ordered
	 */
	protected String exportName = EXPORT_NAME_EDEFAULT;

	/**
	 * The default value of the '{@link #getLocalName() <em>Local Name</em>}' attribute.
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @see #getLocalName()
	 * @generated
	 * @ordered
	 */
	protected static final String LOCAL_NAME_EDEFAULT = null;

	/**
	 * The cached value of the '{@link #getLocalName() <em>Local Name</em>}' attribute.
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @see #getLocalName()
	 * @generated
	 * @ordered
	 */
	protected String localName = LOCAL_NAME_EDEFAULT;

	/**
	 * The default value of the '{@link #getTineEquipmentModule() <em>Tine Equipment Module</em>}' attribute.
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @see #getTineEquipmentModule()
	 * @generated
	 * @ordered
	 */
	protected static final TEquipmentModule TINE_EQUIPMENT_MODULE_EDEFAULT = null;

	/**
	 * The cached value of the '{@link #getTineEquipmentModule() <em>Tine Equipment Module</em>}' attribute.
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @see #getTineEquipmentModule()
	 * @generated
	 * @ordered
	 */
	protected TEquipmentModule tineEquipmentModule = TINE_EQUIPMENT_MODULE_EDEFAULT;

	/**
	 * The cached value of the '{@link #getDevices() <em>Devices</em>}' containment reference list.
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @see #getDevices()
	 * @generated
	 * @ordered
	 */
	protected EList<Device> devices;

	/**
	 * The cached value of the '{@link #getProperties() <em>Properties</em>}' containment reference list.
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @see #getProperties()
	 * @generated
	 * @ordered
	 */
	protected EList<Property> properties;

	/**
	 * The default value of the '{@link #getTineBackgroundTask() <em>Tine Background Task</em>}' attribute.
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @see #getTineBackgroundTask()
	 * @generated
	 * @ordered
	 */
	protected static final TEquipmentBackgroundTask TINE_BACKGROUND_TASK_EDEFAULT = null;

	/**
	 * The cached value of the '{@link #getTineBackgroundTask() <em>Tine Background Task</em>}' attribute.
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @see #getTineBackgroundTask()
	 * @generated
	 * @ordered
	 */
	protected TEquipmentBackgroundTask tineBackgroundTask = TINE_BACKGROUND_TASK_EDEFAULT;

	/**
	 * The default value of the '{@link #getDevicePollingRate() <em>Device Polling Rate</em>}' attribute.
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @see #getDevicePollingRate()
	 * @generated
	 * @ordered
	 */
	protected static final int DEVICE_POLLING_RATE_EDEFAULT = 100;

	/**
	 * The cached value of the '{@link #getDevicePollingRate() <em>Device Polling Rate</em>}' attribute.
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @see #getDevicePollingRate()
	 * @generated
	 * @ordered
	 */
	protected int devicePollingRate = DEVICE_POLLING_RATE_EDEFAULT;

	/**
	 * The default value of the '{@link #getDriverType() <em>Driver Type</em>}' attribute.
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @see #getDriverType()
	 * @generated
	 * @ordered
	 */
	protected static final String DRIVER_TYPE_EDEFAULT = null;

	/**
	 * The cached value of the '{@link #getDriverType() <em>Driver Type</em>}' attribute.
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @see #getDriverType()
	 * @generated
	 * @ordered
	 */
	protected String driverType = DRIVER_TYPE_EDEFAULT;

	/**
	 * The default value of the '{@link #getByteOrder() <em>Byte Order</em>}' attribute.
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @see #getByteOrder()
	 * @generated
	 * @ordered
	 */
	protected static final ByteOrder BYTE_ORDER_EDEFAULT = ByteOrder.BIG_ENDIAN;
	/**
	 * The cached value of the '{@link #getByteOrder() <em>Byte Order</em>}' attribute.
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @see #getByteOrder()
	 * @generated
	 * @ordered
	 */
	protected ByteOrder byteOrder = BYTE_ORDER_EDEFAULT;

	/**
	 * The default value of the '{@link #getTinePath() <em>Tine Path</em>}' attribute.
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @see #getTinePath()
	 * @generated
	 * @ordered
	 */
	protected static final String TINE_PATH_EDEFAULT = null;

	/**
	 * The cached value of the '{@link #getActualState() <em>Actual State</em>}' reference.
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @see #getActualState()
	 * @generated
	 * @ordered
	 */
	protected PropertyNumeric actualState;
	/**
	 * The cached value of the '{@link #getTargetState() <em>Target State</em>}' reference.
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @see #getTargetState()
	 * @generated
	 * @ordered
	 */
	protected PropertyNumeric targetState;

	/**
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @generated
	 */
	protected EquipmentModuleImpl() {
		super();
	}

	/**
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @generated
	 */
	@Override
	protected EClass eStaticClass() {
		return ServerPackage.Literals.EQUIPMENT_MODULE;
	}

	/**
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @generated
	 */
	public FrontEndController getFrontEndController() {
		if (eContainerFeatureID() != ServerPackage.EQUIPMENT_MODULE__FRONT_END_CONTROLLER) return null;
		return (FrontEndController)eContainer();
	}

	/**
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @generated
	 */
	public NotificationChain basicSetFrontEndController(FrontEndController newFrontEndController, NotificationChain msgs) {
		msgs = eBasicSetContainer((InternalEObject)newFrontEndController, ServerPackage.EQUIPMENT_MODULE__FRONT_END_CONTROLLER, msgs);
		return msgs;
	}

	/**
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @generated
	 */
	public void setFrontEndController(FrontEndController newFrontEndController) {
		if (newFrontEndController != eInternalContainer() || (eContainerFeatureID() != ServerPackage.EQUIPMENT_MODULE__FRONT_END_CONTROLLER && newFrontEndController != null)) {
			if (EcoreUtil.isAncestor(this, newFrontEndController))
				throw new IllegalArgumentException("Recursive containment not allowed for " + toString());
			NotificationChain msgs = null;
			if (eInternalContainer() != null)
				msgs = eBasicRemoveFromContainer(msgs);
			if (newFrontEndController != null)
				msgs = ((InternalEObject)newFrontEndController).eInverseAdd(this, ServerPackage.FRONT_END_CONTROLLER__EQUIPMENT_MODULES, FrontEndController.class, msgs);
			msgs = basicSetFrontEndController(newFrontEndController, msgs);
			if (msgs != null) msgs.dispatch();
		}
		else if (eNotificationRequired())
			eNotify(new ENotificationImpl(this, Notification.SET, ServerPackage.EQUIPMENT_MODULE__FRONT_END_CONTROLLER, newFrontEndController, newFrontEndController));
	}

	/**
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @generated
	 */
	public String getExportName() {
		return exportName;
	}

	/**
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @generated
	 */
	public void setExportName(String newExportName) {
		String oldExportName = exportName;
		exportName = newExportName;
		if (eNotificationRequired())
			eNotify(new ENotificationImpl(this, Notification.SET, ServerPackage.EQUIPMENT_MODULE__EXPORT_NAME, oldExportName, exportName));
	}

	/**
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @generated
	 */
	public String getLocalName() {
		return localName;
	}

	/**
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @generated
	 */
	public void setLocalName(String newLocalName) {
		String oldLocalName = localName;
		localName = newLocalName;
		if (eNotificationRequired())
			eNotify(new ENotificationImpl(this, Notification.SET, ServerPackage.EQUIPMENT_MODULE__LOCAL_NAME, oldLocalName, localName));
	}

	/**
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @generated
	 */
	public TEquipmentModule getTineEquipmentModule() {
		return tineEquipmentModule;
	}

	/**
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @generated
	 */
	public EList<Device> getDevices() {
		if (devices == null) {
			devices = new EObjectContainmentWithInverseEList<Device>(Device.class, this, ServerPackage.EQUIPMENT_MODULE__DEVICES, ServerPackage.DEVICE__EQUIPMENT_MODULE);
		}
		return devices;
	}

	/**
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @generated
	 */
	public EList<Property> getProperties() {
		if (properties == null) {
			properties = new EObjectContainmentWithInverseEList<Property>(Property.class, this, ServerPackage.EQUIPMENT_MODULE__PROPERTIES, ServerPackage.PROPERTY__EQUIPMENT_MODULE);
		}
		return properties;
	}

	/**
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @generated
	 */
	public TEquipmentBackgroundTask getTineBackgroundTask() {
		return tineBackgroundTask;
	}

	/**
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @generated
	 */
	public int getDevicePollingRate() {
		return devicePollingRate;
	}

	/**
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @generated
	 */
	public void setDevicePollingRate(int newDevicePollingRate) {
		int oldDevicePollingRate = devicePollingRate;
		devicePollingRate = newDevicePollingRate;
		if (eNotificationRequired())
			eNotify(new ENotificationImpl(this, Notification.SET, ServerPackage.EQUIPMENT_MODULE__DEVICE_POLLING_RATE, oldDevicePollingRate, devicePollingRate));
	}

	/**
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @generated
	 */
	public String getDriverType() {
		return driverType;
	}

	/**
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @generated
	 */
	public void setDriverType(String newDriverType) {
		String oldDriverType = driverType;
		driverType = newDriverType;
		if (eNotificationRequired())
			eNotify(new ENotificationImpl(this, Notification.SET, ServerPackage.EQUIPMENT_MODULE__DRIVER_TYPE, oldDriverType, driverType));
	}

	/**
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @generated
	 */
	public ByteOrder getByteOrder() {
		return byteOrder;
	}

	/**
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @generated
	 */
	public void setByteOrder(ByteOrder newByteOrder) {
		ByteOrder oldByteOrder = byteOrder;
		byteOrder = newByteOrder == null ? BYTE_ORDER_EDEFAULT : newByteOrder;
		if (eNotificationRequired())
			eNotify(new ENotificationImpl(this, Notification.SET, ServerPackage.EQUIPMENT_MODULE__BYTE_ORDER, oldByteOrder, byteOrder));
	}

	/**
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @generated NOT
	 */
	public String getTinePath() {
		return "/" + this.getFrontEndController().getContextName() + "/" + this.getExportName() + "/";
	}

	/**
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @generated
	 */
	public PropertyNumeric getActualState() {
		if (actualState != null && actualState.eIsProxy()) {
			InternalEObject oldActualState = (InternalEObject)actualState;
			actualState = (PropertyNumeric)eResolveProxy(oldActualState);
			if (actualState != oldActualState) {
				if (eNotificationRequired())
					eNotify(new ENotificationImpl(this, Notification.RESOLVE, ServerPackage.EQUIPMENT_MODULE__ACTUAL_STATE, oldActualState, actualState));
			}
		}
		return actualState;
	}

	/**
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @generated
	 */
	public PropertyNumeric basicGetActualState() {
		return actualState;
	}

	/**
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @generated
	 */
	public void setActualState(PropertyNumeric newActualState) {
		PropertyNumeric oldActualState = actualState;
		actualState = newActualState;
		if (eNotificationRequired())
			eNotify(new ENotificationImpl(this, Notification.SET, ServerPackage.EQUIPMENT_MODULE__ACTUAL_STATE, oldActualState, actualState));
	}

	/**
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @generated
	 */
	public PropertyNumeric getTargetState() {
		if (targetState != null && targetState.eIsProxy()) {
			InternalEObject oldTargetState = (InternalEObject)targetState;
			targetState = (PropertyNumeric)eResolveProxy(oldTargetState);
			if (targetState != oldTargetState) {
				if (eNotificationRequired())
					eNotify(new ENotificationImpl(this, Notification.RESOLVE, ServerPackage.EQUIPMENT_MODULE__TARGET_STATE, oldTargetState, targetState));
			}
		}
		return targetState;
	}

	/**
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @generated
	 */
	public PropertyNumeric basicGetTargetState() {
		return targetState;
	}

	/**
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @generated
	 */
	public void setTargetState(PropertyNumeric newTargetState) {
		PropertyNumeric oldTargetState = targetState;
		targetState = newTargetState;
		if (eNotificationRequired())
			eNotify(new ENotificationImpl(this, Notification.SET, ServerPackage.EQUIPMENT_MODULE__TARGET_STATE, oldTargetState, targetState));
	}

	/**
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @generated NOT
	 */
	public void initialize() throws Exception {
		
		/* 	Die Namen fr die TINE Registrierung werden vom externen Ini-File geholt. Sie drfen aber nur
			fr die TINE-Umgebung sichtbar sein, innerhalb des EMF-Modells wird mit den Namen aus dem 
			XML-File gearbeitet (Primary Key fr Datenpersistenz!!!)
		*/
		UCSConfig config = UCSConfig.getInstance();
		this.tineEquipmentModule = new TEquipmentModule(config.getEQMLocalName(this.getLocalName()));
		this.tineEquipmentModule.setExportName(config.getEQMExportName(this.getExportName()));
		
		Vector<TDevice> tDevices = new Vector<TDevice>();
		
		for(Device device : this.getDevices()) {
			device.initialize();
			tDevices.add(device.getTineDevice());
		}
		
		getTineEquipmentModule().registerDevices(tDevices.toArray(new TDevice[0]));
		
		// Bei einem Dummy Modul werden alle Properties, ungeachtet den
		// Einstellungen in der Signalliste, auf Read/Write gesetzt. Dies
		// erleichtert das Testen in einer Debug-Umgebung, in welcher keine
		// Hardware vorhanden ist.
		boolean dummyModule = "dummy".equals(this.getDriverType());
		for(Property property : this.getProperties()) {
			if(dummyModule && property instanceof RealProperty)
				property.setTineAccess(TineAccess.READ_WRITE);			
			property.initialize();
		}
		
		// Backgroundtask erzeugen
		this.tineBackgroundTask = new TEquipmentBackgroundTask() {
/*
			long highest; //us
			int counter;
			long sum; //us
			long lastStart;
			long cycleSum; //us
			long highestCycle; //us
			int cycleCounter;
			DeviceImpl firstDevice;
			EquipmentModule eqmFast;
*/			
			
			@Override
			public void call() {
/*				
				if(firstDevice == null) 
					firstDevice = (DeviceImpl) getDevices().get(0);
				
				if(eqmFast == null) {
					if("FAST_CHANNEL".equals(getExportName()))
						eqmFast = EquipmentModuleImpl.this;
				}
				
				long start = System.nanoTime();
				
				if(lastStart != 0) {
					long cycle = ((start - lastStart) / 1000);
					//System.out.println(getExportName() + ": " + cycle / 1000.0);
					cycleSum += cycle;
					cycleCounter++;
					if(cycle > highestCycle)
						highestCycle = cycle;
				}
			
				lastStart = start;
*/					
				for(Device device : getDevices()) {
					device.run();
				}
				
				if(scheduleTimeHighest > 50000) // 50ms
					Activator.getLogger().error("ScheduleProperty for property " + scheduleTimeHighestProperty.getName() + " lasts very long: blocking time= " + (scheduleTimeHighest / 1000.0) + "ms");
				
				scheduleTimeHighest = 0;
				scheduleTimeHighestProperty = null;
/*				
				long duration = (System.nanoTime() - start) / 1000; // us
				sum += duration;
				counter++;
				
				if(duration > highest) {
					highest = duration;
				}
				
				if(counter == 1000) {
					if(eqmFast == EquipmentModuleImpl.this)
						System.out.println(getExportName() + ": worker-avg=" + (sum / counter / 1000.0) + "ms, worker-highest=" + (highest / 1000.0) + "ms, " + "cycle-avg=" + (cycleSum / cycleCounter / 1000.0) + "ms, cycle-highest=" + (highestCycle / 1000.0) + "ms, driver-highest=" + (firstDevice.highDriver / 1000.0) + "ms, schedule-highest=" + (scheduleTimeHighest / 1000.0) + "ms");
					counter = 0;
					sum = 0;
					highest = 0;
					highestCycle = 0;
					cycleSum = 0;
					cycleCounter = 0;
					//firstDevice.highDriver = 0;
					scheduleTimeHighest = 0;
				}
*/				
			}
		};
		
		this.getTineBackgroundTask().setBackgroundTaskInterval(this.getDevicePollingRate());
		this.getTineEquipmentModule().addEquipmentBackgroundTask(this.getTineBackgroundTask());
	}

	private Map<String, Device> deviceCacheMap = null;
	
	/**
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @generated NOT
	 */
	public Device findDevice(String name) {
		if(name == null)
			return null;
		
		if(deviceCacheMap == null) {
			deviceCacheMap = new HashMap<String, Device>();
		}
		
		Device device = deviceCacheMap.get(name);
		
		if(device != null)
			return device;
		
		// Nicht im Cache, muss also gesucht werden, und abgelegt werden.
		for(Device d : this.getDevices()) {
			if(d.getName().equals(name)) {
				deviceCacheMap.put(d.getName(), d);
				device = d;
				break;
			}
		}
		
		return device;
	}

	/**
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @generated NOT
	 */
	public EList<RealProperty> collectRealProperties() {
		EList<RealProperty> realProperties = new BasicEList<RealProperty>();
		
		for(Property property : this.getProperties()) {
			if(property instanceof RealProperty) {
				realProperties.add((RealProperty)property);
			}
		}
		
		return realProperties;
	}

	//private ExecutorService schedulePool = Executors.newFixedThreadPool(10);
	
	long scheduleTimeHighest; //us
	Property scheduleTimeHighestProperty;
	
	/**
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @generated NOT
	 */
	public void scheduleProperty(Property property) {
		
		long start = System.nanoTime();
		getFrontEndController().getTineEquipmentModuleFactory().systemScheduleProperty(getTineEquipmentModule(), property.getName());
		long duration = (System.nanoTime() - start) / 1000;
		
		if(duration > scheduleTimeHighest) {
			scheduleTimeHighest = duration;
			scheduleTimeHighestProperty = property;
		}
	}
	
	private Map<String, Property> propertyCacheMap = null;

	/**
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @generated NOT
	 */
	public Property findProperty(String name) {
		if(propertyCacheMap == null) {
			propertyCacheMap = new HashMap<String, Property>();
		}
		
		Property property = propertyCacheMap.get(name);
		
		if(property != null)
			return property;
		
		// Nicht im Cache, muss also gesucht werden, und abgelegt werden.
		for(Property p : this.getProperties()) {
			if(p.getName().equals(name)) {
				propertyCacheMap.put(p.getName(), p);
				property = p;
				break;
			}
		}
		
		return property;
	}

	/**
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @generated NOT
	 */
	public void loadData(RealPropertyDAO dao) throws Exception {
		for(Property property : this.getProperties()) {
			if(property instanceof Persistable) {
				((Persistable) property).loadData(dao);
			}
		}
	}

	/**
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @generated NOT
	 */
	public void saveData(RealPropertyDAO dao) throws Exception {
		for(Property property : this.getProperties()) {
			if(property instanceof Persistable) {
				((Persistable) property).saveData(dao);
			}
		}
	}

	/**
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @generated
	 */
	@SuppressWarnings("unchecked")
	@Override
	public NotificationChain eInverseAdd(InternalEObject otherEnd, int featureID, NotificationChain msgs) {
		switch (featureID) {
			case ServerPackage.EQUIPMENT_MODULE__FRONT_END_CONTROLLER:
				if (eInternalContainer() != null)
					msgs = eBasicRemoveFromContainer(msgs);
				return basicSetFrontEndController((FrontEndController)otherEnd, msgs);
			case ServerPackage.EQUIPMENT_MODULE__DEVICES:
				return ((InternalEList<InternalEObject>)(InternalEList<?>)getDevices()).basicAdd(otherEnd, msgs);
			case ServerPackage.EQUIPMENT_MODULE__PROPERTIES:
				return ((InternalEList<InternalEObject>)(InternalEList<?>)getProperties()).basicAdd(otherEnd, msgs);
		}
		return super.eInverseAdd(otherEnd, featureID, msgs);
	}

	/**
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @generated
	 */
	@Override
	public NotificationChain eInverseRemove(InternalEObject otherEnd, int featureID, NotificationChain msgs) {
		switch (featureID) {
			case ServerPackage.EQUIPMENT_MODULE__FRONT_END_CONTROLLER:
				return basicSetFrontEndController(null, msgs);
			case ServerPackage.EQUIPMENT_MODULE__DEVICES:
				return ((InternalEList<?>)getDevices()).basicRemove(otherEnd, msgs);
			case ServerPackage.EQUIPMENT_MODULE__PROPERTIES:
				return ((InternalEList<?>)getProperties()).basicRemove(otherEnd, msgs);
		}
		return super.eInverseRemove(otherEnd, featureID, msgs);
	}

	/**
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @generated
	 */
	@Override
	public NotificationChain eBasicRemoveFromContainerFeature(NotificationChain msgs) {
		switch (eContainerFeatureID()) {
			case ServerPackage.EQUIPMENT_MODULE__FRONT_END_CONTROLLER:
				return eInternalContainer().eInverseRemove(this, ServerPackage.FRONT_END_CONTROLLER__EQUIPMENT_MODULES, FrontEndController.class, msgs);
		}
		return super.eBasicRemoveFromContainerFeature(msgs);
	}

	/**
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @generated
	 */
	@Override
	public Object eGet(int featureID, boolean resolve, boolean coreType) {
		switch (featureID) {
			case ServerPackage.EQUIPMENT_MODULE__FRONT_END_CONTROLLER:
				return getFrontEndController();
			case ServerPackage.EQUIPMENT_MODULE__EXPORT_NAME:
				return getExportName();
			case ServerPackage.EQUIPMENT_MODULE__LOCAL_NAME:
				return getLocalName();
			case ServerPackage.EQUIPMENT_MODULE__TINE_EQUIPMENT_MODULE:
				return getTineEquipmentModule();
			case ServerPackage.EQUIPMENT_MODULE__DEVICES:
				return getDevices();
			case ServerPackage.EQUIPMENT_MODULE__PROPERTIES:
				return getProperties();
			case ServerPackage.EQUIPMENT_MODULE__TINE_BACKGROUND_TASK:
				return getTineBackgroundTask();
			case ServerPackage.EQUIPMENT_MODULE__DEVICE_POLLING_RATE:
				return getDevicePollingRate();
			case ServerPackage.EQUIPMENT_MODULE__DRIVER_TYPE:
				return getDriverType();
			case ServerPackage.EQUIPMENT_MODULE__BYTE_ORDER:
				return getByteOrder();
			case ServerPackage.EQUIPMENT_MODULE__TINE_PATH:
				return getTinePath();
			case ServerPackage.EQUIPMENT_MODULE__ACTUAL_STATE:
				if (resolve) return getActualState();
				return basicGetActualState();
			case ServerPackage.EQUIPMENT_MODULE__TARGET_STATE:
				if (resolve) return getTargetState();
				return basicGetTargetState();
		}
		return super.eGet(featureID, resolve, coreType);
	}

	/**
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @generated
	 */
	@SuppressWarnings("unchecked")
	@Override
	public void eSet(int featureID, Object newValue) {
		switch (featureID) {
			case ServerPackage.EQUIPMENT_MODULE__FRONT_END_CONTROLLER:
				setFrontEndController((FrontEndController)newValue);
				return;
			case ServerPackage.EQUIPMENT_MODULE__EXPORT_NAME:
				setExportName((String)newValue);
				return;
			case ServerPackage.EQUIPMENT_MODULE__LOCAL_NAME:
				setLocalName((String)newValue);
				return;
			case ServerPackage.EQUIPMENT_MODULE__DEVICES:
				getDevices().clear();
				getDevices().addAll((Collection<? extends Device>)newValue);
				return;
			case ServerPackage.EQUIPMENT_MODULE__PROPERTIES:
				getProperties().clear();
				getProperties().addAll((Collection<? extends Property>)newValue);
				return;
			case ServerPackage.EQUIPMENT_MODULE__DEVICE_POLLING_RATE:
				setDevicePollingRate((Integer)newValue);
				return;
			case ServerPackage.EQUIPMENT_MODULE__DRIVER_TYPE:
				setDriverType((String)newValue);
				return;
			case ServerPackage.EQUIPMENT_MODULE__BYTE_ORDER:
				setByteOrder((ByteOrder)newValue);
				return;
			case ServerPackage.EQUIPMENT_MODULE__ACTUAL_STATE:
				setActualState((PropertyNumeric)newValue);
				return;
			case ServerPackage.EQUIPMENT_MODULE__TARGET_STATE:
				setTargetState((PropertyNumeric)newValue);
				return;
		}
		super.eSet(featureID, newValue);
	}

	/**
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @generated
	 */
	@Override
	public void eUnset(int featureID) {
		switch (featureID) {
			case ServerPackage.EQUIPMENT_MODULE__FRONT_END_CONTROLLER:
				setFrontEndController((FrontEndController)null);
				return;
			case ServerPackage.EQUIPMENT_MODULE__EXPORT_NAME:
				setExportName(EXPORT_NAME_EDEFAULT);
				return;
			case ServerPackage.EQUIPMENT_MODULE__LOCAL_NAME:
				setLocalName(LOCAL_NAME_EDEFAULT);
				return;
			case ServerPackage.EQUIPMENT_MODULE__DEVICES:
				getDevices().clear();
				return;
			case ServerPackage.EQUIPMENT_MODULE__PROPERTIES:
				getProperties().clear();
				return;
			case ServerPackage.EQUIPMENT_MODULE__DEVICE_POLLING_RATE:
				setDevicePollingRate(DEVICE_POLLING_RATE_EDEFAULT);
				return;
			case ServerPackage.EQUIPMENT_MODULE__DRIVER_TYPE:
				setDriverType(DRIVER_TYPE_EDEFAULT);
				return;
			case ServerPackage.EQUIPMENT_MODULE__BYTE_ORDER:
				setByteOrder(BYTE_ORDER_EDEFAULT);
				return;
			case ServerPackage.EQUIPMENT_MODULE__ACTUAL_STATE:
				setActualState((PropertyNumeric)null);
				return;
			case ServerPackage.EQUIPMENT_MODULE__TARGET_STATE:
				setTargetState((PropertyNumeric)null);
				return;
		}
		super.eUnset(featureID);
	}

	/**
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @generated
	 */
	@Override
	public boolean eIsSet(int featureID) {
		switch (featureID) {
			case ServerPackage.EQUIPMENT_MODULE__FRONT_END_CONTROLLER:
				return getFrontEndController() != null;
			case ServerPackage.EQUIPMENT_MODULE__EXPORT_NAME:
				return EXPORT_NAME_EDEFAULT == null ? exportName != null : !EXPORT_NAME_EDEFAULT.equals(exportName);
			case ServerPackage.EQUIPMENT_MODULE__LOCAL_NAME:
				return LOCAL_NAME_EDEFAULT == null ? localName != null : !LOCAL_NAME_EDEFAULT.equals(localName);
			case ServerPackage.EQUIPMENT_MODULE__TINE_EQUIPMENT_MODULE:
				return TINE_EQUIPMENT_MODULE_EDEFAULT == null ? tineEquipmentModule != null : !TINE_EQUIPMENT_MODULE_EDEFAULT.equals(tineEquipmentModule);
			case ServerPackage.EQUIPMENT_MODULE__DEVICES:
				return devices != null && !devices.isEmpty();
			case ServerPackage.EQUIPMENT_MODULE__PROPERTIES:
				return properties != null && !properties.isEmpty();
			case ServerPackage.EQUIPMENT_MODULE__TINE_BACKGROUND_TASK:
				return TINE_BACKGROUND_TASK_EDEFAULT == null ? tineBackgroundTask != null : !TINE_BACKGROUND_TASK_EDEFAULT.equals(tineBackgroundTask);
			case ServerPackage.EQUIPMENT_MODULE__DEVICE_POLLING_RATE:
				return devicePollingRate != DEVICE_POLLING_RATE_EDEFAULT;
			case ServerPackage.EQUIPMENT_MODULE__DRIVER_TYPE:
				return DRIVER_TYPE_EDEFAULT == null ? driverType != null : !DRIVER_TYPE_EDEFAULT.equals(driverType);
			case ServerPackage.EQUIPMENT_MODULE__BYTE_ORDER:
				return byteOrder != BYTE_ORDER_EDEFAULT;
			case ServerPackage.EQUIPMENT_MODULE__TINE_PATH:
				return TINE_PATH_EDEFAULT == null ? getTinePath() != null : !TINE_PATH_EDEFAULT.equals(getTinePath());
			case ServerPackage.EQUIPMENT_MODULE__ACTUAL_STATE:
				return actualState != null;
			case ServerPackage.EQUIPMENT_MODULE__TARGET_STATE:
				return targetState != null;
		}
		return super.eIsSet(featureID);
	}

	/**
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @generated
	 */
	@Override
	public String toString() {
		if (eIsProxy()) return super.toString();

		StringBuffer result = new StringBuffer(super.toString());
		result.append(" (exportName: ");
		result.append(exportName);
		result.append(", localName: ");
		result.append(localName);
		result.append(", tineEquipmentModule: ");
		result.append(tineEquipmentModule);
		result.append(", tineBackgroundTask: ");
		result.append(tineBackgroundTask);
		result.append(", devicePollingRate: ");
		result.append(devicePollingRate);
		result.append(", driverType: ");
		result.append(driverType);
		result.append(", byteOrder: ");
		result.append(byteOrder);
		result.append(')');
		return result.toString();
	}

} //EquipmentModuleImpl
