/**
 * <copyright>
 * </copyright>
 *
 * $Id$
 */
package ch.thomson.ucs.tine.server.impl;

import java.io.IOException;

import org.eclipse.emf.ecore.EClass;

import ch.thomson.ucs.tine.server.PropertyInt32;
import ch.thomson.ucs.tine.server.ServerPackage;
import ch.thomson.ucs.utils.logic.ByteOrderedDataInputStream;
import ch.thomson.ucs.utils.logic.ByteOrderedDataOutputStream;
import de.desy.tine.dataUtils.TDataType;
import de.desy.tine.definitions.TErrorList;
import de.desy.tine.definitions.TFormat;

/**
 * <!-- begin-user-doc -->
 * An implementation of the model object '<em><b>Property Int32</b></em>'.
 * <!-- end-user-doc -->
 * <p>
 * </p>
 *
 * @generated
 */
public class PropertyInt32Impl extends PropertyNumericImpl implements PropertyInt32 {
	/**
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @generated
	 */
	protected PropertyInt32Impl() {
		super();
	}

	/**
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @generated
	 */
	@Override
	protected EClass eStaticClass() {
		return ServerPackage.Literals.PROPERTY_INT32;
	}

	@Override
	public int write(TDataType din, ByteOrderedDataOutputStream dos) throws IOException {
		int result = 0;
		
		// Keine Kommastelle? -> Durchgngige Integerbehandlung (wichtig fr Bit-Vectoren)
		if(this.getDecimalPlaces() == 0) {
			int[] values = new int[din.getArrayLength()];
			if ((result = din.getData(values)) == 0) {
				for (int val : values) {
					// Range Check
					if(isOutOfRange((float) val))
						return TErrorList.out_of_range;
					
					dos.writeInt((int)(val * getFactor() + getOffset()));
				}
			}
			return result;
		}
		
		// Hier entsteht zwangslufig eine Gleitkommazahl, die gerundet werden muss (ungeeignet fr Bit-Vectoren)
		double[] values = new double[din.getArrayLength()];
		if ((result = din.getData(values)) == 0) {
			for (double val : values) {
				
				// Range Check
				if(isOutOfRange((float) val))
					return TErrorList.out_of_range;
				
				dos.writeInt((int) Math.round((val * getFactor() + getOffset())));
			}
		}
		
		return result;
	}
	

	@Override
	public void read(ByteOrderedDataInputStream source, Object target) throws IOException {
		// Knnte auch "target instanceof int[]" sein, int Vergleich ist aber performanter.
		if(this.getDecimalPlaces() == 0) {
			int[] values = (int[]) target;
			for(int i=0; i<values.length; ++i) {
				values[i] = (int) ((source.readInt() - getOffset()) / getFactor());
			}
		}
		
		// Hier entsteht zwangslufig eine Gleitkommazahl (ungeeignet fr Bit-Vectoren)
		else {
			double[] values = (double[]) target; 
			for(int i=0; i<values.length; i++) {
				values[i] = roundScale(((double) source.readInt() - getOffset()) / getFactor());
			}
		}
	}

	@Override
	public short getTineTypeCode() {
		return TFormat.CF_INT32;
	}

	@Override
	public TDataType createTDataType() {
		return new TDataType(new int[this.getLength()]);
	}

	@Override
	public Object createDataCacheBuffer() {
		// Ohne Kommastellen soll zwingend ein Integer-Array verwendet werden (z.B. wegen Bit-Vectoren)
		return getDecimalPlaces() == 0 ? new int[getLength()] : new double[getLength()];
	}

} //PropertyInt32Impl
