package ch.thomson.ucs.tine.server.internal.client;

import java.util.HashSet;
import java.util.Set;

import ch.thomson.ucs.tine.server.Device;
import ch.thomson.ucs.tine.server.Observable;
import ch.thomson.ucs.tine.server.PropertyNumeric;
import de.desy.tine.dataUtils.TDataType;

public class PropertyFloatValueProvider extends PropertyValueProvider {
	protected float[] readBuffer;
	protected float[] writeBuffer;
	protected int size;
	protected Set<IPropertyFloatValueListener> listeners = new HashSet<IPropertyFloatValueListener>();
	
	public PropertyFloatValueProvider(Device device, PropertyNumeric property, boolean notificated) {
		super(device, property, notificated);
		readBuffer = new float[property.getLength()];
		writeBuffer = new float[property.getLength()];
		dout = new TDataType(readBuffer);
		din = new TDataType(writeBuffer);
		refreshValues();
	}
	
	public float[] getValues() {
		if(isNotificated())
			return readBuffer; // gecachter Wert zurckgeben
		
		refreshValues();
		return readBuffer;
	}
	
	public float getValue() {
		return getValues()[0];
	}
	
	public float getValue(int index) {
		return getValues()[index];
	}
	
	public void setValues(float[] values) {
		System.arraycopy(values, 0, writeBuffer, 0, values.length);
		writeValues(values.length);
	}
	
	public void setValue(float value) {
		writeBuffer[0] = value;
		writeValues(1);
	}
	
	public void setValue(int index, float value) {
		System.arraycopy(getValues(), 0, writeBuffer, 0, getLength());
		writeBuffer[index] = value;
		writeValues();
	}

	@Override
	public void update(Observable observable) {
		if(observable == property) {
			refreshValues();
			for(IPropertyFloatValueListener l : listeners)
				l.valuesChanged(readBuffer.clone(), getValues());
		}
	}
	
	public void addChangeListener(IPropertyFloatValueListener l) {
		this.listeners.add(l);
	}
	
	public void removeChangeListener(IPropertyFloatValueListener l) {
		this.listeners.remove(l);
	}
}
