package ch.thomson.ucs.tine.server.internal.client;

import java.util.HashSet;
import java.util.Set;

import ch.thomson.ucs.tine.server.Device;
import ch.thomson.ucs.tine.server.Observable;
import ch.thomson.ucs.tine.server.PropertyNumeric;
import de.desy.tine.dataUtils.TDataType;

public class PropertyIntValueProvider extends PropertyValueProvider {
	protected int[] readBuffer;
	protected int[] writeBuffer;
	protected int size;
	protected Set<IPropertyIntValueListener> listeners = new HashSet<IPropertyIntValueListener>();
	
	public PropertyIntValueProvider(Device device, PropertyNumeric property, boolean notificated) {
		super(device, property, notificated);
		readBuffer = new int[property.getLength()];
		writeBuffer = new int[property.getLength()];
		dout = new TDataType(readBuffer);
		din = new TDataType(writeBuffer);
		refreshValues();
	}
	
	public int[] getValues() {
		if(isNotificated())
			return readBuffer; // gecachter Wert zurckgeben
		
		refreshValues();
		return readBuffer;
	}
	
	public int getValue() {
		return getValues()[0];
	}
	
	public int getValue(int index) {
		return getValues()[index];
	}
	
	public void setValues(int[] values) {
		System.arraycopy(values, 0, writeBuffer, 0, values.length);
		writeValues(values.length);
	}
	
	public void setValue(int value) {
		writeBuffer[0] = value;
		writeValues(1);
	}
	
	public void setValue(int index, int value) {
		System.arraycopy(getValues(), 0, writeBuffer, 0, getLength());
		writeBuffer[index] = value;
		writeValues();
	}

	@Override
	public void update(Observable observable) {
		if(observable == property) {
			refreshValues();
			for(IPropertyIntValueListener l : listeners)
				l.valuesChanged(readBuffer.clone(), getValues());
		}
	}
	
	public void addChangeListener(IPropertyIntValueListener l) {
		this.listeners.add(l);
	}
	
	public void removeChangeListener(IPropertyIntValueListener l) {
		this.listeners.remove(l);
	}
}
