package ch.thomson.ucs.tine.server.util;

import java.net.URL;

import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.resource.ResourceSet;
import org.eclipse.emf.ecore.resource.impl.ResourceSetImpl;
import org.eclipse.emf.ecore.xmi.impl.XMIResourceFactoryImpl;

import ch.thomson.ucs.tine.server.FrontEndController;
import ch.thomson.ucs.tine.server.ServerPackage;

public class ServerResourceLoader {
	private static FrontEndController fec;
	private static URL url;
	
	/**
	 * Diese Methode gibt lediglich eine FrontEndController-Instanz aus dem
	 * "SignalList.server" File zurck. Weil diese Methode auch von den
	 * Client-Plugins benutzt wird, um die Server-Konfiguration zu lesen, wurde
	 * auf dieser Instanz die initialize() und run() Methoden nicht aufgerufen,
	 * welche die TINE-Maschinerie in Gang setzen wrde. Dies bedeutet, dass
	 * dies bei der wirklichen Server-Instanz vom Aufrufer gemacht werden muss.
	 * 
	 * @return Einzige Front End Controller Instanz (Singleton)
	 * @throws Exception
	 */
	public static synchronized FrontEndController getFrontEndControllerInstance() throws Exception {
		if(fec == null) {
			// Modell initialisieren
			ServerPackage.eINSTANCE.eClass();

			// Neuen ResourceSet erzeugen
			ResourceSet resourceSet = new ResourceSetImpl();
			resourceSet.getResourceFactoryRegistry().getExtensionToFactoryMap().put("server", new XMIResourceFactoryImpl());

			// Konfiguration aus dem Fragment  laden
			url = ServerResourceLoader.class.getResource("/config/SignalList.server");			
			Resource resource = resourceSet.getResource(URI.createURI(url.toString()), true);
			fec = (FrontEndController) resource.getContents().get(0);			
		}		
		return fec;
	}
	
	public static URI createURI() throws Exception {
		if(url == null) 
			getFrontEndControllerInstance();
		return URI.createURI(url.toString());
	}
}
