package ch.thomson.ucs.utils.logging;

import org.apache.log4j.Logger;
import org.eclipse.core.runtime.ILogListener;
import org.eclipse.core.runtime.IStatus;

/**
 * Diese Klasse leitet die Meldungen aus der Eclipse Logging API an Log4J
 * weiter. Typischerweise wird diese Klasse in der Start-Methode der Activator
 * Klasse instanziert und anschliessend mit
 * <code>Platform.addLogListener(log4jTransfer)</code> bei der Plattform als Beobachter
 * angemeldet.
 * 
 * @author Urs Fischer
 */
public class Log4jTransfer implements ILogListener {

	private Logger logger = Logger.getLogger(Log4jTransfer.class);
	private static Log4jTransfer instance;
	
	private Log4jTransfer(){};
	
	public static synchronized Log4jTransfer getInstance() {
		if(instance == null) {
			instance = new Log4jTransfer();
		}
		return instance;
	}

	@Override
	public void logging(IStatus status, String plugin) {
		String message = "[" + plugin + "] " + status.getMessage(); //$NON-NLS-1$ //$NON-NLS-2$
		Throwable exception = status.getException();
		switch (status.getSeverity()) {
		case IStatus.CANCEL:
			logger.error(message, exception);
			break;
		case IStatus.ERROR:
			logger.error(message, exception);
			break;
		case IStatus.INFO:
			logger.info(message, exception);
			break;
		case IStatus.OK:
			logger.info(message, exception);
			break;
		case IStatus.WARNING:
			logger.warn(message, exception);
			break;
		default:
			logger.info(message, exception);
			break;
		}
	}
}