package ch.thomson.ucs.utils.logic;

import java.io.Closeable;
import java.io.DataInput;
import java.io.DataInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.nio.ByteOrder;

/**
 * Diese Klasse dekoriert die Java-Klasse DataInputStream, da eine direkte
 * Vererbung aufgrund der als Final deklarierten Methoden nicht mglich ist.
 * Standardmssig interpretiert die Java-Klasse die Byte Order als Big Endian.
 * Mit Hilfe dieser Erweiterung kann dies beliebig umgestellt werden.
 * 
 * @author FischerU
 * 
 */
public class ByteOrderedDataInputStream implements DataInput, Closeable {
	protected ByteOrder byteOrder = ByteOrder.BIG_ENDIAN; // Default
	protected DataInputStream dis;
	
	public ByteOrderedDataInputStream(InputStream is) {
		dis = new DataInputStream(is);
	}
	
	public ByteOrderedDataInputStream(InputStream is, ByteOrder byteOrder) {
		this(is);
		this.byteOrder = byteOrder;
	}
	
	public ByteOrder getByteOrder() {
		return byteOrder;
	}

	public void setByteOrder(ByteOrder byteOrder) {
		this.byteOrder = byteOrder;
	}

	@Override
	public boolean readBoolean() throws IOException {
		return dis.readBoolean();
	}

	@Override
	public byte readByte() throws IOException {
		return dis.readByte();
	}

	@Override
	public char readChar() throws IOException {
		return this.readChar();
	}

	@Override
	public double readDouble() throws IOException {
		return byteOrder == ByteOrder.LITTLE_ENDIAN ? ByteSwapper.swap(dis.readDouble()) : dis.readDouble();
	}

	@Override
	public float readFloat() throws IOException {
		return byteOrder == ByteOrder.LITTLE_ENDIAN ? ByteSwapper.swap(dis.readFloat()) : dis.readFloat();
	}

	@Override
	public void readFully(byte[] b) throws IOException {
		dis.readFully(b);
	}

	@Override
	public void readFully(byte[] b, int off, int len) throws IOException {
		dis.readFully(b, off, len);
	}

	@Override
	public int readInt() throws IOException {
		return byteOrder == ByteOrder.LITTLE_ENDIAN ? ByteSwapper.swap(dis.readInt()) : dis.readInt();
	}

	@Deprecated
	@Override
	public String readLine() throws IOException {
		return dis.readLine();
	}

	@Override
	public long readLong() throws IOException {
		return byteOrder == ByteOrder.LITTLE_ENDIAN ? ByteSwapper.swap(dis.readLong()) : dis.readLong();
	}

	@Override
	public short readShort() throws IOException {
		return byteOrder == ByteOrder.LITTLE_ENDIAN ? ByteSwapper.swap(dis.readShort()) : dis.readShort();
	}

	@Override
	public String readUTF() throws IOException {
		return dis.readUTF();
	}

	@Override
	public int readUnsignedByte() throws IOException {
		return byteOrder == ByteOrder.LITTLE_ENDIAN ? ByteSwapper.swap(dis.readUnsignedByte()) : dis.readUnsignedByte();
	}

	@Override
	public int readUnsignedShort() throws IOException {
		return byteOrder == ByteOrder.LITTLE_ENDIAN ? ByteSwapper.swap(dis.readUnsignedShort()) : dis.readUnsignedShort();
	}

	@Override
	public int skipBytes(int n) throws IOException {
		return dis.skipBytes(n);
	}

	@Override
	public void close() throws IOException {
		dis.close();
	}
}
