/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of Java-Common.
 *
 * Java-Common is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Java-Common is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Java-Common.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.application.state.impl;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.Iterator;
import java.util.List;

import org.xml.sax.InputSource;

import com.cosylab.application.state.State;
import com.cosylab.application.state.StateStorage;
/**
 * This class is an implementation of the <code>StateStorage</code> that uses a
 * XML file to store the array of the <code>State</code> objects.
 *
 * Use the <code>csl.XMLStateStorage.useNew=true</code> system property to switch
 * to the new implementation of the state storage (faster, with nicer output, uses SAX instead of DOM).
 *  
 * @author dvitas
 * @author mkadunc
 */
public class XMLStateStorage extends DefaultStateStorage {
    private static final boolean USE_NEW="true".equalsIgnoreCase(System.getProperty("csl.XMLStateStorage.useNew"));
    
    /**
     * DOCUMENT ME!
     *
     * @param is DOCUMENT ME!
     *
     * @return DOCUMENT ME!
     *
     * @throws IOException DOCUMENT ME!
     */
    public static final State[] loadStates(InputSource is) throws IOException
    {
        if (USE_NEW) return NewXMLStateStorage.loadStates(is);
        else return OldXMLStateStorage.loadStates(is);
    }

    /**
     * DOCUMENT ME!
     *
     * @param outStates DOCUMENT ME!
     * @param os DOCUMENT ME!
     *
     * @throws IOException DOCUMENT ME!
     */
    public static final void storeStates(State[] outStates, OutputStream os)
    	throws IOException
    {
        if (USE_NEW) NewXMLStateStorage.storeStates(outStates, os);
        else OldXMLStateStorage.storeStates(outStates, os);
    }

    /**
    		             *
    		             */
    private DefaultStateStorage ref=null;
    public XMLStateStorage()
    {
    	super();
    	ref=USE_NEW?new NewXMLStateStorage():new OldXMLStateStorage();
    }

    /**
     * Creates a new XMLStateStorage object.
     *
     * @param ss DOCUMENT ME!
     */
    public XMLStateStorage(StateStorage ss)
    {
    	super(ss);
        ref=USE_NEW?new NewXMLStateStorage(ss):new OldXMLStateStorage(ss);
    }

    public void add(State state) {
        ref.add(state);
    }

    public void addAll(List states) {
        ref.addAll(states);
    }

    public boolean contains(State st) {
        return ref.contains(st);
    }

    public boolean equals(Object obj) {
        return ref.equals(obj);
    }

    public List getStates() {
        return ref.getStates();
    }

    public int hashCode() {
        return ref.hashCode();
    }

    public Iterator iterator() {
        return ref.iterator();
    }

    public void load(InputStream stream) throws IOException {
        ref.load(stream);
    }

    public void load(String filePath, String applicationName)
            throws IOException {
        ref.load(filePath, applicationName);
    }

    public void remove(State st) {
        ref.remove(st);
    }

    public void store(OutputStream stream) throws IOException {
        ref.store(stream);
    }

    public void store(String filePath, String applicationName)
            throws IOException {
        ref.store(filePath, applicationName);
    }

    public String toString() {
        return ref.toString();
    }
}
