/**
 *  Dec 16, 2005 - Dec 22, 2005	ws:	Created from scratch
 *  Feb 10, 2005 - Feb 10, 2005	ws: added getMaschineByString(..)
 *  Nov 29, 2007 - Nov 29, 2007	ws: added Maschinen 6 - 10, support for sorted
 *  Dec  5, 2007 - Dec  5, 2007 kh, ws: Logbuch support verbessert, = flash
 *	Feb 15, 2008 - Feb 15, 2008 ws: added "@Override"es
 *	Jan  9, 2009 - Jan  9, 2009	ws: added hasLogbuch()
 *	Sep  1, 2009 - Sep  2, 2009	ws: added bkrWikiName support
 *	Sep 14, 2009 - Sep 14, 2009	ws: added Maschinen Com, Pit AnLa Wunsch in AppMeeting&eMail von Heute
 */
package de.desy.acop.print.logbook;

import java.util.Arrays;
import java.util.concurrent.atomic.AtomicReference;

/**
 * Enumeration aller gltigen Maschinen:			<br>
 * L2(1), D2(2), Pe(3), Do(4)						<br>
 * falls nicht bekannt oder nicht anwendbar: NA(0)	<br>
 * 													<br>
 * Die Applikation bekommt es durch die Applikations Ini Datei.
 * Einfach zu den properties versionUID, webAddress usw. folgendes einfgen:	<br>
 * <pre>
 * &lt;void property="maschinenNr"&gt;
 *    &lt;int&gt;1&lt;/int&gt;
 * &lt;/void&gt;
 *</pre>											<br>
 *(mit der richtigen maschinenNr natrlich).		<br>
 * Die im Augenblick gewhlte Maschine findet man durch <code>Maschine.getMaschine()</code>
 * ein Beispiel ist zu finden in:<br>
 * &linkhttp //mstweb/ini/~~testAppMitAppIniOnWeb/~~testAppMitAppIniOnWeb.xml
 * @author ws
 * @version 1.0.0
 */
public enum Maschine {							  //langer-, 	logBuch-, bkrWiki-Name
	  /**nicht anwendbar:	maschinenNr  0*/	NA ("NA",		""		, "")
	, /**Linac 2: 			maschinenNr  1*/	L2 ("Linac 2",	"LINAC2", "LINAC_II")
	, /**Desy 2:  			maschinenNr  2*/	D2 ("Desy 2",	"DESY2" , "DESY_II")
	, /**Petra:   			maschinenNr  3*/	Pe ("Petra",	"PETRA" , "PETRA_III")
	, /**Doris:   			maschinenNr  4*/	Do ("Doris",	"DORIS" , "DORIS_III")
	, /**Mst:				maschinenNr  5*/	Mst("Mst",		""      , "")
	, /**Pia:				maschinenNr  6*/	Pia("Pia",		"LINAC2", "PIA")
	, /**L-Weg:				maschinenNr  7*/	Lw ("L-Weg",	"DESY2" , "DESY_II")
	, /**E-Weg:				maschinenNr  8*/	Ew ("E-Weg",	"PETRA" , "PETRA_III")
	, /**R-Weg:				maschinenNr  9*/	Rw ("R-Weg",	"DORIS" , "DORIS_III")
	, /**Flash:				maschinenNr 10*/	Fl ("Flash",	"TTF"   , "FLASH"){
		@Override public String getLogBuchNameBild(){return "//adprint/ttflog";}
		@Override public String getLogBuchNameText(){return "//adprint/ttflog";}
	  }
	, /**Common-Service:	maschinenNr 11*/	CS ("Common-Service",	"", "")	//AnLa 2009-09-14
	, /**Pitz:				maschinenNr 12*/	PZ ("Pitz",		"", 		"")	//AnLa 2009-09-14
	, /**Regae:				maschinenNr 13*/	Rg ("Regae",	"REGAE", 	"")	//KiHi 2010-10-27
	;

//-----------------------------------------------------------------------------
	private static AtomicReference<Maschine> maschine = new AtomicReference<Maschine>(Maschine.NA);
	private static Maschine[] _valuesSorted = {
			L2, Pia
		  , D2, Lw
		  , Pe, Ew
		  , Do, Rw
		  , Fl
		  , Rg
		  , PZ
		  , CS , Mst, NA
											 };
	public final String langerName;
	public final String logBuchName;
	public final String bkrWikiName;
//=============================================================================
	/**@throws NullPointerException, falls mindestens einer der Argumente null ist.*/
	private Maschine(final String langerName, final String logBuchName, final String bkrWikiName) {
		if(langerName  == null){throw new NullPointerException("Der lange Name kann nicht null sein.");}
		if(logBuchName == null){throw new NullPointerException("Der Logbuchname kann nicht null sein.");}
		this.langerName  = langerName;
		this.logBuchName = logBuchName;
		this.bkrWikiName = bkrWikiName;
	}
//=============================================================================
	/**true falls es eine spezifische Maschine wie PETRA ist*/
	public final boolean istMaschine(){return Maschine.istMaschine(this);}
	/**int Maschinen Nummer*/
	public int getMaschineNr() {return this.ordinal();}
//=============================================================================
	/**Maschine dieser Applikation, nie null use MaschineMgr Object .getMaschine() instead*/
	@Deprecated public static final Maschine getMaschine() {return maschine.get();}
	/**Maschine dieser Applikation @throws NullPointerException*/
	static final void setMaschine(final Maschine maschine) {
		if(maschine == null)
			throw new NullPointerException("Die Maschine kann nicht null sein.");
		Maschine.maschine.set(maschine);
	}
	/**true falls es eine spezifische Maschine wie PETRA ist*/
	public static boolean istMaschine(final Maschine maschine) {
		return (maschine != null)
			&& (maschine != Maschine.NA)
			&& (maschine != Maschine.Mst);
	}
	/**Nimmt das Framework falls die Maschine im Mgr oder via ini xml Datei
	 * nicht anders gesetzt wird.*/
	public static Maschine getDefaultMaschine() {return Maschine.NA;}
	/**Sinnvoll Sortiert fuer Menues etc. (Reihenfolge KH)*/
	public static final Maschine[] valuesSorted(){
		return Arrays.copyOf(_valuesSorted, _valuesSorted.length);
	}
	//-------------------------------------------------------------------------
	/**Implementation fr MaschineMgr @throws IndexOutOfBoundsException,
	 * if not isValidMaschineNr(maschineNr)*/
	static Maschine getMaschineByNr(final int maschineNr){
		return Maschine.values()[maschineNr];
	}
	/**Implementation fr MaschineMgr*/
	public static Maschine getMaschineByString(final String maschinenString){
		if(maschinenString == null || maschinenString.isEmpty())
			return Maschine.getDefaultMaschine();
		final String candMaschine = toLowerAndClean(maschinenString);
		for (final Maschine maschine: Maschine.values())
			if(maschine.hasSuchName(candMaschine))
				return maschine;
		return Maschine.getDefaultMaschine();
	}
	/**candMaschine toLowerAndClean was called on it*/
	private boolean hasSuchName(final String candMaschine) {
		return normalizedAliasEquals(this.name(),	 		candMaschine)
			|| normalizedAliasEquals(this.getLangenNamen(), candMaschine)
			|| normalizedAliasEquals(this.getLogBuchName(), candMaschine)
			|| normalizedAliasEquals(this.getBkrWikiName(), candMaschine)
			;
	}
	private static boolean normalizedAliasEquals(final String maschineNameAlias, final String candMaschine) {
		return toLowerAndClean(maschineNameAlias).equals(candMaschine);
	}
	/**maschinenString may not be null<br>
	 * removes blanks and capitalizes elements*/
	static String toLowerAndClean(final String maschinenString) {
		return maschinenString.toLowerCase().replace(" ","");
	}
	static boolean isValidMaschineNr(final int maschineNr){
		return (0 <= maschineNr)   &&   (maschineNr < Maschine.values().length);
	}
//=============================================================================
	public final String getLogBuchName() {return this.logBuchName;}
	public final boolean hasLogbuch() {return !getLogBuchName().trim().isEmpty();}
	public String getLogBuchNameBild(){
		return "//adprint/" + getLogBuchName() + "log";
	}
	public String getLogBuchNameText(){
		return "//adprint/" + getLogBuchName() + "log";
	}
	public final String getLangenNamen() {return this.langerName;}
	public final String getBkrWikiName() {return this.bkrWikiName;}
	public final int ordinalSorted() {
		for (int i = 0; i < Maschine._valuesSorted.length; i++) {
			if(this == Maschine._valuesSorted[i]){
				return i;
			}
		}
		return values().length;	//falls Maschine vergessen nach ganz hinten
	}
	public boolean hasBKRWikiAcceleratorWithControlsPage() {
		return this.bkrWikiName != null && !this.bkrWikiName.isEmpty();
	} 
	
	/**
	 * Returns the maschine that matches the given ens string. ENS 
	 * string is the name of the machine as it is registered in the ENS.
	 * 
	 * @param ensString
	 * @return
	 */
	public static Maschine ensStringToMachine(String ensString) {
		if ("DESY2".equalsIgnoreCase(ensString)) {
			return Maschine.D2;
		} else if ("DESY3".equalsIgnoreCase(ensString)) {
			
		} else if ("DORIS".equalsIgnoreCase(ensString)) {
			return Maschine.Do;
		} else if ("FLASH".equalsIgnoreCase(ensString)) {
			return Maschine.Fl;
		} else if ("HERA".equalsIgnoreCase(ensString)) {
			
		} else if ("LINAC2".equalsIgnoreCase(ensString)) {
			return Maschine.L2;
		} else if ("MTF".equalsIgnoreCase(ensString)) {
			
		} else if ("PETRA".equalsIgnoreCase(ensString)) {
			return Maschine.Pe;
		} else if ("PETRA2".equalsIgnoreCase(ensString)) {
			return Maschine.Pe;
		} else if ("PETRA.EXT".equalsIgnoreCase(ensString)) {
			return Maschine.Pe;
		} else if ("PITZ".equalsIgnoreCase(ensString)) {
			return Maschine.PZ;
		} else if ("REGAE".equalsIgnoreCase(ensString)) {
			return Maschine.Rg;
		} else if ("TTF".equalsIgnoreCase(ensString)) {
			return Maschine.Fl;
		} else if ("FLASH".equalsIgnoreCase(ensString)) {
			return Maschine.Fl;
		} else if ("SERVICE".equalsIgnoreCase(ensString)) {
			return Maschine.CS;
		}
		return Maschine.NA;
	}
}
