package de.desy.acop.video;

import java.awt.Color;

import de.desy.acop.displayers.tools.AcopDisplayerParameters;
import de.desy.acop.transport.ConnectionParameters;
import de.desy.acop.video.displayer.ColorMap;
import de.desy.acop.video.displayer.GridMode;
import de.desy.acop.video.displayer.ImageDisplayer;
import de.desy.acop.video.displayer.ImageZoom;
import de.desy.acop.video.displayer.OverlayState;

/**
 * 
 * <code>AcopVideoParameters</code> are parameters used by the AcopVideo.
 * <p>
 * They hold information about the {@link ConnectionParameters} and the specific
 * AcopVideo features.
 * 
 * @author <a href="mailto:jaka.bobnar@cosylab.com">Jaka Bobnar</a>
 * @version $Id: Templates.xml,v 1.10 2004/01/13 16:17:13 jbobnar Exp $
 * 
 */
public class AcopVideoParameters extends AcopDisplayerParameters {

	private static final long serialVersionUID = 1L;

	/**
	 * Constructs a new AcopVideoParameters with all properties false and
	 * grayscale color map.
	 * 
	 * @param parameters
	 */
	public AcopVideoParameters(ConnectionParameters parameters) {
		this(parameters, //
				ImageDisplayer.DEFAULT_AOI_ZOOM, //
				ImageDisplayer.DEFAULT_IMAGE_ZOOM, //
				ImageDisplayer.DEFAULT_OVERLAY_STATE, //
				ImageDisplayer.DEFAULT_KEEP_ASPECT_RATIO, //
				ImageDisplayer.DEFAULT_HISTOGRAM_EQUALISATION, //
				ImageDisplayer.DEFAULT_HISTOGRAM_MIN, // 
				ImageDisplayer.DEFAULT_HISTOGRAM_MAX, // 
				ImageDisplayer.DEFAULT_COLOR_MAP, //
				ImageDisplayer.DEFAULT_GRID_COLOR, // 
				ImageDisplayer.DEFAULT_GRID_MODE, //
				ImageDisplayer.DEFAULT_GRID_SIZE, //
				ImageDisplayer.DEFAULT_GRID_VISIBLE, //
				ImageDisplayer.DEFAULT_SHOW_DROPPED_FRAMES_NUMBER, //
				false);
	}

	/**
	 * Constructs a new AcopVideoParameters.
	 * 
	 * @param parameters
	 * @param aoiZoom
	 * @param imageZoom
	 * @param overlayState
	 * @param keepAspectRatio
	 * @param histogramEqualisation
	 * @param histogramMin
	 * @param histogramMax
	 * @param colorMap
	 * @param gridColor
	 * @param gridMode
	 * @param gridSize
	 * @param gridVisible
	 * @param showDroppedFramesNumber
	 * @param displayClutHeader
	 */
	public AcopVideoParameters(ConnectionParameters parameters, //
			boolean aoiZoom, //
			ImageZoom imageZoom, //
			OverlayState overlayState, //
			boolean keepAspectRatio, //
			boolean histogramEqualisation, //
			int histogramMin, //
			int histogramMax, //
			ColorMap colorMap, //
			Color gridColor, //
			GridMode gridMode, //
			double gridSize, //
			boolean gridVisible, //
			boolean showDroppedFramesNumber, //
			boolean displayClutHeader) {

		super("", 0, parameters, 0, 0, 0, "", "%f");

		addParameter(ImageDisplayer.PROPERTY_AOI_ZOOM, aoiZoom);
		addParameter(ImageDisplayer.PROPERTY_IMAGE_ZOOM, imageZoom);
		addParameter(ImageDisplayer.PROPERTY_OVERLAY_STATE, overlayState);
		addParameter(ImageDisplayer.PROPERTY_KEEP_ASPECT_RATIO, keepAspectRatio);
		addParameter(ImageDisplayer.PROPERTY_HISTOGRAM_EQUALISATION, histogramEqualisation);
		addParameter(ImageDisplayer.PROPERTY_HISTOGRAM_MIN, histogramMin);
		addParameter(ImageDisplayer.PROPERTY_HISTOGRAM_MAX, histogramMax);
		addParameter(ImageDisplayer.PROPERTY_COLOR_MAP, colorMap);
		addParameter(ImageDisplayer.PROPERTY_GRID_COLOR, gridColor);
		addParameter(ImageDisplayer.PROPERTY_GRID_MODE, gridMode);
		addParameter(ImageDisplayer.PROPERTY_GRID_SIZE, gridSize);
		addParameter(ImageDisplayer.PROPERTY_GRID_VISIBLE, gridVisible);
		addParameter(ImageDisplayer.PROPERTY_SHOW_DROPPED_FRAMES_NUMBER, showDroppedFramesNumber);
		addParameter(AcopVideo.PROPERTY_DISPLAY_CLUT_HEADER, displayClutHeader);
	}

	/**
	 * Returns the value of aoiZoom property.
	 * 
	 * @return
	 */
	public boolean isAOIZoom() {
		return getBool(ImageDisplayer.PROPERTY_AOI_ZOOM);
	}

	/**
	 * Returns the value of the overlay state. mdavid
	 * 
	 * @return
	 */
	public ImageZoom getImageZoom() {
		return (ImageZoom) getParameter(ImageDisplayer.PROPERTY_IMAGE_ZOOM);
	}

	/**
	 * Returns the value of the overlay state. mdavid
	 * 
	 * @return
	 */
	public OverlayState getOverlayState() {
		return (OverlayState) getParameter(ImageDisplayer.PROPERTY_OVERLAY_STATE);
	}

	/**
	 * Returns the value of the keep aspect ratio property.
	 * 
	 * @return
	 */
	public boolean isKeepAspectRatio() {
		return getBool(ImageDisplayer.PROPERTY_KEEP_ASPECT_RATIO);
	}

	/**
	 * Returns the value of the histogram equalization.
	 * 
	 * @return
	 */
	public boolean isHistogramEqualisation() {
		return getBool(ImageDisplayer.PROPERTY_HISTOGRAM_EQUALISATION);
	}

	/**
	 * Returns min value of the histogram.
	 * 
	 * @return
	 */
	public int getHistogramMin() {
		return (Integer) getParameter(ImageDisplayer.PROPERTY_HISTOGRAM_MIN);
	}

	/**
	 * Returns max value of the histogram.
	 * 
	 * @return
	 */
	public int getHistogramMax() {
		return (Integer) getParameter(ImageDisplayer.PROPERTY_HISTOGRAM_MAX);
	}

	/**
	 * Returns the color map.
	 * 
	 * @return
	 */
	public ColorMap getColorMap() {
		return (ColorMap) getParameter(ImageDisplayer.PROPERTY_COLOR_MAP);
	}

	/** @see {@link ImageDisplayer#getGridColor()} */
	public Color getGridColor() {
		return (Color) getParameter(ImageDisplayer.PROPERTY_GRID_COLOR);
	}

	/** @see {@link ImageDisplayer#getGridMode()} */
	public GridMode getGridMode() {
		return (GridMode) getParameter(ImageDisplayer.PROPERTY_GRID_MODE);
	}

	/** @see {@link ImageDisplayer#getGridSize()} */
	public double getGridSize() {
		return (Double) getParameter(ImageDisplayer.PROPERTY_GRID_SIZE);
	}

	/** @see {@link ImageDisplayer#isGridVisible()} */
	public boolean isGridVisible() {
		return getBool(ImageDisplayer.PROPERTY_GRID_VISIBLE);
	}

	/** @see {@link ImageDisplayer#isShowDroppedFramesNumber()} */
	public boolean isShowDroppedFramesNumber() {
		return getBool(ImageDisplayer.PROPERTY_SHOW_DROPPED_FRAMES_NUMBER);
	}

	/**
	 * Returns the value of the visible CLUT header property.
	 * 
	 * @return true if the CLUT header is visible, otherwise false
	 */
	public boolean isDisplayClutHeader() {
		return getBool(AcopVideo.PROPERTY_DISPLAY_CLUT_HEADER);
	}
}
