/**
 * 
 */
package de.desy.acop.video;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.Rectangle;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.awt.image.BufferedImage;
import java.beans.Customizer;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.File;
import java.io.IOException;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;
import java.util.concurrent.Executor;
import java.util.concurrent.Executors;

import javax.imageio.ImageIO;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JFileChooser;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JTabbedPane;
import javax.swing.JTable;
import javax.swing.SwingConstants;
import javax.swing.SwingUtilities;
import javax.swing.border.LineBorder;
import javax.swing.border.TitledBorder;
import javax.swing.filechooser.FileFilter;
import javax.swing.table.DefaultTableModel;

import com.cosylab.gui.IconCustomizer;
import com.cosylab.gui.components.NumberField;
import com.cosylab.gui.components.util.RunnerHelper;

import de.desy.acop.video.analysis.AnalysisImage;
import de.desy.acop.video.analysis.AnalysisMode;
import de.desy.acop.video.analysis.AnalysisType;
import de.desy.acop.video.analysis.ImageAnalysisEngine;
import de.desy.acop.video.timageio.TImageIO;
import de.desy.tine.types.IMAGE;
import de.desy.tine.types.IMAGE.FrameHeader;

/**
 * </code>AnalysisCustomizer</code> allows customization of analysis parameters
 * for an </code>AcopVideo</code>.
 * 
 * @author Tilen Kusterle, Cosylab
 *
 */
public class AnalysisCustomizer extends JPanel implements Customizer {

	private class ServerSettingsModel extends DefaultTableModel {
		private static final long serialVersionUID = 1L;
		private AnalysisImage image;
		public void setAImage(AnalysisImage image) {
			this.image = image;
			fireTableDataChanged();
		}
		@Override
		public Class<?> getColumnClass(int columnIndex) {
			return String.class;
		}
		@Override
		public int getColumnCount() {
			return 2;
		}
		@Override
		public int getRowCount() {
			return 11;
		}
		@Override
		public boolean isCellEditable(int row, int column) {
			return false;
		}
		@Override
		public String getColumnName(int column) {
			switch (column) {
				case 0 : return "Parameter";
				case 1 : return "Value";
				default : return null;
			}
		}
		@Override
		public Object getValueAt(int row, int column) {
			if (column == 0) {
				switch (row) {
					case 0 : return "ROI X";
					case 1 : return "ROI Y";
					case 2 : return "ROI W";
					case 3 : return "ROI H";
					case 4 : return "Threshold ROI X";
					case 5 : return "Threshold ROI Y";
					case 6 : return "Threshold ROI W";
					case 7 : return "Threshold ROI H";
					case 8 : return "Threshold";
					case 9 : return "Calculate Threshold";
					case 10 : return "Perform Fit";
				}
			} else if (column == 1 && image != null) {
				switch (row) {
    				case 0 : return image.getRoiX();
    				case 1 : return image.getRoiY();
    				case 2 : return image.getRoiW();
    				case 3 : return image.getRoiH();
    				case 4 : return image.getRoi2X();
    				case 5 : return image.getRoi2Y();
    				case 6 : return image.getRoi2W();
    				case 7 : return image.getRoi2H();
    				case 8 : return image.getThreshold();
    				case 9 : return image.isCalculateThreshold();
    				case 10 : return image.isPerformFit();
    			}
			}
			return null;
		}
	}
	
	private static final long serialVersionUID = 1L;
	private static final String NO_BACKGROUND = "No Background";
	private static final String INVALID_BACKGROUND = "Invalid Background";
	private static final String VIDEO_BACKGROUND = "from video";
	private static final int PREVIEW_DIMENSION = 100;
	private static final Dimension FIELD_DIMENSION = new Dimension(80,21);
	private static final Dimension LABEL_DIMENSION = new Dimension(200,25);
	private static final String[] FILE_TYPES = new String[] {"jpg", "jpeg", "png"};
	
	private Executor executor;
	private AcopVideo acopVideo;
	
	private JComboBox analysisModeCombo;
	private JComboBox analysisImageTypeCombo;
	private JButton applyToServerButton;
	private JButton synchronizeButton;
	private NumberField thresholdField;
	private NumberField roiXField;
	private NumberField roiYField;
	private NumberField roiWField;
	private NumberField roiHField;
	private NumberField roi2XField;
	private NumberField roi2YField;
	private NumberField roi2WField;
	private NumberField roi2HField;
	private JCheckBox calculateThresholdCheckBox;
	private JCheckBox useThresholdCheckBox;
	private JLabel numberOfPixelsLabel;
	private JButton allButton;
	private JButton allThresholdButton;
	private JButton duplicateThresholdButton;
	private JButton currentButton;
	private JButton saveButton;
	private JButton noneButton;
	private JButton browseButton;
	private JLabel backgroundLabel;
	private JLabel fileNameLabel;
	private JFileChooser fileChooser;
	private JCheckBox displayNumericalValuesBox;
	private JCheckBox performFitCheckBox;
	private JCheckBox performSmoothCheckBox;
	private JCheckBox displayStatisticsCheckBox;
	
	private JTabbedPane tabPane;
	private JPanel analysisPanel;
	private JPanel backgroundPanel;
	private JPanel thresholdPanel;
	private JPanel serverSettingsPanel;
	private ServerSettingsModel serverModel;
	
	private File backgroundFile;
	private double[] preciseBackground;
	private int preciseBackgroundWidth;
	private IMAGE backgroundImage;
	private boolean localChanges = false;
	private String fileName;
	private int roiX;
	private int roiY;
	private int roiW;
	private int roiH;
	private int roi2X;
	private int roi2Y;
	private int roi2W;
	private int roi2H;
	private int imageW;
	private int imageH;
	private double threshold;
	private int thresholdPoints;
	private boolean calculateThreshold;
	private boolean performFit;
	private boolean performSmoothing;
	
	private boolean appliedToServer = true;
	private boolean suppressChanges = false;
	private double lastThreshold = 0;
	
	private Font plainFont;
	private Font boldFont;
	private Font labelPlainFont;
	private Font labelBoldFont;
	
	private static final String PIXELS_USED = "Pixels Used: ";
	
	// TODO: mdavid - can be instantiated as local variable, instead of using as class member variable (?)
//	private TImageConverter imageConverter = new TImageConverter(); // mdavid: added

	/* (non-Javadoc)
	 * @see java.beans.Customizer#setObject(java.lang.Object)
	 */
	@Override
	public void setObject(Object bean) {
		if (!(bean instanceof AcopVideo)) {
			throw new IllegalArgumentException("Only AcopVideo can use this Customizer.");
		}
		acopVideo = (AcopVideo) bean;
		
		AnalysisMode am = acopVideo.getAnalysisMode();
		if (am == AnalysisMode.REMOTE_ANALYSIS) fileName = VIDEO_BACKGROUND;
		else fileName = NO_BACKGROUND;
		getAnalysisModeCombo().setSelectedItem(am);
		AnalysisType type = acopVideo.getAnalysisType();
		getAnalysisImageTypeCombo().setSelectedItem(type);
		getAnalysisImageTypeCombo().setEnabled(am != AnalysisMode.NO_ANALYSIS);
		getDisplayStatisticsCheckBox().setEnabled(type == AnalysisType.GAUSSIAN);
		getPerformFitCheckBox().setSelected(type == AnalysisType.INTERFEROMETER_V ||
				type == AnalysisType.INTERFEROMETER_H);
		
		getDisplayNumericalValuesBox().setSelected(acopVideo.isDisplayNumericalValues());
		
		AnalysisImage ai = acopVideo.getAImage();
		if (ai != null) {
			updateAnalysisData(ai);
		} else {
			resetData();
		}
		updateBackground();
		acopVideo.addPropertyChangeListener(new PropertyChangeListener() {
			@Override
			public void propertyChange(final PropertyChangeEvent evt) {
				SwingUtilities.invokeLater(new Runnable() {
					public void run() {
						if (evt.getPropertyName() == AcopVideo.PROPERTY_AIMAGE) {
							AnalysisImage ai = (AnalysisImage) evt.getNewValue();
							if (ai != null) {
								updateAnalysisData(ai);
							} else {
								resetData();
							}
						} else if (evt.getPropertyName().equals(AcopVideo.DISPLAY_STATISTISC)) {
							getDisplayStatisticsCheckBox().setSelected(acopVideo.isDisplayStatistics());
						}
					
					}
				});
				
			}
		});
		getDisplayStatisticsCheckBox().setSelected(acopVideo.isDisplayStatistics());
		acopVideo.getImageAnalysisEngine().addPropertyChangeListener(ImageAnalysisEngine.PROPERTY_PRECISE_BACKGROUND, new PropertyChangeListener() {
			@Override
			public void propertyChange(PropertyChangeEvent evt) {
				updateBackground();
			}
		});
		acopVideo.getDisplayerManager().addPropertyChangeListener(DisplayerManager.PROPERTY_ON_SCREEN_ROI, new PropertyChangeListener() {
			@Override
			public void propertyChange(PropertyChangeEvent evt) {
				Rectangle r = (Rectangle) evt.getNewValue();
				if (r != null) {
					updateData(true, imageW, imageH, r.x, r.y, r.width, r.height, roi2X, roi2Y, roi2W, roi2H, threshold,thresholdPoints,calculateThreshold, performFit,performSmoothing);
				}
			}
		});
	}
	
	public AnalysisCustomizer() {
		initialize();
	}
	
	private void initialize() {
		setLayout(new GridBagLayout());
		
		add(getAnalysisModeCombo(), new GridBagConstraints(0,0,1,1,1,0,GridBagConstraints.EAST,GridBagConstraints.NONE,new Insets(11,11,2,11),0,0));
		add(getAnalysisImageTypeCombo(), new GridBagConstraints(1,0,1,1,1,0,GridBagConstraints.WEST,GridBagConstraints.NONE,new Insets(11,11,2,11),0,0));
		
		JPanel parametersPanel = new JPanel(new GridBagLayout());
		parametersPanel.setBorder(new TitledBorder("Analysis Parameters"));
		parametersPanel.add(getTabPane(),new GridBagConstraints(0,0,2,1,1,1,GridBagConstraints.CENTER,GridBagConstraints.BOTH,new Insets(2,2,2,2),0,0));
		parametersPanel.add(new JLabel("Synchronize With Server:"), new GridBagConstraints(0,14,2,1,0,0,GridBagConstraints.CENTER,GridBagConstraints.NONE,new Insets(2,2,2,2),0,0));
		JPanel p = new JPanel(new GridBagLayout());
		p.add(getApplyToServerButton(), new GridBagConstraints(0,15,1,1,0,0,GridBagConstraints.CENTER,GridBagConstraints.NONE,new Insets(2,2,2,2),0,0));
		p.add(getSynchronizeButton(), new GridBagConstraints(1,15,1,1,0,0,GridBagConstraints.CENTER,GridBagConstraints.NONE,new Insets(2,2,2,2),0,0));
		parametersPanel.add(p, new GridBagConstraints(0,15,2,1,0,0,GridBagConstraints.CENTER,GridBagConstraints.NONE,new Insets(0,0,0,0),0,0));
		
		add(parametersPanel, new GridBagConstraints(0,3,2,1,1,1,GridBagConstraints.CENTER,GridBagConstraints.BOTH,new Insets(4,11,4,11),0,0));
		
		add(getDisplayNumericalValuesBox(), new GridBagConstraints(0,4,2,1,0,0,GridBagConstraints.CENTER,GridBagConstraints.HORIZONTAL,new Insets(2,11,11,11),0,0));
	}
	
	private JTabbedPane getTabPane() {
		if (tabPane == null) {
			tabPane = new JTabbedPane();
			tabPane.add("Parameters", getAnalysisPanel());
			tabPane.add("Background", getBackgroundPanel());
			tabPane.add("Threshold", getThresholdPanel());
			tabPane.add("Applied Settings", getServerSettingsPanel());
		}
		return tabPane;
	}
	
	private JPanel getServerSettingsPanel() {
		if (serverSettingsPanel == null) {
			serverSettingsPanel = new JPanel(new BorderLayout());
			JTable table = new JTable(getServerModel());
			serverSettingsPanel.add(table.getTableHeader(), BorderLayout.NORTH);
			serverSettingsPanel.add(table, BorderLayout.CENTER);
			table.setMinimumSize(new Dimension(0,0));
			serverSettingsPanel.setMinimumSize(new Dimension(0,0));
		}
		return serverSettingsPanel;
	}
	
	private ServerSettingsModel getServerModel() {
		if (serverModel == null) {
			serverModel = new ServerSettingsModel();
		}
		return serverModel;
	}
	
	private JPanel getAnalysisPanel() {
		if (analysisPanel == null) {
			analysisPanel = new JPanel(new GridBagLayout());
			analysisPanel.add(getDisplayStatisticsCheckBox(), new GridBagConstraints(0,0,1,1,0,1,GridBagConstraints.EAST,GridBagConstraints.NONE,new Insets(5,5,2,2),0,0));
			analysisPanel.add(getPerformFitCheckBox(), new GridBagConstraints(1,0,1,1,0,1,GridBagConstraints.WEST,GridBagConstraints.NONE,new Insets(5,2,2,5),0,0));
			analysisPanel.add(getPerformSmoothCheckBox(), new GridBagConstraints(0,2,2,1,0,1,GridBagConstraints.WEST,GridBagConstraints.NONE,new Insets(5,5,2,5),0,0));
			
			analysisPanel.add(new JLabel("Region of Interest Parameters [pixels]:"), new GridBagConstraints(0,3,2,1,0,1,GridBagConstraints.CENTER,GridBagConstraints.NONE,new Insets(2,5,2,10),0,0));
			
			analysisPanel.add(new JLabel("Start X:"), new GridBagConstraints(0,4,1,1,0,1,GridBagConstraints.WEST,GridBagConstraints.NONE,new Insets(2,5,2,2),0,0));
			analysisPanel.add(getRoiXField(), new GridBagConstraints(1,4,1,1,0,1,GridBagConstraints.WEST,GridBagConstraints.NONE,new Insets(2,2,2,2),0,0));
			analysisPanel.add(new JLabel("Start Y:"), new GridBagConstraints(0,5,1,1,0,1,GridBagConstraints.WEST,GridBagConstraints.NONE,new Insets(2,5,2,2),0,0));
			analysisPanel.add(getRoiYField(), new GridBagConstraints(1,5,1,1,0,1,GridBagConstraints.WEST,GridBagConstraints.NONE,new Insets(2,2,2,2),0,0));
			analysisPanel.add(new JLabel("Width:"), new GridBagConstraints(0,6,1,1,0,1,GridBagConstraints.WEST,GridBagConstraints.NONE,new Insets(2,5,2,2),0,0));
			analysisPanel.add(getRoiWField(), new GridBagConstraints(1,6,1,1,0,1,GridBagConstraints.WEST,GridBagConstraints.NONE,new Insets(2,2,2,2),0,0));
			analysisPanel.add(new JLabel("Height:"), new GridBagConstraints(0,7,1,1,0,1,GridBagConstraints.NORTHWEST,GridBagConstraints.NONE,new Insets(2,5,2,2),0,0));
			analysisPanel.add(getRoiHField(), new GridBagConstraints(1,7,1,1,0,1,GridBagConstraints.NORTHWEST,GridBagConstraints.NONE,new Insets(2,2,2,2),0,0));
			analysisPanel.add(getAllButton(), new GridBagConstraints(0,8,2,1,0,1,GridBagConstraints.NORTH,GridBagConstraints.NONE,new Insets(2,2,2,2),0,0));
		}
		return analysisPanel;
	}
	
	private JPanel getThresholdPanel() {
		if (thresholdPanel == null) {
			thresholdPanel = new JPanel(new GridBagLayout());
			JPanel p1 = new JPanel(new GridBagLayout());
			p1.add(new JLabel("Threshold:"), new GridBagConstraints(2,0,1,1,0,1,GridBagConstraints.WEST,GridBagConstraints.NONE,new Insets(5,2,2,5),0,0));
			p1.add(getThresholdField(), new GridBagConstraints(3,0,1,1,0,1,GridBagConstraints.WEST,GridBagConstraints.NONE,new Insets(5,2,2,5),0,0));
			p1.add(getUseTresholdCheckBox(), new GridBagConstraints(2,1,2,1,0,1,GridBagConstraints.WEST,GridBagConstraints.NONE,new Insets(2,2,0,5),0,0));
			p1.add(getCalculateTresholdCheckBox(), new GridBagConstraints(2,2,2,1,0,1,GridBagConstraints.WEST,GridBagConstraints.NONE,new Insets(2,2,0,5),0,0));
			p1.add(getNumberOfPixelsLabel(), new GridBagConstraints(2,3,2,1,0,1,GridBagConstraints.WEST,GridBagConstraints.NONE,new Insets(2,23,4,5),0,0));
			p1.add(new JLabel("Threshold Calculation Region [pixels]:"), new GridBagConstraints(2,4,2,1,0,1,GridBagConstraints.CENTER,GridBagConstraints.NONE,new Insets(2,2,2,5),0,0));
			p1.add(new JLabel("Start X:"), new GridBagConstraints(2,5,1,1,0,1,GridBagConstraints.WEST,GridBagConstraints.NONE,new Insets(2,2,2,2),0,0));
			p1.add(getRoi2XField(), new GridBagConstraints(3,5,1,1,0,1,GridBagConstraints.WEST,GridBagConstraints.NONE,new Insets(2,2,2,5),0,0));
			p1.add(new JLabel("Start Y:"), new GridBagConstraints(2,6,1,1,0,1,GridBagConstraints.WEST,GridBagConstraints.NONE,new Insets(2,2,2,2),0,0));
			p1.add(getRoi2YField(), new GridBagConstraints(3,6,1,1,0,1,GridBagConstraints.WEST,GridBagConstraints.NONE,new Insets(2,2,2,5),0,0));
			p1.add(new JLabel("Width:"), new GridBagConstraints(2,7,1,1,0,1,GridBagConstraints.WEST,GridBagConstraints.NONE,new Insets(2,2,2,2),0,0));
			p1.add(getRoi2WField(), new GridBagConstraints(3,7,1,1,0,1,GridBagConstraints.WEST,GridBagConstraints.NONE,new Insets(2,2,2,5),0,0));
			p1.add(new JLabel("Height:"), new GridBagConstraints(2,8,1,1,0,1,GridBagConstraints.NORTHWEST,GridBagConstraints.NONE,new Insets(2,2,2,2),0,0));
			p1.add(getRoi2HField(), new GridBagConstraints(3,8,1,1,0,1,GridBagConstraints.NORTHWEST,GridBagConstraints.NONE,new Insets(2,2,2,5),0,0));
			
			JPanel p = new JPanel(new GridBagLayout());
			
			p.add(getAllThresholdButton(), new GridBagConstraints(0,0,1,1,0,0,GridBagConstraints.NORTHEAST,GridBagConstraints.NONE,new Insets(2,2,2,2),0,0));
			p.add(getDuplicateThresholdButton(), new GridBagConstraints(1,0,1,1,0,0,GridBagConstraints.NORTHWEST,GridBagConstraints.NONE,new Insets(2,2,2,2),0,0));
			thresholdPanel.add(p1, new GridBagConstraints(0,0,1,1,0,1,GridBagConstraints.NORTH,GridBagConstraints.BOTH,new Insets(2,2,2,2),0,0));
			thresholdPanel.add(p, new GridBagConstraints(0,1,1,1,0,0,GridBagConstraints.NORTH,GridBagConstraints.NONE,new Insets(2,2,2,2),0,0));
			
//			thresholdPanel.add(new JLabel("Threshold:"), new GridBagConstraints(0,0,1,1,0,0,GridBagConstraints.WEST,GridBagConstraints.NONE,new Insets(5,2,2,5),0,0));
//			thresholdPanel.add(getThresholdField(), new GridBagConstraints(1,0,1,1,0,0,GridBagConstraints.WEST,GridBagConstraints.NONE,new Insets(5,2,2,2),0,0));
//			thresholdPanel.add(getCalculateTresholdCheckBox(), new GridBagConstraints(0,1,2,1,0,0,GridBagConstraints.WEST,GridBagConstraints.NONE,new Insets(2,2,0,2),0,0));
//			thresholdPanel.add(getNumberOfPixelsLabel(), new GridBagConstraints(0,2,2,1,0,0,GridBagConstraints.WEST,GridBagConstraints.NONE,new Insets(2,23,4,2),0,0));
//			
//			thresholdPanel.add(new JLabel("Threshold Calculation Region [pixels]:"), new GridBagConstraints(0,3,2,1,0,0,GridBagConstraints.CENTER,GridBagConstraints.NONE,new Insets(2,2,2,2),0,0));
//			
//			thresholdPanel.add(new JLabel("Start X:"), new GridBagConstraints(0,4,1,1,0,0,GridBagConstraints.WEST,GridBagConstraints.NONE,new Insets(2,2,2,2),0,0));
//			thresholdPanel.add(getRoi2XField(), new GridBagConstraints(1,4,1,1,0,0,GridBagConstraints.WEST,GridBagConstraints.NONE,new Insets(2,2,2,2),0,0));
//			thresholdPanel.add(new JLabel("Start Y:"), new GridBagConstraints(0,5,1,1,0,0,GridBagConstraints.WEST,GridBagConstraints.NONE,new Insets(2,2,2,2),0,0));
//			thresholdPanel.add(getRoi2YField(), new GridBagConstraints(1,5,1,1,0,0,GridBagConstraints.WEST,GridBagConstraints.NONE,new Insets(2,2,2,2),0,0));
//			thresholdPanel.add(new JLabel("Width:"), new GridBagConstraints(0,6,1,1,0,0,GridBagConstraints.WEST,GridBagConstraints.NONE,new Insets(2,2,2,2),0,0));
//			thresholdPanel.add(getRoi2WField(), new GridBagConstraints(1,6,1,1,0,0,GridBagConstraints.WEST,GridBagConstraints.NONE,new Insets(2,2,2,2),0,0));
//			thresholdPanel.add(new JLabel("Height:"), new GridBagConstraints(0,7,1,1,0,0,GridBagConstraints.NORTHWEST,GridBagConstraints.NONE,new Insets(2,2,2,2),0,0));
//			thresholdPanel.add(getRoi2HField(), new GridBagConstraints(1,7,1,1,0,1,GridBagConstraints.NORTHWEST,GridBagConstraints.NONE,new Insets(2,2,2,2),0,0));
		}
		return thresholdPanel;
	}
	
	private JLabel getNumberOfPixelsLabel() {
		if (numberOfPixelsLabel == null) {
			numberOfPixelsLabel = new JLabel(PIXELS_USED + (0));
		}
		return numberOfPixelsLabel;
	}
	
	private JPanel getBackgroundPanel() {
		if (backgroundPanel == null) {
			backgroundPanel = new JPanel(new GridBagLayout());
			backgroundPanel.add(new JLabel("Background Image:"), new GridBagConstraints(0,8,2,1,0,0,GridBagConstraints.CENTER,GridBagConstraints.NONE,new Insets(4,4,4,4),0,0));
			backgroundPanel.add(getFileNameLabel(), new GridBagConstraints(0,9,2,1,0,0,GridBagConstraints.CENTER,GridBagConstraints.HORIZONTAL,new Insets(4,4,8,4),0,0));
			backgroundPanel.add(getBackgroundLabel(), new GridBagConstraints(0,10,1,4,1,1,GridBagConstraints.CENTER,GridBagConstraints.BOTH,new Insets(4,4,4,4),0,0));
			backgroundPanel.add(getCurrentButton(), new GridBagConstraints(1,10,1,1,0,0,GridBagConstraints.NORTH,GridBagConstraints.HORIZONTAL,new Insets(4,4,4,4),0,0));
			backgroundPanel.add(getNoneButton(), new GridBagConstraints(1,11,1,1,0,0,GridBagConstraints.NORTH,GridBagConstraints.HORIZONTAL,new Insets(4,4,4,4),0,0));
			backgroundPanel.add(getBrowseButton(), new GridBagConstraints(1,12,1,1,0,0,GridBagConstraints.NORTH,GridBagConstraints.HORIZONTAL,new Insets(4,4,4,4),0,0));
			backgroundPanel.add(getSaveButton(), new GridBagConstraints(1,13,1,1,0,1,GridBagConstraints.NORTH,GridBagConstraints.HORIZONTAL,new Insets(4,4,4,4),0,0));			
		}
		return backgroundPanel;
	}

	private JComboBox getAnalysisModeCombo() {
		if (analysisModeCombo == null) {
			analysisModeCombo = new JComboBox(AnalysisMode.values());
			analysisModeCombo.addActionListener(new ActionListener() {
				@Override
				public void actionPerformed(ActionEvent e) {
					AnalysisMode mode = (AnalysisMode) analysisModeCombo.getSelectedItem();
					setAnalysisMode(mode);
					
					if (mode == AnalysisMode.LOCAL_ANALYSIS) {
						if (backgroundFile != null)
							updateFileNameLabel(backgroundFile.getAbsolutePath());
						else if (fileName == VIDEO_BACKGROUND) updateFileNameLabel(VIDEO_BACKGROUND);
						else updateFileNameLabel(NO_BACKGROUND);
					}
					else if (mode == AnalysisMode.REMOTE_ANALYSIS) updateFileNameLabel(VIDEO_BACKGROUND);
					else updateFileNameLabel(NO_BACKGROUND);
					getAnalysisImageTypeCombo().setEnabled(mode != AnalysisMode.NO_ANALYSIS);
				}
			});
		}
		return analysisModeCombo;
	}
	
	private JComboBox getAnalysisImageTypeCombo() {
		if (analysisImageTypeCombo == null) {
			analysisImageTypeCombo = new JComboBox(AnalysisType.values());
			analysisImageTypeCombo.addActionListener(new ActionListener() {
				@Override
				public void actionPerformed(ActionEvent e) {
					AnalysisType mode = (AnalysisType) analysisImageTypeCombo.getSelectedItem();
					acopVideo.setAnalysisType(mode);
					getDisplayStatisticsCheckBox().setEnabled(mode == AnalysisType.GAUSSIAN);
					if (mode == AnalysisType.INTERFEROMETER_V || mode == AnalysisType.INTERFEROMETER_H) {
						getPerformFitCheckBox().setSelected(true);
					}
				}
			});
		}
		return analysisImageTypeCombo;
	}
	
	
	
	private JButton getApplyToServerButton() {
		if (applyToServerButton == null) {
			applyToServerButton = new JButton("Apply");
			applyToServerButton.setToolTipText("Apply Parameters to Server");
			applyToServerButton.setEnabled(false);
			applyToServerButton.addActionListener(new ActionListener() {
				@Override
				public void actionPerformed(ActionEvent e) {
					applyToServer();
				}
			});
		}
		return applyToServerButton;
	}
	
	private JButton getSynchronizeButton() {
		if (synchronizeButton == null) {
			synchronizeButton = new JButton("Reload");
			synchronizeButton.setToolTipText("Reload Parameters from Server");
			synchronizeButton.setEnabled(false);
			synchronizeButton.addActionListener(new ActionListener() {
				@Override
				public void actionPerformed(ActionEvent e) {
					appliedToServer = true;
					updateData(true, imageW, imageH, roiX, roiY, roiW, roiH, roi2X, roi2Y, roi2W, roi2H, threshold, thresholdPoints, calculateThreshold, performFit,performSmoothing);
				}
			});
		}

		return synchronizeButton;
	}
	
	private NumberField getThresholdField() {
		if (thresholdField == null) {
			thresholdField = new NumberField();
			thresholdField.setNumberType(Double.class);
			thresholdField.setFormat("%3.2f");
			thresholdField.setPreferredSize(FIELD_DIMENSION);
			thresholdField.setMinimumSize(FIELD_DIMENSION);
			thresholdField.setMaximumSize(FIELD_DIMENSION);
			thresholdField.addPropertyChangeListener("value", new PropertyChangeListener() {
				@Override
				public void propertyChange(PropertyChangeEvent evt) {
					if (suppressChanges) return;
					final double value = (Double) evt.getNewValue();
					if (!Double.isNaN(value)) lastThreshold = value;
					if (threshold == value) {
						if (localChanges) thresholdField.setFont(getPlainFont());
						return;
					}
					if (!localChanges) {
						acopVideo.getImageAnalysisEngine().setThreshold(value);
					} 
					else {
						thresholdField.setFont(getBoldFont());
						appliedToServer = false;
					}
				}
			});
			suppressChanges = true;
			thresholdField.setValue(Double.NaN);
			suppressChanges = false;
		}
		return thresholdField;
	}
	
	private JCheckBox getCalculateTresholdCheckBox() {
		if (calculateThresholdCheckBox == null) {
			calculateThresholdCheckBox = new JCheckBox("Automatic Threshold Calculation");
			calculateThresholdCheckBox.setToolTipText("Automatic Threshold Calculation (Average Pixel Value in [Threshold ROI - ROI])");
			calculateThresholdCheckBox.addItemListener(new ItemListener() {
				public void itemStateChanged(ItemEvent e) {
					if (suppressChanges) return;
					boolean value = e.getStateChange() == ItemEvent.SELECTED;
					if (!localChanges) {
						acopVideo.getImageAnalysisEngine().setCalculateThreshold(value);
					} else {
						calculateThresholdCheckBox.setFont(getLabelBoldFont());
						appliedToServer = false;
					}
					getThresholdField().setEditable(!value && getUseTresholdCheckBox().isSelected());
					getUseTresholdCheckBox().setEnabled(!value);
				}
			});
		}
		return calculateThresholdCheckBox;
	}
	
	private JCheckBox getUseTresholdCheckBox() {
		if (useThresholdCheckBox == null) {
			useThresholdCheckBox = new JCheckBox("Apply Threshold");
			useThresholdCheckBox.setToolTipText("Select Whether Set Threshold is Used in Calculation or Not");
			useThresholdCheckBox.addItemListener(new ItemListener() {
				public void itemStateChanged(ItemEvent e) {
					if (suppressChanges) return;
					boolean value = e.getStateChange() == ItemEvent.SELECTED;
					if (value) {
						acopVideo.getImageAnalysisEngine().setThreshold(lastThreshold);
					} else {
						acopVideo.getImageAnalysisEngine().setThreshold(Double.NaN);
					}
					getThresholdField().setEditable(!value);
				}
			});
			useThresholdCheckBox.setEnabled(getCalculateTresholdCheckBox().isSelected());
		}
		return useThresholdCheckBox;
	}
	
	
	
	private JCheckBox getPerformFitCheckBox() {
		if (performFitCheckBox == null) {
			performFitCheckBox = new JCheckBox("Perform Fitting");
			performFitCheckBox.setToolTipText("Perform Exponential + Linear Function Fit on Side View Data");
			performFitCheckBox.addItemListener(new ItemListener() {
				public void itemStateChanged(ItemEvent e) {
					if (suppressChanges) return;
					boolean value = e.getStateChange() == ItemEvent.SELECTED;
					if (!localChanges) {
						acopVideo.getImageAnalysisEngine().setPerformFit(value);
					} else {
						performFitCheckBox.setFont(getLabelBoldFont());
						appliedToServer = false;
					}
				}
			});
		}
		return performFitCheckBox;
	}
	
	private JCheckBox getPerformSmoothCheckBox() {
		if (performSmoothCheckBox == null) {
			performSmoothCheckBox = new JCheckBox("Perform Smoothing");
			performSmoothCheckBox.setToolTipText("Smooth Side View Data by Averaging Values With Neighbours"); 
			performSmoothCheckBox.addItemListener(new ItemListener() {
				public void itemStateChanged(ItemEvent e) {
					if (suppressChanges) return;
					boolean value = e.getStateChange() == ItemEvent.SELECTED;
					if (!localChanges) {
						acopVideo.getImageAnalysisEngine().setPerformSmoothing(value);
					} else {
						performSmoothCheckBox.setFont(getLabelBoldFont());
						appliedToServer = false;
					}
				}
			});
		}
		return performSmoothCheckBox;
	}
	
	private JCheckBox getDisplayStatisticsCheckBox() {
		if (displayStatisticsCheckBox == null) {
			displayStatisticsCheckBox = new JCheckBox("Display Statisics");
			displayStatisticsCheckBox.setToolTipText("Display Statistical Analysis Results");
			displayStatisticsCheckBox.addItemListener(new ItemListener() {
				public void itemStateChanged(ItemEvent e) {
					if (suppressChanges) return;
					boolean value = e.getStateChange() == ItemEvent.SELECTED;
					acopVideo.setDisplayStatistics(value);
				}
			});
		}
		return displayStatisticsCheckBox;
	}
		
	private NumberField getRoiXField() {
		if (roiXField == null) {
			roiXField = new NumberField();
			roiXField.setNumberType(Integer.class);
			roiXField.setFormat("%d");
			roiXField.setPreferredSize(FIELD_DIMENSION);
			roiXField.setMinimumSize(FIELD_DIMENSION);
			roiXField.setMaximumSize(FIELD_DIMENSION);
			roiXField.setMinimum(0);
			roiXField.addPropertyChangeListener("value", new PropertyChangeListener() {
				@Override
				public void propertyChange(PropertyChangeEvent evt) {
					if (suppressChanges) return;
					int value = ((Number) evt.getNewValue()).intValue();
					if (roiX == value) {
						if (localChanges) roiXField.setFont(getPlainFont());
						return;
					}
					if (!localChanges) {
						roiX = value;
						setData();
					}
					else {
						roiXField.setFont(getBoldFont());
						appliedToServer = false;
					}
				}
			});
		}
		return roiXField;
	}
	
	private NumberField getRoiYField() {
		if (roiYField == null) {
			roiYField = new NumberField();
			roiYField.setNumberType(Integer.class);
			roiYField.setFormat("%d");
			roiYField.setPreferredSize(FIELD_DIMENSION);
			roiYField.setMinimumSize(FIELD_DIMENSION);
			roiYField.setMaximumSize(FIELD_DIMENSION);
			roiYField.setMinimum(0);
			roiYField.addPropertyChangeListener("value", new PropertyChangeListener() {
				@Override
				public void propertyChange(PropertyChangeEvent evt) {
					if (suppressChanges) return;
					int value = ((Number) evt.getNewValue()).intValue();
					if (roiY == value) {
						if (localChanges) roiYField.setFont(getPlainFont());
						return;
					}
					if (!localChanges) {
						roiY = value;
						setData();
					}
					else {
						roiYField.setFont(getBoldFont());
						appliedToServer = false;
					}
				}
			});
		}
		return roiYField;
	}
	
	private NumberField getRoiWField() {
		if (roiWField == null) {
			roiWField = new NumberField();
			roiWField.setNumberType(Integer.class);
			roiWField.setFormat("%d");
			roiWField.setPreferredSize(FIELD_DIMENSION);
			roiWField.setMinimumSize(FIELD_DIMENSION);
			roiWField.setMaximumSize(FIELD_DIMENSION);
			roiWField.setMinimum(0);
			roiWField.addPropertyChangeListener("value", new PropertyChangeListener() {
				@Override
				public void propertyChange(PropertyChangeEvent evt) {
					if (suppressChanges) return;
					int value = ((Number) evt.getNewValue()).intValue();
					if (roiW == value) {
						if (localChanges) roiWField.setFont(getPlainFont());
						return;
					}
					if (!localChanges) {
						roiW = value;
						setData();
					}
					else {
						roiWField.setFont(getBoldFont());
						appliedToServer = false;
					}
				}
			});
		}
		return roiWField;
	}
	
	private NumberField getRoiHField() {
		if (roiHField == null) {
			roiHField = new NumberField();
			roiHField.setNumberType(Integer.class);
			roiHField.setFormat("%d");
			roiHField.setPreferredSize(FIELD_DIMENSION);
			roiHField.setMinimumSize(FIELD_DIMENSION);
			roiHField.setMaximumSize(FIELD_DIMENSION);
			roiHField.setMinimum(0);
			roiHField.addPropertyChangeListener("value", new PropertyChangeListener() {
				@Override
				public void propertyChange(PropertyChangeEvent evt) {
					if (suppressChanges) return;
					int value = ((Number) evt.getNewValue()).intValue();
					if (roiH == value) {
						if (localChanges) roiHField.setFont(getPlainFont());
						return;
					}
					if (!localChanges) {
						roiH = value;
						setData();
					}
					else {
						roiHField.setFont(getBoldFont());
						appliedToServer = false;
					}
					
				}
			});
		}
		return roiHField;
	}
	
	
	private NumberField getRoi2XField() {
		if (roi2XField == null) {
			roi2XField = new NumberField();
			roi2XField.setNumberType(Integer.class);
			roi2XField.setFormat("%d");
			roi2XField.setPreferredSize(FIELD_DIMENSION);
			roi2XField.setMinimumSize(FIELD_DIMENSION);
			roi2XField.setMaximumSize(FIELD_DIMENSION);
			roi2XField.setMinimum(0);
			roi2XField.addPropertyChangeListener("value", new PropertyChangeListener() {
				@Override
				public void propertyChange(PropertyChangeEvent evt) {
					if (suppressChanges) return;
					int value = ((Number) evt.getNewValue()).intValue();
					if (roi2X == value) {
						if (localChanges) roi2XField.setFont(getPlainFont());
						return;
					}
					if (!localChanges) {
						roi2X = value;
						setData2();
					}
					else {
						roi2XField.setFont(getBoldFont());
						appliedToServer = false;
					}
				}
			});
		}
		return roi2XField;
	}
	
	private NumberField getRoi2YField() {
		if (roi2YField == null) {
			roi2YField = new NumberField();
			roi2YField.setNumberType(Integer.class);
			roi2YField.setFormat("%d");
			roi2YField.setPreferredSize(FIELD_DIMENSION);
			roi2YField.setMinimumSize(FIELD_DIMENSION);
			roi2YField.setMaximumSize(FIELD_DIMENSION);
			roi2YField.setMinimum(0);
			roi2YField.addPropertyChangeListener("value", new PropertyChangeListener() {
				@Override
				public void propertyChange(PropertyChangeEvent evt) {
					if (suppressChanges) return;
					int value = ((Number) evt.getNewValue()).intValue();
					if (roi2Y == value) {
						if (localChanges) roi2YField.setFont(getPlainFont());
						return;
					}
					if (!localChanges) {
						roi2Y = value;
						setData2();
					}
					else {
						roi2YField.setFont(getBoldFont());
						appliedToServer = false;
					}
				}
			});
		}
		return roi2YField;
	}
	
	private NumberField getRoi2WField() {
		if (roi2WField == null) {
			roi2WField = new NumberField();
			roi2WField.setNumberType(Integer.class);
			roi2WField.setFormat("%d");
			roi2WField.setPreferredSize(FIELD_DIMENSION);
			roi2WField.setMinimumSize(FIELD_DIMENSION);
			roi2WField.setMaximumSize(FIELD_DIMENSION);
			roi2WField.setMinimum(0);
			roi2WField.addPropertyChangeListener("value", new PropertyChangeListener() {
				@Override
				public void propertyChange(PropertyChangeEvent evt) {
					if (suppressChanges) return;
					int value = ((Number) evt.getNewValue()).intValue();
					if (roi2W == value) {
						if (localChanges) roi2WField.setFont(getPlainFont());
						return;
					}
					if (!localChanges) {
						roi2W = value;
						setData2();
					}
					else {
						roi2WField.setFont(getBoldFont());
						appliedToServer = false;
					}
				}
			});
		}
		return roi2WField;
	}
	
	private NumberField getRoi2HField() {
		if (roi2HField == null) {
			roi2HField = new NumberField();
			roi2HField.setNumberType(Integer.class);
			roi2HField.setFormat("%d");
			roi2HField.setPreferredSize(FIELD_DIMENSION);
			roi2HField.setMinimumSize(FIELD_DIMENSION);
			roi2HField.setMaximumSize(FIELD_DIMENSION);
			roi2HField.setMinimum(0);
			roi2HField.addPropertyChangeListener("value", new PropertyChangeListener() {
				@Override
				public void propertyChange(PropertyChangeEvent evt) {
					if (suppressChanges) return;
					int value = ((Number) evt.getNewValue()).intValue();
					if (roi2H == value) {
						if (localChanges) roi2HField.setFont(getPlainFont());
						return;
					}
					if (!localChanges) {
						roi2H = value;
						setData2();
					}
					else {
						roi2HField.setFont(getBoldFont());
						appliedToServer = false;
					}
					
				}
			});
		}
		return roi2HField;
	}
	
	private JButton getAllButton() {
		if (allButton == null) {
			allButton = new JButton("Select Entire Image");
			allButton.setToolTipText("Select Entire Image for Region of Interest");
			allButton.addActionListener(new ActionListener() {

				@Override
				public void actionPerformed(ActionEvent e) {
					updateData(localChanges, imageW, imageH, 0, 0, imageW, imageH, roi2X, roi2Y, roi2W, roi2H, threshold, thresholdPoints,calculateThreshold,performFit,performSmoothing);
					if (!localChanges) setData();
				}});
		}
		return allButton;
	}
	
	private JButton getAllThresholdButton() {
		if (allThresholdButton == null) {
			allThresholdButton = new JButton("Select Entire Image");
			allThresholdButton.setToolTipText("Threshold Region Is Whole Image");
			allThresholdButton.addActionListener(new ActionListener() {

				@Override
				public void actionPerformed(ActionEvent e) {
					updateData(localChanges, imageW, imageH, roiX, roiY, roiW, roiH, 0, 0, imageW, imageH, threshold,thresholdPoints,calculateThreshold,performFit,performSmoothing);
					if (!localChanges) setData2();
				}});
		}
		return allThresholdButton;
	}
	
	private JButton getDuplicateThresholdButton() {
		if (duplicateThresholdButton == null) {
			duplicateThresholdButton = new JButton("Copy Values from ROI");
			duplicateThresholdButton.setToolTipText("Threshold Region Is the Same As ROI");
			duplicateThresholdButton.addActionListener(new ActionListener() {

				@Override
				public void actionPerformed(ActionEvent e) {
					updateData(localChanges, imageW, imageH, roiX, roiY, roiW, roiH, roiX, roiY, roiW, roiH, threshold,thresholdPoints,calculateThreshold,performFit,performSmoothing);
					if (!localChanges) setData2();
				}});
		}
		return duplicateThresholdButton;
	}
	
	private JButton getCurrentButton() {
		if (currentButton == null) {
			currentButton = new JButton("Use Current");
			currentButton.addActionListener(new ActionListener() {
				@Override
				public void actionPerformed(ActionEvent e) {
					executeTask(new Runnable() {
						@Override
						public void run() {
							IMAGE im = acopVideo.getImageAnalysisEngine().getImage();
							if (im == null) return;
							im = im.clone();
							fileName = VIDEO_BACKGROUND;
							setBackgroundImage(im, false);
							backgroundFile = null;
						}
					});
				}
			});
		}
		return currentButton;
	}
	
	private JButton getBrowseButton() {
		if (browseButton == null) {
			browseButton = new JButton("Load...");
			browseButton.addActionListener(new ActionListener() {
				@Override
				public void actionPerformed(ActionEvent e) {
					int option = getFileChooser().showOpenDialog(AnalysisCustomizer.this);
					if (option == JFileChooser.APPROVE_OPTION) {
						final File f = getFileChooser().getSelectedFile();
						executeTask(new Runnable() {
							@Override
							public void run() {
								try {
//									BufferedImage img = ImageIO.read(f);
									fileName = f.getAbsolutePath();
									setBackgroundImage(TImageIO.read(f).toIMAGE(false),false);
									backgroundFile = f;
								} catch (IOException e1) {
									e1.printStackTrace();
								}
							}
						});
					}
				}
			});
		}
		return browseButton;
	}
	
	private JButton getSaveButton() {
		if (saveButton == null) {
			saveButton = new JButton("Save...");
			saveButton.addActionListener(new ActionListener() {
				@Override
				public void actionPerformed(ActionEvent e) {
					int option = getFileChooser().showSaveDialog(AnalysisCustomizer.this);
					if (option == JFileChooser.APPROVE_OPTION) {
						final File f = getFileChooser().getSelectedFile();
						executeTask(new Runnable() {
							@Override
							public void run() {
								if (backgroundImage == null) return;
								BufferedImage bi = ImageAnalysisEngine.toBufferedImage(backgroundImage, acopVideo.getColorMap());
								
								List<String> ext = Arrays.asList(FILE_TYPES);
								int dotIndex = f.getName().lastIndexOf('.');
								String extension = FILE_TYPES[0];
								if (dotIndex > 0 && dotIndex < f.getName().length()-2) {
									String ex = f.getName().substring(dotIndex+1);
									if (ext.contains(ex)) extension = ex;
								}
								
								try {
									ImageIO.write(bi, extension, f);
								} catch (IOException e1) {
									e1.printStackTrace();
								}
							}
						});
					}
				}
			});
			saveButton.setEnabled(false);
		}
		return saveButton;
	}
	
	private JButton getNoneButton() {
		if (noneButton == null) {
			noneButton = new JButton("Use None");
			noneButton.addActionListener(new ActionListener() {
				@Override
				public void actionPerformed(ActionEvent e) {
					fileName = NO_BACKGROUND;
					setBackgroundImage((double[])null, 0, false);
					backgroundFile = null;
				}
			});
		}
		return noneButton;
	}
	
	private JLabel getBackgroundLabel() {
		if (backgroundLabel == null) {
			backgroundLabel = new JLabel();
			backgroundLabel.setBorder(new LineBorder(Color.BLACK));
			backgroundLabel.setHorizontalAlignment(SwingConstants.CENTER);
			backgroundLabel.setText(NO_BACKGROUND);
			backgroundLabel.setPreferredSize(new Dimension(PREVIEW_DIMENSION, PREVIEW_DIMENSION));
		}
		return backgroundLabel;
	}
	
	private JLabel getFileNameLabel() {
		if (fileNameLabel == null) {
			fileNameLabel = new JLabel("File: N/A");
			fileNameLabel.setHorizontalAlignment(SwingConstants.CENTER);
			fileNameLabel.setPreferredSize(LABEL_DIMENSION);
			fileNameLabel.setMinimumSize(LABEL_DIMENSION);
			fileNameLabel.setMaximumSize(LABEL_DIMENSION);
		}
		return fileNameLabel;
	}
	
	private JFileChooser getFileChooser() {
		if (fileChooser == null) {
			fileChooser = new JFileChooser();
			fileChooser.setFileSelectionMode(JFileChooser.FILES_ONLY);
			fileChooser.setMultiSelectionEnabled(false);
			fileChooser.setFileFilter(new FileFilter() {
				List<String> ext = Arrays.asList(FILE_TYPES);
				@Override
				public boolean accept(File f) {
					if (f.isDirectory()) return true;
					String name = f.getName();
					int index = name.lastIndexOf('.');
					if (index > 0 && index < name.length()-2) {
						name = name.substring(index+1);
						if (ext.contains(name)) return true;
					}
					return false;
				}
				@Override
				public String getDescription() {
					String s = "Images (";
					for (Iterator<String> iterator = ext.iterator(); iterator.hasNext();) {
						String e = iterator.next();
						s += e;
						if (iterator.hasNext()) s += ", ";
					}
					s += ")";
					return s;
				}
			});
		}
		return fileChooser;
	}
	
	private JCheckBox getDisplayNumericalValuesBox() {
		if (displayNumericalValuesBox == null) {
			displayNumericalValuesBox = new JCheckBox("Display Values");
			displayNumericalValuesBox.addActionListener(new ActionListener() {
				@Override
				public void actionPerformed(ActionEvent e) {
					acopVideo.setDisplayNumericalValues(displayNumericalValuesBox.isSelected());
				}
			});
		}
		return displayNumericalValuesBox;
	}
	
	private void setAnalysisMode(AnalysisMode mode) {
		getApplyToServerButton().setEnabled(mode == AnalysisMode.REMOTE_ANALYSIS);
		getSynchronizeButton().setEnabled(mode == AnalysisMode.REMOTE_ANALYSIS);		
		boolean enable = mode != AnalysisMode.NO_ANALYSIS;
		setSettingsEnabled(enable);
		localChanges = mode == AnalysisMode.REMOTE_ANALYSIS;
		if (!localChanges) {
			getRoiXField().setFont(getPlainFont());
			getRoiYField().setFont(getPlainFont());
			getRoiWField().setFont(getPlainFont());
			getRoiHField().setFont(getPlainFont());
			getRoi2XField().setFont(getPlainFont());
			getRoi2YField().setFont(getPlainFont());
			getRoi2WField().setFont(getPlainFont());
			getRoi2HField().setFont(getPlainFont());
			getThresholdField().setFont(getPlainFont());
			getCalculateTresholdCheckBox().setFont(getLabelPlainFont());
			getPerformFitCheckBox().setFont(getLabelPlainFont());
			getPerformSmoothCheckBox().setFont(getLabelPlainFont());
		}
		acopVideo.setAnalysisMode(mode);
		appliedToServer = true;
	}
	
	private void setSettingsEnabled(boolean enable) {
		getThresholdField().setEnabled(enable);
		getCalculateTresholdCheckBox().setEnabled(enable);
		getPerformFitCheckBox().setEnabled(enable);
		getPerformSmoothCheckBox().setEnabled(enable);
		getDisplayStatisticsCheckBox().setEnabled(enable);
		getRoiXField().setEnabled(enable);
		getRoiYField().setEnabled(enable);
		getRoiWField().setEnabled(enable);
		getRoiHField().setEnabled(enable);
		getRoi2XField().setEnabled(enable);
		getRoi2YField().setEnabled(enable);
		getRoi2WField().setEnabled(enable);
		getRoi2HField().setEnabled(enable);
		getAllButton().setEnabled(enable);
		getAllThresholdButton().setEnabled(enable);
		getDuplicateThresholdButton().setEnabled(enable);
		getCurrentButton().setEnabled(enable);
		getBrowseButton().setEnabled(enable);
		getSaveButton().setEnabled(enable);
		getNoneButton().setEnabled(enable);
		getBackgroundLabel().setEnabled(enable);
		acopVideo.setCanChangeSettings(enable);
	}
	
	private void applyToServer() {
		int rX = getRoiXField().getValue().intValue();
		int rY = getRoiYField().getValue().intValue();
		int rW = getRoiWField().getValue().intValue();
		int rH = getRoiHField().getValue().intValue();
		int r2X = getRoi2XField().getValue().intValue();
		int r2Y = getRoi2YField().getValue().intValue();
		int r2W = getRoi2WField().getValue().intValue();
		int r2H = getRoi2HField().getValue().intValue();
		double thr = getThresholdField().getValue().doubleValue();
		boolean cth = getCalculateTresholdCheckBox().isSelected();
		boolean fit = getPerformFitCheckBox().isSelected();
		boolean smooth = getPerformSmoothCheckBox().isSelected();
		
		performFit = fit;
		performSmoothing = smooth;
		calculateThreshold = cth;
		threshold = thr;
		
		roiX = rX;
		roiY = rY;
		roiW = rW;
		roiY = rY;
		acopVideo.getImageAnalysisEngine().setAnalysisParameters(
				threshold,rX,rY,rW,rH,r2X,r2Y,r2W,r2H,cth,fit,smooth);
		appliedToServer = true;
	}
	
	private boolean updateData(boolean localChanges, int imageWidth, int imageHeight, int roiX, int roiY, int roiW, int roiH,
			int roi2X, int roi2Y, int roi2W, int roi2H,
			double threshold,int thresholdPoints,boolean calculateThreshold, boolean performFit, boolean performSmooth) {
		if (!appliedToServer && !localChanges) return false;
		suppressChanges = true;
		if (!localChanges) {
			this.roiX = roiX;
			this.roiY = roiY;
			this.roiW = roiW;
			this.roiH = roiH;
			this.roi2X = roi2X;
			this.roi2Y = roi2Y;
			this.roi2W = roi2W;
			this.roi2H = roi2H;
			this.threshold = threshold;
			this.calculateThreshold = calculateThreshold;
			this.performFit = performFit;
			this.performSmoothing = performSmooth;
		}
		this.thresholdPoints = thresholdPoints;
		this.imageW = imageWidth;
		this.imageH = imageHeight;
		getCalculateTresholdCheckBox().setSelected(calculateThreshold);
		getThresholdField().setEditable(!calculateThreshold && getUseTresholdCheckBox().isSelected());
		getPerformFitCheckBox().setSelected(performFit);
		getPerformSmoothCheckBox().setSelected(performSmooth);
		if (getUseTresholdCheckBox().isSelected()) {
			getThresholdField().setValue(threshold);
		} else {
			getThresholdField().setValue(Double.NaN);
		}
		if (!calculateThreshold) {
			getUseTresholdCheckBox().setEnabled(true);
			if (Double.isNaN(threshold)) {
				getUseTresholdCheckBox().setSelected(false);
				getThresholdField().setValue(Double.NaN);
			} else {
				getUseTresholdCheckBox().setSelected(true);
				getThresholdField().setValue(threshold);
			}
		} else {
			getUseTresholdCheckBox().setEnabled(false);
		}
		getRoiXField().setMaximum(imageWidth);
		getRoiYField().setMaximum(imageHeight);
		getRoiWField().setMaximum(imageWidth-roiX);
		getRoiHField().setMaximum(imageHeight-roiY);
		getRoi2XField().setMaximum(imageWidth);
		getRoi2YField().setMaximum(imageHeight);
		getRoi2WField().setMaximum(imageWidth-roi2X);
		getRoi2HField().setMaximum(imageHeight-roi2Y);
		getRoiXField().setValue(roiX);
		getRoiYField().setValue(roiY);
		getRoiWField().setValue(roiW);
		getRoiHField().setValue(roiH);
		getRoi2XField().setValue(roi2X);
		getRoi2YField().setValue(roi2Y);
		getRoi2WField().setValue(roi2W);
		getRoi2HField().setValue(roi2H);
		
		getNumberOfPixelsLabel().setText(PIXELS_USED + (thresholdPoints));

		getRoiXField().setFont(getPlainFont());
		getRoiYField().setFont(getPlainFont());
		getRoiWField().setFont(getPlainFont());
		getRoiHField().setFont(getPlainFont());
		getRoi2XField().setFont(getPlainFont());
		getRoi2YField().setFont(getPlainFont());
		getRoi2WField().setFont(getPlainFont());
		getRoi2HField().setFont(getPlainFont());
		getThresholdField().setFont(getPlainFont());
		getCalculateTresholdCheckBox().setFont(getLabelPlainFont());
		getPerformFitCheckBox().setFont(getLabelPlainFont());
		getPerformSmoothCheckBox().setFont(getLabelPlainFont());
		suppressChanges = false;
		return true;
	}
	
	private void resetData() {
		suppressChanges = true;
		this.roiX = 0;
		this.roiY = 0;
		this.roiW = 0;
		this.roiH = 0;
		this.roi2X = 0;
		this.roi2Y = 0;
		this.roi2W = 0;
		this.roi2H = 0;
		this.imageW = 0;
		this.imageH = 0;
		this.threshold = 0;
		this.calculateThreshold = true;
		getThresholdField().setValue(Double.NaN);
		getCalculateTresholdCheckBox().setSelected(true);
		getPerformFitCheckBox().setSelected(false);
		getPerformSmoothCheckBox().setSelected(true);
		getThresholdField().setEditable(false);
		getUseTresholdCheckBox().setSelected(false);
		getRoiXField().setMaximum(0);
		getRoiYField().setMaximum(0);
		getRoiWField().setMaximum(0);
		getRoiHField().setMaximum(0);
		getRoiXField().setValue(0);
		getRoiYField().setValue(0);
		getRoiWField().setValue(0);
		getRoiHField().setValue(0);
		getRoi2XField().setMaximum(0);
		getRoi2YField().setMaximum(0);
		getRoi2WField().setMaximum(0);
		getRoi2HField().setMaximum(0);
		getRoi2XField().setValue(0);
		getRoi2YField().setValue(0);
		getRoi2WField().setValue(0);
		getRoi2HField().setValue(0);
		getNumberOfPixelsLabel().setText(PIXELS_USED + (0));
		suppressChanges = false;
		appliedToServer = true;
	}
		
	private synchronized void setBackgroundImage(IMAGE img, boolean local) {
		if (img != null) {
			FrameHeader hdr = img.getFrameHeader();
			int w = hdr.aoiWidth == -1 ? hdr.sourceWidth : hdr.aoiWidth;
			int h = hdr.aoiHeight == -1 ? hdr.sourceHeight : hdr.aoiHeight;
			int[] rgb = ImageAnalysisEngine.toBufferedImage(img,acopVideo.getColorMap()).getRGB(0,0,w,
					h,null,0,w);
			backgroundImage = img;
			preciseBackground = acopVideo.getImageAnalysisEngine().getDecoder().transform(rgb);
			preciseBackgroundWidth = w;
		} else {
			backgroundImage = null;
			preciseBackground = null;
			preciseBackgroundWidth = 0;
		}		
		processBackground(local);
	}
	
	private synchronized void setBackgroundImage(double[] image, int width, boolean local) {
		preciseBackground = image;
		backgroundImage = ImageAnalysisEngine.toImage(preciseBackground,width);
		preciseBackgroundWidth = width;
		processBackground(local);
	}
	
	private void processBackground(boolean local) {
		updateFileNameLabel(fileName);
		if (preciseBackground == null || backgroundImage == null) {
			getBackgroundLabel().setText(NO_BACKGROUND);
			getBackgroundLabel().setIcon(null);
		} else {
			BufferedImage bi = ImageAnalysisEngine.toBufferedImage(preciseBackground,preciseBackgroundWidth);	
			if (bi.getWidth() != 0 || bi.getHeight() != 0) {
    			getSaveButton().setEnabled(true);
    			int w1 = getBackgroundLabel().getWidth();
    			int h1 = getBackgroundLabel().getHeight();
    			h1 = Math.max(h1,100);
    			w1 = Math.max(w1,100);
    			ImageIcon icon = IconCustomizer.getScaledImage(bi, h1, w1);
    			getBackgroundLabel().setIcon(icon);
    			getBackgroundLabel().setText(null);
			} else {
				getBackgroundLabel().setText(INVALID_BACKGROUND);
				getBackgroundLabel().setIcon(null);
			}
			
		}
		if (!local) acopVideo.getImageAnalysisEngine().setPreciseBackground(preciseBackground,preciseBackgroundWidth);
	}
	
	private void updateFileNameLabel(String text) {
		getFileNameLabel().setText(text);
		getFileNameLabel().setToolTipText(text);
	}
	
	private Font getPlainFont() {
		if (plainFont == null) {
			plainFont = getThresholdField().getFont();	
		}
		return plainFont;
	}
	
	private Font getLabelPlainFont() {
		if (labelPlainFont == null) {
			labelPlainFont = getNumberOfPixelsLabel().getFont();
		}
		return labelPlainFont;
	}
	
	private Font getBoldFont() {
		if (boldFont == null) {
			boldFont = getPlainFont().deriveFont(Font.BOLD);	
		}
		return boldFont;
	}
	
	private Font getLabelBoldFont() {
		if (labelBoldFont == null) {
			labelBoldFont = getLabelPlainFont().deriveFont(Font.BOLD);
		}
		return labelBoldFont;
	}
	
	private void setData() {
		acopVideo.getImageAnalysisEngine().setRoi(roiX, roiY, roiW, roiH);
	}
	
	private void setData2() {
		acopVideo.getImageAnalysisEngine().setRoi2(roi2X, roi2Y, roi2W, roi2H);
	}
	
	private void updateAnalysisData(AnalysisImage ai) {
		appliedToServer = updateData(false, ai.getImageW(), ai.getImageH(), ai.getRoiX(), ai.getRoiY(), ai.getRoiW(), ai.getRoiH(),
				ai.getRoi2X(), ai.getRoi2Y(), ai.getRoi2W(), ai.getRoi2H(), ai.getThreshold(),ai.getThresholdPoints(),ai.isCalculateThreshold(), ai.isPerformFit(), ai.isPerformSmoothing());
		getServerModel().setAImage(ai);
	}
	
	private void updateBackground() {
		executeTask(new Runnable() {
			@Override
			public void run() {
				if (backgroundFile != null) return;
				double[] im = acopVideo.getImageAnalysisEngine().getPreciseBackground();
				int width = acopVideo.getImageAnalysisEngine().getBackgroundWidth();
				fileName = VIDEO_BACKGROUND;
				setBackgroundImage(im,width, true);
			}
		});
	}
	
	private void executeTask(Runnable r) {
		if (executor == null) {
			executor = Executors.newSingleThreadExecutor();
		}
		executor.execute(r);
	}
	
	public static void main(String[] args) {
		
		AnalysisCustomizer customizer = new AnalysisCustomizer();
		RunnerHelper.runComponent(customizer,500,500);
	}
}
