/**
 * 
 */
package de.desy.acop.video;

import java.awt.Color;
import java.awt.Component;
import java.awt.Font;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.text.DecimalFormat;

import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JTable;
import javax.swing.SwingConstants;
import javax.swing.border.LineBorder;
import javax.swing.table.DefaultTableCellRenderer;
import javax.swing.table.DefaultTableModel;

import de.desy.acop.video.analysis.AnalysisImage;
import de.desy.acop.video.analysis.gaussian.AImage;
import de.desy.acop.video.analysis.interferometer.IFMImage;

/**
 * </code>NumericDataDisplayer</code> displays numerical image analysis data in form of
 * a table.
 * 
 * @author Tilen Kusterle, Cosylab
 *
 */
public class NumericDataDisplayer extends JPanel {

	private static final long serialVersionUID = 1L;
	private static final DecimalFormat decimalFormat = new DecimalFormat("#0.0");
	private static final DecimalFormat decimalFormat2 = new DecimalFormat("#0.0000");
	
	private String type = "AISr4";

	private DefaultTableModel tmGaussian3C;
	private DefaultTableModel tmGaussian2C;
	private JPanel gaussianPanel;
	
	private DefaultTableModel tmInterferometer2C; 
	private JPanel interferometerPanel;
	
	private class CellRenderer extends DefaultTableCellRenderer {

		private static final long serialVersionUID = 1L;
		
		private int[] headerRows;
		
		public CellRenderer(int[] headerRows) {
			this.headerRows = headerRows;
		}

		/* (non-Javadoc)
		 * @see javax.swing.table.DefaultTableCellRenderer#getTableCellRendererComponent(javax.swing.JTable, java.lang.Object, boolean, boolean, int, int)
		 */
		@Override
		public Component getTableCellRendererComponent(JTable table,
				Object value, boolean isSelected, boolean hasFocus, int row,
				int column) {
			JLabel label =  (JLabel) super.getTableCellRendererComponent(table, value, isSelected, hasFocus,
					row, column);
			for (int i = 0; i < headerRows.length; i++) {
				if (headerRows[i] == row) {
					label.setHorizontalAlignment(SwingConstants.CENTER);
					label.setFont(label.getFont().deriveFont(Font.BOLD));
					return label;
				}
			}
			if (column == 0) label.setFont(label.getFont().deriveFont(Font.ITALIC));
			label.setHorizontalAlignment(SwingConstants.LEADING);
			return label;
		}
		
	}
	
	/**
	 * Constructs this </code>NumericDataDisplayer</code>..
	 */
	public NumericDataDisplayer() {
		super(new GridBagLayout());		
		add(getGaussianPanel(), new GridBagConstraints(0,0,1,1,1,1,GridBagConstraints.CENTER,GridBagConstraints.BOTH,new Insets(0,0,0,0),0,0));
		add(getInterferometerPanel(), new GridBagConstraints(0,0,1,1,1,1,GridBagConstraints.CENTER,GridBagConstraints.BOTH,new Insets(0,0,0,0),0,0));
		
	}
	
	private JPanel getInterferometerPanel() {
		if (interferometerPanel == null) {
			interferometerPanel = new JPanel(new GridBagLayout());
			
			tmInterferometer2C = new DefaultTableModel(16, 2) {
			    
    			private static final long serialVersionUID = 1L;
    
    			@Override
    			public boolean isCellEditable(int row, int column) {
    				return false;
    			}
    		};
    		tmInterferometer2C.setValueAt("Y Projection", 0, 0);
    		tmInterferometer2C.setValueAt("Value", 0, 1);
    		tmInterferometer2C.setValueAt("Background", 1, 0);
    		tmInterferometer2C.setValueAt("Amplitude", 2, 0);
    		tmInterferometer2C.setValueAt("Asinc", 3, 0);
    		tmInterferometer2C.setValueAt("PSinc", 4, 0);
    		tmInterferometer2C.setValueAt("ACos", 5, 0);
    		tmInterferometer2C.setValueAt("PCos", 6, 0);
    		tmInterferometer2C.setValueAt("Vis", 7, 0);
    		tmGaussian2C.setValueAt("IMAGE DATA", 8, 0);
    		tmGaussian2C.setValueAt("VALUE", 8, 1);
    		tmInterferometer2C.setValueAt("width [pixels]", 9, 0);
    		tmInterferometer2C.setValueAt("height [pixels]", 10, 0);
    		tmInterferometer2C.setValueAt("ROI X [pixels]", 11, 0);
    		tmInterferometer2C.setValueAt("ROI Y [pixels]", 12, 0);
    		tmInterferometer2C.setValueAt("ROI width [pixels]", 13, 0);
    		tmInterferometer2C.setValueAt("ROI height [pixels]", 14, 0);
    		tmInterferometer2C.setValueAt("threshold", 15, 0);
    		JTable interferometerTable = new JTable(tmInterferometer2C);
    		interferometerTable.setDefaultRenderer(Object.class, new CellRenderer(new int[]{0,8}));
    		
    		interferometerPanel.add(interferometerTable, new GridBagConstraints(0,0,1,1,1,0,GridBagConstraints.CENTER,GridBagConstraints.BOTH,new Insets(0,0,0,0),0,0));
    		
			interferometerPanel.setBorder(new LineBorder(Color.BLACK));
		}
		 return interferometerPanel;
	}
	
	private JPanel getGaussianPanel() {
		if (gaussianPanel == null) {
    		gaussianPanel = new JPanel(new GridBagLayout());
    		
    		tmGaussian3C = new DefaultTableModel(6, 3) {
    
    			private static final long serialVersionUID = 1L;
    
    			@Override
    			public boolean isCellEditable(int row, int column) {
    				return false;
    			}
    		};
    		tmGaussian3C.setValueAt("XY PROJECTION", 0, 0);
    		tmGaussian3C.setValueAt("X VALUE (stat./fit)", 0, 1);
    		tmGaussian3C.setValueAt("Y VALUE (stat./fit)", 0, 2);
    		tmGaussian3C.setValueAt("mean [pixels]", 1, 0);
    		tmGaussian3C.setValueAt("STD [pixels]", 2, 0);
    		tmGaussian3C.setValueAt("amplitude", 3, 0);
    		tmGaussian3C.setValueAt("constant", 4, 0);
    		tmGaussian3C.setValueAt("slope", 5, 0);
    		JTable gaussianTable3C = new JTable(tmGaussian3C);
    		gaussianTable3C.setDefaultRenderer(Object.class, new CellRenderer(new int[]{0}));
    		
    		gaussianPanel.add(gaussianTable3C, new GridBagConstraints(0,0,1,1,1,0,GridBagConstraints.CENTER,GridBagConstraints.BOTH,new Insets(0,0,0,0),0,0));
    		
    		tmGaussian2C = new DefaultTableModel(14, 2) {
    
    			private static final long serialVersionUID = 1L;
    
    			@Override
    			public boolean isCellEditable(int row, int column) {
    				return false;
    			}
    		};
    		tmGaussian2C.setValueAt("2D DATA", 0, 0);
    		tmGaussian2C.setValueAt("VALUE", 0, 1);
    		tmGaussian2C.setValueAt("STD A [pixels]", 1, 0);
    		tmGaussian2C.setValueAt("STD B [pixels]", 2, 0);
    		tmGaussian2C.setValueAt("rotation [degrees]", 3, 0);
    		tmGaussian2C.setValueAt("amplitude", 4, 0);
    		tmGaussian2C.setValueAt("minimum", 5, 0);
    		tmGaussian2C.setValueAt("IMAGE DATA", 6, 0);
    		tmGaussian2C.setValueAt("VALUE", 6, 1);
    		tmGaussian2C.setValueAt("width [pixels]", 7, 0);
    		tmGaussian2C.setValueAt("height [pixels]", 8, 0);
    		tmGaussian2C.setValueAt("ROI X [pixels]", 9, 0);
    		tmGaussian2C.setValueAt("ROI Y [pixels]", 10, 0);
    		tmGaussian2C.setValueAt("ROI width [pixels]", 11, 0);
    		tmGaussian2C.setValueAt("ROI height [pixels]", 12, 0);
    		tmGaussian2C.setValueAt("threshold", 13, 0);
    		JTable gaussianTable2C = new JTable(tmGaussian2C);
    		gaussianTable2C.setDefaultRenderer(Object.class, new CellRenderer(new int[]{0, 6}));
    		
    		gaussianPanel.add(gaussianTable2C, new GridBagConstraints(0,1,1,1,1,0,GridBagConstraints.CENTER,GridBagConstraints.BOTH,new Insets(0,0,0,0),0,0));
    		
    		gaussianPanel.setBorder(new LineBorder(Color.BLACK));
		}
		return gaussianPanel;
	}
	
	private void restructure(AnalysisImage image) {
		type = image.getName();
		if (image instanceof AImage) {
			getGaussianPanel().setVisible(true);
			getInterferometerPanel().setVisible(false);
		} else if (image instanceof IFMImage) {
			getGaussianPanel().setVisible(false);
			getInterferometerPanel().setVisible(true);
		}
	}
	
	/**
	 * Sets table data.
	 */
	public void setData(AnalysisImage aImage) {
		if (this.type == null || !this.type.equals(aImage.getName())) {
			restructure(aImage);
		}
		
		if (aImage instanceof AImage) {
			setGaussian((AImage)aImage);
		} else if (aImage instanceof IFMImage) {
			setInterferometer((IFMImage)aImage);
		}
	}
	
	private void setInterferometer(IFMImage image) {		
		tmInterferometer2C.setValueAt(decimalFormat2.format(image.getConstantY()), 1, 1);
		tmInterferometer2C.setValueAt(decimalFormat2.format(image.getAmplitudeY()), 2, 1);
		tmInterferometer2C.setValueAt(decimalFormat2.format(image.getAsincY()), 3, 1);
		tmInterferometer2C.setValueAt(decimalFormat2.format(image.getPsincY()), 4, 1);
		tmInterferometer2C.setValueAt(decimalFormat2.format(image.getAcosY()), 5, 1);
		tmInterferometer2C.setValueAt(decimalFormat2.format(image.getPcosY()), 6, 1);
		tmInterferometer2C.setValueAt(decimalFormat2.format(image.getVisY()), 7, 1);
		tmInterferometer2C.setValueAt(image.getImageW(), 9, 1);
		tmInterferometer2C.setValueAt(image.getImageH(), 10, 1);
		tmInterferometer2C.setValueAt(image.getRoiX(), 11, 1);
		tmInterferometer2C.setValueAt(image.getRoiY(), 12, 1);
		tmInterferometer2C.setValueAt(image.getRoiW(), 13, 1);
		tmInterferometer2C.setValueAt(image.getRoiH(), 14, 1);
		tmInterferometer2C.setValueAt(decimalFormat.format(image.getThreshold()), 15, 1);
	}
	
	private void setGaussian(AImage image) {
		tmGaussian3C.setValueAt(decimalFormat.format(image.getMeanX()) + " / " + decimalFormat.format(image.getMeanXFit()), 1, 1);
		tmGaussian3C.setValueAt(decimalFormat.format(image.getMeanY()) + " / " + decimalFormat.format(image.getMeanYFit()), 1, 2);
		tmGaussian3C.setValueAt(decimalFormat.format(image.getStdX()) + " / " + decimalFormat.format(image.getStdXFit()), 2, 1);
		tmGaussian3C.setValueAt(decimalFormat.format(image.getStdY()) + " / " + decimalFormat.format(image.getStdYFit()), 2, 2);
		tmGaussian3C.setValueAt(decimalFormat.format(image.getSideViewAmplitudeX()) + " / " + decimalFormat.format(image.getAmplitudeXFit()), 3, 1);
		tmGaussian3C.setValueAt(decimalFormat.format(image.getSideViewAmplitudeY()) + " / " + decimalFormat.format(image.getAmplitudeYFit()), 3, 2);
		tmGaussian3C.setValueAt(decimalFormat.format(image.getSideViewConstX()) + " / " + decimalFormat.format(image.getConstXFit()), 4, 1);
		tmGaussian3C.setValueAt(decimalFormat.format(image.getSideViewConstY()) + " / " + decimalFormat.format(image.getConstYFit()), 4, 2);
		tmGaussian3C.setValueAt(decimalFormat.format(0) + " / " + decimalFormat.format(image.getSlopeXFit()), 5, 1);
		tmGaussian3C.setValueAt(decimalFormat.format(0) + " / " + decimalFormat.format(image.getSlopeYFit()), 5, 2);

		
		tmGaussian2C.setValueAt(decimalFormat.format(image.getStdA()), 1, 1);
		tmGaussian2C.setValueAt(decimalFormat.format(image.getStdB()), 2, 1);
		tmGaussian2C.setValueAt(decimalFormat.format(image.getAngle2D()), 3, 1);
		tmGaussian2C.setValueAt(decimalFormat.format(image.getAmplitude2D()), 4, 1);
		tmGaussian2C.setValueAt(decimalFormat.format(image.getConst2D()), 5, 1);
		tmGaussian2C.setValueAt(image.getImageW(), 7, 1);
		tmGaussian2C.setValueAt(image.getImageH(), 8, 1);
		tmGaussian2C.setValueAt(image.getRoiX(), 9, 1);
		tmGaussian2C.setValueAt(image.getRoiY(), 10, 1);
		tmGaussian2C.setValueAt(image.getRoiW(), 11, 1);
		tmGaussian2C.setValueAt(image.getRoiH(), 12, 1);
		tmGaussian2C.setValueAt(decimalFormat.format(image.getThreshold()), 13, 1);
	}
	
	/**
	 * Clears table contents.
	 */
	public void clearData() {
		tmGaussian3C.setValueAt(null, 1, 1);
		tmGaussian3C.setValueAt(null, 1, 2);
		tmGaussian3C.setValueAt(null, 2, 1);
		tmGaussian3C.setValueAt(null, 2, 2);
		tmGaussian3C.setValueAt(null, 3, 1);
		tmGaussian3C.setValueAt(null, 3, 2);
		tmGaussian3C.setValueAt(null, 4, 1);
		tmGaussian3C.setValueAt(null, 4, 2);
		tmGaussian3C.setValueAt(null, 5, 1);
		tmGaussian3C.setValueAt(null, 5, 2);
		
		tmGaussian2C.setValueAt(null, 1, 1);
		tmGaussian2C.setValueAt(null, 2, 1);
		tmGaussian2C.setValueAt(null, 3, 1);
		tmGaussian2C.setValueAt(null, 4, 1);
		tmGaussian2C.setValueAt(null, 5, 1);
		tmGaussian2C.setValueAt(null, 7, 1);
		tmGaussian2C.setValueAt(null, 8, 1);
		tmGaussian2C.setValueAt(null, 9, 1);
		tmGaussian2C.setValueAt(null, 10, 1);
		tmGaussian2C.setValueAt(null, 11, 1);
		tmGaussian2C.setValueAt(null, 12, 1);
		tmGaussian2C.setValueAt(null, 13, 1);
		
		tmInterferometer2C.setValueAt(null, 1, 1);
		tmInterferometer2C.setValueAt(null, 2, 1);
		tmInterferometer2C.setValueAt(null, 3, 1);
		tmInterferometer2C.setValueAt(null, 4, 1);
		tmInterferometer2C.setValueAt(null, 5, 1);
		tmInterferometer2C.setValueAt(null, 6, 1);
		tmInterferometer2C.setValueAt(null, 7, 1);
		tmInterferometer2C.setValueAt(null, 9, 1);
		tmInterferometer2C.setValueAt(null, 10, 1);
		tmInterferometer2C.setValueAt(null, 11, 1);
		tmInterferometer2C.setValueAt(null, 12, 1);
		tmInterferometer2C.setValueAt(null, 13, 1);
		tmInterferometer2C.setValueAt(null, 14, 1);
		tmInterferometer2C.setValueAt(null, 15, 1);
	}
}
