/**
 * 
 */
package de.desy.acop.video;

import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.util.logging.Level;

import de.desy.acop.transport.AccessMode;
import de.desy.acop.transport.ConnectionFailed;
import de.desy.acop.transport.ConnectionParameters;
import de.desy.acop.video.analysis.gaussian.AImage;
import de.desy.acop.video.displayer.TineHandler;
//import de.desy.acop.video.displayer.VideoHeaderV3; // mdavid: commented
import de.desy.tine.client.TCallback;
import de.desy.tine.client.TLink;
import de.desy.tine.client.TLinkCallback;
import de.desy.tine.dataUtils.TDataType;
import de.desy.tine.definitions.TAccess;
import de.desy.tine.definitions.TErrorList;
import de.desy.tine.definitions.TMode;
import de.desy.tine.types.IMAGE;

/**
 * </code>VideoServerConnection</code> provides the interface for accessing video
 * analysis server.
 * 
 * @author Tilen Kusterle, Cosylab
 *
 */
public class VideoServerConnection {
	
	//FIXME add handling of precise background
	public static final String PROPERTY_A_IMAGE = "aImage";
	public static final String PROPERTY_BACKGROUND = "background";
	public static final String PROPERTY_PRECISE_BACKGROUND = "preciseBackground";
	public static final String PROPERTY_PRECISE_BACKGROUND_WIDTH = "preciseBackgroundWidth";
	private static final String ANALYSIS_PROPERTY = "Analysis";
	private static final String BACKGROUND_PROPERTY = "Background";
	private static final String PRECISE_BACKGROUND_PROPERTY = "PreciseBackground";
	private static final String BACKGROUND_ID_PROPERTY = "BackgroundId";
	private static final String ANALYSIS_PARAMETERS_PROPERTY = "AnalysisParameters";
	
	private class AnalysisHandler implements TLinkCallback {
		
		private TLink tlink;
		private boolean isTransfering;
		private AImage image;
		
		public AnalysisHandler() {
			IMAGE image = new IMAGE();
			this.image = new AImage(image, 0, 0, 0, 0, 
					0, 0, 0, 0, 0.0,
					new double[]{}, 0.0, 0.0, 0.0, 0.0,0.0,
					0.0,0.0,0.0,0.0,
					new double[]{}, 0.0, 0.0, 0.0, 0.0,0.0,
					0.0,0.0,0.0,0.0,
					0.0, 0.0, 0.0, 0.0, 0.0,0,true,false,true);
			
		}
	
		
		/* (non-Javadoc)
		 * @see de.desy.tine.client.TLinkCallback#callback(de.desy.tine.client.TLink)
		 */
		@Override
		public void callback(TLink link) {
			if (!isTransfering) return;
			if (link.getLinkStatus() != 0) {
				System.out.println("Link Error : code=" + link.getLinkStatus() + " msg=" + link.getLastError());
				return;
			}
			int id = getBackgroundId();
			if (backgroundId != id) {
				try {
					backgroundId = id;
					loadBackground();
					loadPreciseBackground();
					//TODO
					getPcSupport().firePropertyChange(PROPERTY_BACKGROUND, null, background);
					getPcSupport().firePropertyChange(PROPERTY_PRECISE_BACKGROUND,null,preciseBackground);
					getPcSupport().firePropertyChange(PROPERTY_PRECISE_BACKGROUND_WIDTH,null,preciseBackgroundWidth);
				} catch (Exception e) {
					e.printStackTrace();
				}
			}
			aImage = new AImage(this.image);
			getPcSupport().firePropertyChange(PROPERTY_A_IMAGE, null, aImage);
		}
		
		public void openLink(int accessRate) throws ConnectionFailed {
			TDataType dataOut = new TDataType(new AImage[]{image});
			tlink = new TLink(server+"/"+device, ANALYSIS_PROPERTY, dataOut, null, TAccess.CA_READ);
			short tMode = TMode.CM_POLL;
			switch (accessMode) {
    			case POLL:
    				tMode = TMode.CM_POLL;
    				break;
    			case POLL_NETWORK:
    				tMode = TMode.CM_POLL | TMode.CM_MCAST;
    				break;
    			case POLL_CONNECT:
    				tMode = TMode.CM_POLL | TMode.CM_CONNECT;
    				break;
    			case READ:
    				tMode = TMode.CM_SINGLE;
    				break;
    			default : 
    				tMode = TMode.CM_POLL;
    				break;
			}
			if (tlink.attach(tMode, this, accessRate) < 0) {
				tlink.close();
				tlink = null;
				connected = false;
				throw new ConnectionFailed("No permanent connection to server (\"" + tlink.getLastError() + "\".", null);
			}
			if (tMode == TMode.CM_SINGLE) {
				connected = false;
			} else {
				isTransfering = true;
			}
		}
		
		public void closeLink() {
			if (tlink != null) {
				tlink.close();
				tlink = null;
			}
			isTransfering = false;
		}
		
	}
	
	private String server = "VideoAnalysis-EQM";
	private String device = "Device";
	private int backgroundId = -1;
	
	private AnalysisHandler analysisHandler;
	
	private AImage aImage;
	private IMAGE background;
	private double[] preciseBackground;
	private int preciseBackgroundWidth = -1;
	private int accessRate = 1000;
	private AccessMode accessMode = AccessMode.POLL;
	private int timeout = 1000;
	private boolean connected = false;
	
	private PropertyChangeSupport pcSupport;
	
	/**
	 * Sets the analysis parameters (threshold and region of interest) to analysis server.
	 * @param threshold the threshold
	 * @param roiX the region's starting X coordinate
	 * @param roiY the region's starting Y coordinate
	 * @param roiW the region's width
	 * @param roiH the region's height
	 * @param calculateThreshold true if the threshold should be calculated by the server or false otherwise
	 * @throws ConnectionFailed if unable to connect
	 */
	public void setAnalysisParameters(double threshold, int roiX, int roiY, int roiW, int roiH, int roi2X, int roi2Y, int roi2W, int roi2H, boolean calculateThreshold,boolean performFit) throws ConnectionFailed {
		TDataType tdt = new TDataType(new double[]{threshold, roiX, roiY, roiW, roiH,roi2X, roi2Y, roi2W,roi2H,calculateThreshold ? 1.0 : 0.0,performFit ? 1.0 : 0.0});
		TLink tl = new TLink(server+"/"+device, ANALYSIS_PARAMETERS_PROPERTY, null, tdt, TAccess.CA_WRITE);
		tl.execute(timeout, true);
		tl.close();
	}
	
	/**
	 * Sets the background image to analysis server.
	 * @param background the background image to set
	 * @throws ConnectionFailed if unable to connect
	 */

	public void setBackground(IMAGE background) throws ConnectionFailed {
		if (background == null) {
			background = new IMAGE(0);
		}
		TDataType tdt = new TDataType(new IMAGE[]{background});
		TLink tl = new TLink(server+"/"+device, BACKGROUND_PROPERTY, null, tdt, TAccess.CA_WRITE);
		tl.execute(timeout, true);
		tl.close();
	}
		
	/**
	 * Gets the background image from analysis server.
	 * @return the background image
	 * @throws ConnectionFailed
	 */
	public IMAGE getBackground() {
		return background;
	}
	
	private void loadBackground() {
		if (background == null) {
			background = new IMAGE(/*VideoHeaderV3.TRANSPORT_LENGTH_V3*/); // mdavid: commented
		}
		TDataType tdt = new TDataType(new IMAGE[]{background});
		TLink tl = new TLink(server+"/"+device, BACKGROUND_PROPERTY, tdt, null, TAccess.CA_READ);
		tl.execute(timeout, true);
		tdt.getData(background);
		tl.close();
//		if (background != null && background.getFrameHeader().sourceWidth == 0 && background.getFrameHeader().sourceHeight == 0) background = null;
	}
	
	private void loadPreciseBackground() {
		//TODO put somewhere the image width;
		if (preciseBackground == null) {
			preciseBackground = new double[0];
		}
		TDataType tdt = new TDataType(preciseBackground);
		TLink tl = new TLink(server+"/"+device, PRECISE_BACKGROUND_PROPERTY, tdt, null, TAccess.CA_READ);
		tl.execute(timeout, true);
		tdt.getData(preciseBackground);
		tl.close();
	}
	
	/**
	 * Sets the precise background image array.
	 * 
	 * @param background the image array
	 * @param width the width of the image
	 * @throws ConnectionFailed
	 */
	public void setPreciseBackground(double[] background, int width) throws ConnectionFailed {
		//TODO put the width somewhere
		if (background == null) {
			background = new double[0];
		}
		TDataType tdt = new TDataType(background);
		TLink tl = new TLink(server+"/"+device, PRECISE_BACKGROUND_PROPERTY, null, tdt, TAccess.CA_WRITE);
		tl.execute(timeout, true);
		tl.close();
	}
	
	/**
	 * Returns the precise background width.
	 * @return precise background image width
	 */
	public int getPreciseBackgroundWidth() {
		return preciseBackgroundWidth;
	}
	
	/**
	 * Returns the precise background array.
	 * 
	 * @return the precise background
	 */
	public double[] getPreciseBackground() {
		return preciseBackground;
	}
	
	public int getBackgroundId() {
		int[] result = new int[] {-1};
		TDataType tdt = new TDataType(result);
		TLink tl = new TLink(server+"/"+device, BACKGROUND_ID_PROPERTY, tdt, null, TAccess.CA_READ);
		tl.execute(timeout, true);
		tl.close();
		return result[0];
	}
	
	/**
	 * Connects this </code>VideoServerConnection</code> to video analysis server.
	 * @throws ConnectionFailed if unable to connect
	 */
	public void connect() throws ConnectionFailed {
		if (connected) return;
		getAnalysisHandler().openLink(accessRate);
		connected = true;
	}
	
	/**
	 * Disconnects this </code>VideoServerConnection</code> from video analysis server.
	 */
	public void disconnect() {
		if (!connected) return;
		TineHandler.logger.log(Level.INFO, "Stop analysis.");
		getAnalysisHandler().closeLink();
		connected = false;
	}
	
	public void setConnectionParameters(ConnectionParameters cp) {
		disconnect();
		server = cp.getDeviceGroup();
		device = cp.getDeviceName();
		accessRate = cp.getAccessRate();
		accessMode = cp.getAccessMode();
	}
	
	/**
	 * Gets the selected access rate.
	 * @return the selected access rate
	 */
	public int getAccessRate() {
		return accessRate;
	}
	
	public void setAccessRate(int accessRate) {
		if (this.accessRate == accessRate) return;
		this.accessRate = accessRate;
		if (connected) {
			disconnect();
			try {
				connect();
			} catch (ConnectionFailed e) {
				e.printStackTrace();
			}
		}
		
	}
	
	/**
	 * Gets current </code>AImage</code>.
	 * @return current </code>AImage</code>
	 */
	public AImage getAImage() {
		return aImage;
	}
	
	private AnalysisHandler getAnalysisHandler() {
		if (analysisHandler == null) {
			analysisHandler = new AnalysisHandler();
		}
		return analysisHandler;
	}
	
	private PropertyChangeSupport getPcSupport() {
		if (pcSupport == null) {
			pcSupport = new PropertyChangeSupport(this);
		}
		return pcSupport;
	}
	
	/**
	 * Adds a </code>PropertyChangeListener</code> for the specified propertyName.
	 * @param propertyName the name of the property
	 * @param listener the </code>PropertyChangeListener</code> to add
	 */
	public void addPropertyChangeListener(String propertyName, PropertyChangeListener listener) {
		getPcSupport().addPropertyChangeListener(propertyName, listener);
	}
	
	/**
	 * Removes a </code>PropertyChangeListener</code> for the specified propertyName.
	 * @param propertyName the name of the property
	 * @param listener the </code>PropertyChangeListener</code> to remove
	 */
	public void removePropertyChangeListener(String propertyName, PropertyChangeListener listener) {
		getPcSupport().removePropertyChangeListener(propertyName, listener);
	}
	
	/**
	 * Adds a </code>PropertyChangeListener</code>.
	 * @param listener the </code>PropertyChangeListener</code> to add
	 */
	public void addPropertyChangeListener(PropertyChangeListener listener) {
		getPcSupport().addPropertyChangeListener(listener);
	}
	
	/**
	 * Removes a </code>PropertyChangeListener</code>.
	 * @param listener the </code>PropertyChangeListener</code> to remove
	 */
	public void removePropertyChangeListener(PropertyChangeListener listener) {
		getPcSupport().removePropertyChangeListener(listener);
	}	
	
	public static void main(String[] args) {
		final IMAGE image = new IMAGE();
		final AImage aImage = new AImage(image, 0, 0, 0, 0, 0, 0, 0, 0, 0.0,
				new double[]{}, 0.0, 0.0, 0.0, 0.0,0.0,
				0.0,0.0,0.0,0.0,
				new double[]{}, 0.0, 0.0, 0.0, 0.0,0.0,
				0.0,0.0,0.0,0.0,
				0.0, 0.0, 0.0, 0.0, 0.0,0,true,false,true);
		TDataType dataOut = new TDataType(aImage);
		TLink tlink = new TLink("/TEST/VideoAnalysis-EQM/Device", "Analysis", dataOut, null, TAccess.CA_READ);
		int c = tlink.attach(TMode.CM_POLL, new TCallback(){
			@Override
			public void callback(int LinkIndex, int LinkStatus) {
				System.out.println(LinkIndex + " " + TErrorList.toString(LinkStatus));
				System.out.println(aImage);
			}
		}, 1000);
		System.out.println(c);
	}
}
