package de.desy.acop.video.analysis;

import de.desy.acop.video.displayer.ColorMap;
import de.desy.tine.types.IMAGE;

/**
 * 
 * <code>ImageAnalyzer</code> implements the image analysis. There can be different implementation
 * of image analysis, depending on the source and type of the image (e.g. beam cross section, interferometer
 * image).
 *
 * @author <a href="mailto:jaka.bobnar@cosylab.com">Jaka Bobnar</a>
 *
 */
public interface ImageAnalyzer {
	
	/**
	 * Perform the analysis and return the analysis image.
	 * @param colorMap
	 * @param decoder
	 * @param image
	 * @param preciseBackground
	 * @param imageWidth
	 * @param imageHeight
	 * @param roiX
	 * @param roiY
	 * @param roiW
	 * @param roiH
	 * @param roi2x
	 * @param roi2y
	 * @param roi2w
	 * @param roi2h
	 * @param threshold
	 * @param calculateThreshold
	 * @param performFit
	 * @param performSmoothing
	 * @return
	 */
	public AnalysisImage performAnalysis(ColorMap colorMap, ColorDecoder decoder, IMAGE image, double[] preciseBackground, int imageWidth, int imageHeight, int roiX, int roiY, int roiW,
			int roiH, int roi2x, int roi2y, int roi2w, int roi2h, double threshold, boolean calculateThreshold, boolean performFit,
			boolean performSmoothing);
}
