package de.desy.tine.client;

import de.desy.tine.dataUtils.*;
import de.desy.tine.definitions.*;
import de.desy.tine.exceptions.TineRuntimeErrorException;

public class TFilterLink implements TLinkCallback
{
  public String toString()
  {
    if (lnk == null) return "no assigned filter";
    return lnk.getFullDeviceNameAndProperty()+" "+getCompString()+" "+sv;
  }
  private TLink lnk = null;
  private static enum comp { neq, eq, lt, gt; };
  private static String[] compToken = { "!=", "==", "=", "<", ">" };
  private comp thisComparator = comp.eq;
  private double dv = 0;
  private String sv = "";
  private boolean active = false;
  public boolean isActive() { return active; }
  public String getCompString()
  {
    switch (thisComparator)
    {
      case eq: return "=";
      case neq: return "!=";
      case lt: return "<";
      case gt: return ">";
    }
    return "?";
  }
  public void remove()
  {
    if (lnk != null) lnk.close();
    lnk = null;
    filterIsValid = true;
    active = false;
  }
  private void makeFilterLink(String addressKey,String comparator,String value)
  {
    short dFmt = TFormat.CF_DEFAULT;
    int dSiz = 128;
    TDataType dout = new TDataType(dSiz,dFmt);
    lnk = new TLink(addressKey,dout,null,TAccess.CA_READ);
    int lid = lnk.attach(TMode.CM_TIMER, this, 1000);
    if (lid < 0)
    {
      throw new TineRuntimeErrorException(-lid);
    }
    sv = value;
    try
    {
      dv = Double.parseDouble(sv);
    }
    catch (Exception x) {};
    TLinkFactory.addFilterLink(this);
    active = true;
  }
  private String filterString = null;
  public String getFilterString() { return filterString; }
  public TFilterLink(String parsableFilterString)
  {
    if (parsableFilterString == null || parsableFilterString.length() == 0)
    {
      throw new TineRuntimeErrorException(TErrorList.argument_list_error);
    }
    int idx = -1, len = 1;
    for (int i=0; i<compToken.length && idx == -1; i++)
    {
      idx = parsableFilterString.indexOf(compToken[i]);
      len = i < 2 ? 2 : 1;
    }
    if (idx < 0) throw new TineRuntimeErrorException(TErrorList.parsing_error);
    String key = parsableFilterString.substring(0, idx).trim();
    String cmp = parsableFilterString.substring(idx, idx+len);
    String val = parsableFilterString.substring(idx+len).trim();
    makeFilterLink(key,cmp,val);
    filterString = parsableFilterString;
  }
  public TFilterLink(String addressKey,String comparator,String value)
  {
    if (addressKey == null || addressKey.length() == 0 ||
        comparator == null || comparator.length() == 0 ||
        value == null || value.length() == 0)
    {
      throw new TineRuntimeErrorException(TErrorList.argument_list_error);
    }
    if      (comparator.compareTo("=") == 0) thisComparator = comp.eq;
    else if (comparator.compareTo("==") == 0) thisComparator = comp.eq;
    else if (comparator.compareTo("!=") == 0) thisComparator = comp.neq;
    else if (comparator.compareTo("<") == 0) thisComparator = comp.lt;
    else if (comparator.compareTo(">") == 0) thisComparator = comp.gt;
    else
    {
      throw new TineRuntimeErrorException(TErrorList.invalid_parameter);      
    }
    makeFilterLink(addressKey,comparator,value);
  }
  private boolean filterIsValid = true;
  public boolean isValid()
  {
    return filterIsValid;
  }
  @Override
  public void callback(TLink link)
  {
    if (!TErrorList.hasData(link.linkStatus))
    { // punt
      filterIsValid = true;
      return;
    }
    TDataType d = link.getOutputDataObject();
    if (d == null)
    {
      filterIsValid = true;
      link.close();
      return;      
    }
    if (TFormat.isNumberFormat(d.dFormat))
    {
      boolean top = false;
      switch (thisComparator)
      {
        case eq: top = true;
        case neq:
        {
          int[] v = new int[1];        
          d.getData(v);
          filterIsValid = top ? (v[0] == (int)dv) : (v[0] != (int)dv);
          return;
        }
        case gt: top = true;
        case lt:
        {
          double[] v = new double[1];
          d.getData(v); 
          filterIsValid = top ? (v[0] > dv) : (v[0] < dv);
          return;
        }          
      }
    }
    if (TFormat.isName(d.dFormat) || TFormat.CF_TEXT == d.dFormat)
    {
      char[] cs = new char[128];
      d.getData(cs);
      String s = new String(cs).trim();
      switch (thisComparator)
      {
        case eq: filterIsValid = sv.compareToIgnoreCase(s) == 0; return;
        case neq: filterIsValid = sv.compareToIgnoreCase(s) != 0; return;
        case gt: filterIsValid = s.compareToIgnoreCase(sv) > 0; return;
        case lt: filterIsValid = s.compareToIgnoreCase(sv) < 0; return;
      }      
    }
    filterIsValid = true;
    link.close();
  }
}
