package de.desy.tine.eventArchive;

import java.io.IOException;

import de.desy.tine.client.TLink;
import de.desy.tine.dataUtils.TDataType;
import de.desy.tine.definitions.TAccess;
import de.desy.tine.definitions.TErrorList;
import de.desy.tine.types.NAME16;

public class TEventArchiveDataBase
{
  private static final int nice_query_size = 256;
  private static ARBQS[] arbLst = new ARBQS[nice_query_size];
  private static DASQS[] dasLst = new DASQS[nice_query_size];
  public static ARBQS[] getTriggerTable(String context) throws IOException
  { // "TRIGGER.TBL"
    if (context == null || context.length() == 0) return null;
    if (arbLst[0] == null)
    {
      for (int i = 0; i < nice_query_size; i++) arbLst[i] = new ARBQS();
    }
    else
    {
      for (int i = 0; i < nice_query_size; i++) arbLst[i].clear();
    }
    TDataType dout = new TDataType(arbLst);
    TLink lnk = new TLink("/"+context+"/EVENTSTORE","TRIGGER.TBL",dout, null,
        TAccess.CA_READ);
    if (lnk.executeAndClose(500) != 0) throw new IOException("link error " + lnk.getLastError());
    int n = dout.getCompletionLength();
    if (n == 0) return null;
    ARBQS[] rtLst = new ARBQS[n];
    for (int i = 0; i < n; i++) rtLst[i] = arbLst[i];
    return rtLst;
  }
  public static int setTriggerTable(String context,ARBQS[] tableItems)
  {
    if (context == null || context.length() == 0) return TErrorList.invalid_parameter;
    TDataType din;
    if (tableItems == null || tableItems.length == 0)
      din = new TDataType();
    else
      din = new TDataType(tableItems);
    TLink lnk = new TLink("/"+context+"/EVENTSTORE","TRIGGER.TBL",null, din,
        TAccess.CA_WRITE);
    return lnk.executeAndClose(500);
  }
  public static boolean isTriggered(String context)
  {
    if (context == null || context.length() == 0) return false;
    int[] sts = new int[1];
    TDataType dout = new TDataType(sts);
    TLink lnk = new TLink("/"+context+"/EVENTSTORE/ANY","TRIGGER.STATUS",dout, null,TAccess.CA_READ);
    if (lnk.executeAndClose(500) != 0) return false;
    return !(sts[0] == 0);
  }
  public static int resetServer(String context)
  {
    if (context == null || context.length() == 0) return TErrorList.invalid_parameter;
    if (isTriggered(context)) return TErrorList.operation_busy;
    TLink lnk = new TLink("/"+context+"/EVENTSTORE","SRVRESET",null, null,TAccess.CA_WRITE);
    return lnk.executeAndClose(500);
  }
  public static DASQS[] getTriggerDataBase(String context, String trigger) throws IOException
  { // "TRIGGER.DB"
    if (context == null || context.length() == 0) return null;
    if (dasLst[0] == null)
    {
      for (int i = 0; i < nice_query_size; i++) dasLst[i] = new DASQS();
    }
    else
    {
      for (int i = 0; i < nice_query_size; i++) dasLst[i].clear();
    }
    TDataType dout = new TDataType(dasLst);
    TLink lnk = new TLink("/"+context+"/EVENTSTORE/"+trigger,"TRIGGER.DB",dout, null,
        TAccess.CA_READ);
    if (lnk.executeAndClose(500) != 0) throw new IOException("link error " + lnk.getLastError());
    int n = dout.getCompletionLength();
    if (n == 0) return null;
    DASQS[] rtLst = new DASQS[n];
    for (int i = 0; i < n; i++) rtLst[i] = dasLst[i];
    return rtLst;
  }
  public static int setTriggerDataBase(String context,String trigger, DASQS[] triggerItems)
  {
    if (context == null || context.length() == 0) return TErrorList.invalid_parameter;
    TDataType din;
    if (triggerItems == null || triggerItems.length == 0)
      din = new TDataType();
    else
      din = new TDataType(triggerItems);
    TLink lnk = new TLink("/"+context+"/EVENTSTORE/"+trigger,"TRIGGER.DB",null, din,
        TAccess.CA_WRITE);
    return lnk.executeAndClose(500);
  }
  public static Object getTriggerInputData(String context,String file,int format) throws IOException
  { // "TRIGGER.ITEM.DATA"
    if (context == null || context.length() == 0) return null;
    if (file == null || file.length() == 0) return null;
    TDataType dout = new TDataType(nice_query_size,(short)format);
    TLink lnk = new TLink("/"+context+"/EVENTSTORE/"+file,"TRIGGER.ITEM.DATA",dout, null,
        TAccess.CA_READ);
    if (lnk.executeAndClose(500) != 0) throw new IOException("link error " + lnk.getLastError());
    int n = dout.getCompletionLength();
    if (n == 0) return null;
    return dout.getDataObject();
  }
  public static int setTriggerInputData(String context,String file,Object data)
  {
    if (context == null || context.length() == 0) return TErrorList.invalid_parameter;
    if (file == null || file.length() == 0) return TErrorList.invalid_parameter;
    TDataType din = null;
    if (data == null)
      din = new TDataType();
    else if (data instanceof byte[])
      din = new TDataType((byte[])data);
    else if (data instanceof short[])
      din = new TDataType((short[])data);
    else if (data instanceof int[])
      din = new TDataType((int[])data);
    else if (data instanceof float[])
      din = new TDataType((float[])data);
    else if (data instanceof double[])
      din = new TDataType((double[])data);
    else 
      return TErrorList.illegal_format;
    TLink lnk = new TLink("/"+context+"/EVENTSTORE/"+file,"TRIGGER.ITEM.DATA",null, din,
        TAccess.CA_WRITE);
    return lnk.executeAndClose(500);
  }
  public static String[] getTriggerDataFiles(String context) throws IOException
  {
    if (context == null || context.length() == 0) return null;
    NAME16[] n16 = new NAME16[nice_query_size];
    TDataType dout = new TDataType(n16);
    TLink lnk = new TLink("/"+context+"/EVENTSTORE","TRIGGER.ITEM.DATAFILES",dout, null,
        TAccess.CA_READ);
    if (lnk.executeAndClose(500) != 0) throw new IOException("link error " + lnk.getLastError());
    int n = dout.getCompletionLength();
    if (n == 0) return null;
    String[] files = new String[n];
    for (int i=0; i<n; i++) files[i] = n16[i].getName();
    return files;
  }
}
