//: TLink.java
package de.desy.tine.headers;
import java.io.ByteArrayOutputStream;
import java.io.ByteArrayInputStream;
import java.io.DataOutputStream;
import java.io.DataInputStream;
import java.io.IOException;

import de.desy.tine.addrUtils.TFecEntry;
import de.desy.tine.client.TLink;
import de.desy.tine.dataUtils.TDataType;
import de.desy.tine.definitions.TErrorList;
import de.desy.tine.definitions.TFormat;
import de.desy.tine.definitions.TMode;
import de.desy.tine.endianUtils.Swap;
import de.desy.tine.server.logger.MsgLog;
import de.desy.tine.types.SPECTRUM;
public final class TSubscription
{
  public short msgSizeInBytes;
  public short extSizeInBytes;
  public short mtu;
  public short numblks;  // careful here!  byte is signed in java !!!!!
  public short blknum;
  public short id;
  public short mode;
  public short blkid;
  public int   pollingInterval;
  public int   starttime;
  public TContract contract;
  // non 'network' fields ...
  public long linkStartTime;
  public long linkLastTime;
  public int dataSizeInBytes = 0;
  public int nextDataSizeInBytes = 0;
  static public final int hdrSizeInBytes = (8*2 + 2*4); // not including contract 
  static public final int hdrSizeInBytesLegacy = (6*2 + 2*4);
  static private short ensLinkId = 0;
  public boolean isLegacy = false;
  TLink link;
  static public int getRenewalMultiplierFromRate(int rate)
  {
    if (rate < 2) rate = 2;
    return (rate > 0 && rate < 1000 ? 1000/rate : 1);
  }
  static public int getRenewalMultiplier(int multiplier,int rate)
  {
    return (multiplier > 1 ? multiplier : getRenewalMultiplierFromRate(rate));
  }
  public void resetCounters()
  {
    blknum = 1;
    nextDataSizeInBytes = dataSizeInBytes > mtu ? mtu : dataSizeInBytes % mtu;    
  }
  public void resetCounters(short blockId)
  {
    blknum = 1;
    blkid = blockId;
    nextDataSizeInBytes = dataSizeInBytes > mtu ? mtu : dataSizeInBytes % mtu;    
  }
  public byte[] toByteArray()
  { // does not put the contract on the byte stream here, just the subscription header!
    try
    {
      short m = (short)(mode & 0x0fff);
      int hdrSize = isLegacy ? hdrSizeInBytesLegacy : hdrSizeInBytes;
      int conSize = isLegacy ? TContractP5.hdrSizeInBytes : TContract.hdrSizeInBytes; 
      msgSizeInBytes = (short)(hdrSize + conSize + extSizeInBytes);
      ByteArrayOutputStream db = new ByteArrayOutputStream(msgSizeInBytes);
      short payloadSizeInBytes = 0;
      if (link.dInput.blknum == link.dInput.numblks) 
      { // finished
        payloadSizeInBytes = (short)(dataSizeInBytes % mtu);
        link.dInput.blkid = (link.dInput.blkid+1) % 1000;
      }
      else                   
      { // move along ...
        payloadSizeInBytes = mtu;
        link.dInput.blknum++;
      }
      msgSizeInBytes += payloadSizeInBytes;
      DataOutputStream ds = new DataOutputStream(db);
      ds.writeShort(Swap.Short(msgSizeInBytes));
      if (!isLegacy)
      { 
        ds.writeShort(Swap.Short(extSizeInBytes));
        ds.writeShort(Swap.Short(mtu));
        ds.writeShort(Swap.Short(numblks));
        ds.writeShort(Swap.Short(blknum));
        ds.writeShort(Swap.Short(id));
        ds.writeShort(Swap.Short(m));
        ds.writeShort(Swap.Short(blkid));
        ds.writeInt(Swap.Long(pollingInterval));
        ds.writeInt(Swap.Long(starttime));
      }
      else
      { // protocol < 6
        ds.writeShort(Swap.Short(mtu));
        ds.writeByte(numblks);
        ds.writeByte(blknum);
        ds.writeShort(Swap.Short(id));
        ds.writeShort(Swap.Short(m));
        ds.writeShort(Swap.Short(blkid));
        ds.writeInt(Swap.Long(pollingInterval));
        ds.writeInt(Swap.Long(starttime));
      }
      ds.close();
      nextDataSizeInBytes = payloadSizeInBytes; 
      blknum = (short)link.dInput.blknum;
      return db.toByteArray(); 
    }
    catch (IOException e)
    {
      e.printStackTrace();
      MsgLog.log("TSubscription.toByteArray",e.toString(),TErrorList.code_failure,e,0);
    }
    return null;  
  }
  //Date  date = new Date();
  // out-going (client-side) ...
  public TSubscription(TContract c, TLink lnk) 
  {
    if (c == null || lnk == null) return;
    contract = c;
    link = lnk; // store the link reference for this subscription
    isLegacy = lnk.getTineProtocol() < 6;
    int hdrSize = isLegacy ? hdrSizeInBytesLegacy : hdrSizeInBytes;
    int conSize = isLegacy ? TContractP5.hdrSizeInBytes : TContract.hdrSizeInBytes;
    mtu = (short)lnk.dInput.getMtu();
    numblks = (short)lnk.dInput.numblks;
    blknum = (short)(lnk.dInput.blknum);
    extSizeInBytes = blknum == 1 ? (short)c.extStringSize : 0;
    int padding = extSizeInBytes % 2; 
    msgSizeInBytes = (short)(hdrSize + conSize + extSizeInBytes + padding);
    id = (short)lnk.linkId;
    mode = TMode.CM_REGISTER;
    dataSizeInBytes = lnk.dInput.getSizeInBytes();
    nextDataSizeInBytes = blknum == numblks ? dataSizeInBytes : mtu;
    blkid = (short)(lnk.dInput.blkid);
    pollingInterval = lnk.devTimeout;
    linkStartTime = System.currentTimeMillis();
    linkLastTime = linkStartTime;
    starttime = (int)(linkStartTime/1000);
    if (id == 0 && lnk.expName.startsWith("ENS"))
    {
      starttime = TFecEntry.ENS_ID;
      ensLinkId = (short)((ensLinkId+1)%10000);
      id = ensLinkId;
    }
    if (isLegacy) contract.setLegacy(lnk);
  }
  // in-coming (server-side)...
  public TSubscription(int protocol, byte[] data, int off, int len) 
  {  
    try
    {
      ByteArrayInputStream is = new ByteArrayInputStream(data,off,len); 
      DataInputStream ds = new DataInputStream(is);
      msgSizeInBytes = Swap.Short(ds.readShort());
      if (protocol < 6)
      {
        mtu = Swap.Short(ds.readShort());
        numblks = TDataType.UnsignedShort(ds.readByte());
        blknum = TDataType.UnsignedShort(ds.readByte());
        id = Swap.Short(ds.readShort());
        mode = Swap.Short(ds.readShort());
        blkid = Swap.Short(ds.readShort());
        pollingInterval = Swap.Long(ds.readInt());
        starttime = Swap.Long(ds.readInt());
        extSizeInBytes = 0;
        contract = new TContract(new TContractP5(data,off+hdrSizeInBytesLegacy,TContractP5.hdrSizeInBytes));        
      }
      else
      {
        extSizeInBytes = Swap.Short(ds.readShort());
        mtu = Swap.Short(ds.readShort());
        numblks = Swap.Short(ds.readShort());
        blknum = Swap.Short(ds.readShort());
        id = Swap.Short(ds.readShort());
        mode = Swap.Short(ds.readShort());
        blkid = Swap.Short(ds.readShort());
        pollingInterval = Swap.Long(ds.readInt());
        starttime = Swap.Long(ds.readInt());
        off += hdrSizeInBytes;
        contract = new TContract(data,off,TContract.hdrSizeInBytes);
        if (extSizeInBytes > 0) 
        {
          off += TContract.hdrSizeInBytes;
          contract.setExtendedStringSpace(data, off, extSizeInBytes);
        }
      }
      ds.close();
      is.close();
    } 
    catch (IOException e)
    {
      e.printStackTrace();
      MsgLog.log("TSubscription",e.toString(),TErrorList.code_failure,e,0);
    }      
  }
  public int getInputDataSize()
  {
    int fmtsiz = 0;
    if ((fmtsiz=TFormat.formatSizeOf(contract.dataFormatIn)) > 0)
    {
      int datasize = fmtsiz * contract.dataSizeIn;
      if (datasize > mtu) 
      {
        datasize = blknum < numblks ? mtu : datasize % mtu;
      }
      return datasize;
    }
    return 0;
  }
  public int getOutputDataSize()
  {
    int fmtsiz = 0;
    if ((fmtsiz=TFormat.formatSizeOf(contract.dataFormatOut)) > 0)
    {
      if (contract.dataFormatOut == TFormat.CF_MDA)
        fmtsiz = TFormat.getCarriedFormatSize(contract.dataTagOut);
      int datasize = fmtsiz * contract.dataSizeOut;
      datasize += TFormat.getFormatHeaderSize(contract.dataFormatOut);
      return datasize;
    }
    return 0;
  }
  public String toString()
  {
    String msg = new String("msg size = " + msgSizeInBytes);
    msg = msg.concat("\next size : " + extSizeInBytes);
    msg = msg.concat("\nmtu : " + mtu);
    msg = msg.concat("\nnumblks : " + numblks);
    msg = msg.concat("\nblknum : " + blknum);
    msg = msg.concat("\nid : " + id);
    msg = msg.concat("\nmode : " + TMode.toString(mode));
    msg = msg.concat("\nblkid : " + blkid);
    msg = msg.concat("\npollingInterval : " + pollingInterval + " ms");
    msg = msg.concat("\nstarttime : " + starttime);
    msg = msg.concat("\n" + contract.toString());
    return msg;
  }
}