/*
 * Created on Aug 31, 2005
 *
 * To change the template for this generated file go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */
package de.desy.tine.server.histories;
import java.util.Arrays;
import java.util.Calendar;
import java.io.*;
import de.desy.tine.client.TFilterLink;
import de.desy.tine.client.TLinkFactory;
import de.desy.tine.dataUtils.*;
import de.desy.tine.definitions.*;
import de.desy.tine.exceptions.TineRuntimeErrorException;
import de.desy.tine.histUtils.THistory;
import de.desy.tine.server.equipment.TEquipmentModule;
import de.desy.tine.server.logger.DbgLog;
import de.desy.tine.server.logger.MsgLog;
import de.desy.tine.server.logger.TFecLog;
import de.desy.tine.server.properties.TPropertyDescription;
import de.desy.tine.server.properties.TPropertyList;
import de.desy.tine.startup.TInitializerFactory;
import de.desy.tine.types.*;

public class THistoryRecord
{
  private static final long msecondsInMonth = (long)3600 * 24 * 31 * 1000;
  private static final int AS_STATE_SIZE = 16;
  public static final int HRS_FREE_RECORD_INDEX_OFFSET = 100;
  private TEquipmentModule eqm;
  private String prp;   
  private String dev; 
  private int siz;
  private byte fmt;
  private String tag = null;
  private int  aType;
  private THistorySpecification hspec;
  public THistorySpecification getHistorySpecification() { return hspec; }
  private int  rngPtr;
  private int  rcdIdx;       // local index identifier
  private int  rngBufferSize;
  private float range = 0;
  private long lastArchiveTime;
  private long lastAccessTime;
  public long getLastAccessTime() { return lastAccessTime; }
  public void setLastAccessTime(long time) { lastAccessTime = time; }
  private String ArchiveFileName; // current working file name
  private int lastArchivedDayOfTheMonth = -1;
  private boolean useMonthlyHistoryFiles = false;
  private boolean useMinimalStorage = false; // changed to false 15.1.09
  private boolean useStandardFileSet = false;
  private boolean storeSummary = false;
  public boolean isStoreSummary() { return storeSummary; }
  private TFilterLink filter = null;
  public TFilterLink getFilter() { return filter; }
  public void setFilter(TFilterLink filter) 
  { 
    String fstr = filter == null ? "none" : filter.toString();
    TFecLog.log("THistoryRecord","set filter to "+fstr);
    this.filter = filter; 
  }
  public boolean isPastFilter()
  {
    if (filter != null && !filter.isActive())
    { // marked for removal;
      filter = null;
    }
    return filter == null ? true : filter.isValid();
  }
  public void setUseStandardFileSet(boolean value)
  {
    useStandardFileSet = true;
  }
  public boolean getUseStandardFileSet() { return useStandardFileSet; }
  private Calendar calendar;
  private long[] timestamp; // timestamp array
  private int[] sysstamp; // system stamp array
  private int[] usrstamp; // user stamp array
  private long lastReadTimestamp = 0;
  private int lastReadSystemStamp = 0;
  private boolean isScheduled = false;
  private boolean keepPointsOfInterest = false;
  private static String arcDbPath = TInitializerFactory.getInstance().getInitializer().getHistoryHome();
  private static String arcDbPathSaved = arcDbPath + "/SAVED";
  public String dump()
  {
    StringBuffer sb = new StringBuffer(512);
    sb.append(prp+" "+dev+" : "+siz+" "+TFormat.toString(fmt)+" values");
    sb.append("("+hspec.getDepthShort()+" deep, "+hspec.getDepthLong()+" months)\n");
    sb.append("\tinitialized : "+ (rngPtr != -1) + "\n");
    sb.append("\tscheduled : "+isScheduled+"\n");
    sb.append("\tkeeps points of interest : "+keepPointsOfInterest+"\n");
    sb.append("\tpolling interval : "+hspec.getPollngRate()+" msec\n");
    sb.append("\tarchive intervals : minimum "+hspec.getArchiveRate()+" msec, maximum "+hspec.getHeartbeat()+" sec\n");
    sb.append("\tarchive tolerance : absolute "+hspec.getAbsoluteTolerance()+", relative "+hspec.getPercentTolerance()+" percent\n");
    sb.append("\trecord length : "+getRecordLength()+" bytes\n");
    sb.append("\tcurrent file name : "+ArchiveFileName+"\n");
    sb.append("\tcurrent data timestamp : "+TDataTime.toString(lastAccessTime)+"\n");
    sb.append("\tlast archived : "+TDataTime.toString(lastArchiveTime)+"\n");
    if (filter != null)
    {
      sb.append("\filter : "+filter.getFilterString()+" : valid = "+filter.isValid()+"\n");
    }
    return sb.toString();
  }
  public static void setHistoryStaticFilesRepository(String path)
  {
    if (path == null) return;
    if (path.length() == 0) path = ".";
    if (path.endsWith("/") || path.endsWith("\\")) path = path.substring(0, path.length()-1);
    arcDbPathSaved = path;
  }
  public static String getHistoryStaticFilesRepository() { return arcDbPathSaved; }
  public void setScheduled(boolean value) { isScheduled = value; }
  public boolean isScheduled() { return isScheduled; }
  private int getRecordPrefixLength() { return useMinimalStorage ? 4 : 16; }
  Object rngBuffer;
  public static boolean isAllowableFormat(String fmt)
  {
    return isAllowableFormat(TFormat.getFormatCode(fmt));
  }
  public static boolean isAllowableFormat(short fmt)
  {
    switch (fmt)
    {
      case TFormat.CF_BYTE:
      case TFormat.CF_DOUBLE:
      case TFormat.CF_SHORT:
      case TFormat.CF_LONG:
      case TFormat.CF_DLONG:          
      case TFormat.CF_FLOAT:
      case TFormat.CF_NAME16:
      case TFormat.CF_NAME32:
      case TFormat.CF_NAME64:
      // as of Release 4.0:
      case TFormat.CF_FLTINT:
      case TFormat.CF_FLTFLT:
      case TFormat.CF_INTINT:
      case TFormat.CF_DBLDBL:
      case TFormat.CF_DBLTIME:
      case TFormat.CF_NAME16I:
      case TFormat.CF_NAME32I:
      case TFormat.CF_NAME64I:
      case TFormat.CF_NAME16FI:
      case TFormat.CF_NAME16DBLDBL:
      case TFormat.CF_NAME32DBLDBL:
      case TFormat.CF_NAME64DBLDBL:
        return true;
    }
    return false;
  }
  private THistoryRecord(THistoryRecord reference)
  {
    eqm = reference.eqm;
    prp = reference.prp;
    dev = reference.dev;
    siz = reference.siz;
    fmt = reference.fmt;
    aType = reference.aType;
    rcdIdx = reference.rcdIdx;
    hspec = reference.hspec;
    rngBufferSize = reference.rngBufferSize;
    rngBuffer = reference.rngBuffer;
    rngPtr = reference.rngPtr;
    timestamp = reference.timestamp;
    sysstamp = reference.sysstamp;
    usrstamp = reference.usrstamp;
    useMinimalStorage = reference.useMinimalStorage;
    useMonthlyHistoryFiles = reference.useMonthlyHistoryFiles;
  }
  public THistoryRecord(TEquipmentModule equipmentModule,String devName,String devProperty,int size,
      int format,int arrayType,int recordIndex,THistorySpecification hstSpecification)
  {
    rngPtr = -1; // ring buffer as yet empty !
    eqm = equipmentModule;
    prp = devProperty;
    dev = devName;
    siz = size;
    fmt = (byte)format;
    aType = arrayType;
    rcdIdx = recordIndex;
    hspec = hstSpecification;
    if (eqm == null)
    {
      TFecLog.log("THistoryRecord: equipment module null pointer!");
      return;
    }
    if (hspec != null)
    {
      int pintvl = hspec.getPollngRate();
      if (pintvl < eqm.getHistoryCycleInterval())
      {
        eqm.setHistoryCycleInterval(pintvl);
      }
    }
    if (!eqm.hasProperty(prp))
    {
      TFecLog.log("THistoryRecord: equipment module does not have property "+prp+"!");
      return;      
    }
    if (siz == 1) keepPointsOfInterest = true;
    TPropertyList plst = eqm.getPropertyList();
    TPropertyDescription dsc = plst.getFirstProperty(prp).getDescription();
    if (dsc != null)
    {
      if (TArrayType.isChannel(dsc.getArrayType())) keepPointsOfInterest = true;
      range = dsc.getYRange().getMaxValue() - dsc.getYRange().getMinValue(); 
    }
    if (hspec.getPollngRate() <= 0)
    {
      TFecLog.log("THistoryRecord: polling interval "+hspec.getPollngRate()+" for property "+prp+" is not allowed");
      return;            
    }
    TFecLog.log("keep history of " + devName + " " + devProperty);
    if (hspec.getDepthLong() == 0 && hspec.getArchiveRate() > hspec.getPollngRate())
    { // signal to keep a data summary in long-term storage
      storeSummary = true;
      keepPointsOfInterest = false;
    }
    rngBufferSize = siz * hspec.getDepthShort();
    rngBufferSize += siz; // add space to store last archived record
    switch (fmt)
    {
      case TFormat.CF_BYTE:
        rngBuffer = new byte[rngBufferSize];
        break;
      case TFormat.CF_SHORT:
        rngBuffer = new short[rngBufferSize];
        break;
      case TFormat.CF_LONG:
        rngBuffer = new int[rngBufferSize];
        break;
      case TFormat.CF_DLONG:
        rngBuffer = new long[rngBufferSize];
        break;        
      case TFormat.CF_FLOAT:
        rngBuffer = new float[rngBufferSize];
        break;
      case TFormat.CF_DOUBLE:
        rngBuffer = new double[rngBufferSize];
        break;
      case TFormat.CF_NAME16:
        rngBuffer = new NAME16[rngBufferSize];
        for (int i=0; i<rngBufferSize; i++)
          ((NAME16[])rngBuffer)[i] = new NAME16();
        break;
      case TFormat.CF_NAME32:
        rngBuffer = new NAME32[rngBufferSize];
        for (int i=0; i<rngBufferSize; i++)
          ((NAME32[])rngBuffer)[i] = new NAME32();
        break;
      case TFormat.CF_NAME64:
        rngBuffer = new NAME64[rngBufferSize];
        for (int i=0; i<rngBufferSize; i++)
          ((NAME64[])rngBuffer)[i] = new NAME64();
        break;
      case TFormat.CF_FLTINT:
        rngBuffer = new FLTINT[rngBufferSize];
        for (int i=0; i<rngBufferSize; i++)
          ((FLTINT[])rngBuffer)[i] = new FLTINT();
        break;
      case TFormat.CF_FLTFLT:
        rngBuffer = new FLTFLT[rngBufferSize];
        for (int i=0; i<rngBufferSize; i++)
          ((FLTFLT[])rngBuffer)[i] = new FLTFLT();
        break;
      case TFormat.CF_INTINT:
        rngBuffer = new INTINT[rngBufferSize];
        for (int i=0; i<rngBufferSize; i++)
          ((INTINT[])rngBuffer)[i] = new INTINT();
        break;
      case TFormat.CF_DBLDBL:
        rngBuffer = new DBLDBL[rngBufferSize];
        for (int i=0; i<rngBufferSize; i++)
          ((DBLDBL[])rngBuffer)[i] = new DBLDBL();
        break;
      case TFormat.CF_DBLTIME:
        rngBuffer = new DBLTIME[rngBufferSize];
        for (int i=0; i<rngBufferSize; i++)
          ((DBLTIME[])rngBuffer)[i] = new DBLTIME();
        break;
      case TFormat.CF_NAME16I:
        rngBuffer = new NAME16I[rngBufferSize];
        for (int i=0; i<rngBufferSize; i++)
          ((NAME16I[])rngBuffer)[i] = new NAME16I();
        break;
      case TFormat.CF_NAME32I:
        rngBuffer = new NAME32I[rngBufferSize];
        for (int i=0; i<rngBufferSize; i++)
          ((NAME32I[])rngBuffer)[i] = new NAME32I();
        break;
      case TFormat.CF_NAME64I:
        rngBuffer = new NAME64I[rngBufferSize];
        for (int i=0; i<rngBufferSize; i++)
          ((NAME64I[])rngBuffer)[i] = new NAME64I();
        break;
      case TFormat.CF_NAME64DBL:
        rngBuffer = new NAME64DBL[rngBufferSize];
        for (int i=0; i<rngBufferSize; i++)
          ((NAME64DBL[])rngBuffer)[i] = new NAME64DBL();
        break;
      case TFormat.CF_NAME64TIME:
        rngBuffer = new NAME64TIME[rngBufferSize];
        for (int i=0; i<rngBufferSize; i++)
          ((NAME64TIME[])rngBuffer)[i] = new NAME64TIME();
        break;
      case TFormat.CF_NAME16FI:
        rngBuffer = new NAME16FI[rngBufferSize];
        for (int i=0; i<rngBufferSize; i++)
          ((NAME16FI[])rngBuffer)[i] = new NAME16FI();
        break;
      case TFormat.CF_NAME16II:
        rngBuffer = new NAME16II[rngBufferSize];
        for (int i=0; i<rngBufferSize; i++)
          ((NAME16II[])rngBuffer)[i] = new NAME16II();
        break;
      case TFormat.CF_NAME16DBLDBL:
        rngBuffer = new NAME16DBLDBL[rngBufferSize];
        for (int i=0; i<rngBufferSize; i++)
          ((NAME16DBLDBL[])rngBuffer)[i] = new NAME16DBLDBL();
        break;
      case TFormat.CF_NAME32DBLDBL:
        rngBuffer = new NAME32DBLDBL[rngBufferSize];
        for (int i=0; i<rngBufferSize; i++)
          ((NAME32DBLDBL[])rngBuffer)[i] = new NAME32DBLDBL();
        break;
      case TFormat.CF_NAME64DBLDBL:
        rngBuffer = new NAME64DBLDBL[rngBufferSize];
        for (int i=0; i<rngBufferSize; i++)
          ((NAME64DBLDBL[])rngBuffer)[i] = new NAME64DBLDBL();
        break;
      default:
        TFecLog.log("history record of format type " + TFormat.toString(fmt) + " not allowed!");
        return;
    }
    rngBufferSize -= siz; // remove extra storage space from official size 
    timestamp = new long[rngBufferSize];
    sysstamp = new int[rngBufferSize];
    usrstamp = new int[rngBufferSize];
  }
  public int writeRecordToPIF(TDataType dout)
  {
    THistoryRecord thr = new THistoryRecord(this);
    char[] afn = this.ArchiveFileName.toCharArray();
    "pi".getChars(0, 2, afn, 0);
    thr.ArchiveFileName = new String(afn);
    thr.useStandardFileSet = false;
    thr.writeRecordToLTS(dout);
    return 0;
  }
  public int writeRecordToLTS(TDataType dout)
  { // write to Long-Term-Storage (i.e. disk)
    Object values = dout.getDataObject();
    long ts = dout.getDataTimeStamp();
    short fmt = getFmt();
    String fileName;
    if (ts < 1000)
    { // either not set or invalid
      DbgLog.log("writeRecordToLTS","invalid timestamp "+ts+" refused");
      return 0;
    }
    try 
    {
      File d = new File(arcDbPath);
      if (!d.exists()) d.mkdirs();
      DSUMMARY ds = null;
      if (storeSummary)
      {
        double si = hspec.getAbsoluteTolerance()/1000.0;
        int cc = 0;
        try { ds = getDataSummaryFromSTS(si); } 
        catch (TineRuntimeErrorException e) { cc = e.getErrorCode(); }
        if (cc != 0) return cc;
        double[] dvals = new double[6];
        dvals[0] = ds.getAverage();
        dvals[1] = ds.getDeviation();
        dvals[2] = ds.getMax();
        dvals[3] = ds.getMin();
        values = dvals;
        fmt = TFormat.CF_DOUBLE;
        siz = 6; // DSUMMARY = effectively 6 doubles
      }
      int nsiz = TFormat.formatSizeOf(fmt);
      ByteArrayOutputStream baos = new ByteArrayOutputStream(getRecordPrefixLength()+siz*nsiz);
      DataOutputStream os = new DataOutputStream(baos);
      //n.b. writeDouble(), etc. writes BIG-Endian to disk
      //     the compound types write LITTLE-Endian to disk
      if (!useMinimalStorage)
      {
        os.writeDouble((double)ts/1000.0);
        os.writeInt(dout.getSystemDataStamp());
        os.writeInt(dout.getUserDataStamp());
      }
      else
      {
        os.writeInt((int)(ts/1000));
      }
      switch (fmt)
      {
        case TFormat.CF_BYTE:
          for (int i=0; i<siz; i++) os.writeByte(((byte[])values)[i]);
          break;
        case TFormat.CF_SHORT:
          for (int i=0; i<siz; i++) os.writeShort(((short[])values)[i]);
          break;
        case TFormat.CF_LONG:
          for (int i=0; i<siz; i++) os.writeInt(((int[])values)[i]);
          break;
        case TFormat.CF_DLONG:
          for (int i=0; i<siz; i++) os.writeLong(((long[])values)[i]);
          break;          
        case TFormat.CF_FLOAT:
          for (int i=0; i<siz; i++) os.writeFloat(((float[])values)[i]);
          break;
        case TFormat.CF_DOUBLE:
          for (int i=0; i<siz; i++) os.writeDouble(((double[])values)[i]);
          break;
        case TFormat.CF_NAME16:
        case TFormat.CF_NAME32:
        case TFormat.CF_NAME64:
          for (int i=0; i<siz; i++) os.write(((NAME[])values)[i].toByteArray(),0,nsiz);
          break;
        default:
          if (values instanceof TCompoundDataObject[])
          {
            for (int i=0; i<siz; i++) 
              os.write(((TCompoundDataObject[])values)[i].toByteArray(),0,nsiz);
          }
      }
      fileName = arcDbPath +  File.separator + ArchiveFileName;
      RandomAccessFile raf = new RandomAccessFile(fileName,"rw");
      int p = getFilePosition(raf);
      raf.seek(p); 
      raf.write(baos.toByteArray());
      setFilePosition(raf, p + getRecordLength());
      raf.close();
      lastArchiveTime = ts;
      // write the 'last archived data into the extended space of the ring buffer :
      writeRecordToRingBuffer(values, hspec.getDepthShort());
    }
    catch (Exception e)
    {
      TFecLog.log("HST: cannot write record " + getDev() + " " + getPrp());
    }
    return 0;
  }
  public long getLastReadTimestamp() { return lastReadTimestamp; }
  public int getLastReadSystemStamp() { return lastReadSystemStamp; }
  public int getNumDaysInMonth(int year,int month)
  {
    boolean isLeapYear = false;
    if ((year % 4) == 0) isLeapYear = true;
    switch (month)
    {
      case 0: case 2: case 4: case 6: case 7: case 9: case 11: return 31;
      case 1: return isLeapYear ? 29 : 28;
      default: return 30;
    }
  }
  public int getNumDaysInTimeSpan(long t1,long t2)
  {
    if (calendar == null) calendar = Calendar.getInstance();    
    int n = 0, nx, y, d, ms, me, i;

    calendar.setTimeInMillis(t1);
    int y1 = calendar.get(Calendar.YEAR);
    int m1 = calendar.get(Calendar.MONTH);
    int d1 = calendar.get(Calendar.DAY_OF_MONTH);
    
    calendar.setTimeInMillis(t2);
    int y2 = calendar.get(Calendar.YEAR);
    int m2 = calendar.get(Calendar.MONTH);
    int d2 = calendar.get(Calendar.DAY_OF_MONTH);

    d = getNumDaysInMonth(y2,m2);
    nx = d1 + (d-d2-1);
    for (y=y1; y<y2+1; y++)
    {
      ms = y == y1 ? m1 : 0;
      me = y == y2 ? m2 : 11;
      for (i=ms; i<me+1; i++) 
      {
        n += getNumDaysInMonth(y,i);
      }
    }
    return n - nx;
  }
  public int getDataFromLTS(int index,long starttime,long endtime,int sysstart,int sysstop,
      float lwrLimit,float uprLimit,Object dataTarget,int arraysize,int arrayOffset,int sample)
  {
    String fn;
    int ps;
    int recsread=0,recsfound=0,rnumber=0,asize,daystotal,daysinmonth;
    boolean finalPass=false, needAnotherPass=false;
    long t = TDataTime.now(),itime = 0,stsTime;
    THistorySpecification spc = getHspec();
    
    if (index < 0 || index > siz) return -TErrorList.out_of_range;
    if (arraysize < 1) return -TErrorList.buffer_too_small;
    int depthShort = spc.getDepthShort();
    if (depthShort <= 0) return -TErrorList.server_redirection;
    int depthLong = spc.getDepthLong();
    if (storeSummary) depthLong = 1;
    stsTime = depthShort*spc.getPollngRate();
    if (depthLong <= 0)
    {
      if (endtime < t - stsTime) return -TErrorList.server_redirection;
      return 0;
    }
    if (starttime >= t - stsTime) return 0;
    if (endtime < t  - depthLong * msecondsInMonth )
    {
      return -TErrorList.server_redirection;
    }
    short targetFmt = TDataType.getObjectFormat(dataTarget);
    if (targetFmt == TFormat.CF_NULL) return -TErrorList.illegal_format;
    
    if (calendar == null) calendar = Calendar.getInstance();    
    calendar.setTimeInMillis(endtime); // end time
    int ye = calendar.get(Calendar.YEAR);
    int de = calendar.get(Calendar.DAY_OF_MONTH);
    int me = calendar.get(Calendar.MONTH);
    calendar.setTimeInMillis(starttime); // start time
    int ys = calendar.get(Calendar.YEAR); 
    int ds = calendar.get(Calendar.DAY_OF_MONTH);
    int ms = calendar.get(Calendar.MONTH);
    int yi, di, mi; // interim time
    boolean isSnapshot = (starttime == endtime);
    if (sysstart > 0 && sysstart == sysstop) isSnapshot = true;
    if (me == ms && de == ds)
    {
      asize = arraysize;
      finalPass = true;
    }
    else if (arraysize == 1)   
    { // just a number query
      asize = arraysize;
    }
    else                       
    { // spans multiple files
      asize = 0;
    }
    daystotal = getNumDaysInTimeSpan(starttime,endtime);  
    while (recsread < arraysize)
    {
      if (asize < arraysize || arraysize == 1)
      {
        yi = ys; mi = ms; di = ds;
        if (yi == ye && mi == me && 
            (di == de || useMonthlyHistoryFiles))
        {
          asize = arraysize;
          finalPass = true;
        }
        else  
        { // move along
          di = (di+1)%(getNumDaysInMonth(yi,mi)+1);
          if (di == 0)
          {
            di++;
            mi = (mi+1)%12;
            if (mi == 0) yi++;
          }
          // start at midnight 
          calendar.set(yi,mi,di,0,0,0);
          itime = calendar.getTimeInMillis();
          daysinmonth = 1;
          if ((daysinmonth =(int)((itime-starttime)/(3600*24*1000))) < 1) daysinmonth = 1;
          if (arraysize > 1) 
          {
            if (TFormat.isSimpleFormat(targetFmt) || isSnapshot) 
            { // almost certainly a snapshot if simple format
              asize = arraysize;
              if (asize >= arraysize) finalPass = true; // just in case
            }
            else if (sample > 0) 
            { // specific sampling raster given
              asize = arraysize;
            }
            else 
            { // estimate the num points in the data set
              int add = (arraysize * daysinmonth)/daystotal;
              if (add < 1) add = 1;
              asize += add;
              if (asize >= arraysize) finalPass = true; // just in case
            }
          }
        }
      }
      if (asize > arraysize) asize = arraysize;
      fn = makeFileNameFromTime(starttime);
      ps = loadLTSData(fn,index,starttime,endtime,sysstart,sysstop,lwrLimit,uprLimit,dataTarget,asize,recsread,sample);
      if (ps < 0) break;
      if (sample > 0) asize = 0;
      if (asize == 1) recsfound += (int)ps; else recsread += (int)ps;
      needAnotherPass = (ps == 0 && isSnapshot);
      if (finalPass && !needAnotherPass) break;
      starttime = itime;
      calendar.setTimeInMillis(starttime);
      ys = calendar.get(Calendar.YEAR);
      ds = calendar.get(Calendar.DAY_OF_MONTH);
      ms = calendar.get(Calendar.MONTH);
    }
    rnumber = recsfound > 0 ? recsfound : recsread;
    if (arraysize == 1) recsread = recsfound;
    if (!isSnapshot && arraysize < 2) assignNumberOfPoints(dataTarget,rnumber);

    return rnumber;
  }
  private String makeFileNameFromTime(long time)
  {
    if (calendar == null) calendar = Calendar.getInstance();
    calendar.setTimeInMillis(time);
    int dd = calendar.get(Calendar.DAY_OF_MONTH);
    int yy, mm;
    lastArchivedDayOfTheMonth = dd;
    yy = calendar.get(Calendar.YEAR) % 100;
    mm = calendar.get(Calendar.MONTH) + 1; // January = 0
    if (useMonthlyHistoryFiles) dd = 1; 
    return getFilePrefix() +  twoDigitString(yy) + twoDigitString(mm) + 
                      twoDigitString(dd) + "." + Integer.toHexString(rcdIdx);
  }
  private int getFilePosition(RandomAccessFile raf)
  {
    int fpos = -1;
    try
    {
      if (useStandardFileSet)
      { // read the current position from the header
        raf.seek(0);
        fpos = raf.readInt() + AS_STATE_SIZE;
      }
      else
      { // old style : SEEK_END
        fpos = (int)raf.length(); 
      }
    }
    catch (IOException e)
    {
      e.printStackTrace();
      MsgLog.log("THistoryRecord.getFilePosition",e.toString(),TErrorList.code_failure,e,0);        
    } 
    return fpos;
  }
  private int setFilePosition(RandomAccessFile raf, int fpos)
  {
    int cc = 0;
    try
    {
      if (useStandardFileSet)
      {
        raf.seek(0);
        raf.writeInt(fpos - AS_STATE_SIZE);
        raf.writeInt((int)(System.currentTimeMillis()/1000));
      }
    }
    catch (IOException e)
    {
      cc = -1;
    }   
    return cc;
  }
  private int getNumberOfRecords(RandomAccessFile raf)
  {
    int nrecs = 0;
    try
    {
      int fpos = 0, rsiz = getRecordLength();
      if (useStandardFileSet)
      {
        raf.seek(0);
        fpos = raf.readInt();
        raf.readInt(); // don't need the file time
        int sz = raf.readInt();
        if (rsiz != sz)
        { // do a consistency check
          if (TLinkFactory.debugLevel > 0)
            DbgLog.log("getNumberOfRecords","stored record size is invalid : " + rsiz + " vs " + sz);
          nrecs = -1;
        }
      }
      else
      { // old-style: we're always at the end
        fpos = (int)raf.length();
      }
      nrecs = fpos/rsiz;
    }
    catch (IOException e)
    {
      nrecs = -1;
    }
    return nrecs;
  }
  private double readTimeStamp(RandomAccessFile raf) throws IOException
  {
    double dt;
    int sstamp, ustamp; //TODO add these in later ...
    if (!useMinimalStorage)
    {
      dt = raf.readDouble(); 
      sstamp = raf.readInt(); ustamp = raf.readInt();
    }
    else
    {
      dt = (double)raf.readInt();
    }
    return dt;
  }
  private int locateRecordFromLTS(String fn,double dt,boolean isPoiFile)
  {
    int lts_offset = useStandardFileSet ? AS_STATE_SIZE : 0;
    double dt0,dt1,dt2;
    int nrecs,p0,p1 = lts_offset,p2;
    double f;
    int rlen = getRecordLength();
    if (isPoiFile)
    { // points-of-interest file is always stored dynamic and not standard
      lts_offset = 0;
    }
    try
    {
      RandomAccessFile raf = new RandomAccessFile(fn,"r");
      if ((nrecs = getNumberOfRecords(raf)) <= 0) nrecs = 0;
      p0 = p1 = lts_offset;
      p2 = (nrecs-1)*rlen + lts_offset;
      raf.seek(p0); 
      dt0 = readTimeStamp(raf);
      if (dt < dt0) p2 = 0;  // stay out of the while loop
      raf.seek(p2);
      dt2 = readTimeStamp(raf);
      while (p2 > p0 + rlen)
      {
        if (dt > (dt1=dt2)) { p1 = p2; break; }
        while (dt1 > dt0)   
        { // seek <--
          f = ((dt-dt0)/(dt1-dt0))*((double)(p2-p0)/rlen);
          if ((p1=p0+rlen*(int)f) >= p2) break;
          raf.seek(p1);
          dt1 = readTimeStamp(raf);
          if (dt1 < dt) break;
          p2 = p1; dt2 = dt1;
        }
        dt0 = dt1; p0 = p1;
        while (dt2 > dt1) 
        {  // seek -->
          f = ((dt2-dt)/(dt2-dt1))*((double)(p2-p0)/rlen);
          if ((p1=p2-rlen*(int)f) <= p0) break;
          raf.seek(p1);
          dt1 = readTimeStamp(raf);
          if (dt1 > dt) break;
          p0 = p1; dt0 = dt1;
        }
        p2 = p1; dt2 = dt1;
      }
      raf.close();
    }
    catch (Exception e)
    {
      DbgLog.log("locateRecordFromLTS",e.toString());
      return -TErrorList.file_error;
    }
    return p1;
  }
  
  private int getFormattedHistoryDataPoint(Object dataTarget,int position,Object dataSource,int index,double tstmp,int sstmp,int ustmp)
  {
    NAME[] n;
    TCompoundDataObject[] cdo;
    if (dataSource instanceof double[])
    {
      double[] v = (double[])dataSource;
      return getFormattedHistoryDataPoint(dataTarget,position,v[index],tstmp,sstmp,ustmp);
    }
    if (dataSource instanceof float[])
    {
      float[] v = (float[])dataSource;
      return getFormattedHistoryDataPoint(dataTarget,position,v[index],tstmp,sstmp,ustmp);
    }
    if (dataSource instanceof int[])
    {
      int[] v = (int[])dataSource;
      return getFormattedHistoryDataPoint(dataTarget,position,v[index],tstmp,sstmp,ustmp);
    }
    if (dataSource instanceof short[])
    {
      short[] v = (short[])dataSource;
      return getFormattedHistoryDataPoint(dataTarget,position,v[index],tstmp,sstmp,ustmp);
    }
    if (dataSource instanceof byte[])
    {
      byte[] v = (byte[])dataSource;
      return getFormattedHistoryDataPoint(dataTarget,position,v[index],tstmp,sstmp,ustmp);
    }
    if (dataSource instanceof long[])
    {
      long[] v = (long[])dataSource;
      return getFormattedHistoryDataPoint(dataTarget,position,v[index],tstmp,sstmp,ustmp);
    }
    if (dataSource instanceof NAME[])
    {
      n = (NAME[])dataSource;
      return getFormattedHistoryDataPoint(dataTarget,position,n[index].getName(),tstmp,sstmp,ustmp);
    }
    if (dataSource instanceof TCompoundDataObject[])
    {
      cdo = (TCompoundDataObject[])dataSource;
      return getFormattedHistoryDataPoint(dataTarget,position,cdo[index],tstmp,sstmp,ustmp);
    }
    return TErrorList.illegal_format;
  }
  private int getFormattedHistoryDataPoint(Object dataTarget,int number,double[] v,NAME[] n,TCompoundDataObject[] cdo,int index,double tstmp,int sstmp,int ustmp)
  {
    if (v != null) return getFormattedHistoryDataPoint(dataTarget,number,v[index],tstmp,sstmp,ustmp);
    if (n != null) return getFormattedHistoryDataPoint(dataTarget,number,n[index].getName(),tstmp,sstmp,ustmp);
    if (cdo != null) return getFormattedHistoryDataPoint(dataTarget,number,cdo[index],tstmp,sstmp,ustmp);
    return TErrorList.illegal_format;
  }
  private int getFormattedHistoryDataPoint(Object dataTarget,int number,double v,double tstmp,int sstmp,int ustmp)
  {
    short fmt;
    if (dataTarget instanceof HISTORY[])
    {
      fmt = TFormat.CF_HISTORY;
    }
    else
    {
      fmt = TDataType.getObjectFormat(dataTarget);
    }
    switch (fmt)
    {
      case TFormat.CF_HISTORY:
        switch (((HISTORY[])dataTarget)[0].getCarriedFormat())
        {
          case TFormat.CF_FLOAT:
            ((HISTORY[])dataTarget)[number] = new HISTORY(tstmp,(float)v,sstmp,ustmp);
            break;
          case TFormat.CF_DOUBLE:
            ((HISTORY[])dataTarget)[number] = new HISTORY(tstmp,v,sstmp,ustmp);
            break;
          case TFormat.CF_INT32:
            ((HISTORY[])dataTarget)[number] = new HISTORY(tstmp,(int)v,sstmp,ustmp);
            break;
          case TFormat.CF_INT64:
            ((HISTORY[])dataTarget)[number] = new HISTORY(tstmp,(long)v,sstmp,ustmp);
            break;
          default: 
            return -TErrorList.illegal_format;
        }        
        break;
      case TFormat.CF_FLOAT:
        ((float[])dataTarget)[number] = (float)v;
        break;
      case TFormat.CF_FLTINT:
        if (((FLTINT[])dataTarget)[number] == null) ((FLTINT[])dataTarget)[number] = new FLTINT();
        ((FLTINT[])dataTarget)[number].ival = (int)tstmp;
        ((FLTINT[])dataTarget)[number].fval = (float)v;
        break;
      case TFormat.CF_INTFLTINT:            
        if (((INTFLTINT[])dataTarget)[number] == null) ((INTFLTINT[])dataTarget)[number] = new INTFLTINT();
        ((INTFLTINT[])dataTarget)[number].i1val = (int)tstmp;
        ((INTFLTINT[])dataTarget)[number].fval = (float)v;
        break;
      case TFormat.CF_INT32:
        ((int[])dataTarget)[number] = (int)v;
        break;
      case TFormat.CF_DLONG:
        ((long[])dataTarget)[number] = (long)v;
        break;            
      case TFormat.CF_INTINT:            
        if (((INTINT[])dataTarget)[number] == null) ((INTINT[])dataTarget)[number] = new INTINT();
        ((INTINT[])dataTarget)[number].i2val = (int)tstmp;
        ((INTINT[])dataTarget)[number].i1val = (int)v;
        break;
      case TFormat.CF_DOUBLE:
        ((double[])dataTarget)[number] = (double)v;
        break;
      case TFormat.CF_DBLTIME:
        if (((DBLTIME[])dataTarget)[number] == null) ((DBLTIME[])dataTarget)[number] = new DBLTIME();
        ((DBLTIME[])dataTarget)[number].d2val = tstmp;
        ((DBLTIME[])dataTarget)[number].d1val = v;
        break;
      case TFormat.CF_DBLDBL:            
        if (((DBLDBL[])dataTarget)[number] == null) ((DBLDBL[])dataTarget)[number] = new DBLDBL();
        ((DBLDBL[])dataTarget)[number].d2val = tstmp;
        ((DBLDBL[])dataTarget)[number].d1val = v;
        break;
      case TFormat.CF_DBLDBLDBL:
        if (((DBLDBLDBL[])dataTarget)[number] == null) ((DBLDBLDBL[])dataTarget)[number] = new DBLDBLDBL();
        ((DBLDBLDBL[])dataTarget)[number].d3val = sstmp;
        ((DBLDBLDBL[])dataTarget)[number].d2val = tstmp;
        ((DBLDBLDBL[])dataTarget)[number].d1val = v;
        break;
      case TFormat.CF_FLTINTINT:
        if (((FLTINTINT[])dataTarget)[number] == null) ((FLTINTINT[])dataTarget)[number] = new FLTINTINT();
        ((FLTINTINT[])dataTarget)[number].i2val = sstmp;
        ((FLTINTINT[])dataTarget)[number].i1val = (int)tstmp;
        ((FLTINTINT[])dataTarget)[number].f1val = (float)v;
        break;
      case TFormat.CF_NAME64DBLDBL:
      case TFormat.CF_NAME64DBLDBLDBL:
        break;
      default:
        return -TErrorList.illegal_format;
    }
    return 0;
  }
  private int getFormattedHistoryDataPoint(Object dataTarget,int number,String n,double tstmp,int sstmp,int ustmp)
  {
    switch (TDataType.getObjectFormat(dataTarget))
    {
      case TFormat.CF_HISTORY:
        switch (((HISTORY[])dataTarget)[0].getCarriedFormat())
        {
          case TFormat.CF_NAME16:
            NAME16 n16 = new NAME16(n);
            ((HISTORY[])dataTarget)[number] = new HISTORY(tstmp,n16,sstmp,ustmp);
            break;
          case TFormat.CF_NAME32:
            NAME32 n32 = new NAME32(n);
            ((HISTORY[])dataTarget)[number] = new HISTORY(tstmp,n32,sstmp,ustmp);
            break;
          case TFormat.CF_NAME64:
            NAME64 n64 = new NAME64(n);
            ((HISTORY[])dataTarget)[number] = new HISTORY(tstmp,n64,sstmp,ustmp);
            break;
          default:
            return -TErrorList.illegal_format;
        }
        break;
    case TFormat.CF_NAME16I:
        if (((NAME16I[])dataTarget)[number] == null) ((NAME16I[])dataTarget)[number] = new NAME16I();
        ((NAME16I[])dataTarget)[number].ival = (int)tstmp;
        ((NAME16I[])dataTarget)[number].name = n;
        break;
      case TFormat.CF_NAME32I:
        if (((NAME32I[])dataTarget)[number] == null) ((NAME32I[])dataTarget)[number] = new NAME32I();
        ((NAME32I[])dataTarget)[number].ival = (int)tstmp;
        ((NAME32I[])dataTarget)[number].name = n;
        break;
      case TFormat.CF_NAME64I:
        if (((NAME64I[])dataTarget)[number] == null) ((NAME64I[])dataTarget)[number] = new NAME64I();
        ((NAME64I[])dataTarget)[number].ival = (int)tstmp;
        ((NAME64I[])dataTarget)[number].name = n;
        break;
      case TFormat.CF_NAME64DBL:
        if (((NAME64DBL[])dataTarget)[number] == null) ((NAME64DBL[])dataTarget)[number] = new NAME64DBL();
        ((NAME64DBL[])dataTarget)[number].dval = tstmp;
        ((NAME64DBL[])dataTarget)[number].name = n;
        break;
      case TFormat.CF_NAME64TIME:
        if (((NAME64TIME[])dataTarget)[number] == null) ((NAME64TIME[])dataTarget)[number] = new NAME64TIME();
        ((NAME64TIME[])dataTarget)[number].dval = tstmp;
        ((NAME64TIME[])dataTarget)[number].name = n;
        break;
      case TFormat.CF_NAME64DBLDBL:
        if (((NAME64DBLDBL[])dataTarget)[number] == null) ((NAME64DBLDBL[])dataTarget)[number] = new NAME64DBLDBL();
        ((NAME64DBLDBL[])dataTarget)[number].d2val = sstmp;
        ((NAME64DBLDBL[])dataTarget)[number].d1val = tstmp;
        ((NAME64DBLDBL[])dataTarget)[number].name = n;
        break;
      case TFormat.CF_NAME64DBLDBLDBL:
        break;
      default:
        return -TErrorList.illegal_format;
    }
    return 0;
  }
  private int getFormattedHistoryDataPoint(Object dataTarget,int number,TCompoundDataObject cdo,double tstmp,int sstmp,int ustmp)
  {
    float f=0; double d1=0; String n = "";
    short fmt = TDataType.getObjectFormat(dataTarget);
    if (TFormat.isTimeFormat(fmt))
    { // handle special instructions ...
      switch (fmt)
      {
        case TFormat.CF_DBLTIME:
        {
          double v = ((DBLTIME)cdo).d1val;
          double ts = ((DBLTIME)cdo).d2val;
          return getFormattedHistoryDataPoint(dataTarget,number,v,ts,sstmp,ustmp);
        }
        case TFormat.CF_NAME64TIME:
        {
          String v = ((NAME64TIME)cdo).name;
          double ts = ((NAME64TIME)cdo).dval;
          return getFormattedHistoryDataPoint(dataTarget,number,v,ts,sstmp,ustmp);
        }
      }
    }
    switch (fmt)
    {
      case TFormat.CF_HISTORY:
        if (cdo != null)
        {
          short sfmt = cdo.getFormat();
          short dfmt = ((HISTORY[])dataTarget)[0].getCarriedFormat();
          if (sfmt != dfmt) return -TErrorList.illegal_format;
          ((HISTORY[])dataTarget)[number] = new HISTORY(tstmp,cdo.clone(),sstmp,ustmp);
        }
        break;
      case TFormat.CF_FLTINTINT:
        if (cdo != null)
        { // data were stored as compound type
          if (cdo instanceof FLTINT)
          {
            FLTINT d = ((FLTINT)cdo);
            f =  d.fval; 
          }
          else if (cdo instanceof INTINT)
          {
            INTINT d = ((INTINT)cdo);
            f =  d.i1val; 
          }
          else if (cdo instanceof DBLDBL)
          {
            DBLDBL d = ((DBLDBL)cdo);
            f =  (float)d.d1val; 
          }
        }
        else 
        { // sorry !
          return -TErrorList.illegal_format;
        }
        if (((FLTINTINT[])dataTarget)[number] == null) ((FLTINTINT[])dataTarget)[number] = new FLTINTINT();
        ((FLTINTINT[])dataTarget)[number].setValues(f, (int)tstmp, sstmp);
        break;
      case TFormat.CF_DBLDBLDBL:
        if (cdo != null)
        { // data were stored as compound type
          if (cdo instanceof FLTINT)
          {
            FLTINT d = ((FLTINT)cdo);
            d1 =  d.fval; 
          }
          if (cdo instanceof FLTFLT)
          {
            FLTFLT d = ((FLTFLT)cdo);
            d1 =  d.f1val; 
          }
          else if (cdo instanceof INTINT)
          {
            INTINT d = ((INTINT)cdo);
            d1 =  d.i1val; 
          }
          else if (cdo instanceof DBLDBL)
          {
            DBLDBL d = ((DBLDBL)cdo);
            d1 =  d.d1val; 
          }
        }
        else 
        { // sorry !
          return -TErrorList.illegal_format;
        }
        if (((DBLDBLDBL[])dataTarget)[number] == null) ((DBLDBLDBL[])dataTarget)[number] = new DBLDBLDBL();
        ((DBLDBLDBL[])dataTarget)[number].setValues(d1, tstmp, sstmp);
        break;
      case TFormat.CF_NAME64DBL:
      case TFormat.CF_NAME64DBLDBL:
        if (cdo != null)
        { // data were stored as compound type
          if (cdo instanceof NAMEINT)
          {
            NAMEINT d = ((NAMEINT)cdo);
            n =  d.name; d1 = d.ival;
          }
          else if (cdo instanceof NAME64DBL)
          {
            NAME64DBL d = ((NAME64DBL)cdo);
            n = d.name; d1 = d.dval;
          }
        }
        else 
        { // sorry !
          return -TErrorList.illegal_format;
        }
        if (fmt == TFormat.CF_NAME64DBLDBL)
        {
          if (((NAME64DBLDBL[])dataTarget)[number] == null) ((NAME64DBLDBL[])dataTarget)[number] = new NAME64DBLDBL();
          ((NAME64DBLDBL[])dataTarget)[number].setValues(n, tstmp, sstmp);
        }
        else
        {
          if (((NAME64DBL[])dataTarget)[number] == null) ((NAME64DBL[])dataTarget)[number] = new NAME64DBL();
          ((NAME64DBL[])dataTarget)[number].setValues(n, tstmp);        
        }
        break;
    }
    return 0;
  }
  private int getFormattedHistoryDataSet(Object dataTarget,double[] v,NAME[] n,TCompoundDataObject[] cdo,double tstmp,int sstmp,int ustmp)
  {
    if (v != null) return getFormattedHistoryDataSet(dataTarget,v,tstmp,sstmp,ustmp);
    if (n != null) return getFormattedHistoryDataSet(dataTarget,n,tstmp,sstmp,ustmp);
    return getFormattedHistoryDataSet(dataTarget,cdo,tstmp,sstmp,ustmp);
  }
  private int getFormattedHistoryDataSet(Object dataTarget,double[] v,double tstmp,int sstmp,int ustmp)
  {
    int number = TDataType.getObjectSize(dataTarget);
    if (number > v.length) number = v.length;
    switch (TDataType.getObjectFormat(dataTarget))
    {
      case TFormat.CF_FLOAT:
        for (int i=0; i<number; i++) ((float[])dataTarget)[i] = (float)v[i];
        return number;          
      case TFormat.CF_LONG:
        for (int i=0; i<number; i++) ((int[])dataTarget)[i] = (int)v[i];
        return number;          
      case TFormat.CF_DLONG:
        for (int i=0; i<number; i++) ((long[])dataTarget)[i] = (long)v[i];
        return number;          
      case TFormat.CF_DOUBLE:
        for (int i=0; i<number; i++) ((double[])dataTarget)[i] = (double)v[i];
        return number;      
    }
    return 0;
  }
  private int getFormattedHistoryDataSet(Object dataTarget,NAME[] n,double tstmp,int sstmp,int ustmp)
  {
    int number = TDataType.getObjectSize(dataTarget);
    if (number > n.length) number = n.length;
    if (TDataType.getObjectFormat(dataTarget) != TDataType.getObjectFormat(n))
      return -TErrorList.illegal_format;
    for (int i=0; i<number; i++) ((NAME[])dataTarget)[i] = n[i];
    return number;
   }
  private int getFormattedHistoryDataSet(Object dataTarget,TCompoundDataObject[] cdo,double tstmp,int sstmp,int ustmp)
  {
    int number = TDataType.getObjectSize(dataTarget);
    if (number > cdo.length) number = cdo.length;
    if (TDataType.getObjectFormat(dataTarget) != TDataType.getObjectFormat(cdo))
      return -TErrorList.illegal_format;
    for (int i=0; i<number; i++) 
      ((TCompoundDataObject[])dataTarget)[i] = cdo[i];
    return number;
  }    
  private int loadLTSData(String fileName,int index,long tsstart,long tsstop,int sysstart,int sysstop,
      float lwrLimit, float uprLimit,Object dataTarget,int arraysize,int arrayoffset,int sample)
  {
    int depth = hspec.getDepthLong();
    if (storeSummary) depth = 1;
    if (arraysize < 1) return 0;
    if (depth <= 0) return 0;
    long tnow = TDataTime.now();
    if ((tnow - tsstart) < hspec.getDepthShort()) return 0;
    if (index < 0 || index >= getArraySize()) return 0;
    double dstarttime = (double)(tsstart/1000.0);
    double dendtime = (double)(tsstop/1000.0);
    double dtnow = (double)(tnow)/1000.0;
    double datatime = 0.0;
    double dlaststamp = -1;
    int datasys = sysstart;
    int n, nrecs,raster = 1;
    int timebin = 1,pstart,pstop,number = 0;
    int sstamp = 0, ustamp = 0;
    boolean isSnapshot = false, useSysStamp = false;
    boolean hasPointsOfInterest = false, isPointsOfInterestFile = false;
    String poiName = null;
    long poiStart,poiStop;

    int rlen = getRecordLength();
    int hlen = getRecordPrefixLength(); 
    if (fileName.startsWith("pi") == false)
    { // isn't already a points-of-interest file 
      poiName = fileName.replace(fileName.substring(0, 2), "pi");
      File poif = new File(arcDbPath +  File.separator + poiName);
      if (!poif.exists())
      { // not there : try the 'static' area
        poif = new File(arcDbPathSaved +  File.separator + poiName);
      }
      if (poif.exists())
      { // there IS a companion points-of-interest file
        if (poif.length() < rlen * 100)
        { // only turn this on if the points-of-interest file doesn't appear to contain a lot of garbage
          hasPointsOfInterest = true;
        }
        else
        {
          MsgLog.log("loadLTSData","ignoring points-of-interest file (too large)",TErrorList.not_allowed,null,0);
        }
      }
    }
    else
    {
      isPointsOfInterestFile = true;
    }
    String fn = arcDbPath +  File.separator + fileName;
    if (!(new File(fn).exists()))
    { // not there : try the 'static' area
      fn = arcDbPathSaved +  File.separator + fileName;
    }
    if (dendtime < dstarttime) dendtime = dtnow;
    if ((pstart=locateRecordFromLTS(fn,dstarttime,isPointsOfInterestFile)) < 0) return 0;
    if ((pstop=locateRecordFromLTS(fn,dendtime,isPointsOfInterestFile)) < 0) return 0;
    isSnapshot = (dendtime-dstarttime) < 0.01;
    if (sysstart > 0 && sysstart == sysstop) isSnapshot = true;
    if (sysstart > 0 && sysstop >= sysstart && !useMinimalStorage) useSysStamp = true;
    nrecs = (int)((pstop-pstart)/rlen);
    if (arraysize == 1 && !isSnapshot) 
    {
      if (isPointsOfInterestFile) return 0;
      lastReadTimestamp = tsstop;
      return nrecs > 0 ? nrecs : 0;
    }
    arraysize -= arrayoffset;  // subtract off the offset
    if (arraysize < nrecs) raster += nrecs/(arraysize+1);
    if (raster > 1 && isPointsOfInterestFile) return 0;
    if (nrecs == 0 && isPointsOfInterestFile) return 0;
    if (sample > 0) raster = sample;
    if (raster == 1) hasPointsOfInterest = false; // no need to check
    short targetFmt = TDataType.getObjectFormat(dataTarget);
    if (targetFmt == TFormat.CF_NULL) return 0;
    if (isSnapshot)
    {
      index = 0;
      if (!useSysStamp) datatime = dstarttime - timebin;
    }
    else if (nrecs == 0)
    {
      return 0;
    }
    try 
    {
      int cc;
      double dts = 0.0;
      short sourceFmt = getFmt();
      String sourceTag = getTag();
      double[] dv = null;
      NAME[] nv = null;
      TCompoundDataObject[] cdo = null;
      int nsiz = TFormat.formatSizeOf(sourceFmt);
      
      RandomAccessFile raf = new RandomAccessFile(fn,"r");
      long flen = raf.length();
      raf.seek(pstart);
      
      while (raf.getFilePointer() < flen) 
      { // read through file
        if (!useMinimalStorage)
        {
          dts = raf.readDouble();
          sstamp = raf.readInt();
          ustamp = raf.readInt();
        }
        else
        {
          dts = (double)raf.readInt();          
        }
        if (hasPointsOfInterest && dlaststamp > 0 && dts > dlaststamp-2)
        {
          poiStart = (long)(dlaststamp*1000.0) + 500;
          poiStop = (long)(dts*1000.0) - 500;
          int poisize = arraysize < 2 ? arraysize : 2;
          n = loadLTSData(poiName,index,poiStart,poiStop,sysstart,sysstop,lwrLimit,uprLimit,dataTarget,poisize,arrayoffset+number,1);
          if (n >= arraysize - number) break;
          if (n > 0) number += n;
        }
        if (isSnapshot)
        {
          if (dts > dendtime) dendtime = dts;
          if (sstamp > sysstop) sysstop = sstamp;
        }
        if (dts > dendtime) return number;
        if (useSysStamp && sstamp > sysstop) return number;
        if ((useSysStamp && sstamp < datasys) ||
            dts < datatime + (double)timebin/1000.0)
        {
          raf.seek(raf.getFilePointer()+rlen-hlen);
          continue;
        }
        datatime = dts;
        datasys = sstamp;
        int dsiz = siz;
        if (storeSummary)
        {
          sourceFmt = TFormat.CF_DOUBLE;
          dsiz = 6;
          if (index > 3) return 0;
        }
        // read the data
        //TODO:  LTS of an array index will NOT need to read all the elements !
        // (this looks really inefficient !
        switch (sourceFmt)
        {
          case TFormat.CF_DOUBLE:
            if (dv == null) dv = new double[dsiz];
            for (int i=0; i<dsiz; i++) dv[i] = raf.readDouble();
            break;
          case TFormat.CF_FLOAT:
            if (dv == null) dv = new double[dsiz];
            for (int i=0; i<dsiz; i++) dv[i] = raf.readFloat(); 
            break;
          case TFormat.CF_DLONG:
            if (dv == null) dv = new double[dsiz];
            for (int i=0; i<dsiz; i++) dv[i] = raf.readLong();
            break;            
          case TFormat.CF_LONG:
            if (dv == null) dv = new double[dsiz];
            for (int i=0; i<dsiz; i++) dv[i] = raf.readInt();
            break;
          case TFormat.CF_SHORT:
            if (dv == null) dv = new double[dsiz];
            for (int i=0; i<dsiz; i++) dv[i] = raf.readShort();
            break;
          case TFormat.CF_BYTE:
            if (dv == null) dv = new double[dsiz];
            for (int i=0; i<dsiz; i++) dv[i] = raf.readByte();
            break;
          case TFormat.CF_NAME16:
          case TFormat.CF_NAME32:
          case TFormat.CF_NAME64:
            if (nv == null) nv = new NAME[dsiz];
            byte[] b = new byte[nsiz];
            for (int i=0; i<dsiz; i++) 
            {
              nv[i] = new NAME(nsiz);
              raf.read(b,0,nsiz);
              nv[i].setName(b);
            }
            break;
          default:
            if (cdo == null) 
            {
              cdo = TFormat.makeCompoundDataObjectArray(sourceFmt, sourceTag, dsiz);
            }
            break;  
        }
        if (cdo != null)
        { // compound data type : still need to read it
          for (int i=0; i<dsiz; i++) 
          {
            raf.read(cdo[i].getByteArray(), 0, nsiz);
            cdo[i].toStruct();
          }          
        }
        if (isSnapshot)
        {
          number = getFormattedHistoryDataSet(dataTarget,dv,nv,cdo,dts,sstamp,ustamp);
          break; // we're finished 
        }
        cc = getFormattedHistoryDataPoint(dataTarget,number+arrayoffset,dv,nv,cdo,index,dts,sstamp,ustamp);
        if (cc < 0)
        {
          number = cc;
          break;
        }
        if (sample < 0 && dv != null && !isInRange((float)dv[index],lwrLimit,uprLimit)) continue;
        number++;
        if (number >= arraysize) break;
        if (raster > 1) // move along through file
        {
          raf.seek(raf.getFilePointer()+rlen*(raster-1));
        }
        dlaststamp = dts;
      }
      raf.close();
      if (datatime > 0.0) lastReadTimestamp = (long)(datatime*1000);
      if (datasys > 0) lastReadSystemStamp = datasys;
      return number;
    }
    catch (Exception e)
    {
      TFecLog.log("Error reading history record : " + e.toString());
    }    
    return 0; 
  }  
  public int getDataFromSTS(int index,long starttime,long endtime,int sysstart,int sysstop,
         float lwrLimit,float uprLimit,Object targetData,int arraysize,int arrayOffset,int sample)
  {
    int depth = hspec.getDepthShort();
    int raster=1,cnt,cc,m;
    int n,number,rnumber=0;
    int prate = hspec.getPollngRate();
    boolean isSnapshot = false, useSysStamp = false;
    long thistimestamp;
    int sds, uds;
    double dts;
    double[] dv = null;
    NAME[] nv = null;
    TCompoundDataObject[] cdo = null;
    
    lastReadTimestamp = 0;
    lastReadSystemStamp = 0;
    if (prate == 0) return 0;
    if (depth == 0) return -TErrorList.un_allocated;
    if (index >= siz) return -TErrorList.out_of_range;
    if (arraysize < 1) return -TErrorList.buffer_too_small;    
    if (sysstart > 0 && sysstop >= sysstart) useSysStamp = true;
    if (starttime < endtime - depth*prate) starttime = endtime - depth*prate;
    raster = (int)(endtime-starttime)/(prate*arraysize);
    m = (int)(endtime-starttime)%(prate*arraysize);
    if (m != 0) raster++;
    if (raster < 1) raster = 1;
    if (arraysize == 1) raster = 1;
    if (sample > 0) raster = sample;
    if (starttime == endtime) isSnapshot = true;
    if (useSysStamp && sysstart == sysstop) isSnapshot = true;
    n = rngPtr;
    /* move raster past the current position to start at the beginning */
    n = (n + raster + depth) % depth;
    if (arraysize == endtime-starttime) 
    { // caller wants the array filled
      if (arraysize <= depth)
      {
        n = (n + (depth-arraysize) + depth) % depth;
        starttime = timestamp[n];
        endtime = TDataTime.now();
      }
    }
    lastReadTimestamp = starttime; lastReadSystemStamp = sysstart;
    for (number=arrayOffset,cnt=0; number < arraysize && number <= depth && cnt < depth; cnt += raster)
    {
      thistimestamp = timestamp[n];
      sds = sysstamp[n];
      uds = usrstamp[n];
      n = (n + raster + depth) % depth;
      if (useSysStamp)
      {
        if (sds < lastReadSystemStamp || sds > sysstop) continue;
      }
      else 
      {
        if (thistimestamp < lastReadTimestamp || thistimestamp > endtime) continue;
      }
      lastReadTimestamp = thistimestamp;
      rnumber++;
      if (!isSnapshot && arraysize < 2) continue;
      nv = getNAMEArrayAtPosition(n);
      if (nv == null) cdo = getCompoundArrayAtPosition(n);
      if (nv == null && cdo == null) dv = getDoubleArrayAtPosition(n);
      dts = ((double)thistimestamp)/1000.0;
      if (isSnapshot)
      {
        number = getFormattedHistoryDataSet(targetData,dv,nv,cdo,dts,sds,uds);
        break; // we're finished 
      }
      cc = getFormattedHistoryDataPoint(targetData,number,dv,nv,cdo,index,dts,sds,uds);
      if (cc < 0)
      {
        number = cc;
        break;
      }
      if (sample < 0 && dv != null && !isInRange((float)dv[index],lwrLimit,uprLimit)) continue;
      number++;    
    }    
    if (!isSnapshot && arraysize < 2) assignNumberOfPoints(targetData,rnumber);
    return number;
  }
  public DSUMMARY getDataSummaryFromSTS(double sampleInterval)
  {
    double si = 1.0; // default sample interval in seconds
    DSUMMARY ds = new DSUMMARY();
    if (sampleInterval > 0) si = sampleInterval;
    
    double pps = 1000.0 / (double)hspec.getPollngRate();
    double depthInSeconds = (double)hspec.getDepthShortInSeconds();
    if (si > depthInSeconds) si = depthInSeconds;
    int npts = (int)(pps * si);
    if (npts < 1) throw new TineRuntimeErrorException(TErrorList.invalid_interval);
    int depth = hspec.getDepthShort();
    if (npts > depth) npts = depth;
    if (npts == 1 && depth > 2) npts = 2; // get it some lee way
    
    int k=0, cp = (rngPtr + 1) % depth;
    int n = (cp + depth - npts) % depth; // beginning of sampling interval
    double tnow = TDataTime.nowInSeconds();
    double ts, v, ra, ave=0, max=0, min=0, del=0;
    double[] dv;
    for (int np=-1; n != cp; n=np)
    {
      np = (n + 1 + depth) % depth;
      ts = (double)timestamp[n]/1000.0;
      if (ts < tnow - si) continue; 
      dv = getDoubleArrayAtPosition(n);
      if (dv == null)  throw new TineRuntimeErrorException(TErrorList.not_implemented);
      v = dv[0];
      ave += v;
      if (k == 0) 
      {
        max = min = v;
      }
      else
      {
        if (v > max) max = v;
        else if (v < min) min = v;
        ra = ave/(double)(k+1);
        if (v > ra) del += v - ra; else del += ra - v;
      }
      k++;
    }
    if (k == 0) throw new TineRuntimeErrorException(TErrorList.no_data_in_range);
    ave /= (double)k;
    del /= (double)k;
    ds.ave[0] = ave;
    ds.max[0] = max;
    ds.min[0] = min;
    ds.dev[0] = del;
    ds.cnt[0] = k;
    ds.ts[0] = (int)(si + (tnow - si)/2.0);
    ds.tMax[0] = ds.ts[0];
    ds.tMin[0] = ds.ts[0];    
    return ds;
  }
  private void assignNumberOfPoints(Object data,int numberOfPoints)
  {
    if (data instanceof float[])
    {
      ((float[])data)[0] = (float)numberOfPoints;
    }
    else if (data instanceof FLTINT[])
    {
      if (((FLTINT[])data)[0] == null) ((FLTINT[])data)[0] = new FLTINT();
      ((FLTINT[])data)[0].ival = numberOfPoints;
      ((FLTINT[])data)[0].fval = (float)numberOfPoints;
    }
    else if (data instanceof INTFLTINT[])
    {
      if (((INTFLTINT[])data)[0] == null) ((INTFLTINT[])data)[0] = new INTFLTINT();
      ((INTFLTINT[])data)[0].i1val = numberOfPoints;
      ((INTFLTINT[])data)[0].fval  = (float)numberOfPoints;
      ((INTFLTINT[])data)[0].i2val = numberOfPoints;
    }
    else if (data instanceof FLTINTINT[])
    {
      if (((FLTINTINT[])data)[0] == null) ((FLTINTINT[])data)[0] = new FLTINTINT();
      ((FLTINTINT[])data)[0].i1val = numberOfPoints;
      ((FLTINTINT[])data)[0].i2val = numberOfPoints;
      ((FLTINTINT[])data)[0].f1val  = (float)numberOfPoints;
    }
    else if (data instanceof int[])
    {
      ((int[])data)[0] = numberOfPoints;
    }
    else if (data instanceof INTINT[])
    {
      if (((INTINT[])data)[0] == null) ((INTINT[])data)[0] = new INTINT();
      ((INTINT[])data)[0].i2val = numberOfPoints;
      ((INTINT[])data)[0].i1val = numberOfPoints;
    }
    else if (data instanceof double[])
    {
      ((double[])data)[0] = (double)numberOfPoints;
    }
    else if (data instanceof DBLDBL[])
    {
      if (((DBLDBL[])data)[0] == null) ((DBLDBL[])data)[0] = new DBLDBL();
      ((DBLDBL[])data)[0].d2val = (double)numberOfPoints;
      ((DBLDBL[])data)[0].d1val = (double)numberOfPoints;
    }
    else if (data instanceof NAME16I[])
    {
      if (((NAME16I[])data)[0] == null) ((NAME16I[])data)[0] = new NAME16I();
      ((NAME16I[])data)[0].ival = numberOfPoints;
      ((NAME16I[])data)[0].name = "";
    }
    else if (data instanceof NAME32I[])
    {
      if (((NAME32I[])data)[0] == null) ((NAME32I[])data)[0] = new NAME32I();
      ((NAME32I[])data)[0].ival = numberOfPoints;
      ((NAME32I[])data)[0].name = "";
    }
    else if (data instanceof NAME64I[])
    {
      if (((NAME64I[])data)[0] == null) ((NAME64I[])data)[0] = new NAME64I();
      ((NAME64I[])data)[0].ival = numberOfPoints;
      ((NAME64I[])data)[0].name = "";
    }    
  }
  private boolean isInRange(float value,float lower, float upper)
  {
    if (value < lower) return false;
    if (value > upper) return false;
    return true;
  }
  public float getFloatValueAtPosition(int ptr,int index)
  {
    if (index < 0 || index >= siz) return -1;
    if (ptr < 0 || ptr >= hspec.getDepthShort()) return -1;
    if (rngBuffer instanceof float[]) return ((float[])rngBuffer)[ptr*siz+index];
    if (rngBuffer instanceof byte[]) return (float)((byte[])rngBuffer)[ptr*siz+index];
    if (rngBuffer instanceof short[]) return (float)((short[])rngBuffer)[ptr*siz+index];
    if (rngBuffer instanceof int[]) return (float)((int[])rngBuffer)[ptr*siz+index];
    if (rngBuffer instanceof double[]) return (float)((double[])rngBuffer)[ptr*siz+index];
    return -1;    
  }  
  public double getDoubleValueAtPosition(int ptr,int index)
  {
    if (index < 0 || index >= siz) return -1;
    if (ptr < 0 || ptr > hspec.getDepthShort()) return -1;
    if (rngBuffer instanceof double[]) return ((double[])rngBuffer)[ptr*siz+index];
    if (rngBuffer instanceof byte[]) return (double)((byte[])rngBuffer)[ptr*siz+index];
    if (rngBuffer instanceof short[]) return (double)((short[])rngBuffer)[ptr*siz+index];
    if (rngBuffer instanceof int[]) return (double)((int[])rngBuffer)[ptr*siz+index];
    if (rngBuffer instanceof float[]) return (double)((float[])rngBuffer)[ptr*siz+index];
    return -1;    
  }
  public int getLongValueAtPosition(int ptr,int index)
  {
    if (index < 0 || index >= siz) return -1;
    if (ptr < 0 || ptr >= hspec.getDepthShort()) return -1;
    if (rngBuffer instanceof int[]) return ((int[])rngBuffer)[ptr*siz+index];
    if (rngBuffer instanceof byte[]) return (int)((byte[])rngBuffer)[ptr*siz+index];
    if (rngBuffer instanceof short[]) return (int)((short[])rngBuffer)[ptr*siz+index];
    if (rngBuffer instanceof double[]) return (int)((double[])rngBuffer)[ptr*siz+index];
    if (rngBuffer instanceof float[]) return (int)((float[])rngBuffer)[ptr*siz+index];
    return -1;    
  }
  public short getShortValueAtPosition(int ptr,int index)  
  {
    if (index < 0 || index >= siz) return -1;
    if (ptr < 0 || ptr >= hspec.getDepthShort()) return -1;
    if (rngBuffer instanceof byte[]) return (short)((byte[])rngBuffer)[ptr*siz+index];
    if (rngBuffer instanceof short[]) return ((short[])rngBuffer)[ptr*siz+index];
    if (rngBuffer instanceof int[]) return (short)((int[])rngBuffer)[ptr*siz+index];
    if (rngBuffer instanceof double[]) return (short)((double[])rngBuffer)[ptr*siz+index];
    if (rngBuffer instanceof float[]) return (short)((float[])rngBuffer)[ptr*siz+index];
    return -1;    
  }
  public NAME16 getNAME16ValueAtPosition(int ptr,int index)
  {
    if (index < 0 || index >= siz) return null;
    if (ptr < 0 || ptr >= hspec.getDepthShort()) return null;
    if (rngBuffer instanceof NAME16[]) return ((NAME16[])rngBuffer)[ptr*siz+index];
    if (rngBuffer instanceof NAME32[]) return new NAME16(((NAME32[])rngBuffer)[ptr*siz+index].toString());
    if (rngBuffer instanceof NAME64[]) return new NAME16(((NAME64[])rngBuffer)[ptr*siz+index].toString());
    return null;    
  }
  public NAME32 getNAME32ValueAtPosition(int ptr,int index)
  {
    if (index < 0 || index >= siz) return null;
    if (ptr < 0 || ptr >= hspec.getDepthShort()) return null;
    if (rngBuffer instanceof NAME32[]) return ((NAME32[])rngBuffer)[ptr*siz+index];
    if (rngBuffer instanceof NAME16[]) return new NAME32(((NAME16[])rngBuffer)[ptr*siz+index].toString());
    if (rngBuffer instanceof NAME64[]) return new NAME32(((NAME64[])rngBuffer)[ptr*siz+index].toString());
    return null;    
  }
  public NAME64 getNAME64ValueAtPosition(int ptr,int index)
  {
    if (index < 0 || index >= siz) return null;
    if (ptr < 0 || ptr >= hspec.getDepthShort()) return null;
    if (rngBuffer instanceof NAME64[]) return ((NAME64[])rngBuffer)[ptr*siz+index];
    if (rngBuffer instanceof NAME16[]) return new NAME64(((NAME16[])rngBuffer)[ptr*siz+index].toString());
    if (rngBuffer instanceof NAME32[]) return new NAME64(((NAME32[])rngBuffer)[ptr*siz+index].toString());
    return null;    
  }
  public TCompoundDataObject[] getCompoundArrayAtPosition(int ptr)
  {
    TCompoundDataObject[] values = new TCompoundDataObject[siz];
    if (rngBuffer instanceof TCompoundDataObject[])
    {
      for (int i=0; i<siz; i++) values[i] = ((TCompoundDataObject[])rngBuffer)[ptr*siz+i]; 
      return values;
    }
    return null;
  }
  public NAME[] getNAMEArrayAtPosition(int ptr)
  {
    NAME[] values = new NAME[siz];
    if (rngBuffer instanceof NAME64[]) 
    {
      for (int i=0; i<siz; i++) values[i] = ((NAME64[])rngBuffer)[ptr*siz+i]; 
      return values;
    }
    if (rngBuffer instanceof NAME16[]) 
    {
      for (int i=0; i<siz; i++) values[i] = ((NAME16[])rngBuffer)[ptr*siz+i]; 
      return values;
    }
    if (rngBuffer instanceof NAME32[])
    {
      for (int i=0; i<siz; i++) values[i] = ((NAME32[])rngBuffer)[ptr*siz+i]; 
      return values;      
    }
    return null;
  }
  public float[] getFloatArrayAtPosition(int ptr)
  {
    float[] values = new float[siz];
    
    if (rngBuffer instanceof float[])
    {
      System.arraycopy((float[])rngBuffer,ptr*siz,values,0,siz);
    }
    else if (rngBuffer instanceof byte[])
    {
      for (int i=0; i<siz; i++) values[i] = (float)((byte[])rngBuffer)[ptr*siz+i];
    }
    else if (rngBuffer instanceof short[])
    {
      for (int i=0; i<siz; i++) values[i] = (float)((short[])rngBuffer)[ptr*siz+i];
    }
    else if (rngBuffer instanceof int[])
    {
      for (int i=0; i<siz; i++) values[i] = (float)((int[])rngBuffer)[ptr*siz+i];
    }
    else if (rngBuffer instanceof double[])
    {
      for (int i=0; i<siz; i++) values[i] = (float)((double[])rngBuffer)[ptr*siz+i];
    }
    else
    {
      return null;
    }
    return values;
  }
  public double[] getDoubleArrayAtPosition(int ptr)
  {
    double[] values = new double[siz];
    
    if (rngBuffer instanceof double[])
    {
      System.arraycopy((double[])rngBuffer,ptr*siz,values,0,siz);
    }
    else if (rngBuffer instanceof byte[])
    {
      for (int i=0; i<siz; i++) values[i] = (double)((byte[])rngBuffer)[ptr*siz+i];
    }
    else if (rngBuffer instanceof short[])
    {
      for (int i=0; i<siz; i++) values[i] = (double)((short[])rngBuffer)[ptr*siz+i];
    }
    else if (rngBuffer instanceof int[])
    {
      for (int i=0; i<siz; i++) values[i] = (double)((int[])rngBuffer)[ptr*siz+i];
    }
    else if (rngBuffer instanceof float[])
    {
      for (int i=0; i<siz; i++) values[i] = (double)((float[])rngBuffer)[ptr*siz+i];
    }
    else if (rngBuffer instanceof TCompoundDataObject[])
    {
      for (int i=0; i<siz; i++) values[i] = ((TCompoundDataObject[])rngBuffer)[ptr*siz+i].getDoubleValue();
    }
    else
    {
      return null;
    }
    return values;
  }
  public int[] getLongArrayAtPosition(int ptr)
  {
    int[] values = new int[siz];
    
    if (rngBuffer instanceof int[])
    {
      System.arraycopy((int[])rngBuffer,ptr*siz,values,0,siz);
    }
    else if (rngBuffer instanceof short[])
    {
      for (int i=0; i<siz; i++) values[i] = (int)((short[])rngBuffer)[ptr*siz+i];
    }
    else if (rngBuffer instanceof byte[])
    {
      for (int i=0; i<siz; i++) values[i] = (int)((byte[])rngBuffer)[ptr*siz+i];
    }
    else if (rngBuffer instanceof double[])
    {
      for (int i=0; i<siz; i++) values[i] = (int)((double[])rngBuffer)[ptr*siz+i];
    }
    else if (rngBuffer instanceof float[])
    {
      for (int i=0; i<siz; i++) values[i] = (int)((float[])rngBuffer)[ptr*siz+i];
    }
    else
    {
      return null;
    }
    return values;
  }
  public short[] getShortArrayAtPosition(int ptr)
  {
    short[] values = new short[siz];
    
    if (rngBuffer instanceof short[])
    {
      System.arraycopy((short[])rngBuffer,ptr*siz,values,0,siz);
    }
    else if (rngBuffer instanceof byte[])
    {
      for (int i=0; i<siz; i++) values[i] = (short)((byte[])rngBuffer)[ptr*siz+i];
    }
    else if (rngBuffer instanceof int[])
    {
      for (int i=0; i<siz; i++) values[i] = (short)((int[])rngBuffer)[ptr*siz+i];
    }
    else if (rngBuffer instanceof double[])
    {
      for (int i=0; i<siz; i++) values[i] = (short)((double[])rngBuffer)[ptr*siz+i];
    }
    else if (rngBuffer instanceof float[])
    {
      for (int i=0; i<siz; i++) values[i] = (short)((float[])rngBuffer)[ptr*siz+i];
    }
    else
    {
      return null;
    }
    return values;
  }
  public NAME16[] getNAME16ArrayAtPosition(int ptr)
  {
    NAME16[] values = new NAME16[siz];
    
    if (rngBuffer instanceof NAME16[])
    {
      for (int i=0; i<siz; i++) values[i] = new NAME16(((NAME16[])rngBuffer)[i].toString());
    }
    if (rngBuffer instanceof NAME32[])
    {
      for (int i=0; i<siz; i++) values[i] = new NAME16(((NAME32[])rngBuffer)[i].toString());
    }
    if (rngBuffer instanceof NAME64[])
    {
      for (int i=0; i<siz; i++) values[i] = new NAME16(((NAME64[])rngBuffer)[i].toString());
    }
   return values;
  }
  public NAME32[] getNAME32ArrayAtPosition(int ptr)
  {
    NAME32[] values = new NAME32[siz];
    
    if (rngBuffer instanceof NAME16[])
    {
      for (int i=0; i<siz; i++) values[i] = new NAME32(((NAME16[])rngBuffer)[ptr*siz+i].toString());
    }
    if (rngBuffer instanceof NAME32[])
    {
      for (int i=0; i<siz; i++) values[i] = new NAME32(((NAME32[])rngBuffer)[ptr*siz+i].toString());
    }
    if (rngBuffer instanceof NAME64[])
    {
      for (int i=0; i<siz; i++) values[i] = new NAME32(((NAME64[])rngBuffer)[ptr*siz+i].toString());
    }
    return values;
  }
  public NAME64[] getNAME64ArrayAtPosition(int ptr)
  {
    NAME64[] values = new NAME64[siz];
    
    if (rngBuffer instanceof NAME16[])
    {
      for (int i=0; i<siz; i++) values[i] = new NAME64(((NAME16[])rngBuffer)[ptr*siz+i].toString());
    }
    if (rngBuffer instanceof NAME32[])
    {
      for (int i=0; i<siz; i++) values[i] = new NAME64(((NAME32[])rngBuffer)[ptr*siz+i].toString());
    }
    if (rngBuffer instanceof NAME64[])
    {
      for (int i=0; i<siz; i++) values[i] = new NAME64(((NAME64[])rngBuffer)[ptr*siz+i].toString());
    }
    return values;
  }
  private int writeRecordToRingBuffer(Object values,int position)
  {
    if (values instanceof byte[]) 
    {
      if (!(rngBuffer instanceof byte[])) return TErrorList.illegal_format;
      int len = ((byte[])values).length;
      if (len > siz) len = siz;
      System.arraycopy(values,0,(byte[])rngBuffer,position*siz,len);            
    }
    else if (values instanceof short[]) 
    {
      if (!(rngBuffer instanceof short[])) return TErrorList.illegal_format;
      int len = ((short[])values).length;
      if (len > siz) len = siz;
      System.arraycopy(values,0,(short[])rngBuffer,position*siz,len);            
    }
    else if (values instanceof int[]) 
    {
      if (!(rngBuffer instanceof int[])) return TErrorList.illegal_format;
      int len = ((int[])values).length;
      if (len > siz) len = siz;
      System.arraycopy(values,0,(int[])rngBuffer,position*siz,len);            
    }
    else if (values instanceof float[]) 
    {
      if (!(rngBuffer instanceof float[])) return TErrorList.illegal_format;
      int len = ((float[])values).length;
      if (len > siz) len = siz;
      System.arraycopy(values,0,(float[])rngBuffer,position*siz,len);      
    }
    else if (values instanceof double[]) 
    {
      if (!(rngBuffer instanceof double[])) return TErrorList.illegal_format;
      int len = ((double[])values).length;
      if (len > siz) len = siz;
      System.arraycopy(values,0,(double[])rngBuffer,position*siz,len);            
    }
    else if (values instanceof TCompoundDataObject[]) 
    {
      if (!(rngBuffer instanceof TCompoundDataObject[])) return TErrorList.illegal_format;
      if (((TCompoundDataObject[])values)[0].getFormat() != 
          ((TCompoundDataObject[])rngBuffer)[0].getFormat())
        return TErrorList.illegal_format;
      int len = ((TCompoundDataObject[])values).length;
      if (len > siz) len = siz;
      System.arraycopy(values,0,rngBuffer,position*siz,len);            
    }
    else
    {
      return TErrorList.illegal_format;
    }
    return 0;
  }
  public int writeRecordToSTS(TDataType d)
  { // write to Short-Term-Storage (i.e. ram)
    rngPtr = (rngPtr + 1) % hspec.getDepthShort();
    int cc = writeRecordToRingBuffer(d.getDataObject(),rngPtr);
    timestamp[rngPtr] = d.getDataTimeStamp();
    sysstamp[rngPtr] = d.getSystemDataStamp();
    usrstamp[rngPtr] = d.getUserDataStamp();
    return cc;
  }
  public boolean needsToFillSTS = false;
  public int fillSTS()
  {
    try
    {
      needsToFillSTS = false;
      if (hspec.getDepthLong() > 1) return 0; // not necessary
      if (!TFormat.isNumberFormat(fmt)) return TErrorList.not_implemented;
      DBLDBL[] dda = new DBLDBL[rngBufferSize];
      String aprp = eqm.getExportName()+"."+prp;
      double stop = TDataTime.getDataTimeStamp(); 
      double start = stop - hspec.getDepthShortInSeconds() - 1800.0;
      int npts = THistory.getArchivedData(eqm.getContext(), "HISTORY", aprp, dev, 0, 0, start, stop, dda,500);
      if (npts > 0)
      {
        TDataType d = new TDataType(1,fmt);
        double[] dd = new double[1];
        for (int i=0; i<npts; i++)
        {
          dd[0] = dda[i].d1val;
          d.putData(dd);
          d.setDataTimeStamp(dda[i].d2val);
          writeRecordToSTS(d);
        }
      }
      return npts < 0 ? -npts : 0;
    }
    catch (Exception e)
    {
      TFecLog.log("THistoryRecord","could not fill short-term storage for "+prp+" : "+e.getMessage());
      return TErrorList.code_failure;
    }
  }
  private String twoDigitString(int n)
  {
    return n < 10 ? "0" + n : "" + n;
  }
  public void setArchiveFileName(long cycleTime)
  {
    if (calendar == null) calendar = Calendar.getInstance();
    calendar.setTimeInMillis(cycleTime);
    int dd = calendar.get(Calendar.DAY_OF_MONTH);
    if (dd == lastArchivedDayOfTheMonth) return;
    int yy, mm;
    lastArchivedDayOfTheMonth = dd;
    yy = calendar.get(Calendar.YEAR) % 100;
    mm = calendar.get(Calendar.MONTH) + 1; // January = 0
    if (useMonthlyHistoryFiles) dd = 1; 
    ArchiveFileName = getFilePrefix() +  twoDigitString(yy) + twoDigitString(mm) + 
                      twoDigitString(dd) + "." + Integer.toHexString(rcdIdx);

    if (useStandardFileSet)
    {
      int mm0 = ((mm+12) - ((hspec.getDepthLong()+1)%12))%12;
      yy -= hspec.getDepthLong()/12;
      if (mm0 > mm) yy--;
      
      String rmvFileName = getFilePrefix() +  twoDigitString(yy) + twoDigitString(mm0) + 
                           twoDigitString(dd) + "." + Integer.toHexString(rcdIdx);
      rotateHistoryFile(rmvFileName,ArchiveFileName);
    }
    else
    {
      removeOutdatedLTSFiles(cycleTime);
    }
  }
  private boolean unlinkHistFile(int yy,int mm,int dd)
  {
    String fileName = null, poiName = null, pfix = getFilePrefix();
    fileName = pfix+twoDigitString(yy)+twoDigitString(mm)+twoDigitString(dd)+"."+Integer.toHexString(rcdIdx);
    poiName = new String(fileName); poiName.replace(pfix, "pi");
    File f = null;
    try
    {
      f = new File(arcDbPath+File.separator+fileName);
      if (!f.exists()) return false;
      f.delete();
      f = new File(arcDbPath+File.separator+poiName);
      if (f.exists()) f.delete();
      return true;
    }
    catch (Exception e)
    {
      String msg = "unable to delete file " + f != null ? f.getName() : "(null file)";
      MsgLog.log("removeOutdatedLTSFiles", msg,TErrorList.file_error,e,0);
      TFecLog.log(msg);
    }
    return false;
  }
  public void removeOutdatedLTSFiles(long cycleTime)
  {
    int yy,mm,mm0,dd;
    calendar.setTimeInMillis(cycleTime);
    yy = calendar.get(Calendar.YEAR) % 100;
    mm0 = calendar.get(Calendar.MONTH) + 1; // January = 0
    mm = ((mm0+12) - ((hspec.getDepthLong()+1)%12))%12;
    yy -= hspec.getDepthLong()/12;
    if (mm > mm0) yy--;
    dd = calendar.get(Calendar.DAY_OF_MONTH);
    
    do
    {
      for (int i=dd-1; i>0; i--)
      {
        if (!unlinkHistFile(yy,mm,i) && i<27) break;
      }
      mm--;
      dd = 32;
    } while (mm > 0);    
  }
  private String getFilePrefix()
  {
    if (useStandardFileSet) return "ts";
    return useMinimalStorage ? "ar" : "ta"; /* "ha" */
  }
  private void rotateHistoryFile(String src, String dst)
  {
    File fdst = new File(arcDbPath +  File.separator + dst);
    if (fdst.exists()) return;
    File fsrc = new File(arcDbPath +  File.separator + src);
    fsrc.renameTo(fdst);
    try
    {
      RandomAccessFile raf = new RandomAccessFile(fdst,"rw");
      setFilePosition(raf,AS_STATE_SIZE);
    }
    catch (Exception e)
    {
      MsgLog.log("rotateHistoryFile","could not set file position in "+fdst.getName(),TErrorList.file_error,e,0);
    }
    TFecLog.log("rotate history file " + src + " to " + dst);
  }
  public String getArchiveFileName()
  {
    return ArchiveFileName;
  }
  public int getArrayType()
  {
    return aType;
  }
  public int getRingBufferPointer()
  {
    return rngPtr;
  }
  public String getDev()
  {
    return dev;
  }
  public TEquipmentModule getEqm()
  {
    return eqm;
  }
  public byte getFmt()
  {
    return fmt;
  }
  public String getTag()
  {
    return tag;
  }
  public String getRemoteHistoryServer() 
  { 
    return hspec.getRemoteServerSpecification();
  }
  public THistorySpecification getHspec()
  {
    return hspec;
  }
  public long getLastArchiveTime()
  {
    return lastArchiveTime;
  }
  public String getPrp()
  {
    return prp;
  }
  public int getRcdIdx()
  {
    return rcdIdx;
  }
  public int getRngBufferSize()
  {
    return rngBufferSize;
  }
  public int getArraySize()
  {
    return siz;
  }
  public int getRecordLength()
  {
    int rlen = storeSummary ? DSUMMARY.sizeInBytes  :  siz * TFormat.formatSizeOf(fmt);
    return rlen + getRecordPrefixLength();
  }
  public long getTimestamp()
  {
    return rngPtr < 0 ? 0 : timestamp[rngPtr];
  }
  public int getSystemStamp()
  {
    return rngPtr < 0 ? 0 : sysstamp[rngPtr];
  }
  public int getUserStamp()
  {
    return rngPtr < 0 ? 0 : usrstamp[rngPtr];
  }
  public boolean isUseMonthlyHistoryFiles()
  {
    return useMonthlyHistoryFiles;
  }
  public void setUseMonthlyHistoryFiles(boolean useMonthlyHistoryFiles)
  {
    this.useMonthlyHistoryFiles = useMonthlyHistoryFiles;
  }
  public boolean isUseMinimalStorage()
  {
    return useMinimalStorage;
  }
  public void setUseMinimalStorage(boolean value)
  {
    useMinimalStorage = value;
  }
  /**
   * @param useMsecHistoryTimeStamps
   * @deprecated
   * use setUseMinimalStorage() instead
   */
  public void setUseMsecHistoryTimeStamps(boolean useMsecHistoryTimeStamps)
  {
    useMinimalStorage = !useMsecHistoryTimeStamps;
  }
  public boolean isWithinTolerance()
  {
    double t, d0, d1;
    String s0, s1;
    float aTol = hspec.getAbsoluteTolerance();
    float pTol = hspec.getPercentTolerance();
    int p;
    int ptr0 = rngPtr;
    int ptr1 = hspec.getDepthShort(); //(ptr0 + rngBufferSize - 1) % rngBufferSize; 
    switch (fmt)
    {
      case TFormat.CF_DOUBLE:
      case TFormat.CF_DLONG:          
      case TFormat.CF_FLOAT:
      case TFormat.CF_SHORT:
      case TFormat.CF_LONG:
        for (p=0; p<siz; p++)
        {
          d0 = getDoubleValueAtPosition(ptr0,p);
          d1 = getDoubleValueAtPosition(ptr1,p);
          t = aTol + pTol * Math.abs(d1);
          if (Math.abs(d0 - d1) > t) return false;
        }
        break;
      case TFormat.CF_NAME16:
      case TFormat.CF_NAME32:
      case TFormat.CF_NAME64:
        for (p=0; p<siz; p++)
        {
          s0 = getNAME64ValueAtPosition(ptr0,p).toString();
          s1 = getNAME64ValueAtPosition(ptr1,p).toString();
          if (s0.compareTo(s1) == 0) continue;
          return false;
        }
        break;
      default:
        try
        {
          TCompoundDataObject[] tco0 = getCompoundArrayAtPosition(ptr0);
          TCompoundDataObject[] tco1 = getCompoundArrayAtPosition(ptr1);
          byte[] ba0, ba1;
          for (p=0; p<siz; p++)
          {
            ba0 = tco0[p].getByteArray();
            ba1 = tco1[p].getByteArray();
            if (!Arrays.equals(ba0, ba1)) return false;
          }
        }
        catch (Exception ignore)
        { // possible an invalid cast exception
          // whatever it is, break and return true
        }
        break;
    }
    return true;
  }
  private static double poiRangeFactor = 0.10;
  private static double poiToleranceFactor = 10.0;
  public boolean isPointOfInterest()
  {
    if (!keepPointsOfInterest) return false;
    if (rngPtr == 1 && timestamp[rngBufferSize-1] == 0)
    { // just starting
      return false;
    }
    double t, d0, d1;
    float aTol = hspec.getAbsoluteTolerance();
    float pTol = hspec.getPercentTolerance();
    boolean useRange = pTol > 0 ? false : true;
    if (useRange && range <= 0) 
    {
      useRange = false;
      if (aTol == 0) return false;
    }
    int p;
    int ptr0 = rngPtr;
    int ptr1 = (ptr0 + rngBufferSize - 1) % rngBufferSize; 
    switch (fmt)
    {
      case TFormat.CF_DBLDBL:
      case TFormat.CF_DBLTIME:  
      case TFormat.CF_FLTINT:
      case TFormat.CF_FLTFLT:
      case TFormat.CF_DOUBLE:
      case TFormat.CF_DLONG:          
      case TFormat.CF_FLOAT:
      case TFormat.CF_SHORT:
      case TFormat.CF_LONG:
        for (p=0; p<siz; p++)
        {
          d0 = getDoubleValueAtPosition(ptr0,p);
          d1 = getDoubleValueAtPosition(ptr1,p);
          t = useRange ? poiRangeFactor * range : poiToleranceFactor *(aTol + pTol * Math.abs(d1));
          if (Math.abs(d0 - d1) > t) return true;
        }
        break;
      default:
        break;
    }
    return false;    
  }
  public static double getPoiRangeFactor()
  {
    return poiRangeFactor;
  }
  public static void setPointsOfInterestRangeFactor(double poiRangeFactor)
  {
    THistoryRecord.poiRangeFactor = poiRangeFactor;
  }
  public static double getPoiToleranceFactor()
  {
    return poiToleranceFactor;
  }
  public static void setPointsOfInterestToleranceFactor(double poiToleranceFactor)
  {
    THistoryRecord.poiToleranceFactor = poiToleranceFactor;
  }
}
