/*
 * Created on Aug 31, 2005
 *
 * To change the template for this generated file go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */
package de.desy.tine.server.logger;

import java.io.*;
import java.util.logging.*;

import de.desy.tine.client.TLinkFactory;
import de.desy.tine.definitions.TErrorList;
import de.desy.tine.server.equipment.TEquipmentModuleFactory;
import de.desy.tine.startup.TInitializer;
import de.desy.tine.startup.TInitializerFactory;

public class TFecLog
{
  public static final int LOG_SCAN_NONE = 0;
  public static final int LOG_SCAN_FEC = 1;
  public static final int LOG_SCAN_FULL = 2;
  private static int logFileAllowScan = LOG_SCAN_FEC;
  /**
   * @brief Sets the allowed range of log file retrieval from a FEC process
   * (via stock property SRVLOGFILE).
   *
   * A running FEC process will attempt to read and return the contents of
   * the file name input to stock property SRVLOGFILE.  The FEC process must
   * have read permission in order to succeed. The setting of
   * the LogFileAllowScan will further restrict the allowed directories from
   * which files may be read.  LogFileAllowScan = LOG_SCAN_FEC will only return
   * the contents of files within the FEC_HOME or FEC_LOG directories and their
   * subdirectories.  LogFileAllowScan = LOG_SCAN_FULL will itself allow the
   * entire file system.  LogFileAllowScan = LOG_SCAN_NONE will not allow any 
   * files to be retrieved other than fec.log.
   *
   * @param value is the desired scan range, one of: LOG_SCAN_NONE,
   * LOG_SCAN_FEC, LOG_SCAN_FULL (default = LOG_SCAN_FEC).  
   */
  public static void setLogFileAllowScan(int value)
  {
    if (value < LOG_SCAN_NONE) value = LOG_SCAN_NONE;
    if (value > LOG_SCAN_FULL) value = LOG_SCAN_FULL;
    logFileAllowScan = value;
  }
  /**
   * @brief Gets the allowed range of log file retrieval from a FEC process
   * (via stock property SRVLOGFILE).
   *
   * A running FEC process will attempt to read and return the contents of
   * the file name input to stock property SRVLOGFILE.  The FEC process must
   * have read permission in order to succeed. The setting of
   * the LogFileAllowScan will further restrict the allowed directories from
   * which files may be read.  LogFileAllowScan = LOG_SCAN_FEC will only return
   * the contents of files within the FEC_HOME or FEC_LOG directories and their
   * subdirectories.  LogFileAllowScan = LOG_SCAN_FULL will itself allow the
   * entire file system.  LogFileAllowScan = LOG_SCAN_NONE will not allow any 
   * files to be retrieved other than fec.log.
   *
   * @return the current scan range setting, one of: LOG_SCAN_NONE,
   * LOG_SCAN_FEC, LOG_SCAN_FULL (default = LOG_SCAN_FEC).  
   */
  public static int getLogFileAllowScan() 
  { 
    return logFileAllowScan; 
  }
  public static boolean isAllowedLogFileAccess(String filename)
  {
    if (logFileAllowScan == LOG_SCAN_FULL) return true;
    if (logFileAllowScan == LOG_SCAN_NONE)
    {
      if (filename.compareTo("fec.log.0") == 0) return true;
      return false;
    }
    if (!isAbsolutePath(filename)) return true;
    String plh = initializer.getLogHome();
    if (plh == null) return false;
    if (plh.length() > 0 && filename.startsWith(plh)) return true;
    String pfh = initializer.getFecHome();
    if (pfh.compareToIgnoreCase(plh) != 0)
    { /* FEC_LOG != FEC_HOME */
      if (pfh.length() > 0 && filename.startsWith(pfh)) return true;
    }
    return false;
  }
  private static TInitializer initializer = TInitializerFactory.getInstance().getInitializer();
  private static FileHandler feclogHandler = null;
  private static int logDepth = 1000;
  private static int debugLevel = 0;
  public static int getDebugLevel() { return debugLevel; }
  public static void setDebugLevel(int debuglevel) { debugLevel = debuglevel; }
  public static int getLogDepth() { return logDepth; }
  public static void setLogDepth(int depthInLines) { if (depthInLines > 0) logDepth = depthInLines; }
  private static String tag = null;
  private static final String fecLogFileName = "fec.log.0";
  public static void close()
  {
  	if (feclogHandler != null) 
  	{
  		feclogHandler.close();
  		feclogHandler =  null;
  	}
  }
  /**
   * @brief logs the entry into the server's 'fec.log' file
   * 
   * The fec.log is rotated with the current file instance named fec.log.0.
   *
   * @param text is the log entry
   * 
   * @return 0 upon success or a TINE error code
   */
  public static int log(String text)
  {
    if (!TLinkFactory.isRunningAsServer()) return TErrorList.not_running;
    if (tag == null) tag = TEquipmentModuleFactory.getInstance().getFecName();
    return log(tag,text);
  }
  /**
   * @brief logs the entry into the server's 'fec.log' file
   * 
   * The fec.log is rotated with the current file instance named fec.log.0.
   * 
   * @param tag is a tag identify preceding the logged entry
   * @param text is the log entry
   * 
   * @return 0 upon success or a TINE error code
   */
  public static int log(String tag, String text)
  {
    if (feclogHandler == null) 
    {
      feclogHandler = initializer.getFecLogResource();
      if (feclogHandler == null)
      { // still null !!!
        DbgLog.log("TFecLog","cannot write to log file : no resource !");
        return -1;
      }
      feclogHandler.setFormatter(new TFecLogFormatter());
    }
    String txt = "[" + tag + "] : " + text;
    try
    {
      
      LogRecord lr = new LogRecord(Level.INFO,txt);
      feclogHandler.publish(lr);
    }
    catch (OutOfMemoryError e)
    {
      e.printStackTrace();      
    }
    catch (Exception e)
    {
      DbgLog.log("TFecLog","cannot write to log file : " + e.toString());
    }
    if (TLinkFactory.debugLevel > 0) DbgLog.log("TFecLog",txt);
    return 0;
  }
  public static String[] getFiles()
  {
    return getFiles(initializer.getLogHome());
  }
  public static String[] getFiles(String filePath)
  {
    File path = new File(filePath);
    File[] rawList = path.listFiles();
    String[] files = new String[rawList.length];
    for (int i=0; i<rawList.length; i++)
    {
      files[i] = rawList[i].getName();      
    }
    return files;
  }
  public static String getLines(int maxNumberOfLines)
  {
    return getLines(fecLogFileName,maxNumberOfLines);
  }
  public static boolean fileExists(String fileName)
  {
    String fn = null;
    if (fileName == null || fileName.length() == 0) return false;
    if (fileName.indexOf('/') == -1 && fileName.indexOf('\\') == -1 )
    { // just the file name given => look in the default log directory
      fn = initializer.getLogHome() + "/" + fileName;
    }
    else
    { // full path given
      fn = fileName;
    }
    return new File(fn).exists();
  }
  public static int writeTextFile(String fileName,char[] fileText)
  {
    String fn = null;
    try
    {
      if (fileName.indexOf('/') == -1 && fileName.indexOf('\\') == -1 )
      { // just the file name given => look in the default log directory
        fn = initializer.getLogHome() + "/" + fileName;
      }
      else
      { // full path given
        fn = fileName;
      }
      RandomAccessFile raf = new RandomAccessFile(fn,"rw");
      raf.write(new String(fileText).getBytes());
      raf.close();
    }
    catch (Exception e)
    {
      DbgLog.log("TFecLog",e.toString());
      return TErrorList.file_error;
    }
    return 0;
  }
  public static boolean isAbsolutePath(String filename)
  {
    if (filename == null || filename.length() == 0) return false;
    if (filename.startsWith("/")) return true;
    if (filename.startsWith("\\\\")) return true;
    if (filename.startsWith(":\\", 1)) return true;
    if (filename.startsWith(":/",1)) return true; 
    return false;
  }
  public static String getLines(String fileName,int maxNumberOfLines)
  {
    String fn = null;
    StringBuffer sb = null;
    String text=null;
    int nl = 0;
    try
    {
      if (fileName.indexOf('/') == -1 && fileName.indexOf('\\') == -1 )
      { // just the file name given => look in the default log directory
        fn = initializer.getLogHome() + "/" + fileName;
      }
      else
      { // full path given
        fn = fileName;
      }
      RandomAccessFile raf = new RandomAccessFile(fn,"r");
      
      int flen = (int)raf.length();
      int fpos = flen - maxNumberOfLines*80;
      if (fpos < 0) fpos = 0;
      raf.seek(fpos);
      sb = new StringBuffer(flen-fpos);
      while ((text = raf.readLine()) != null)
      {
        sb.append(text);
        sb.append("\n");
        nl++;
      }
    }
    catch (Exception e)
    {
      DbgLog.log("TFecLog",e.toString());
    }
    if (sb == null) sb = new StringBuffer("file unavailable");
    text = sb.toString();
    if (debugLevel > 0) DbgLog.log("TFecLog",text);    
    return text;
  }
}
